/*LINTLIBRARY*/
/*  misc9.c

    This code provides simple colourmap generation routines.

    Copyright (C) 1998 Richard Gooch  API design and general maintenance
    Copyright (C) 2004 Robert Reid    Colour tables

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Robert Reid may be reached by email at    rob.reid@nrc-cnrc.gc.ca
    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for writing simple colour
    values.


    Written by      Richard Gooch   19-MAY-1998

    Last updated by Robert Reid     19-FEB-2004


*/
#include <stdio.h>
#include <math.h>
#include <karma_cf.h>
#include <karma_a.h>
#include <karma_n.h>
#include <karma.h>

#define LUT_SIZE      256

/* Three colour tables: Sunbow, Mutant, and Aberration.

   Sunbow is designed for printing Stokes Q, U, or V (or anything else that
   is ~50% negative) on colour plotters without using a lot of toner.

   Mutant and Aberration should get new names based on what they look like.

         Sunbow               Mutant             Aberration	
     R    RG  GB   B      G    GR  RB   B      R    RB  BG   G 
   1|________________   1|________________   1|________________
    |   /   /\   \       |   /   /\   \       |   /   /\   \   
    | G/   /  \   \G     | R/   /  \   \R     | B/   /  \   \B 
    | /   /    \   \     | /   /    \   \     | /   /    \   \ 
    |/   /B    R\   \    |/   /B    G\   \    |/   /G    R\   \
   0+----------------   0+----------------   0+----------------
    0               1    0               1    0               1


   These are half of a family of six similar tables in which the order of
   the colours is swapped:

		RG3D   Sunbow		
		   \   /		
		    \ /			
	     Hulk----*----Mutant	
		    / \			
		   /   \		
		RB3D   Aberration		

   Only those on the right half of the circle are given here, since the
   others can be replicated by reversing and inverting them.  

   The format of the tables below is based upon those found in 
   $KARMAROOT/source/lib/karma/master/cf_misc2.c, where LUT_SIZE = 256

   Rob Reid, 2/19/2004.
*/

static Colour_struct sunbow[LUT_SIZE] = 
{
    {1.00000, 0.00000, 0.00000},
    {1.00000, 0.01562, 0.00000},
    {1.00000, 0.03125, 0.00000},
    {1.00000, 0.04688, 0.00000},
    {1.00000, 0.06250, 0.00000},
    {1.00000, 0.07812, 0.00000},
    {1.00000, 0.09375, 0.00000},
    {1.00000, 0.10938, 0.00000},
    {1.00000, 0.12500, 0.00000},
    {1.00000, 0.14062, 0.00000},
    {1.00000, 0.15625, 0.00000},
    {1.00000, 0.17188, 0.00000},
    {1.00000, 0.18750, 0.00000},
    {1.00000, 0.20312, 0.00000},
    {1.00000, 0.21875, 0.00000},
    {1.00000, 0.23438, 0.00000},
    {1.00000, 0.25000, 0.00000},
    {1.00000, 0.26562, 0.00000},
    {1.00000, 0.28125, 0.00000},
    {1.00000, 0.29688, 0.00000},
    {1.00000, 0.31250, 0.00000},
    {1.00000, 0.32812, 0.00000},
    {1.00000, 0.34375, 0.00000},
    {1.00000, 0.35938, 0.00000},
    {1.00000, 0.37500, 0.00000},
    {1.00000, 0.39062, 0.00000},
    {1.00000, 0.40625, 0.00000},
    {1.00000, 0.42188, 0.00000},
    {1.00000, 0.43750, 0.00000},
    {1.00000, 0.45312, 0.00000},
    {1.00000, 0.46875, 0.00000},
    {1.00000, 0.48438, 0.00000},
    {1.00000, 0.50000, 0.00000},
    {1.00000, 0.51562, 0.00000},
    {1.00000, 0.53125, 0.00000},
    {1.00000, 0.54688, 0.00000},
    {1.00000, 0.56250, 0.00000},
    {1.00000, 0.57812, 0.00000},
    {1.00000, 0.59375, 0.00000},
    {1.00000, 0.60938, 0.00000},
    {1.00000, 0.62500, 0.00000},
    {1.00000, 0.64062, 0.00000},
    {1.00000, 0.65625, 0.00000},
    {1.00000, 0.67188, 0.00000},
    {1.00000, 0.68750, 0.00000},
    {1.00000, 0.70312, 0.00000},
    {1.00000, 0.71875, 0.00000},
    {1.00000, 0.73438, 0.00000},
    {1.00000, 0.75000, 0.00000},
    {1.00000, 0.76562, 0.00000},
    {1.00000, 0.78125, 0.00000},
    {1.00000, 0.79688, 0.00000},
    {1.00000, 0.81250, 0.00000},
    {1.00000, 0.82812, 0.00000},
    {1.00000, 0.84375, 0.00000},
    {1.00000, 0.85938, 0.00000},
    {1.00000, 0.87500, 0.00000},
    {1.00000, 0.89062, 0.00000},
    {1.00000, 0.90625, 0.00000},
    {1.00000, 0.92188, 0.00000},
    {1.00000, 0.93750, 0.00000},
    {1.00000, 0.95312, 0.00000},
    {1.00000, 0.96875, 0.00000},
    {1.00000, 0.98438, 0.00000},
    {1.00000, 1.00000, 0.01562},
    {1.00000, 1.00000, 0.03125},
    {1.00000, 1.00000, 0.04688},
    {1.00000, 1.00000, 0.06250},
    {1.00000, 1.00000, 0.07812},
    {1.00000, 1.00000, 0.09375},
    {1.00000, 1.00000, 0.10938},
    {1.00000, 1.00000, 0.12500},
    {1.00000, 1.00000, 0.14062},
    {1.00000, 1.00000, 0.15625},
    {1.00000, 1.00000, 0.17188},
    {1.00000, 1.00000, 0.18750},
    {1.00000, 1.00000, 0.20312},
    {1.00000, 1.00000, 0.21875},
    {1.00000, 1.00000, 0.23438},
    {1.00000, 1.00000, 0.25000},
    {1.00000, 1.00000, 0.26562},
    {1.00000, 1.00000, 0.28125},
    {1.00000, 1.00000, 0.29688},
    {1.00000, 1.00000, 0.31250},
    {1.00000, 1.00000, 0.32812},
    {1.00000, 1.00000, 0.34375},
    {1.00000, 1.00000, 0.35938},
    {1.00000, 1.00000, 0.37500},
    {1.00000, 1.00000, 0.39062},
    {1.00000, 1.00000, 0.40625},
    {1.00000, 1.00000, 0.42188},
    {1.00000, 1.00000, 0.43750},
    {1.00000, 1.00000, 0.45312},
    {1.00000, 1.00000, 0.46875},
    {1.00000, 1.00000, 0.48438},
    {1.00000, 1.00000, 0.50000},
    {1.00000, 1.00000, 0.51562},
    {1.00000, 1.00000, 0.53125},
    {1.00000, 1.00000, 0.54688},
    {1.00000, 1.00000, 0.56250},
    {1.00000, 1.00000, 0.57812},
    {1.00000, 1.00000, 0.59375},
    {1.00000, 1.00000, 0.60938},
    {1.00000, 1.00000, 0.62500},
    {1.00000, 1.00000, 0.64062},
    {1.00000, 1.00000, 0.65625},
    {1.00000, 1.00000, 0.67188},
    {1.00000, 1.00000, 0.68750},
    {1.00000, 1.00000, 0.70312},
    {1.00000, 1.00000, 0.71875},
    {1.00000, 1.00000, 0.73438},
    {1.00000, 1.00000, 0.75000},
    {1.00000, 1.00000, 0.76562},
    {1.00000, 1.00000, 0.78125},
    {1.00000, 1.00000, 0.79688},
    {1.00000, 1.00000, 0.81250},
    {1.00000, 1.00000, 0.82812},
    {1.00000, 1.00000, 0.84375},
    {1.00000, 1.00000, 0.85938},
    {1.00000, 1.00000, 0.87500},
    {1.00000, 1.00000, 0.89062},
    {1.00000, 1.00000, 0.90625},
    {1.00000, 1.00000, 0.92188},
    {1.00000, 1.00000, 0.93750},
    {1.00000, 1.00000, 0.95312},
    {1.00000, 1.00000, 0.96875},
    {1.00000, 1.00000, 0.98438},
    {1.00000, 1.00000, 1.00000},
    {0.98438, 1.00000, 1.00000},
    {0.96875, 1.00000, 1.00000},
    {0.95312, 1.00000, 1.00000},
    {0.93750, 1.00000, 1.00000},
    {0.92188, 1.00000, 1.00000},
    {0.90625, 1.00000, 1.00000},
    {0.89062, 1.00000, 1.00000},
    {0.87500, 1.00000, 1.00000},
    {0.85938, 1.00000, 1.00000},
    {0.84375, 1.00000, 1.00000},
    {0.82812, 1.00000, 1.00000},
    {0.81250, 1.00000, 1.00000},
    {0.79688, 1.00000, 1.00000},
    {0.78125, 1.00000, 1.00000},
    {0.76562, 1.00000, 1.00000},
    {0.75000, 1.00000, 1.00000},
    {0.73438, 1.00000, 1.00000},
    {0.71875, 1.00000, 1.00000},
    {0.70312, 1.00000, 1.00000},
    {0.68750, 1.00000, 1.00000},
    {0.67188, 1.00000, 1.00000},
    {0.65625, 1.00000, 1.00000},
    {0.64062, 1.00000, 1.00000},
    {0.62500, 1.00000, 1.00000},
    {0.60938, 1.00000, 1.00000},
    {0.59375, 1.00000, 1.00000},
    {0.57812, 1.00000, 1.00000},
    {0.56250, 1.00000, 1.00000},
    {0.54688, 1.00000, 1.00000},
    {0.53125, 1.00000, 1.00000},
    {0.51562, 1.00000, 1.00000},
    {0.50000, 1.00000, 1.00000},
    {0.48438, 1.00000, 1.00000},
    {0.46875, 1.00000, 1.00000},
    {0.45312, 1.00000, 1.00000},
    {0.43750, 1.00000, 1.00000},
    {0.42188, 1.00000, 1.00000},
    {0.40625, 1.00000, 1.00000},
    {0.39062, 1.00000, 1.00000},
    {0.37500, 1.00000, 1.00000},
    {0.35938, 1.00000, 1.00000},
    {0.34375, 1.00000, 1.00000},
    {0.32812, 1.00000, 1.00000},
    {0.31250, 1.00000, 1.00000},
    {0.29688, 1.00000, 1.00000},
    {0.28125, 1.00000, 1.00000},
    {0.26562, 1.00000, 1.00000},
    {0.25000, 1.00000, 1.00000},
    {0.23438, 1.00000, 1.00000},
    {0.21875, 1.00000, 1.00000},
    {0.20312, 1.00000, 1.00000},
    {0.18750, 1.00000, 1.00000},
    {0.17188, 1.00000, 1.00000},
    {0.15625, 1.00000, 1.00000},
    {0.14062, 1.00000, 1.00000},
    {0.12500, 1.00000, 1.00000},
    {0.10938, 1.00000, 1.00000},
    {0.09375, 1.00000, 1.00000},
    {0.07812, 1.00000, 1.00000},
    {0.06250, 1.00000, 1.00000},
    {0.04688, 1.00000, 1.00000},
    {0.03125, 1.00000, 1.00000},
    {0.01562, 1.00000, 1.00000},
    {0.00000, 1.00000, 1.00000},
    {0.00000, 0.98438, 1.00000},
    {0.00000, 0.96875, 1.00000},
    {0.00000, 0.95312, 1.00000},
    {0.00000, 0.93750, 1.00000},
    {0.00000, 0.92188, 1.00000},
    {0.00000, 0.90625, 1.00000},
    {0.00000, 0.89062, 1.00000},
    {0.00000, 0.87500, 1.00000},
    {0.00000, 0.85938, 1.00000},
    {0.00000, 0.84375, 1.00000},
    {0.00000, 0.82812, 1.00000},
    {0.00000, 0.81250, 1.00000},
    {0.00000, 0.79688, 1.00000},
    {0.00000, 0.78125, 1.00000},
    {0.00000, 0.76562, 1.00000},
    {0.00000, 0.75000, 1.00000},
    {0.00000, 0.73438, 1.00000},
    {0.00000, 0.71875, 1.00000},
    {0.00000, 0.70312, 1.00000},
    {0.00000, 0.68750, 1.00000},
    {0.00000, 0.67188, 1.00000},
    {0.00000, 0.65625, 1.00000},
    {0.00000, 0.64062, 1.00000},
    {0.00000, 0.62500, 1.00000},
    {0.00000, 0.60938, 1.00000},
    {0.00000, 0.59375, 1.00000},
    {0.00000, 0.57812, 1.00000},
    {0.00000, 0.56250, 1.00000},
    {0.00000, 0.54688, 1.00000},
    {0.00000, 0.53125, 1.00000},
    {0.00000, 0.51562, 1.00000},
    {0.00000, 0.50000, 1.00000},
    {0.00000, 0.48438, 1.00000},
    {0.00000, 0.46875, 1.00000},
    {0.00000, 0.45312, 1.00000},
    {0.00000, 0.43750, 1.00000},
    {0.00000, 0.42188, 1.00000},
    {0.00000, 0.40625, 1.00000},
    {0.00000, 0.39062, 1.00000},
    {0.00000, 0.37500, 1.00000},
    {0.00000, 0.35938, 1.00000},
    {0.00000, 0.34375, 1.00000},
    {0.00000, 0.32812, 1.00000},
    {0.00000, 0.31250, 1.00000},
    {0.00000, 0.29688, 1.00000},
    {0.00000, 0.28125, 1.00000},
    {0.00000, 0.26562, 1.00000},
    {0.00000, 0.25000, 1.00000},
    {0.00000, 0.23438, 1.00000},
    {0.00000, 0.21875, 1.00000},
    {0.00000, 0.20312, 1.00000},
    {0.00000, 0.18750, 1.00000},
    {0.00000, 0.17188, 1.00000},
    {0.00000, 0.15625, 1.00000},
    {0.00000, 0.14062, 1.00000},
    {0.00000, 0.12500, 1.00000},
    {0.00000, 0.10938, 1.00000},
    {0.00000, 0.09375, 1.00000},
    {0.00000, 0.07812, 1.00000},
    {0.00000, 0.06250, 1.00000},
    {0.00000, 0.04688, 1.00000},
    {0.00000, 0.03125, 1.00000},
    {0.00000, 0.01562, 1.00000},
    {0.00000, 0.00000, 1.00000} 
};		       

static Colour_struct mutant[LUT_SIZE] = 
{
    {0.00000, 1.00000, 0.00000},
    {0.01562, 1.00000, 0.00000},
    {0.03125, 1.00000, 0.00000},
    {0.04688, 1.00000, 0.00000},
    {0.06250, 1.00000, 0.00000},
    {0.07812, 1.00000, 0.00000},
    {0.09375, 1.00000, 0.00000},
    {0.10938, 1.00000, 0.00000},
    {0.12500, 1.00000, 0.00000},
    {0.14062, 1.00000, 0.00000},
    {0.15625, 1.00000, 0.00000},
    {0.17188, 1.00000, 0.00000},
    {0.18750, 1.00000, 0.00000},
    {0.20312, 1.00000, 0.00000},
    {0.21875, 1.00000, 0.00000},
    {0.23438, 1.00000, 0.00000},
    {0.25000, 1.00000, 0.00000},
    {0.26562, 1.00000, 0.00000},
    {0.28125, 1.00000, 0.00000},
    {0.29688, 1.00000, 0.00000},
    {0.31250, 1.00000, 0.00000},
    {0.32812, 1.00000, 0.00000},
    {0.34375, 1.00000, 0.00000},
    {0.35938, 1.00000, 0.00000},
    {0.37500, 1.00000, 0.00000},
    {0.39062, 1.00000, 0.00000},
    {0.40625, 1.00000, 0.00000},
    {0.42188, 1.00000, 0.00000},
    {0.43750, 1.00000, 0.00000},
    {0.45312, 1.00000, 0.00000},
    {0.46875, 1.00000, 0.00000},
    {0.48438, 1.00000, 0.00000},
    {0.50000, 1.00000, 0.00000},
    {0.51562, 1.00000, 0.00000},
    {0.53125, 1.00000, 0.00000},
    {0.54688, 1.00000, 0.00000},
    {0.56250, 1.00000, 0.00000},
    {0.57812, 1.00000, 0.00000},
    {0.59375, 1.00000, 0.00000},
    {0.60938, 1.00000, 0.00000},
    {0.62500, 1.00000, 0.00000},
    {0.64062, 1.00000, 0.00000},
    {0.65625, 1.00000, 0.00000},
    {0.67188, 1.00000, 0.00000},
    {0.68750, 1.00000, 0.00000},
    {0.70312, 1.00000, 0.00000},
    {0.71875, 1.00000, 0.00000},
    {0.73438, 1.00000, 0.00000},
    {0.75000, 1.00000, 0.00000},
    {0.76562, 1.00000, 0.00000},
    {0.78125, 1.00000, 0.00000},
    {0.79688, 1.00000, 0.00000},
    {0.81250, 1.00000, 0.00000},
    {0.82812, 1.00000, 0.00000},
    {0.84375, 1.00000, 0.00000},
    {0.85938, 1.00000, 0.00000},
    {0.87500, 1.00000, 0.00000},
    {0.89062, 1.00000, 0.00000},
    {0.90625, 1.00000, 0.00000},
    {0.92188, 1.00000, 0.00000},
    {0.93750, 1.00000, 0.00000},
    {0.95312, 1.00000, 0.00000},
    {0.96875, 1.00000, 0.00000},
    {0.98438, 1.00000, 0.00000},
    {1.00000, 1.00000, 0.01562},
    {1.00000, 1.00000, 0.03125},
    {1.00000, 1.00000, 0.04688},
    {1.00000, 1.00000, 0.06250},
    {1.00000, 1.00000, 0.07812},
    {1.00000, 1.00000, 0.09375},
    {1.00000, 1.00000, 0.10938},
    {1.00000, 1.00000, 0.12500},
    {1.00000, 1.00000, 0.14062},
    {1.00000, 1.00000, 0.15625},
    {1.00000, 1.00000, 0.17188},
    {1.00000, 1.00000, 0.18750},
    {1.00000, 1.00000, 0.20312},
    {1.00000, 1.00000, 0.21875},
    {1.00000, 1.00000, 0.23438},
    {1.00000, 1.00000, 0.25000},
    {1.00000, 1.00000, 0.26562},
    {1.00000, 1.00000, 0.28125},
    {1.00000, 1.00000, 0.29688},
    {1.00000, 1.00000, 0.31250},
    {1.00000, 1.00000, 0.32812},
    {1.00000, 1.00000, 0.34375},
    {1.00000, 1.00000, 0.35938},
    {1.00000, 1.00000, 0.37500},
    {1.00000, 1.00000, 0.39062},
    {1.00000, 1.00000, 0.40625},
    {1.00000, 1.00000, 0.42188},
    {1.00000, 1.00000, 0.43750},
    {1.00000, 1.00000, 0.45312},
    {1.00000, 1.00000, 0.46875},
    {1.00000, 1.00000, 0.48438},
    {1.00000, 1.00000, 0.50000},
    {1.00000, 1.00000, 0.51562},
    {1.00000, 1.00000, 0.53125},
    {1.00000, 1.00000, 0.54688},
    {1.00000, 1.00000, 0.56250},
    {1.00000, 1.00000, 0.57812},
    {1.00000, 1.00000, 0.59375},
    {1.00000, 1.00000, 0.60938},
    {1.00000, 1.00000, 0.62500},
    {1.00000, 1.00000, 0.64062},
    {1.00000, 1.00000, 0.65625},
    {1.00000, 1.00000, 0.67188},
    {1.00000, 1.00000, 0.68750},
    {1.00000, 1.00000, 0.70312},
    {1.00000, 1.00000, 0.71875},
    {1.00000, 1.00000, 0.73438},
    {1.00000, 1.00000, 0.75000},
    {1.00000, 1.00000, 0.76562},
    {1.00000, 1.00000, 0.78125},
    {1.00000, 1.00000, 0.79688},
    {1.00000, 1.00000, 0.81250},
    {1.00000, 1.00000, 0.82812},
    {1.00000, 1.00000, 0.84375},
    {1.00000, 1.00000, 0.85938},
    {1.00000, 1.00000, 0.87500},
    {1.00000, 1.00000, 0.89062},
    {1.00000, 1.00000, 0.90625},
    {1.00000, 1.00000, 0.92188},
    {1.00000, 1.00000, 0.93750},
    {1.00000, 1.00000, 0.95312},
    {1.00000, 1.00000, 0.96875},
    {1.00000, 1.00000, 0.98438},
    {1.00000, 1.00000, 1.00000},
    {1.00000, 0.98438, 1.00000},
    {1.00000, 0.96875, 1.00000},
    {1.00000, 0.95312, 1.00000},
    {1.00000, 0.93750, 1.00000},
    {1.00000, 0.92188, 1.00000},
    {1.00000, 0.90625, 1.00000},
    {1.00000, 0.89062, 1.00000},
    {1.00000, 0.87500, 1.00000},
    {1.00000, 0.85938, 1.00000},
    {1.00000, 0.84375, 1.00000},
    {1.00000, 0.82812, 1.00000},
    {1.00000, 0.81250, 1.00000},
    {1.00000, 0.79688, 1.00000},
    {1.00000, 0.78125, 1.00000},
    {1.00000, 0.76562, 1.00000},
    {1.00000, 0.75000, 1.00000},
    {1.00000, 0.73438, 1.00000},
    {1.00000, 0.71875, 1.00000},
    {1.00000, 0.70312, 1.00000},
    {1.00000, 0.68750, 1.00000},
    {1.00000, 0.67188, 1.00000},
    {1.00000, 0.65625, 1.00000},
    {1.00000, 0.64062, 1.00000},
    {1.00000, 0.62500, 1.00000},
    {1.00000, 0.60938, 1.00000},
    {1.00000, 0.59375, 1.00000},
    {1.00000, 0.57812, 1.00000},
    {1.00000, 0.56250, 1.00000},
    {1.00000, 0.54688, 1.00000},
    {1.00000, 0.53125, 1.00000},
    {1.00000, 0.51562, 1.00000},
    {1.00000, 0.50000, 1.00000},
    {1.00000, 0.48438, 1.00000},
    {1.00000, 0.46875, 1.00000},
    {1.00000, 0.45312, 1.00000},
    {1.00000, 0.43750, 1.00000},
    {1.00000, 0.42188, 1.00000},
    {1.00000, 0.40625, 1.00000},
    {1.00000, 0.39062, 1.00000},
    {1.00000, 0.37500, 1.00000},
    {1.00000, 0.35938, 1.00000},
    {1.00000, 0.34375, 1.00000},
    {1.00000, 0.32812, 1.00000},
    {1.00000, 0.31250, 1.00000},
    {1.00000, 0.29688, 1.00000},
    {1.00000, 0.28125, 1.00000},
    {1.00000, 0.26562, 1.00000},
    {1.00000, 0.25000, 1.00000},
    {1.00000, 0.23438, 1.00000},
    {1.00000, 0.21875, 1.00000},
    {1.00000, 0.20312, 1.00000},
    {1.00000, 0.18750, 1.00000},
    {1.00000, 0.17188, 1.00000},
    {1.00000, 0.15625, 1.00000},
    {1.00000, 0.14062, 1.00000},
    {1.00000, 0.12500, 1.00000},
    {1.00000, 0.10938, 1.00000},
    {1.00000, 0.09375, 1.00000},
    {1.00000, 0.07812, 1.00000},
    {1.00000, 0.06250, 1.00000},
    {1.00000, 0.04688, 1.00000},
    {1.00000, 0.03125, 1.00000},
    {1.00000, 0.01562, 1.00000},
    {1.00000, 0.00000, 1.00000},
    {0.98438, 0.00000, 1.00000},
    {0.96875, 0.00000, 1.00000},
    {0.95312, 0.00000, 1.00000},
    {0.93750, 0.00000, 1.00000},
    {0.92188, 0.00000, 1.00000},
    {0.90625, 0.00000, 1.00000},
    {0.89062, 0.00000, 1.00000},
    {0.87500, 0.00000, 1.00000},
    {0.85938, 0.00000, 1.00000},
    {0.84375, 0.00000, 1.00000},
    {0.82812, 0.00000, 1.00000},
    {0.81250, 0.00000, 1.00000},
    {0.79688, 0.00000, 1.00000},
    {0.78125, 0.00000, 1.00000},
    {0.76562, 0.00000, 1.00000},
    {0.75000, 0.00000, 1.00000},
    {0.73438, 0.00000, 1.00000},
    {0.71875, 0.00000, 1.00000},
    {0.70312, 0.00000, 1.00000},
    {0.68750, 0.00000, 1.00000},
    {0.67188, 0.00000, 1.00000},
    {0.65625, 0.00000, 1.00000},
    {0.64062, 0.00000, 1.00000},
    {0.62500, 0.00000, 1.00000},
    {0.60938, 0.00000, 1.00000},
    {0.59375, 0.00000, 1.00000},
    {0.57812, 0.00000, 1.00000},
    {0.56250, 0.00000, 1.00000},
    {0.54688, 0.00000, 1.00000},
    {0.53125, 0.00000, 1.00000},
    {0.51562, 0.00000, 1.00000},
    {0.50000, 0.00000, 1.00000},
    {0.48438, 0.00000, 1.00000},
    {0.46875, 0.00000, 1.00000},
    {0.45312, 0.00000, 1.00000},
    {0.43750, 0.00000, 1.00000},
    {0.42188, 0.00000, 1.00000},
    {0.40625, 0.00000, 1.00000},
    {0.39062, 0.00000, 1.00000},
    {0.37500, 0.00000, 1.00000},
    {0.35938, 0.00000, 1.00000},
    {0.34375, 0.00000, 1.00000},
    {0.32812, 0.00000, 1.00000},
    {0.31250, 0.00000, 1.00000},
    {0.29688, 0.00000, 1.00000},
    {0.28125, 0.00000, 1.00000},
    {0.26562, 0.00000, 1.00000},
    {0.25000, 0.00000, 1.00000},
    {0.23438, 0.00000, 1.00000},
    {0.21875, 0.00000, 1.00000},
    {0.20312, 0.00000, 1.00000},
    {0.18750, 0.00000, 1.00000},
    {0.17188, 0.00000, 1.00000},
    {0.15625, 0.00000, 1.00000},
    {0.14062, 0.00000, 1.00000},
    {0.12500, 0.00000, 1.00000},
    {0.10938, 0.00000, 1.00000},
    {0.09375, 0.00000, 1.00000},
    {0.07812, 0.00000, 1.00000},
    {0.06250, 0.00000, 1.00000},
    {0.04688, 0.00000, 1.00000},
    {0.03125, 0.00000, 1.00000},
    {0.01562, 0.00000, 1.00000},
    {0.00000, 0.00000, 1.00000}
};

static Colour_struct aberration[LUT_SIZE] = 
{
    {1.00000, 0.00000, 0.00000},
    {1.00000, 0.00000, 0.01562},
    {1.00000, 0.00000, 0.03125},
    {1.00000, 0.00000, 0.04688},
    {1.00000, 0.00000, 0.06250},
    {1.00000, 0.00000, 0.07812},
    {1.00000, 0.00000, 0.09375},
    {1.00000, 0.00000, 0.10938},
    {1.00000, 0.00000, 0.12500},
    {1.00000, 0.00000, 0.14062},
    {1.00000, 0.00000, 0.15625},
    {1.00000, 0.00000, 0.17188},
    {1.00000, 0.00000, 0.18750},
    {1.00000, 0.00000, 0.20312},
    {1.00000, 0.00000, 0.21875},
    {1.00000, 0.00000, 0.23438},
    {1.00000, 0.00000, 0.25000},
    {1.00000, 0.00000, 0.26562},
    {1.00000, 0.00000, 0.28125},
    {1.00000, 0.00000, 0.29688},
    {1.00000, 0.00000, 0.31250},
    {1.00000, 0.00000, 0.32812},
    {1.00000, 0.00000, 0.34375},
    {1.00000, 0.00000, 0.35938},
    {1.00000, 0.00000, 0.37500},
    {1.00000, 0.00000, 0.39062},
    {1.00000, 0.00000, 0.40625},
    {1.00000, 0.00000, 0.42188},
    {1.00000, 0.00000, 0.43750},
    {1.00000, 0.00000, 0.45312},
    {1.00000, 0.00000, 0.46875},
    {1.00000, 0.00000, 0.48438},
    {1.00000, 0.00000, 0.50000},
    {1.00000, 0.00000, 0.51562},
    {1.00000, 0.00000, 0.53125},
    {1.00000, 0.00000, 0.54688},
    {1.00000, 0.00000, 0.56250},
    {1.00000, 0.00000, 0.57812},
    {1.00000, 0.00000, 0.59375},
    {1.00000, 0.00000, 0.60938},
    {1.00000, 0.00000, 0.62500},
    {1.00000, 0.00000, 0.64062},
    {1.00000, 0.00000, 0.65625},
    {1.00000, 0.00000, 0.67188},
    {1.00000, 0.00000, 0.68750},
    {1.00000, 0.00000, 0.70312},
    {1.00000, 0.00000, 0.71875},
    {1.00000, 0.00000, 0.73438},
    {1.00000, 0.00000, 0.75000},
    {1.00000, 0.00000, 0.76562},
    {1.00000, 0.00000, 0.78125},
    {1.00000, 0.00000, 0.79688},
    {1.00000, 0.00000, 0.81250},
    {1.00000, 0.00000, 0.82812},
    {1.00000, 0.00000, 0.84375},
    {1.00000, 0.00000, 0.85938},
    {1.00000, 0.00000, 0.87500},
    {1.00000, 0.00000, 0.89062},
    {1.00000, 0.00000, 0.90625},
    {1.00000, 0.00000, 0.92188},
    {1.00000, 0.00000, 0.93750},
    {1.00000, 0.00000, 0.95312},
    {1.00000, 0.00000, 0.96875},
    {1.00000, 0.00000, 0.98438},
    {1.00000, 0.01562, 1.00000},
    {1.00000, 0.03125, 1.00000},
    {1.00000, 0.04688, 1.00000},
    {1.00000, 0.06250, 1.00000},
    {1.00000, 0.07812, 1.00000},
    {1.00000, 0.09375, 1.00000},
    {1.00000, 0.10938, 1.00000},
    {1.00000, 0.12500, 1.00000},
    {1.00000, 0.14062, 1.00000},
    {1.00000, 0.15625, 1.00000},
    {1.00000, 0.17188, 1.00000},
    {1.00000, 0.18750, 1.00000},
    {1.00000, 0.20312, 1.00000},
    {1.00000, 0.21875, 1.00000},
    {1.00000, 0.23438, 1.00000},
    {1.00000, 0.25000, 1.00000},
    {1.00000, 0.26562, 1.00000},
    {1.00000, 0.28125, 1.00000},
    {1.00000, 0.29688, 1.00000},
    {1.00000, 0.31250, 1.00000},
    {1.00000, 0.32812, 1.00000},
    {1.00000, 0.34375, 1.00000},
    {1.00000, 0.35938, 1.00000},
    {1.00000, 0.37500, 1.00000},
    {1.00000, 0.39062, 1.00000},
    {1.00000, 0.40625, 1.00000},
    {1.00000, 0.42188, 1.00000},
    {1.00000, 0.43750, 1.00000},
    {1.00000, 0.45312, 1.00000},
    {1.00000, 0.46875, 1.00000},
    {1.00000, 0.48438, 1.00000},
    {1.00000, 0.50000, 1.00000},
    {1.00000, 0.51562, 1.00000},
    {1.00000, 0.53125, 1.00000},
    {1.00000, 0.54688, 1.00000},
    {1.00000, 0.56250, 1.00000},
    {1.00000, 0.57812, 1.00000},
    {1.00000, 0.59375, 1.00000},
    {1.00000, 0.60938, 1.00000},
    {1.00000, 0.62500, 1.00000},
    {1.00000, 0.64062, 1.00000},
    {1.00000, 0.65625, 1.00000},
    {1.00000, 0.67188, 1.00000},
    {1.00000, 0.68750, 1.00000},
    {1.00000, 0.70312, 1.00000},
    {1.00000, 0.71875, 1.00000},
    {1.00000, 0.73438, 1.00000},
    {1.00000, 0.75000, 1.00000},
    {1.00000, 0.76562, 1.00000},
    {1.00000, 0.78125, 1.00000},
    {1.00000, 0.79688, 1.00000},
    {1.00000, 0.81250, 1.00000},
    {1.00000, 0.82812, 1.00000},
    {1.00000, 0.84375, 1.00000},
    {1.00000, 0.85938, 1.00000},
    {1.00000, 0.87500, 1.00000},
    {1.00000, 0.89062, 1.00000},
    {1.00000, 0.90625, 1.00000},
    {1.00000, 0.92188, 1.00000},
    {1.00000, 0.93750, 1.00000},
    {1.00000, 0.95312, 1.00000},
    {1.00000, 0.96875, 1.00000},
    {1.00000, 0.98438, 1.00000},
    {1.00000, 1.00000, 1.00000},
    {0.98438, 1.00000, 1.00000},
    {0.96875, 1.00000, 1.00000},
    {0.95312, 1.00000, 1.00000},
    {0.93750, 1.00000, 1.00000},
    {0.92188, 1.00000, 1.00000},
    {0.90625, 1.00000, 1.00000},
    {0.89062, 1.00000, 1.00000},
    {0.87500, 1.00000, 1.00000},
    {0.85938, 1.00000, 1.00000},
    {0.84375, 1.00000, 1.00000},
    {0.82812, 1.00000, 1.00000},
    {0.81250, 1.00000, 1.00000},
    {0.79688, 1.00000, 1.00000},
    {0.78125, 1.00000, 1.00000},
    {0.76562, 1.00000, 1.00000},
    {0.75000, 1.00000, 1.00000},
    {0.73438, 1.00000, 1.00000},
    {0.71875, 1.00000, 1.00000},
    {0.70312, 1.00000, 1.00000},
    {0.68750, 1.00000, 1.00000},
    {0.67188, 1.00000, 1.00000},
    {0.65625, 1.00000, 1.00000},
    {0.64062, 1.00000, 1.00000},
    {0.62500, 1.00000, 1.00000},
    {0.60938, 1.00000, 1.00000},
    {0.59375, 1.00000, 1.00000},
    {0.57812, 1.00000, 1.00000},
    {0.56250, 1.00000, 1.00000},
    {0.54688, 1.00000, 1.00000},
    {0.53125, 1.00000, 1.00000},
    {0.51562, 1.00000, 1.00000},
    {0.50000, 1.00000, 1.00000},
    {0.48438, 1.00000, 1.00000},
    {0.46875, 1.00000, 1.00000},
    {0.45312, 1.00000, 1.00000},
    {0.43750, 1.00000, 1.00000},
    {0.42188, 1.00000, 1.00000},
    {0.40625, 1.00000, 1.00000},
    {0.39062, 1.00000, 1.00000},
    {0.37500, 1.00000, 1.00000},
    {0.35938, 1.00000, 1.00000},
    {0.34375, 1.00000, 1.00000},
    {0.32812, 1.00000, 1.00000},
    {0.31250, 1.00000, 1.00000},
    {0.29688, 1.00000, 1.00000},
    {0.28125, 1.00000, 1.00000},
    {0.26562, 1.00000, 1.00000},
    {0.25000, 1.00000, 1.00000},
    {0.23438, 1.00000, 1.00000},
    {0.21875, 1.00000, 1.00000},
    {0.20312, 1.00000, 1.00000},
    {0.18750, 1.00000, 1.00000},
    {0.17188, 1.00000, 1.00000},
    {0.15625, 1.00000, 1.00000},
    {0.14062, 1.00000, 1.00000},
    {0.12500, 1.00000, 1.00000},
    {0.10938, 1.00000, 1.00000},
    {0.09375, 1.00000, 1.00000},
    {0.07812, 1.00000, 1.00000},
    {0.06250, 1.00000, 1.00000},
    {0.04688, 1.00000, 1.00000},
    {0.03125, 1.00000, 1.00000},
    {0.01562, 1.00000, 1.00000},
    {0.00000, 1.00000, 1.00000},
    {0.00000, 1.00000, 0.98438},
    {0.00000, 1.00000, 0.96875},
    {0.00000, 1.00000, 0.95312},
    {0.00000, 1.00000, 0.93750},
    {0.00000, 1.00000, 0.92188},
    {0.00000, 1.00000, 0.90625},
    {0.00000, 1.00000, 0.89062},
    {0.00000, 1.00000, 0.87500},
    {0.00000, 1.00000, 0.85938},
    {0.00000, 1.00000, 0.84375},
    {0.00000, 1.00000, 0.82812},
    {0.00000, 1.00000, 0.81250},
    {0.00000, 1.00000, 0.79688},
    {0.00000, 1.00000, 0.78125},
    {0.00000, 1.00000, 0.76562},
    {0.00000, 1.00000, 0.75000},
    {0.00000, 1.00000, 0.73438},
    {0.00000, 1.00000, 0.71875},
    {0.00000, 1.00000, 0.70312},
    {0.00000, 1.00000, 0.68750},
    {0.00000, 1.00000, 0.67188},
    {0.00000, 1.00000, 0.65625},
    {0.00000, 1.00000, 0.64062},
    {0.00000, 1.00000, 0.62500},
    {0.00000, 1.00000, 0.60938},
    {0.00000, 1.00000, 0.59375},
    {0.00000, 1.00000, 0.57812},
    {0.00000, 1.00000, 0.56250},
    {0.00000, 1.00000, 0.54688},
    {0.00000, 1.00000, 0.53125},
    {0.00000, 1.00000, 0.51562},
    {0.00000, 1.00000, 0.50000},
    {0.00000, 1.00000, 0.48438},
    {0.00000, 1.00000, 0.46875},
    {0.00000, 1.00000, 0.45312},
    {0.00000, 1.00000, 0.43750},
    {0.00000, 1.00000, 0.42188},
    {0.00000, 1.00000, 0.40625},
    {0.00000, 1.00000, 0.39062},
    {0.00000, 1.00000, 0.37500},
    {0.00000, 1.00000, 0.35938},
    {0.00000, 1.00000, 0.34375},
    {0.00000, 1.00000, 0.32812},
    {0.00000, 1.00000, 0.31250},
    {0.00000, 1.00000, 0.29688},
    {0.00000, 1.00000, 0.28125},
    {0.00000, 1.00000, 0.26562},
    {0.00000, 1.00000, 0.25000},
    {0.00000, 1.00000, 0.23438},
    {0.00000, 1.00000, 0.21875},
    {0.00000, 1.00000, 0.20312},
    {0.00000, 1.00000, 0.18750},
    {0.00000, 1.00000, 0.17188},
    {0.00000, 1.00000, 0.15625},
    {0.00000, 1.00000, 0.14062},
    {0.00000, 1.00000, 0.12500},
    {0.00000, 1.00000, 0.10938},
    {0.00000, 1.00000, 0.09375},
    {0.00000, 1.00000, 0.07812},
    {0.00000, 1.00000, 0.06250},
    {0.00000, 1.00000, 0.04688},
    {0.00000, 1.00000, 0.03125},
    {0.00000, 1.00000, 0.01562},
    {0.00000, 1.00000, 0.00000} 
};

/*EXPERIMENTAL_FUNCTION*/
void cf_sunbow (unsigned int num_cells, unsigned short *reds,
		unsigned short *greens, unsigned short *blues,
		unsigned int stride, double x, double y, void *var_param)
/*  [SUMMARY] Compute a "Sunbow" colourmap.
    <num_cells> The number of colour cells to modify.
    <reds> The red intensity values.
    <greens> The green intensity values.
    <blues> The blue intensity values.
    <stride> The stride (in unsigned shorts) between intensity values.
    <x> A parameter used to compute the colour values, ranging from 0.0 to 1.0.
    <y> A parameter used to compute the colour values, ranging from 0.0 to 1.0.
    <var_param> A parameter used to compute the colour values. Ignored.
    [RETURNS] Nothing.
*/
{
    cf_colour_table (num_cells, reds, greens, blues, stride, x, y, sunbow,
		     LUT_SIZE);
}   /*  End Function cf_sunbow  */

/*EXPERIMENTAL_FUNCTION*/
void cf_mutant (unsigned int num_cells, unsigned short *reds,
		unsigned short *greens, unsigned short *blues,
		unsigned int stride, double x, double y, void *var_param)
/*  [SUMMARY] Compute a "Mutant" colourmap.
    <num_cells> The number of colour cells to modify.
    <reds> The red intensity values.
    <greens> The green intensity values.
    <blues> The blue intensity values.
    <stride> The stride (in unsigned shorts) between intensity values.
    <x> A parameter used to compute the colour values, ranging from 0.0 to 1.0.
    <y> A parameter used to compute the colour values, ranging from 0.0 to 1.0.
    <var_param> A parameter used to compute the colour values. Ignored.
    [RETURNS] Nothing.
*/
{
    cf_colour_table (num_cells, reds, greens, blues, stride, x, y, mutant,
		     LUT_SIZE);
}   /*  End Function cf_mutant  */

/*EXPERIMENTAL_FUNCTION*/
void cf_aberration (unsigned int num_cells, unsigned short *reds,
		    unsigned short *greens, unsigned short *blues,
		    unsigned int stride, double x, double y, void *var_param)
/*  [SUMMARY] Compute an "Aberration" colourmap.
    <num_cells> The number of colour cells to modify.
    <reds> The red intensity values.
    <greens> The green intensity values.
    <blues> The blue intensity values.
    <stride> The stride (in unsigned shorts) between intensity values.
    <x> A parameter used to compute the colour values, ranging from 0.0 to 1.0.
    <y> A parameter used to compute the colour values, ranging from 0.0 to 1.0.
    <var_param> A parameter used to compute the colour values. Ignored.
    [RETURNS] Nothing.
*/
{
    cf_colour_table (num_cells, reds, greens, blues, stride, x, y, aberration,
		     LUT_SIZE);
}   /*  End Function cf_aberration  */
