/*LINTLIBRARY*/
/*  gildas_read.c

    This code provides a GILDAS read facility.

    Copyright (C) 2004  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for reading data files in
  the GILDAS format.


    Written by      Richard Gooch   2-JUL-2004

    Updated by      Richard Gooch   6-JUL-2004: Mostly working, waiting for
  information about different projection type encodings.

    Last updated by Richard Gooch   12-JUL-2004: Scrub unknown co-ordinates
  until I get more information.

    $Id: gildas_read.c,v 2.1 2011/04/21 04:21:11 cal103 Exp $

*/
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_dsrw.h>
#include <karma_pio.h>
#include <karma_chs.h>
#include <karma_ch.h>
#include <karma_ds.h>
#include <karma_m.h>
#include <karma_p.h>
#include <karma_r.h>
#include <karma_a.h>

#define MAGIC_NUMBER   1980723098
#define HEADER_BYTES          512
#define FORMAT_BLOCK         0x00  /*  Can read in raw block                 */
#define FORMAT_HARD          0x01  /*  Use <p> package for reading           */
#define FORMAT_SWAP          0x02  /*  Need byte swapping                    */
#define GET_VALUE(out, offset, type) \
    get_value ((out), header + ( ( (offset)-1 ) << 2 ), (type), format)

#define VERIFY_CONTEXT(ctx) if (ctx == NULL) \
{fprintf (stderr, "NULL context passed\n"); \
 a_prog_bug (function_name); } \
if (ctx->magic_number != MAGIC_NUMBER) \
{fprintf (stderr, "Invalid context object\n"); \
 a_prog_bug (function_name); }


/*  Private structures  */
struct gildas_data_context_type
{
    unsigned int magic_number;
    Channel channel;
    unsigned int format;
    array_desc *arr_desc;
    unsigned int packet_size;
    unsigned int file_element_size;
    int blank_value;
    unsigned long blank_count;
};

struct gildas_header
{
    int     form;
    int     nvb;
    int     gene;
    int     ndim;
    int     dim[4];
    double  ref[4];
    double  val[4];
    double  inc[4];
    int     blan;
    float   bval;
    float   eval;
    float   rmin;
    float   rmax;
    char    data_name[13];
    char    axis_names[4][13];
    int     posi;
    char    isou[13];
    double  ra;
    double  dec;
    double  lii;
    double  bii;
    float   epoc;
    int     proj;
    int     ptyp;
    double  pang;
    float   majo;
    float   mino;
    float   posa;
};


/*  Private functions  */
STATIC_FUNCTION (void strip_trailing_whitespace, (char *string) );
STATIC_FUNCTION (void get_value,
		 (void *dest, CONST void *source, unsigned int type,
		  unsigned int format) );


/*  Private data  */


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
multi_array *foreign_gildas_read (CONST char *setname,unsigned int mmap_option,
				  flag writable)
/*  [SUMMARY] Read a GILDAS dataset.
    <setname> The GILDAS dataset name.
    <mmap_option> Control value which determines whether the channel is opened
    as an ordinary disc file or is mapped. See [<ch_MAP_CONTROLS>] for legal
    values. If the file is not mapped then the routine will attempt to open an
    ordinary disc channel. If the file is opened as a disc channel the access
    mode is: "rz".
    <writable> If the mapped pages are to be writable, this must be TRUE. If
    this is FALSE and the memory pages are written to, a segmentation fault
    occurs.
    [RETURNS] A multi_array descriptor on success, else NULL.
*/
{
    KGILDASDataContext context;
    uaddr arr_size;
    array_pointer arrayp;
    multi_array *multi_desc;
    static char function_name[] = "foreign_gildas_read";

    if ( ( multi_desc = foreign_gildas_read_header (setname, TRUE, mmap_option,
						    writable, &context) )
	 == NULL ) return (NULL);
    /*  Now read data  */
    arr_size = ds_get_array_size (context->arr_desc);
    m_copy (&arrayp, multi_desc->data[0], sizeof arrayp);
    if (arrayp.array)
    {
	if ( !foreign_gildas_read_data (context, arrayp.array, arr_size) )
	{
	    ds_dealloc_multi (multi_desc);
	    foreign_gildas_close_data_context (context);
	    return (NULL);
	}
    }
    else
    {
	uaddr array_bytes = arr_size * context->packet_size;
	uaddr rpos, wpos;

	ch_tell (context->channel, &rpos, &wpos);
	arrayp.array = ch_get_mmap_addr (context->channel) + rpos;
	arrayp.free = ( void (*) () ) ch_close;
	arrayp.info = context->channel;
	arrayp.mem_type = K_ARRAY_MEM_TYPE_MMAP;
	arrayp.u.mmap.fd = ch_get_descriptor (context->channel);
	arrayp.u.mmap.size = ch_get_mem_size (context->channel);
	arrayp.u.mmap.base = ch_get_mmap_addr (context->channel);
	m_copy (multi_desc->data[0], &arrayp, sizeof arrayp);
	/*  Need to skip over the data so that the read pointer is updated  */
	if ( !ch_skip (context->channel, array_bytes) )
	{
	    fprintf (stderr, "%s: error skipping: %lu bytes\t%s\n",
		     function_name, array_bytes, ERRSTRING);
	    ds_dealloc_multi (multi_desc);
	    foreign_gildas_close_data_context (context);
	    return (NULL);
	}
	context->channel = NULL;
    }
    if (context->blank_count > arr_size >> 4)
    {
	fprintf (stderr,
		 "WARNING: %lu out of %lu values are blank: bad data?\n",
		 context->blank_count, arr_size);
    }
    foreign_gildas_close_data_context (context);
    return (multi_desc);
}   /*  End Function foreign_gildas_read  */

/*EXPERIMENTAL_FUNCTION*/
multi_array *foreign_gildas_read_header (CONST char *setname, flag data_alloc,
					 unsigned int mmap_option,
					 flag writable,
					 KGILDASDataContext *context)
/*  [SUMMARY] Read a GILDAS dataset header.
    <setname> The GILDAS dataset name.
    <data_alloc> If TRUE, the data space is allocated.
    <mmap_option> Control value which determines whether the channel is opened
    as an ordinary disc file or is mapped. See [<ch_MAP_CONTROLS>] for legal
    values. If the file is not mapped then the routine will attempt to open an
    ordinary disc channel. If the file is opened as a disc channel the access
    mode is: "rz".
    <writable> If the mapped pages are to be writable, this must be TRUE. If
    this is FALSE and the memory pages are written to, a segmentation fault
    occurs.
    <context> The data reading context is written here, to be used in
    subsequent calls to [<foreign_gildas_read_data>]. This may be NULL.
    [RETURNS] A multi_array descriptor on success, else NULL.
*/
{
    Channel channel;
    unsigned int format, num_axes, axis_index, count, elem_type;
    struct gildas_header gheader;
    array_pointer arrayp;
    array_desc *arr_desc;
    multi_array *multi_desc;
    char *ptr;
    CONST char *header;
    CONST char *dim_names[4];
    char header_buf[HEADER_BYTES];
    uaddr axis_lengths[4];
    double value[2];
    extern char network_type_bytes[NUMTYPES];
    static char function_name[] = "foreign_gildas_read_header";

    if ( ds_can_transfer_element_as_block (K_FLOAT) ) format = FORMAT_BLOCK;
    else if ( ds_can_swaptransfer_element (K_FLOAT) )
	format = FORMAT_BLOCK | FORMAT_SWAP;
    else format = FORMAT_HARD;
    mmap_option |= K_CH_NOMAP_COMPRESSED;
    if ( ( channel = ch_map_disc (setname, mmap_option, writable, TRUE) )
	 == NULL )
    {
	fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		 function_name, setname, ERRSTRING);
	return (NULL);
    }
    /*  Read the header  */
    if (ch_read2 (channel, header_buf, &header, 1, HEADER_BYTES)
	< HEADER_BYTES)
    {
	fprintf (stderr, "%s: error reading header\t%s\n",
		 function_name, ERRSTRING);
	ch_close (channel);
	return (NULL);
    }
    if (strncmp (header, "GILDAS", 6) != 0)
    {
	fprintf (stderr, "%s: file: \"%s\" is not a GILDAS file\n",
		 function_name, setname);
	ch_close (channel);
	return (NULL);
    }
    switch (header[6])
    {
      case '_':  /*  VAX format          */
	fprintf (stderr, "%s: file: \"%s\" is in VAX format: not supported\n",
		 function_name, setname);
	ch_close (channel);
	return (NULL);
	/*break;*/
      case '.':  /*  IEEE big-endian     */
	break;
      case '-':  /*  IEEE little-endian  */
	format = (format & ~FORMAT_SWAP)
	    | (format & FORMAT_SWAP) ? 0 : FORMAT_SWAP;
	break;
      default:
	fprintf (stderr, "%s: file: \"%s\" has unknown data format\n",
		 function_name, setname);
	ch_close (channel);
	return (NULL);
	break;
    }
    if (strncmp (header + 7, "IMAGE", 5) != 0)
    {
	fprintf (stderr, "%s: file: \"%s\" is not a GILDAS image\n",
		 function_name, setname);
	ch_close (channel);
	return (NULL);
    }
    GET_VALUE (&gheader.form,    4, K_INT);
    GET_VALUE (&gheader.nvb,     5, K_INT);
    GET_VALUE (&gheader.gene,   11, K_INT);
    GET_VALUE (&gheader.ndim,   12, K_INT);
    GET_VALUE (&gheader.dim[0], 13, K_INT);
    GET_VALUE (&gheader.dim[1], 14, K_INT);
    GET_VALUE (&gheader.dim[2], 15, K_INT);
    GET_VALUE (&gheader.dim[3], 16, K_INT);
    GET_VALUE (&gheader.ref[0], 17, K_DOUBLE);
    GET_VALUE (&gheader.val[0], 19, K_DOUBLE);
    GET_VALUE (&gheader.inc[0], 21, K_DOUBLE);
    GET_VALUE (&gheader.ref[1], 23, K_DOUBLE);
    GET_VALUE (&gheader.val[1], 25, K_DOUBLE);
    GET_VALUE (&gheader.inc[1], 27, K_DOUBLE);
    GET_VALUE (&gheader.ref[2], 29, K_DOUBLE);
    GET_VALUE (&gheader.val[2], 31, K_DOUBLE);
    GET_VALUE (&gheader.inc[2], 33, K_DOUBLE);
    GET_VALUE (&gheader.ref[3], 35, K_DOUBLE);
    GET_VALUE (&gheader.val[3], 37, K_DOUBLE);
    GET_VALUE (&gheader.inc[3], 39, K_DOUBLE);
    GET_VALUE (&gheader.blan,   41, K_INT);
    GET_VALUE (&gheader.bval,   42, K_FLOAT);
    GET_VALUE (&gheader.eval,   43, K_FLOAT);
    GET_VALUE (&gheader.rmin,   45, K_FLOAT);
    GET_VALUE (&gheader.rmax,   46, K_FLOAT);
    m_copy (&gheader.data_name, header + ( (56 - 1) << 2 ), 12);
    gheader.data_name[12] = '\0';
    strip_trailing_whitespace (gheader.data_name);
    m_copy (gheader.axis_names[0], header + ( (59 - 1) << 2 ), 12);
    gheader.axis_names[0][12] = '\0';
    strip_trailing_whitespace (gheader.axis_names[0]);
    m_copy (gheader.axis_names[1], header + ( (62 - 1) << 2 ), 12);
    gheader.axis_names[1][12] = '\0';
    strip_trailing_whitespace (gheader.axis_names[1]);
    m_copy (gheader.axis_names[2], header + ( (65 - 1) << 2 ), 12);
    gheader.axis_names[2][12] = '\0';
    strip_trailing_whitespace (gheader.axis_names[2]);
    m_copy (gheader.axis_names[3], header + ( (68 - 1) << 2 ), 12);
    gheader.axis_names[3][12] = '\0';
    strip_trailing_whitespace (gheader.axis_names[3]);
    GET_VALUE (&gheader.posi,   74, K_FLOAT);
    m_copy (gheader.isou, header + ( (75 - 1) << 2 ), 12);
    gheader.isou[12] = '\0';
    strip_trailing_whitespace (gheader.isou);
    GET_VALUE (&gheader.ra,     78, K_DOUBLE);
    GET_VALUE (&gheader.dec,    80, K_DOUBLE);
    GET_VALUE (&gheader.lii,    82, K_DOUBLE);
    GET_VALUE (&gheader.bii,    84, K_DOUBLE);
    GET_VALUE (&gheader.epoc,   86, K_FLOAT);
    GET_VALUE (&gheader.proj,   87, K_INT);
    GET_VALUE (&gheader.ptyp,   88, K_INT);
    GET_VALUE (&gheader.pang,   93, K_DOUBLE);
    GET_VALUE (&gheader.majo,  111, K_FLOAT);
    GET_VALUE (&gheader.mino,  112, K_FLOAT);
    GET_VALUE (&gheader.posa,  113, K_FLOAT);
    fprintf (stderr, "form: %d  nvb: %d  gene: %d\n",
	     gheader.form, gheader.nvb, gheader.gene);
    fprintf (stderr, "blan: %d  gval: %g  eval: %g\n",
	     gheader.blan, gheader.bval, gheader.eval);
    fprintf (stderr, "posi: %d\n", gheader.posi);
    fprintf (stderr, "proj: %d  ptyp: %d  pang: %g\n",
	     gheader.proj, gheader.ptyp, gheader.pang);
    /*  First count number of non-degenerate axes  */
    for (count = 0, num_axes = 0; count < gheader.ndim; ++count)
    {
	if (gheader.dim[count] < 2) continue;
	++num_axes;
    }
    /*  Now construct dimension lengths and names  */
    for (count = 0, axis_index = num_axes - 1; count < gheader.ndim; ++count)
    {
	char *name = gheader.axis_names[count];

	/*  Scrub unknown co-ordinates until I get more information  */
	if (gheader.ptyp != 7) name[0] = '\0';
	if (name[0] == '\0') sprintf (name, "naxis%u", count + 1);
	else if ( (gheader.ptyp == 7) && (strcmp (name, "RA") == 0) )
	    strcat (name, "---GLS");
	else if ( (gheader.ptyp == 7) && (strcmp (name, "DEC") == 0) )
	    strcat (name, "--GLS");
	if (gheader.dim[count] < 2) continue;
	axis_lengths[axis_index] = gheader.dim[count];
	dim_names[axis_index] = name;
	--axis_index;
    }
    elem_type = K_FLOAT;
    ptr = gheader.data_name;
    arr_desc = ds_easy_alloc_array_desc (num_axes, axis_lengths,
					 NULL, NULL, NULL, dim_names,
					 1, &elem_type,
					 (CONST char **) &ptr);
    if (arr_desc == NULL)
    {
	ch_close (channel);
	return (NULL);
    }
    if ( (format == FORMAT_BLOCK) && ch_test_for_mmap (channel) )
	data_alloc = FALSE;
    if (data_alloc)
    {
	if ( !ds_alloc_vm_array (&arrayp, arr_desc, FALSE, FALSE) &&
	     !ds_alloc_mmap_array (-1, 0, 0, TRUE, &arrayp, arr_desc,
				   FALSE, FALSE) )
	{
	    m_error_notify (function_name, "array data");
	    ds_dealloc_array_desc (arr_desc);
	    ch_close (channel);
	    return (NULL);
	}
    }
    else m_clear (&arrayp, sizeof arrayp);
    if ( ( multi_desc =
	   ds_easy_alloc_array_from_array_desc (arr_desc, &arrayp, TRUE) )
	 == NULL )
    {
	ds_dealloc_array_desc (arr_desc);
	if (arrayp.free) (*arrayp.free) (arrayp.info, &arrayp);
	ch_close (channel);
	return (NULL);
    }
    for (count = 0; count < gheader.ndim; ++count)
    {
	double crpix = gheader.ref[count];
	double crval = gheader.val[count];
	double cdelt = gheader.inc[count];
	char txt[16];

	if (gheader.dim[count] < 1) continue;
	if (strncmp (gheader.axis_names[count], "RA", 2) == 0)
	{
	    crpix -= crval / cdelt;
	    crval = gheader.ra / PION180;
	    cdelt = cdelt / PION180;
	}
	else if (strncmp (gheader.axis_names[count], "DEC", 3) == 0)
	{
	    crpix -= crval / cdelt;
	    crval = gheader.dec / PION180;
	    cdelt = cdelt / PION180;
	}
	else if (strcmp (gheader.axis_names[count], "VELOCITY") == 0)
	{
	    crval *= 1e3;
	    cdelt *= 1e3;
	}
	sprintf (txt, "CTYPE%u", count + 1);
	ds_put_unique_named_string (multi_desc->headers[0], multi_desc->data,
				    txt, gheader.axis_names[count], FALSE);
	sprintf (txt, "CRVAL%u", count + 1);
	ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
				   txt, K_DOUBLE, &crval, FALSE);
	sprintf (txt, "CRPIX%u", count + 1);
	ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
				   txt, K_DOUBLE, &crpix, FALSE);
	sprintf (txt, "CDELT%u", count + 1);
	ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
				   txt, K_DOUBLE, &cdelt, FALSE);
    }
    value[1] = 0.0;
    if (gheader.epoc != 0.0)
    {
	value[0] = gheader.epoc;
	ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
				   "EQUINOX", K_DOUBLE, value, FALSE);
    }
    if (gheader.ra != 0.0)
    {
	value[0] = gheader.ra / PION180;
	ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
				   "OBSRA", K_DOUBLE, value, FALSE);
    }
    if (gheader.dec != 0.0)
    {
	value[0] = gheader.dec / PION180;
	ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
				   "OBSDEC", K_DOUBLE, value, FALSE);
    }
    if (gheader.lii != 0.0)
    {
	value[0] = gheader.lii / PION180;
	ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
				   "OBSGLON", K_DOUBLE, value, FALSE);
    }
    if (gheader.bii != 0.0)
    {
	value[0] = gheader.bii / PION180;
	ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
				   "OBSGLAT", K_DOUBLE, value, FALSE);
    }
    if (gheader.isou[0])
	ds_put_unique_named_string (multi_desc->headers[0], multi_desc->data,
				    "OBJECT", gheader.isou, FALSE);
    if ( (gheader.majo != 0.0) && (gheader.mino != 0.0) )
    {
	value[0] = gheader.majo / PION180;
	ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
				   "BMAJ", K_DOUBLE, value, FALSE);
	value[0] = gheader.mino / PION180;
	ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
				   "BMIN", K_DOUBLE, value, FALSE);
	value[0] = gheader.posa / PION180;
	ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
				   "BPA", K_DOUBLE, value, FALSE);
    }
    value[0] = gheader.rmin;
    ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
			       "DATAMIN", K_DOUBLE, value, FALSE);
    value[0] = gheader.rmax;
    ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
			       "DATAMAX", K_DOUBLE, value, FALSE);
    if (context == NULL)
    {
	ch_close (channel);
	return (multi_desc);
    }
    if ( ( *context = m_calloc (sizeof **context) ) == NULL )
    {
	m_error_notify (function_name, "data context");
	ds_dealloc_multi (multi_desc);
	ch_close (channel);
	return (NULL);
    }
    (*context)->magic_number = MAGIC_NUMBER;
    (*context)->channel = channel;
    (*context)->format = format;
    (*context)->arr_desc = arr_desc;
    (*context)->packet_size = ds_get_packet_size (arr_desc->packet);
    (*context)->file_element_size = network_type_bytes[elem_type];
    return (multi_desc);
}   /*  End Function foreign_gildas_read_header  */

/*EXPERIMENTAL_FUNCTION*/
flag foreign_gildas_read_data (KGILDASDataContext context, char *data,
			       uaddr num_values)
/*  [SUMMARY] Read data in a GILDAS file.
    [PURPOSE] This routine will read the data of a GILDAS dataset from a
    KGILDASDataContext object. The header section is NOT read.
    <context> The context to read from.
    <data> The array to write the data into. If this is NULL, input data are
    skipped and nothing is written.
    <num_values> The number of values to read.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    uaddr bytes_to_read;
    uaddr bytes_read = 0;
    double buffer;
    static char function_name[] = "foreign_gildas_read_data";

    VERIFY_CONTEXT (context);
    bytes_to_read = num_values * context->file_element_size;
    switch (context->format)
    {
      case FORMAT_BLOCK:
	bytes_read = ch_read (context->channel, data, bytes_to_read);
	break;
      case FORMAT_BLOCK | FORMAT_SWAP:
	bytes_read = ch_read_and_swap_blocks (context->channel, data,
					      num_values,context->packet_size);
	break;
      case FORMAT_HARD:
	if ( dsrw_read_packets (context->channel, context->arr_desc->packet,
				data, num_values) )
	    bytes_read = bytes_to_read;
	break;
      case FORMAT_HARD | FORMAT_SWAP:
	if (context->file_element_size > sizeof buffer)
	{
	    fprintf (stderr, "Data size: %u greater than double size: %zu\n",
		     context->file_element_size, sizeof buffer);
	    a_prog_bug (function_name);
	}
	for (; num_values > 0; --num_values, data += context->packet_size)
	{
	    if ( ( bytes_read = ch_read (context->channel, (char *) &buffer,
					 context->file_element_size) )
		 < context->file_element_size)
	    {
		fprintf (stderr,
			 "%s: error reading: %u bytes, got: %lu\t%s\n",
			 function_name, context->file_element_size, bytes_read,
			 ERRSTRING);
		return (FALSE);
	    }
	}
	bytes_read = bytes_to_read;
	break;
      default:
	fprintf (stderr, "Illegal format: %u\n", context->format);
	a_prog_bug (function_name);
	break;
    }
    if (bytes_read < bytes_to_read)
    {
	fprintf (stderr, "%s: error reading: %lu bytes, got: %lu\t%s\n",
		 function_name, bytes_to_read, bytes_read, ERRSTRING);
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function foreign_gildas_read_data  */

/*EXPERIMENTAL_FUNCTION*/
void foreign_gildas_close_data_context (KGILDASDataContext context)
/*  [SUMMARY] Close a KGILDASDataContext object.
    <context> The context.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "foreign_gildas_close_data_context";

    VERIFY_CONTEXT (context);
    if (context->channel) ch_close (context->channel);
    context->magic_number = 0;
    m_free (context);
}   /*  End Function foreign_gildas_close_data_context  */


/*  Private functions follow  */

static void strip_trailing_whitespace (char *string)
/*  [SUMMARY] Strip trailing whitespace.
    <string> The string to strip.
    [RETURNS] Nothing.
*/
{
    char *ptr;

    for (ptr = string + strlen (string) - 1; (ptr > string) && isspace (*ptr);
	 --ptr) *ptr = '\0';
}   /*  End Function strip_trailing_whitespace  */

static void get_value (void *dest, CONST void *source, unsigned int type,
		       unsigned int format)
{
    unsigned int size;
    double buffer;
    extern char network_type_bytes[NUMTYPES];
    static char function_name[] = "_foreign_gildas_read_get_value";

    size = network_type_bytes[type];
    if ( (format & ~FORMAT_SWAP) == FORMAT_BLOCK )
    {
	if (format & FORMAT_SWAP)
	    m_copy_and_swap_blocks (dest, source, size, size, size, 1);
	else m_copy (dest, source, size);
	return;
    }
    /*  Do it the hard way  */
    if (format & FORMAT_SWAP)
    {
	if (size > sizeof buffer)
	{
	    fprintf (stderr, "Data size: %u greater than double size: %zu\n",
		     size, sizeof buffer);
	    a_prog_bug (function_name);
	}
	m_copy_and_swap_blocks (&buffer, source, size, size, size, 1);
	source = &buffer;
    }
    switch (type)
    {
      case K_FLOAT:
	p_read_buf_float (source, dest);
	break;
      case K_DOUBLE:
	p_read_buf_double (source, dest);
	break;
      case K_INT:
	p_read_buf32s (source, dest);
	break;
      case K_UINT:
	p_read_buf32 (source, dest);
	break;
      case K_SHORT:
	p_read_buf16s (source, dest);
	break;
      case K_USHORT:
	p_read_buf16 (source, dest);
	break;
      default:
	fprintf (stderr, "Unsupported type: %u\n", type);
	a_prog_bug (function_name);
	break;
    }
}   /*  End Function get_value  */
