/*LINTLIBRARY*/
/*  misc.c

    This code provides miscellaneous routines for world co-ordinates.

    Copyright (C) 2001-2003  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains miscellaneous routines for world co-ordinates.


    Written by      Richard Gooch   6-DEC-2001

    Updated by      Richard Gooch   13-DEC-2002: Changed interface to function
  <wcs_compare_grids_2D> so that transform error can be reported and
  restriction information is passed in.

    Last updated by Richard Gooch   23-MAR-2003: Fixed bug: was not comparing
  dimension lengths. Required for images of different sizes but same origin.


*/
#include <stdio.h>
#include <math.h>
#include <karma.h>
#include <karma_wcs.h>
#include <karma_ds_func.h>
#include <karma_m.h>


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
flag wcs_compare_grids_2D (KwcsAstro ap1,
			   CONST dim_desc *hd1, CONST dim_desc *vd1,
			   unsigned int num_restr1, CONST char **restr_names1,
			   CONST double *restr_values1,
			   KwcsAstro ap2,
			   CONST dim_desc *hd2, CONST dim_desc *vd2,
			   unsigned int num_restr2, CONST char **restr_names2,
			   CONST double *restr_values2,
			   flag *same)
/*  [SUMMARY] Compare if two images have the same world co-ordinate grid.
    <ap1> The KwcsAstro object for the first image.
    <hd1> The horizontal dimension of the first image.
    <vd1> The vertical dimension of the first image.
    <ap2> The KwcsAstro object for the second image.
    <hd2> The horizontal dimension of the second image.
    <vd2> The vertical dimension of the second image.
    <same> TRUE is written here if the grids are the same, else FALSE is
    written here.
    [RETURNS] TRUE on success, else FALSE, indicating that comparison failed.
*/
{
    unsigned int count;
    double hcoords1[5], vcoords1[5], hcoords2[5], vcoords2[5];

    *same = FALSE;
    if ( (hd1->length != hd2->length) || (vd1->length != vd2->length) )
	return (TRUE);
    hcoords1[0] = 0.0;
    vcoords1[0] = 0.0;
    hcoords1[1] = hd1->length - 1;
    vcoords1[1] = 0.0;
    hcoords1[2] = hd1->length - 1;
    vcoords1[2] = vd1->length - 1;
    hcoords1[3] = 0.0;
    vcoords1[3] = vd1->length - 1;
    hcoords1[4] = (double) hd1->length * 0.5;
    vcoords1[4] = (double) vd1->length * 0.5;
    m_copy (hcoords2, hcoords1, sizeof hcoords2);
    m_copy (vcoords2, vcoords1, sizeof vcoords2);
    ds_convert_coordinates (hd1, 5, hcoords2, FALSE, FALSE, FALSE);    
    ds_convert_coordinates (vd1, 5, vcoords2, FALSE, FALSE, FALSE);    
    if (ap1)
	if ( !wcs_astro_transform3 (ap1, FALSE, 5,
				    hd1->name, hcoords2, FALSE,
				    vd1->name, vcoords2, FALSE,
				    NULL, NULL, FALSE,
				    num_restr1, restr_names1, restr_values1) )
	    return (FALSE);
    if (ap2)
	if ( !wcs_astro_transform3 (ap2, FALSE, 5,
				    hd2->name, hcoords2, TRUE,
				    vd2->name, vcoords2, TRUE,
				    NULL, NULL, FALSE,
				    num_restr2, restr_names2, restr_values2) )
	    return (FALSE);
    ds_convert_coordinates (hd2, 5, hcoords2, TRUE, FALSE, FALSE);
    ds_convert_coordinates (vd2, 5, vcoords2, TRUE, FALSE, FALSE);
    for (count = 0; count < 5; ++count)
    {
	if (fabs (hcoords1[count] - hcoords2[count]) > 1e-3) return (TRUE);
	if (fabs (vcoords1[count] - vcoords2[count]) > 1e-3) return (TRUE);
    }
    *same = TRUE;
    return (TRUE);
}   /*  End Function wcs_compare_grids_2D  */
