/*LINTLIBRARY*/
/*  array_rgb.c

    This code provides RGB KConstructArrayManager objects.

    Copyright (C) 2004  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for managing the construction of
    RGB array data.


    Written by      Richard Gooch   16-MAR-2004

    Updated by      Richard Gooch   17-MAR-2004: Created function
  <_construct_array_rgb_check_pseudo>.

    Updated by      Richard Gooch   26-MAR-2004: Supported toggling on/off
  components. Supported one or two component combining.

    Updated by      Richard Gooch   31-MAR-2004: Made use of new experimental
  function <iarray_compare_sizes>.

    Last updated by Richard Gooch   10-APR-2004: Made use of new experimental
  function <iarray_register_data_change>.


*/
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_construct.h>
#include <karma_dispdata.h>
#include <karma_iarray.h>
#include <karma_listw.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_m.h>
#include <karma_c.h>
#include <karma_a.h>


#define NUM_RESERVED_CHARS            2
#define MAGIC_NUMBER         1018609574U

#define VERIFY_OBJECT(object) {if (object == NULL) \
{fprintf (stderr, "NULL object passed\n"); \
 a_prog_bug (function_name); } \
if (object->magic_number != MAGIC_NUMBER) \
{fprintf (stderr, "Invalid object\n"); \
 a_prog_bug (function_name); } }


struct list_component_struct
{
    unsigned int index;
    KDisplayDataObject dataobject;
};

struct construct_array_rgb
{
    unsigned int magic_number;
    KConstructArrayManager manager;
    KListWidget listw;
    KDisplayDataClass dataclass;
    unsigned int num_entries;
    KDisplayDataPrivateNameList namelist;
    char **names;
    KDisplayDataObject *dataobjects;
    struct list_component_struct red;
    struct list_component_struct green;
    struct list_component_struct blue;
};

struct callbacks
{
    KCallbackFunc range_change_callback;
    KCallbackFunc destroy_callback;
};

struct component
{
    KDisplayDataObject dataobject;
    iarray arr;
    struct callbacks callbacks;
};

struct composite_array
{
    struct component red;
    struct component green;
    struct component blue;
};


/*  Private functions  */
STATIC_FUNCTION (void _construct_array_rgb_update_list,
		 (struct construct_array_rgb *rgb) );
STATIC_FUNCTION (flag _construct_array_rgb_check_pseudo,
		 (void *info, KDisplayDataObject dataobject) );
STATIC_FUNCTION (void _construct_array_rgb_set_markers,
		 (struct construct_array_rgb *rgb) );
STATIC_FUNCTION (void _construct_array_rgb_change_name,
		 (struct construct_array_rgb *rgb) );
STATIC_FUNCTION (void _construct_array_rgb_make,
		 (KDisplayDataObject red, KDisplayDataObject green,
		  KDisplayDataObject blue, CONST char *name) );
STATIC_FUNCTION (void _construct_array_rgb_range_change_func,
		 (KDisplayDataObject dataobject, void *info) );
STATIC_FUNCTION (void _construct_array_rgb_destroy_func, (void *info) );


/*  Private data  */


/* Public functions follow */

/*UNPUBLISHED_FUNCTION*/
void *construct_array_rgb_init (KConstructArrayManager manager,
				KListWidget list)
{
    struct construct_array_rgb *rgb;
    static char function_name[] = "construct_array_rgb_init";

    if ( ( rgb = m_calloc (sizeof *rgb) ) == NULL )
    {
	m_error_notify (function_name, "RGB manager");
	return (NULL);
    }
    rgb->magic_number = MAGIC_NUMBER;
    rgb->manager = manager;
    rgb->listw = list;
    rgb->dataclass = dispdata_array_create_dataclass ();
    _construct_array_rgb_update_list (rgb);
    _construct_array_rgb_set_markers (rgb);
    return (rgb);
}   /*  End Function construct_array_rgb_init  */

/*UNPUBLISHED_FUNCTION*/
void construct_array_rgb_cleanup (void *algorithm_info)
{
    struct construct_array_rgb *rgb = algorithm_info;
    static char function_name[] = "construct_array_rgb_cleanup";

    VERIFY_OBJECT (rgb);
    dispdata_free_namelist (rgb->namelist);
    rgb->magic_number = 0;
    m_free (rgb);
}   /*  End Function construct_array_rgb_cleanup  */

/*UNPUBLISHED_FUNCTION*/
flag construct_array_rgb_position (void *algorithm_info, unsigned int index,
				   unsigned int event_code, void *e_info)
{
    KDisplayDataObject dataobject;
    char colour;
    struct construct_array_rgb *rgb = algorithm_info;
    static char function_name[] = "construct_array_rgb_position";

    VERIFY_OBJECT (rgb);
    switch (event_code)
    {
      case K_CANVAS_EVENT_LEFT_MOUSE_RELEASE:
	colour = 'R';
	break;
      case K_CANVAS_EVENT_MIDDLE_MOUSE_RELEASE:
	colour = 'G';
	break;
      case K_CANVAS_EVENT_RIGHT_MOUSE_RELEASE:
	colour = 'B';
	break;
      case K_CANVAS_EVENT_PLAIN_KEY_PRESS:
	colour = toupper (*(char *) e_info);
	break;
      default:
	return (FALSE);
	/*break;*/
    }
    dataobject = rgb->dataobjects[index];
    switch (colour)
    {
      case 'R':
	if ( (dataobject == rgb->green.dataobject)
	     || (dataobject == rgb->blue.dataobject) ) return (TRUE);
	if (rgb->red.dataobject) rgb->names[rgb->red.index][0] = ' ';
	if (dataobject == rgb->red.dataobject)
	{
	    rgb->red.dataobject = NULL;
	    break;
	}
	rgb->red.index = index;
	rgb->red.dataobject = dataobject;
	break;
      case 'G':
	if ( (dataobject == rgb->red.dataobject)
	     || (dataobject == rgb->blue.dataobject) ) return (TRUE);
	if (rgb->green.dataobject) rgb->names[rgb->green.index][0] = ' ';
	if (dataobject == rgb->green.dataobject)
	{
	    rgb->green.dataobject = NULL;
	    break;
	}
	rgb->green.index = index;
	rgb->green.dataobject = dataobject;
	break;
      case 'B':
	if ( (dataobject == rgb->red.dataobject)
	     || (dataobject == rgb->green.dataobject) ) return (TRUE);
	if (rgb->blue.dataobject) rgb->names[rgb->blue.index][0] = ' ';
	if (dataobject == rgb->blue.dataobject)
	{
	    rgb->blue.dataobject = NULL;
	    break;
	}
	rgb->blue.index = index;
	rgb->blue.dataobject = dataobject;
	break;
      default:
	return (FALSE);
	/*break;*/
    }
    _construct_array_rgb_set_markers (rgb);
    _construct_array_rgb_change_name (rgb);
    return (TRUE);
}   /*  End Function construct_array_rgb_position  */

/*UNPUBLISHED_FUNCTION*/
flag construct_array_rgb_data_create (void *algorithm_info,
				      KDisplayDataObject dataobject)
{
    struct construct_array_rgb *rgb = algorithm_info;
    static char function_name[] = "construct_array_rgb_data_create";

    VERIFY_OBJECT (rgb);
    _construct_array_rgb_update_list (rgb);
    _construct_array_rgb_set_markers (rgb);
    return (TRUE);
}   /*  End Function construct_array_rgb_data_create  */

/*UNPUBLISHED_FUNCTION*/
flag construct_array_rgb_data_destroy (void *algorithm_info)
{
    struct construct_array_rgb *rgb = algorithm_info;
    static char function_name[] = "construct_array_rgb_data_destroy";

    VERIFY_OBJECT (rgb);
    _construct_array_rgb_update_list (rgb);
    _construct_array_rgb_set_markers (rgb);
    _construct_array_rgb_change_name (rgb);
    return (TRUE);
}   /*  End Function construct_array_rgb_data_destroy  */

/*UNPUBLISHED_FUNCTION*/
void construct_array_rgb_create (void *algorithm_info, CONST char *name)
{
    struct construct_array_rgb *rgb = algorithm_info;
    static char function_name[] = "construct_array_rgb_create";

    VERIFY_OBJECT (rgb);
    if (!rgb->red.dataobject && !rgb->green.dataobject
	&& !rgb->blue.dataobject)
    {
	fprintf (stderr, "%s: insufficient data\n", function_name);
	return;
    }
    _construct_array_rgb_make (rgb->red.dataobject, rgb->green.dataobject,
			       rgb->blue.dataobject, name);
}   /*  End Function construct_array_rgb_create  */


/*  Private functions follow  */

static void _construct_array_rgb_update_list (struct construct_array_rgb *rgb)
{
    unsigned int count;

    rgb->names = dispdata_regenerate_namelist
	(rgb->dataclass, &rgb->namelist, NUM_RESERVED_CHARS, &rgb->num_entries,
	 &rgb->dataobjects, _construct_array_rgb_check_pseudo, NULL);
    rgb->red.index = rgb->num_entries;
    rgb->green.index = rgb->num_entries;
    rgb->blue.index = rgb->num_entries;
    for (count = 0; count < rgb->num_entries; ++count)
    {
	KDisplayDataObject dataobject = rgb->dataobjects[count];

	if (dataobject == rgb->red.dataobject) rgb->red.index = count;
	if (dataobject == rgb->green.dataobject) rgb->green.index = count;
	if (dataobject == rgb->blue.dataobject) rgb->blue.index = count;
    }
    if (rgb->red.index >= rgb->num_entries) rgb->red.dataobject = NULL;
    if (rgb->green.index >= rgb->num_entries) rgb->green.dataobject = NULL;
    if (rgb->blue.index >= rgb->num_entries) rgb->blue.dataobject = NULL;
}   /*  End Function _construct_array_rgb_update_list  */

static flag _construct_array_rgb_check_pseudo (void *info,
					       KDisplayDataObject dataobject)
{
    iarray arr;

    dispdata_get_data_attributes (dataobject,
				  KDISPLAYDATA_DATA_ATT_PSEUDO_ARRAY, &arr,
				  KDISPLAYDATA_DATA_ATT_END);
    return (arr ? TRUE : FALSE);
}   /*  End Function _construct_array_rgb_check_pseudo  */

static void _construct_array_rgb_set_markers (struct construct_array_rgb *rgb)
{
    if (rgb->red.dataobject) rgb->names[rgb->red.index][0] = 'R';
    if (rgb->green.dataobject) rgb->names[rgb->green.index][0] = 'G';
    if (rgb->blue.dataobject) rgb->names[rgb->blue.index][0] = 'B';
    listw_set_attributes (rgb->listw, TRUE,
			  LISTW_ATT_NUM_STRINGS, rgb->num_entries,
			  LISTW_ATT_STRINGS, rgb->names,
			  LISTW_ATT_REVERSE_VIDEO_INDEX, rgb->num_entries,
			  LISTW_ATT_END);
}   /*  End Function _construct_array_rgb_set_markers  */

static void _construct_array_rgb_change_name (struct construct_array_rgb *rgb)
{
    CONST char *name_r = "nored";
    CONST char *name_g = "nogreen";
    CONST char *name_b = "noblue";
    char name[STRING_LENGTH];

    if (!rgb->red.dataobject && !rgb->green.dataobject
	&& !rgb->blue.dataobject) return;
    if (rgb->red.dataobject)
	dispdata_get_data_attributes (rgb->red.dataobject,
				      KDISPLAYDATA_DATA_ATT_NAME, &name_r,
				      KDISPLAYDATA_DATA_ATT_END);
    if (rgb->green.dataobject)
	dispdata_get_data_attributes (rgb->green.dataobject,
				      KDISPLAYDATA_DATA_ATT_NAME, &name_g,
				      KDISPLAYDATA_DATA_ATT_END);
    if (rgb->blue.dataobject)
	dispdata_get_data_attributes (rgb->blue.dataobject,
				      KDISPLAYDATA_DATA_ATT_NAME, &name_b,
				      KDISPLAYDATA_DATA_ATT_END);
    sprintf (name, "%s+%s+%s", name_r, name_g, name_b);
    construct_array_change_name (rgb->manager, name);
}   /*  End Function _construct_array_rgb_change_name  */

static void _construct_array_rgb_make (KDisplayDataObject red,
				       KDisplayDataObject green,
				       KDisplayDataObject blue,
				       CONST char *name)
/*  [SUMMARY] Construct a composite RGB array from three input arrays.
    <red> The array which will be used for the red component.
    <green> The array which will be used for the green component.
    <blue> The array which will be used for the blue component.
    <name> The name of the new array.
*/
{
    iarray arr_r, arr_g, arr_b, arr_any, template;
    unsigned int count, index;
    array_pointer arrayp;
    multi_array *multi_desc;
    array_desc *arr_desc;
    packet_desc *inp_top_pack_desc;
    packet_desc *pack_desc;
    struct composite_array *new;
    char *inp_top_packet, *name_r, *name_g, *name_b;
    char domain[STRING_LENGTH];
    static char *elem_names[3] = {"Red Intensity", "Green Intensity",
				  "Blue Intensity"};
    static char function_name[] = "_construct_array_rgb_make";

    if (red)
	dispdata_get_data_attributes (red,
				      KDISPLAYDATA_DATA_ATT_NAME, &name_r,
				      KDISPLAYDATA_DATA_ATT_PSEUDO_ARRAY,
				          &arr_r,
				      KDISPLAYDATA_DATA_ATT_END);
    else arr_r = NULL;
    if (green)
	dispdata_get_data_attributes (green,
				      KDISPLAYDATA_DATA_ATT_NAME, &name_g,
				      KDISPLAYDATA_DATA_ATT_PSEUDO_ARRAY,
				          &arr_g,
				      KDISPLAYDATA_DATA_ATT_END);
    else arr_g = NULL;
    if (blue)
	dispdata_get_data_attributes (blue,
				      KDISPLAYDATA_DATA_ATT_NAME, &name_b,
				      KDISPLAYDATA_DATA_ATT_PSEUDO_ARRAY,
				          &arr_b,
				      KDISPLAYDATA_DATA_ATT_END);
    else arr_b = NULL;
    arr_any = arr_r;
    if (!arr_any) arr_any = arr_g;
    if (!arr_any) arr_any = arr_b;
    /*  First do sanity checks  */
    if ( ( arr_r && !iarray_compare_sizes (arr_any, arr_r, TRUE) )
	 || ( arr_g && !iarray_compare_sizes (arr_any, arr_g, TRUE) )
	 || ( arr_b && !iarray_compare_sizes (arr_any, arr_b, TRUE) ) )
	return;
    /*  Now create the new RGB array  */
    if ( ( new = m_calloc (sizeof *new) ) == NULL )
    {
	m_error_notify (function_name, "composite array structure");
	return;
    }
    /*  Construct RGB array using template  */
    template = arr_any;
    index = iarray_multi_index (template);
    inp_top_pack_desc = iarray_multi_array (template)->headers[index];
    inp_top_packet = iarray_multi_array (template)->data[index];
    for (index = 0; index < inp_top_pack_desc->num_elements; ++index)
    {
	if (iarray_array_desc (template) ==
	    (array_desc *) inp_top_pack_desc->element_desc[index]) break;
    }
    /*  Copy descriptor  */
    if ( ( multi_desc = ds_alloc_multi (1) ) == NULL )
    {
	m_error_notify (function_name, "multi_array");
	_construct_array_rgb_destroy_func (new);
	return;
    }
    c_register_callback (&multi_desc->destroy_callbacks,
			 ( flag (*) () )
			 _construct_array_rgb_destroy_func, new,
			 NULL, FALSE, NULL, FALSE, FALSE);
    multi_desc->headers[0] = ds_copy_desc_until (inp_top_pack_desc, NULL);
    if (multi_desc->headers[0] == NULL)
    {
	m_error_notify (function_name, "multi_array");
	ds_dealloc_multi (multi_desc);
	return;
    }
    /*  Go in and frob the array packet descriptor  */
    arr_desc = (array_desc *) multi_desc->headers[0]->element_desc[index];
    ds_dealloc_packet (arr_desc->packet, NULL);
    if ( ( arr_desc->packet = ds_alloc_packet_desc (3) ) == NULL )
    {
	m_error_notify (function_name, "packet descriptor");
	ds_dealloc_multi (multi_desc);
	return;
    }
    pack_desc = arr_desc->packet;
    for (count = 0; count < 3; ++count)
    {
	if ( ( pack_desc->element_desc[count] =
	       st_dup (elem_names[count]) ) == NULL )
	{
	    m_error_notify (function_name, "element name");
	    ds_dealloc_multi (multi_desc);
	    return;
	}
	pack_desc->element_types[count] = K_UBYTE;
    }
    /*  Allocate data  */
    multi_desc->data[0] = ds_alloc_data (multi_desc->headers[0], TRUE, TRUE);
    if (multi_desc->data[0] == NULL)
    {
	m_error_notify (function_name, "element name");
	ds_dealloc_multi (multi_desc);
	return;
    }
    /*  Copy part data  */
    ds_copy_data (inp_top_pack_desc, inp_top_packet,
		  multi_desc->headers[0], multi_desc->data[0]);
    m_copy (&arrayp,
	    multi_desc->data[0] +
	    ds_get_element_offset (multi_desc->headers[0], index),
	    sizeof arrayp);
    /*  Get rid of crud  */
    ds_remove_unique_named_element (multi_desc->headers[0],
				    multi_desc->data, "DATAMIN", TRUE);
    ds_remove_unique_named_element (multi_desc->headers[0],
				    multi_desc->data, "DATAMAX", TRUE);
    /*  Set up the colour plane intelligent arrays  */
    if ( ( new->red.arr =
	   iarray_get_from_multi_array (multi_desc, NULL, 0, NULL,
					elem_names[0]) ) == NULL )
    {
	m_error_notify (function_name, "red iarray");
	ds_dealloc_multi (multi_desc);
	return;
    }
    ds_dealloc_multi (multi_desc);  /*  Decrease attachment count  */
    if ( ( new->green.arr =
	   iarray_get_from_multi_array (multi_desc, NULL, 0, NULL,
					elem_names[1]) ) == NULL )
    {
	m_error_notify (function_name, "green iarray");
	ds_dealloc_multi (multi_desc);
	return;
    }
    ds_dealloc_multi (multi_desc);  /*  Decrease attachment count  */
    if ( ( new->blue.arr =
	   iarray_get_from_multi_array (multi_desc, NULL, 0, NULL,
					elem_names[2]) ) == NULL )
    {
	m_error_notify (function_name, "blue iarray");
	ds_dealloc_multi (multi_desc);
	return;
    }
    ds_dealloc_multi (multi_desc);  /*  Decrease attachment count  */
    if (red)
    {
	new->red.dataobject = red;
	new->red.callbacks.range_change_callback =
	    dispdata_array_register_irange_change_func
	    (red, _construct_array_rgb_range_change_func, new);
	new->red.callbacks.destroy_callback =
	    dispdata_register_dataobject_destroy_func (red,
						       c_write_2nulls_func,
						       &new->red.callbacks);
    }
    if (green)
    {
	new->green.dataobject = green;
	new->green.callbacks.range_change_callback =
	    dispdata_array_register_irange_change_func
	    (green, _construct_array_rgb_range_change_func, new);
	new->green.callbacks.destroy_callback =
	    dispdata_register_dataobject_destroy_func (green,
						       c_write_2nulls_func,
						       &new->green.callbacks);
    }
    if (blue)
    {
	new->blue.dataobject = blue;
	new->blue.callbacks.range_change_callback =
	    dispdata_array_register_irange_change_func
	    (blue, _construct_array_rgb_range_change_func, new);
	new->blue.callbacks.destroy_callback =
	    dispdata_register_dataobject_destroy_func (blue,
						       c_write_2nulls_func,
						       &new->blue.callbacks);
    }
    if (red) _construct_array_rgb_range_change_func (red, new);
    if (green) _construct_array_rgb_range_change_func (green, new);
    if (blue) _construct_array_rgb_range_change_func (blue, new);
    sprintf (domain, "COMPOSITE:Red: %s  Green: %s  Blue: %s",
	     name_r, name_g, name_b);
    ds_event_dispatch (multi_desc, domain, name);
    ds_dealloc_multi (multi_desc);
}   /*  End Function _construct_array_rgb_make  */

static void _construct_array_rgb_range_change_func
    (KDisplayDataObject dataobject, void *info)
{
    iarray inp_arr;
    double low, high, scale, offset;
    struct composite_array *arr = info;
    struct component *component;

    if (dataobject == arr->red.dataobject) component = &arr->red;
    else if (dataobject == arr->green.dataobject) component = &arr->green;
    else component = &arr->blue;
    dispdata_get_data_attributes (dataobject,
				  KDISPLAYDATA_DATA_ATT_PSEUDO_ARRAY, &inp_arr,
				  KDISPLAYDATA_DATA_ATT_RAW_LOWER_CLIP, &low,
				  KDISPLAYDATA_DATA_ATT_RAW_UPPER_CLIP, &high,
				  KDISPLAYDATA_DATA_ATT_END);
    scale = 255.0 / (high - low);
    offset = -low * scale;
    iarray_clip_scale_and_offset (component->arr, inp_arr,
				  scale, offset, low, high, FALSE);
    iarray_register_data_change (component->arr, NULL);
}   /*  End Function _construct_array_rgb_range_change_func  */

static void _construct_array_rgb_destroy_func (void *info)
{
    struct composite_array *arr = info;

    c_unregister_callback2 (arr->red.callbacks.range_change_callback);
    c_unregister_callback2 (arr->red.callbacks.destroy_callback);
    c_unregister_callback2 (arr->green.callbacks.range_change_callback);
    c_unregister_callback2 (arr->green.callbacks.destroy_callback);
    c_unregister_callback2 (arr->blue.callbacks.range_change_callback);
    c_unregister_callback2 (arr->blue.callbacks.destroy_callback);
    if (arr->red.arr) iarray_multi_array (arr->red.arr) = NULL;
    if (arr->green.arr) iarray_multi_array (arr->green.arr) = NULL;
    if (arr->blue.arr) iarray_multi_array (arr->blue.arr) = NULL;
    iarray_dealloc2 (arr->red.arr);
    iarray_dealloc2 (arr->green.arr);
    iarray_dealloc2 (arr->blue.arr);
    m_free (arr);
}   /*  End Function _construct_array_rgb_destroy_func  */
