/*LINTLIBRARY*/
/*  main.c

    This code provides KContourImage objects.

    Copyright (C) 1996-2004  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for manipulating contour images.


    Written by      Richard Gooch   20-JUL-1996

    Updated by      Richard Gooch   22-JUL-1996: Created
  <contour_create_sequence_from_iarray>.

    Updated by      Richard Gooch   4-SEP-1996: More than one contour image may
  be active per canvas.

    Updated by      Richard Gooch   5-SEP-1996: Optimisations when number of
  segments less than NUM_SECTIONS. Fixed bug where number of pixel segments
  less than NUM_SECTIONS was not trapped.

    Updated by      Richard Gooch   13-OCT-1996: Trapped NULL <<pixel_values>>
  in <contour_set_levels>.

    Updated by      Richard Gooch   15-OCT-1996: Trapped NULL
  <<contour_levels>> in several places.

    Updated by      Richard Gooch   15-JAN-1997: Switched to
  <wcs_astro_transform3> in order to deal with non RA/DEC axes.

    Updated by      Richard Gooch   17-JAN-1997: Created
  <contour_set_attributes>.

    Updated by      Richard Gooch   27-JAN-1997: Make use of restriction
  information when doing astronomical projections. Do quick check to see if
  canvas X and Y labels match.

    Updated by      Richard Gooch   29-JAN-1997: Only create a single KwcsAstro
  object for a sequence.

    Updated by      Richard Gooch   1-FEB-1997: Made use of new
  <wcs_astro_get_associated> function.

    Updated by      Richard Gooch   2-FEB-1997: Removed quick check for
  matching X and Y labels: this would reject similar axes which have different
  names (i.e. "RA---SIN" and "RA---TAN" or "Axis 1").

    Updated by      Richard Gooch   11-JUL-1997: Separated caching and drawing
  of levels to support later definition of level attributes (i.e. line styles).

    Updated by      Richard Gooch   21-JUL-1997: Made use of new precompute
  event handlers.

    Updated by      Richard Gooch   31-JUL-1997: Took account of change to
  precompute function interface: pointer to <honoured_areas> flag dropped.

    Updated by      Richard Gooch   11-OCT-1997: Added dashing support.

    Updated by      Richard Gooch   29-OCT-1997: Hacked dashing support to
  use grey instead.

    Updated by      Richard Gooch   18-NOV-1997: Fixed bug where segments
  outside canvas could be drawn when number of segments was low.

    Updated by      Richard Gooch   19-NOV-1997: Fixed bug where segments with
  one end outside canvas could be drawn.

    Updated by      Richard Gooch   7-DEC-1997: Added linewidth support.

    Updated by      Richard Gooch   5-MAR-1998: Removed unnecessary cast in
  <_contour_precompute_image>.

    Updated by      Richard Gooch   6-MAR-1998: Removed casts for memory
  functions. Fixed bug in <contour_destroy> which was deallocating levels which
  are also deallocated in <dealloc_levels>.

    Updated by      Richard Gooch   16-SEP-1998: Added grouping key.

    Updated by      Richard Gooch   17-SEP-1998: Added contour image resource
  CONTOUR_CIMAGE_ATT_COLOURNAME. Changed from inserting new image at front of
  list to appending to end of list.

    Updated by      Richard Gooch   5-JUL-1999: Removed sequence holder.
  Created freelist.

    Updated by      Richard Gooch   15-AUG-1999: Made egcs happy.

    Updated by      Richard Gooch   11-SEP-1999: Check for failure return from
  <ds_contour>.

    Updated by      Richard Gooch   27-JUN-2000: Made use of new function
  <canvas_register_destroy_func>. Bugfix in <contour_set_active>: was not
  forcing refresh if other images deactivated and <<refresh_if_changed>> TRUE.

    Updated by      Richard Gooch   7-JUL-2000: Took account of interface
  change to <canvas_register_destroy_func>.

    Updated by      Richard Gooch   11-JUL-2000: Added #include <karma_kwin.h>

    Updated by      Richard Gooch   22-NOV-2001: Removed obsolete external
  function declaration of <canvas_register_precompute_func>.

    Updated by      Richard Gooch   23-NOV-2001: Cosmetic name changes.

    Updated by      Richard Gooch   6-DEC-2001: Took account of interface
  change to <wcs_astro_transform*>.

    Updated by      Richard Gooch   12-DEC-2001: Removed unused position list.

    Updated by      Richard Gooch   29-MAY-2002: Generate warning instead of
  error and abort for non-regular dimensions.

    Updated by      Richard Gooch   8-JUN-2002: Made use of new KLIST_ macros.

    Updated by      Richard Gooch   14-AUG-2002: Added #include <stdlib.h>.

    Updated by      Richard Gooch   19-NOV-2002: Minor documentation change.

    Updated by      Richard Gooch   22-DEC-2002: Stupid compiler warning fix.

    Updated by      Richard Gooch   24-DEC-2002: Fixed bug in function
  <_contour_get_gen_index>: duplicate names in different general data structues
  caused failure.

    Updated by      Richard Gooch   17-MAR-2004: Fixed functions which take an
  Intelligent Array to accept sub-dimensional slices.

    Updated by      Richard Gooch   10-APR-2004: Added <<data_change_callback>>

    Updated by      Richard Gooch   30-APR-2004: Split original KContourImage
  class into new KContourImage and KViewableContourImage classes.

    Updated by      Richard Gooch   2-MAY-2004: Created KContourImageGroup and
  KViewableContourImageGroup classes.

    Updated by      Richard Gooch   9-MAY-2004: Share linear segments between
  canvases with the same world<->linear transformation. Allocate and deallocate
  area pixel segments as needed (only a small fraction of vcimages will need
  this at any one time).

    Updated by      Richard Gooch   11-MAY-2004: Minor memory optimisation:
  free excessive over-allocations.

    Updated by      Richard Gooch   27-JUN-2004: Trap for setting level styles
  before levels are set.

    Updated by      Richard Gooch   7-JUL-2004: Fixed bug in function
  <_contour_find_linsegs_cache>: sole-use linsegs cache was being reused for
  update without being marked invalid.

    Updated by      Richard Gooch   20-AUG-2004: Prepended "_contour_" to all
  remaining static functions.

    Last updated by Richard Gooch   21-NOV-2004: Fixed bug introduced when
  KContourImageGroup was added: missing braces around VERIFY_* macros led to
  NULL dereference when <<group>> is NULL.


*/
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdarg.h>
#include <karma.h>
#include <k_list.h>
#define NEW_WIN_SCALE
#include <karma_contour.h>
#include <karma_canvas.h>
#include <karma_iarray.h>
#include <karma_kwin.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_c.h>


#define CIMAGE_USED_MAGIC_NUMBER    1342908534
#define CIMAGE_FREE_MAGIC_NUMBER    1349879874
#define VCIMAGE_USED_MAGIC_NUMBER   1444841859
#define VCIMAGE_FREE_MAGIC_NUMBER   1174286699
#define HOLDER_MAGIC_NUMBER         1653453498
#define CIMAGE_GROUP_MAGIC_NUMBER   1777684373
#define VCIMAGE_GROUP_MAGIC_NUMBER  2042131812
#define DEFAULT_COLOUR                 "green"
#define NUM_SECTIONS                        64

#define VERIFY_HOLDER(holder) { if (holder == NULL) \
{fprintf (stderr, "NULL canvas holder passed\n"); \
 a_prog_bug (function_name); } \
if (holder->magic_number != HOLDER_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid canvas holder object\n"); \
 a_prog_bug (function_name); } }

#define VERIFY_USED_CIMAGE(cimage) { if (cimage == NULL) \
{fprintf (stderr, "NULL contourable image passed\n"); \
 a_prog_bug (function_name); } \
if (cimage->magic_number != CIMAGE_USED_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid contourable image object\n"); \
 a_prog_bug (function_name); } }

#define VERIFY_FREE_CIMAGE(cimage) { if (cimage == NULL) \
{fprintf (stderr, "NULL contourable image passed\n"); \
 a_prog_bug (function_name); } \
if (cimage->magic_number != CIMAGE_FREE_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid freed contourable image object\n"); \
 a_prog_bug (function_name); } }

#define VERIFY_USED_VCIMAGE(vcimage) { if (vcimage == NULL) \
{fprintf (stderr, "NULL viewable contour image passed\n"); \
 a_prog_bug (function_name); } \
if (vcimage->magic_number != VCIMAGE_USED_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid viewable contour image object\n"); \
 a_prog_bug (function_name); } }

#define VERIFY_FREE_VCIMAGE(vcimage) { if (vcimage == NULL) \
{fprintf (stderr, "NULL viewable contour image passed\n"); \
 a_prog_bug (function_name); } \
if (vcimage->magic_number != VCIMAGE_FREE_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid freed viewable contour image object\n"); \
 a_prog_bug (function_name); } }

#define VERIFY_CIMAGE_GROUP(group) { if (group == NULL) \
{fprintf (stderr, "NULL KContourImageGroup passed\n"); \
 a_prog_bug (function_name); } \
if (group->magic_number != CIMAGE_GROUP_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid KContourImageGroup object\n"); \
 a_prog_bug (function_name); } }

#define VERIFY_VCIMAGE_GROUP(group) { if (group == NULL) \
{fprintf (stderr, "NULL KViewableContourImageGroup passed\n"); \
 a_prog_bug (function_name); } \
if (group->magic_number != VCIMAGE_GROUP_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid KViewableContourImageGroup object\n"); \
 a_prog_bug (function_name); } }


typedef struct canvas_holder_type * CanvasHolder;

struct canvas_holder_type
{
    unsigned int magic_number;
    KWorldCanvas canvas;
    flag needs_refresh;
    KViewableContourImage first;
    KViewableContourImage last;
    CanvasHolder prev;
    CanvasHolder next;
    char *colourname;
};

struct worldsegments_level_struct
{
    uaddr coord_buf_size;
    uaddr num_segments;
    double *x0;
    double *y0;
    double *x1;
    double *y1;
};

struct section_limits_type
{
    double xmin[NUM_SECTIONS];
    double xmax[NUM_SECTIONS];
    double ymin[NUM_SECTIONS];
    double ymax[NUM_SECTIONS];
};

struct linearsegments_level_struct
{
    uaddr coord_buf_size;
    double *x0;
    double *y0;
    double *x1;
    double *y1;
    struct section_limits_type limits;
};

struct linearsegments_cache_struct
{
    KContourImage cimage;
    unsigned int refcount;                             /*  Number of users   */
    flag valid;
    struct linearsegments_level_struct *levels;
    double test_x[5];
    double test_y[5];
    struct linearsegments_cache_struct *prev;
    struct linearsegments_cache_struct *next;
};

struct contourableimage_group
{
    unsigned int magic_number;
    KContourImage first;
    KContourImage last;
};

struct cimage_chain_type
{
    KContourImage prev;          /*  Previous member of list for this chain  */
    KContourImage next;          /*  Next member of list for this chain      */
};

struct contourableimage_type
{
    unsigned int magic_number;
    KContourImageGroup group;
    /*  Image data  */
    multi_array *multi_desc;
    array_desc *arr_desc;
    char *slice;
    unsigned int hdim;
    unsigned int vdim;
    unsigned int elem_index;
    KwcsAstro astro_projection;
    /*  Contour levels  */
    unsigned int num_levels;
    double *contour_levels;
    /*  Other information  */
    flag world_segs_valid; /*Whether the world co-ordinate segments are valid*/
    unsigned int num_restrictions;
    char **restriction_names;
    double *restriction_values;
    char *inline_restr_names[1];
    double inline_restr_values[1];
    /*  Segment data  */
    struct worldsegments_level_struct *worldsegs;
    struct linearsegments_cache_struct *first_linsegs_cache;
    struct linearsegments_cache_struct *last_linsegs_cache;
    KCallbackFunc data_change_callback;
    KViewableContourImage first_vcimage;
    KViewableContourImage last_vcimage;
    KContourImage next_free;
    struct cimage_chain_type group_chain;
};

struct viewable_levelinfo_type
{
    unsigned long pixel_value;
    flag pixel_defined;
    flag dash;
    double linewidth;
    uaddr coord_buf_size;
    unsigned int num_pixel_segments;
    double *pix_x0;
    double *pix_y0;
    double *pix_x1;
    double *pix_y1;
    struct section_limits_type pix_limits;
    unsigned int num_area_segments;
    double *pix_area_x0;  /*  All 4 allocated in a single block  */
    double *pix_area_y0;
    double *pix_area_x1;
    double *pix_area_y1;
};

struct viewable_contourimage_group
{
    unsigned int magic_number;
    KViewableContourImage first;
    KViewableContourImage last;
};

struct vcimage_chain_type
{
    KViewableContourImage prev;  /*  Previous member of list for this chain  */
    KViewableContourImage next;  /*  Next member of list for this chain      */
};

struct viewable_contourimage_type
{
    unsigned int magic_number;
    CanvasHolder canvas_holder;
    KContourImage cimage;
    KViewableContourImageGroup group;
    flag active;
    /*  Contour levels  */
    struct viewable_levelinfo_type *levels;
    /*  Cached segments  */
    struct linearsegments_cache_struct *linsegs_cache;
    flag pixel_segments_valid;
    int old_x_offset;
    int old_y_offset;
    int old_x_pixels;
    int old_y_pixels;
    double old_left_x;
    double old_right_x;
    double old_bottom_y;
    double old_top_y;
    struct vcimage_chain_type holder_chain;  /*  Also used for free chain  */
    struct vcimage_chain_type cimage_chain;
    struct vcimage_chain_type group_chain;
};


/*  Private data  */
static CanvasHolder first_canvas_holder = NULL;
static CanvasHolder last_canvas_holder = NULL;
static KContourImage first_free_cimage = NULL;
static KViewableContourImage first_free_vcimage = NULL;


/*  Private functions  */
STATIC_FUNCTION (KContourImage _contour_create_restr,
		 (multi_array *multi_desc, flag try_ap,
		  array_desc *arr_desc, char *slice,
		  unsigned int hdim, unsigned int vdim,
		  unsigned int elem_index,
		  unsigned int num_restr, char **restr_names,
		  double *restr_values, KContourImageGroup group) );
STATIC_FUNCTION (unsigned int _contour_get_gen_index,
		 (CONST multi_array *multi_desc, CONST array_desc *arr_desc) );
STATIC_FUNCTION (CanvasHolder _contour_get_canvas_holder,
		 (KWorldCanvas canvas, char *func_name) );
STATIC_FUNCTION (CanvasHolder _contour_alloc_canvas_holder,
		 (KWorldCanvas canvas) );
STATIC_FUNCTION (void _contour_free_canvas_holder, (CanvasHolder holder) );
STATIC_FUNCTION (void _contour_process_canvas_attributes,
		 (CanvasHolder holder, va_list argp) );
STATIC_FUNCTION (void _contour_worldcanvas_precompute_func,
		 (KWorldCanvas canvas, int width, int height,
		  CONST struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize,
		  void **info, PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void _contour_precompute_image,
		 (KViewableContourImage vcimage,
		  CONST struct win_scale_type *win_scale,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void _contour_precompute_level,
		 (KViewableContourImage vcimage, unsigned int level_index,
		  CONST struct win_scale_type *win_scale,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void _contour_compute_lin_limits,
		 (KViewableContourImage vcimage, unsigned int level_index,
		  CONST struct win_scale_type *win_scale,
		  double win_x_min, double win_y_min,
		  double win_x_max, double win_y_max) );
STATIC_FUNCTION (void _contour_compute_pix_segs,
		 (KViewableContourImage vcimage, unsigned int level_index,
		  CONST struct win_scale_type *win_scale,
		  double win_x_min, double win_y_min,
		  double win_x_max, double win_y_max) );
STATIC_FUNCTION (void _contour_compute_pixel_coords,
		 (KWorldCanvas canvas, struct viewable_levelinfo_type *level,
		  CONST struct linearsegments_level_struct *linsegs_level,
		  CONST struct win_scale_type *win_scale,
		  unsigned int num_world_segs, unsigned int num_pix_segs) );
STATIC_FUNCTION (void _contour_worldcanvas_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  CONST struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize,
		  void **info, PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (void _contour_refresh_image,
		 (KViewableContourImage vcimage,
		  CONST struct win_scale_type *win_scale,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void _contour_draw_level,
		 (KViewableContourImage vcimage, unsigned int level_index,
		  CONST struct win_scale_type *win_scale,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void _contour_dealloc_levels, (KContourImage cimage) );
STATIC_FUNCTION (void _contour_dealloc_viewable_levels,
		 (KViewableContourImage vcimage) );
STATIC_FUNCTION (void _contour_find_linsegs_cache,
		 (KViewableContourImage vcimage,
		  CONST struct win_scale_type *win_scale) );
STATIC_FUNCTION (void _contour_release_linsegs_cache,
		 (struct linearsegments_cache_struct *cache) );


/* Public functions follow */

/*EXPERIMENTAL_FUNCTION*/
void contour_init (KWorldCanvas canvas, ...)
/*  [SUMMARY] Initialise the package for a particular canvas.
    [PURPOSE] This routine will initialise the [<contour>] package for a
    particular world canvas. Calling this routine causes a number of callback
    routines internal to the package to be registered with the canvas (such
    as refresh and position event callbacks). This routine must be called
    before creating contour images.
    <canvas> The world canvas object.
    [VARARGS] The list of parameter attribute-key attribute-value-ptr pairs
    must follow. This list must be terminated with the CONTOUR_CANVAS_ATT_END.
    See [<contour_CANVAS_ATTRIBUTES>] for the list of attributes.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    CanvasHolder holder;
    static char function_name[] = "contour_init";

    va_start (argp, canvas);
    if (canvas == NULL)
    {
	fprintf (stderr, "NULL world canvas passed\n");
	a_prog_bug (function_name);
    }
    if (_contour_get_canvas_holder (canvas, NULL) != NULL)
    {
	fprintf (stderr, "Canvas already initialised\n");
	a_prog_bug (function_name);
    }
    if ( ( holder = _contour_alloc_canvas_holder (canvas) ) == NULL )
	m_abort (function_name, "canvas holder");
    _contour_process_canvas_attributes (holder, argp);
    va_end (argp);
}   /*  End Function contour_init  */

/*EXPERIMENTAL_FUNCTION*/
KContourImageGroup contour_create_group ()
/*  [SUMMARY] Create a container to hold a group of KContourImage objects.
    [RETURNS] The group container on success, else NULL.
*/
{
    KContourImageGroup group;
    static char function_name[] = "contour_create_group";

    if ( ( group = m_calloc (sizeof *group) ) == NULL )
    {
	m_error_notify (function_name, "cimage group");
	return (NULL);
    }
    group->magic_number = CIMAGE_GROUP_MAGIC_NUMBER;
    return (group);
}   /*  End Function contour_create_group  */

/*EXPERIMENTAL_FUNCTION*/
void contour_destroy_group (KContourImageGroup group)
/*  [SUMMARY] Destroy a group of KContourImage objects.
    <group> The group to destroy. The group container is also destroyed.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "contour_destroy_group";

    VERIFY_CIMAGE_GROUP (group);
    while (group->first) contour_destroy (group->first);
    group->magic_number = 0;
    m_free (group);
}   /*  End Function contour_destroy_group  */

/*EXPERIMENTAL_FUNCTION*/
KContourImage contour_create_restr (multi_array *multi_desc,
				    array_desc *arr_desc, char *slice,
				    unsigned int hdim, unsigned int vdim,
				    unsigned int elem_index,
				    unsigned int num_restr,
				    char **restr_names, double *restr_values,
				    KContourImageGroup group)
/*  [SUMMARY] Create contour image from 2D slice with restrictions.
    [PURPOSE] This routine will create a contour image object from a
    2-dimensional slice of a Karma data structure. At a later time, this
    contour image may be used to create a viewable contour image.
    <multi_desc> The  multi_array  descriptor which contains the Karma data
    structure. The routine increments the attachment count on the descriptor
    on successful completion. This may be NULL.
    <arr_desc> The array descriptor.
    <slice> The start of the slice data.
    <hdim> The dimension index of the horizontal dimension.
    <vdim> The dimension index of the vertical dimension.
    <elem_index> The element index in the data packets.
    <num_restr> The number of matched restrictions.
    <restr_names> The restriction names.
    <restr_values> The restriction values.
    <group> The KContourImageGroup to add the new object too. If NULL, it is
    not added to any group.
    [NOTE] Restriction information is automatically deallocated when
    [<contour_destroy>] is called.
    [RETURNS] A KContourImage object on success, NULL.
*/
{
    static char function_name[] = "contour_create_restr";

    if (group) VERIFY_CIMAGE_GROUP (group);
    if (arr_desc == NULL)
    {
	fprintf (stderr, "NULL array descriptor pointer passed\n");
	a_prog_bug (function_name);
    }
    if (slice == NULL)
    {
	fprintf (stderr, "NULL slice pointer passed\n");
	a_prog_bug (function_name);
    }
    /*  Sanity checks  */
    if (hdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "hdim: %u greater than number of dimensions: %u\n",
		 hdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    if (arr_desc->dimensions[hdim]->coordinates)
	fprintf (stderr, "WARNING: hdim: %u not regularly spaced\n", hdim);
    if (vdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "vdim: %u greater than number of dimensions: %u\n",
		 vdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    if (arr_desc->dimensions[vdim]->coordinates)
	fprintf (stderr, "WARNING: vdim: %u not regularly spaced\n", vdim);
    if (elem_index >= arr_desc->packet->num_elements)
    {
	fprintf (stderr,"elem_index: %u greater than number of elements: %u\n",
		 elem_index, arr_desc->packet->num_elements);
	a_prog_bug (function_name);
    }
    if (arr_desc->offsets == NULL)
    {
	if ( !ds_compute_array_offsets (arr_desc) )
	{
	    fprintf (stderr, "Error computing address offsets\n");
	    a_prog_bug (function_name);
	}
    }
    return _contour_create_restr (multi_desc, TRUE, arr_desc, slice,
				  hdim, vdim, elem_index,
				  num_restr, restr_names, restr_values, group);
}   /*  End Function contour_create_restr  */

/*EXPERIMENTAL_FUNCTION*/
KContourImage contour_create_from_iarray (iarray array, flag swap,
					  KContourImageGroup group)
/*  [SUMMARY] Create a contour image from an Intelligent Array.
    [PURPOSE] This routine will create a contour image object from a
    2-dimensional Intelligent Array. At a later time, this contour image may be
    used to create a viewable contour image.
    <array> The Intelligent Array. The underlying <<multi_array>> data
    structure will have its attachment count incremented upon successful
    completion.
    <swap> If TRUE the y axis will be displayed horizontally.
    <group> The KContourImageGroup to add the new object too. If NULL, it is
    not added to any group.
    [RETURNS] A KContourImage object on success, else NULL.
*/
{
    KContourImage cimage;
    unsigned int num_restr, xdim, ydim;
    dim_desc *xdim_desc, *ydim_desc;
    char **restr_names;
    double *restr_values;
    static char function_name[] = "contour_create_from_iarray";

    if (array == NULL)
    {
	fprintf (stderr, "NULL Intelligent Array passed\n");
	a_prog_bug (function_name);
    }
    FLAG_VERIFY (swap);
    if (group) VERIFY_CIMAGE_GROUP (group);
    if (iarray_num_dim (array) != 2)
    {
	fprintf (stderr, "Intelligent Array must have exactly 2 dimensions\n");
	return (NULL);
    }
    if (swap) xdim = 0, ydim = 1;
    else xdim = 1, ydim = 0;
    xdim_desc = iarray_get_dim_desc (array, xdim);
    ydim_desc = iarray_get_dim_desc (array, ydim);
    if ( (iarray_dim_length (array, xdim) != xdim_desc->length)
	 || (iarray_dim_length (array, ydim) != ydim_desc->length) )
    {
	fprintf (stderr, "%s: Intelligent Array must not be a sub-array\n",
		 function_name);
	return (NULL);
    }
    num_restr = iarray_get_restrictions (array, &restr_names, &restr_values);
    cimage = _contour_create_restr (array->multi_desc, TRUE,
				    array->arr_desc, array->data,
				    array->orig_dim_indices[xdim],
				    array->orig_dim_indices[ydim],
				    array->elem_index,
				    num_restr, restr_names, restr_values,
				    group);
    if (!cimage) return (NULL);
    cimage->data_change_callback =
	iarray_register_data_change_func (array, contour_register_data_change,
					  cimage);
    return (cimage);
}   /*  End Function contour_create_from_iarray  */

/*EXPERIMENTAL_FUNCTION*/
KContourImage *contour_create_sequence (multi_array *multi_desc,
					array_desc *arr_desc, char *cube,
					unsigned int hdim, unsigned int vdim,
					unsigned int fdim,
					unsigned int elem_index,
					KContourImageGroup group)
/*  [SUMMARY] Create a sequence of contour images from a 3D slice.
    [PURPOSE] This routine will create a sequence of contour image objects
    from a 3-dimensional cube of a Karma data structure. At a later time, this
    sequence of contour images may be used to create a sequence of viewable
    contour images.
    <multi_desc> The  multi_array  descriptor which contains the Karma data
    structure. The routine increments the attachment count on the descriptor
    on successful completion. This may be NULL.
    <arr_desc> The array descriptor.
    <cube> The start of the cube data.
    <hdim> The dimension index of the horizontal dimension.
    <vdim> The dimension index of the vertical dimension.
    <fdim> The dimension index of the frame dimension (dimension containing the
    sequence). The number of frames is the same as the length of this
    dimension.
    <elem_index> The element index in the data packets.
    <group> The KContourImageGroup to add the new objects too. If NULL, it is
    not added to any group.
    [RETURNS] A pointer to a dynamically allocated array of contour image
    objects on success, else NULL.
*/
{
    KwcsAstro ap;
    KContourImage cimage;
    unsigned int num_frames, count, index;
    KContourImage *cimages;
    uaddr *foffsets;
    static char function_name[] = "contour_create_sequence";

    if (arr_desc == NULL)
    {
	fprintf (stderr, "NULL array descriptor pointer passed\n");
	a_prog_bug (function_name);
    }
    if (cube == NULL)
    {
	fprintf (stderr, "NULL slice pointer passed\n");
	a_prog_bug (function_name);
    }
    if (group) VERIFY_CIMAGE_GROUP (group);
    /*  Sanity checks  */
    if (fdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "fdim: %u greater than number of dimensions: %u\n",
		 fdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    /*  Sanity checks  */
    if (hdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "hdim: %u greater than number of dimensions: %u\n",
		 hdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    if (arr_desc->dimensions[hdim]->coordinates)
	fprintf (stderr, "WARNING: hdim: %u not regularly spaced\n", hdim);
    if (vdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "vdim: %u greater than number of dimensions: %u\n",
		 vdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    if (arr_desc->dimensions[vdim]->coordinates)
	fprintf (stderr, "WARNING: vdim: %u not regularly spaced\n", vdim);
    if (elem_index >= arr_desc->packet->num_elements)
    {
	fprintf (stderr,"elem_index: %u greater than number of elements: %u\n",
		 elem_index, arr_desc->packet->num_elements);
	a_prog_bug (function_name);
    }
    if (arr_desc->offsets == NULL)
    {
	if ( !ds_compute_array_offsets (arr_desc) )
	{
	    fprintf (stderr, "Error computing address offsets\n");
	    a_prog_bug (function_name);
	}
    }
    foffsets = arr_desc->offsets[fdim];
    num_frames = arr_desc->dimensions[fdim]->length;
    if ( ( cimages = m_alloc (sizeof *cimages * num_frames) ) == NULL )
    {
	m_error_notify (function_name, "array of contourable images");
	return (NULL);
    }
    if (multi_desc == NULL) ap = NULL;
    else
    {
	index = _contour_get_gen_index (multi_desc, arr_desc);
	ap = wcs_astro_get_associated (multi_desc, index);
    }
    for (count = 0; count < num_frames; ++count)
    {
	if ( ( cimage =
	       _contour_create_restr (multi_desc, FALSE, arr_desc,
				      cube + foffsets[count],
				      hdim, vdim, elem_index, 1, NULL, NULL,
				      group) )
	    == NULL )
	{
	    for (; count > 0; --count) contour_destroy (cimages[count - 1]);
	    m_free (cimages);
	    return (NULL);
	}
	cimages[count] = cimage;
	cimages[count]->astro_projection = ap;
	cimage->num_restrictions = 1;
	cimage->inline_restr_names[0] = arr_desc->dimensions[fdim]->name;
	cimage->inline_restr_values[0] =
	    ds_get_coordinate (arr_desc->dimensions[fdim], count);
	cimage->restriction_names = cimage->inline_restr_names;
	cimage->restriction_values = cimage->inline_restr_values;
    }
    return (cimages);
}   /*  End Function contour_create_sequence  */

/*EXPERIMENTAL_FUNCTION*/
KContourImage *contour_create_sequence_from_iarray (iarray array,
						    unsigned int hdim,
						    unsigned int vdim,
						    unsigned int fdim,
						    KContourImageGroup group)
/*  [SUMMARY] Create a sequence of contour images from an Intelligent Array.
    [PURPOSE] This routine will create a sequence of contour image objects
    from a 3-dimensional Intelligent Array. At a later time, this
    sequence of contour images may be used to create a sequence of viewable
    contour images.
    <array> The Intelligent Array. The underlying <<multi_array>> data
    structure will have its attachment count incremented upon successful
    completion.
    <hdim> The dimension index of the horizontal dimension.
    <vdim> The dimension index of the vertical dimension.
    <fdim> The dimension index of the frame dimension (dimension containing the
    sequence). The number of frames is the same as the length of this
    dimension.
    <group> The KContourImageGroup to add the new objects too. If NULL, it is
    not added to any group.
    [RETURNS] A pointer to a dynamically allocated array of contour image
    objects on success, else NULL.
*/
{
    unsigned int fcount;
    KContourImage *cimages;
    dim_desc *hdim_desc, *vdim_desc, *fdim_desc;
    static char function_name[] = "contour_create_sequence_from_iarray";

    if (array == NULL)
    {
	fprintf (stderr, "NULL Intelligent Array passed\n");
	a_prog_bug (function_name);
    }
    if (group) VERIFY_CIMAGE_GROUP (group);
    if (iarray_num_dim (array) != 3)
    {
	fprintf (stderr, "Intelligent Array must have exactly 3 dimensions\n");
	return (NULL);
    }
    if ( hdim >= iarray_num_dim (array) )
    {
	fprintf (stderr, "hdim: %u too large\n", hdim);
	a_prog_bug (function_name);
    }
    if ( vdim >= iarray_num_dim (array) )
    {
	fprintf (stderr, "vdim: %u too large\n", vdim);
	a_prog_bug (function_name);
    }
    if ( fdim >= iarray_num_dim (array) )
    {
	fprintf (stderr, "fdim: %u too large\n", fdim);
	a_prog_bug (function_name);
    }
    hdim_desc = iarray_get_dim_desc (array, hdim);
    vdim_desc = iarray_get_dim_desc (array, vdim);
    fdim_desc = iarray_get_dim_desc (array, fdim);
    if ( (iarray_dim_length (array, hdim) != hdim_desc->length)
	 || (iarray_dim_length (array, vdim) != vdim_desc->length)
	 || (iarray_dim_length (array, fdim) != fdim_desc->length) )
    {
	fprintf (stderr, "%s: Intelligent Array must not be a sub-array\n",
		 function_name);
	return (NULL);
    }
    cimages = contour_create_sequence (array->multi_desc,
				       array->arr_desc, array->data,
				       array->orig_dim_indices[hdim],
				       array->orig_dim_indices[vdim],
				       array->orig_dim_indices[fdim],
				       array->elem_index, group);
    if (!cimages) return (NULL);
    for (fcount = 0; fcount < fdim_desc->length; ++fcount)
	cimages[fcount]->data_change_callback =
	    iarray_register_data_change_func (array,
					      contour_register_data_change,
					      cimages[fcount]);
    return (cimages);
}   /*  End Function contour_create_sequence_from_iarray  */

/*EXPERIMENTAL_FUNCTION*/
void contour_destroy (KContourImage cimage)
/*  [SUMMARY] Destroy contour image.
    [PURPOSE] This routine will destroy a contour image.
    <cimage> The contour image.
    [NOTE] The associated <<multi_array>> descriptor is also deallocated (or
    at least, the attachment count is decreased).
    [RETURNS] Nothing.
*/
{
    KViewableContourImage vcimage, next;
    unsigned int count;
    static char function_name[] = "contour_destroy";

    VERIFY_USED_CIMAGE (cimage);
    for (vcimage = cimage->first_vcimage; vcimage; vcimage = next)
    {
	next = vcimage->cimage_chain.next;
	contour_destroy_viewable (vcimage);
    }
    if (cimage->group) KLIST_REMOVE3 (cimage->group, group_chain, cimage);
    c_unregister_callback2 (cimage->data_change_callback);
    ds_dealloc_multi (cimage->multi_desc);
    _contour_dealloc_levels (cimage);
    if ( cimage->restriction_names &&
	 (cimage->restriction_names != cimage->inline_restr_names) )
    {
	for (count = 0; count < cimage->num_restrictions; ++count)
	    m_free2 (cimage->restriction_names[count]);
	m_free (cimage->restriction_names);
    }
    if ( cimage->restriction_values &&
	 (cimage->restriction_values != cimage->inline_restr_values) )
	m_free (cimage->restriction_values);
    cimage->magic_number = CIMAGE_FREE_MAGIC_NUMBER;
    cimage->next_free = first_free_cimage;
    first_free_cimage = cimage;
}   /*  End Function contour_destroy  */

/*EXPERIMENTAL_FUNCTION*/
KViewableContourImageGroup contour_create_viewable_group ()
/*  [SUMMARY] Create container to hold group of KViewableContourImage objects.
    [RETURNS] The group container on success, else NULL.
*/
{
    KViewableContourImageGroup group;
    static char function_name[] = "contour_create_viewable_group";

    if ( ( group = m_calloc (sizeof *group) ) == NULL )
    {
	m_error_notify (function_name, "vcimage group");
	return (NULL);
    }
    group->magic_number = VCIMAGE_GROUP_MAGIC_NUMBER;
    return (group);
}   /*  End Function contour_create_viewable_group  */

/*EXPERIMENTAL_FUNCTION*/
void contour_destroy_viewable_group (KViewableContourImageGroup group)
/*  [SUMMARY] Destroy a group of KViewableContourImage objects.
    <group> The group to destroy. The group container is also destroyed.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "contour_destroy_viewable_group";

    VERIFY_VCIMAGE_GROUP (group);
    while (group->first) contour_destroy_viewable (group->first);
    group->magic_number = 0;
    m_free (group);
}   /*  End Function contour_destroy_viewable_group  */

/*EXPERIMENTAL_FUNCTION*/
KViewableContourImage contour_create_viewable(KContourImage cimage,
					      KWorldCanvas canvas,
					      KViewableContourImageGroup group)
/*  [SUMMARY] Create viewable contour image.
    [PURPOSE] This routine will create a viewable contour image object from a
    contour image. At a later time, this viewable contour image may be made
    visible. This routine will not cause the canvas to be refreshed.
    <cimage> The contour image.
    <canvas> The world canvas onto which the viewable contour image may be
    drawn. Many viewable contour images may be associated with a single canvas.
    <group> The KViewableContourImageGroup to add the new object too. If NULL,
    it is not added to any group.
    [NOTE] The KViewableContourImage will be automatically destroyed if the
    canvas is destroyed.
    [RETURNS] A KViewableContourImage object on success, NULL.
*/
{
    KViewableContourImage vcimage;
    CanvasHolder holder;
    static char function_name[] = "contour_create_viewable";

    VERIFY_USED_CIMAGE (cimage);
    if (group) VERIFY_VCIMAGE_GROUP (group);
    if (canvas == NULL)
    {
	fprintf (stderr, "NULL world canvas passed\n");
	a_prog_bug (function_name);
    }
    holder = _contour_get_canvas_holder (canvas, function_name);
    if (first_free_vcimage == NULL)
    {
	if ( ( vcimage = m_alloc (sizeof *vcimage) ) == NULL ) return (NULL);
    }
    else
    {
	vcimage = first_free_vcimage;
	first_free_vcimage = vcimage->holder_chain.next;
	VERIFY_FREE_VCIMAGE (vcimage);
    }
    m_clear (vcimage, sizeof *vcimage);
    vcimage->magic_number = VCIMAGE_USED_MAGIC_NUMBER;
    vcimage->canvas_holder = holder;
    vcimage->cimage = cimage;
    vcimage->group = group;
    vcimage->old_x_offset = -1;
    vcimage->old_y_offset = -1;
    vcimage->old_x_pixels = -1;
    vcimage->old_y_pixels = -1;
    vcimage->old_left_x = TOOBIG;
    vcimage->old_right_x = TOOBIG;
    vcimage->old_bottom_y = TOOBIG;
    vcimage->old_top_y = TOOBIG;
    KLIST_APPEND3 (holder, holder_chain, vcimage);
    KLIST_APPEND4 (cimage->first_vcimage, cimage->last_vcimage, cimage_chain,
		   vcimage);
    if (group) KLIST_APPEND3 (group, group_chain, vcimage);
    /*  Do not make the first contour image for this canvas active, else
	the  contour_make_active  routine will not be able to detect a change
	in the active contour image and will do nothing if called with this
	newly created contour image.
	*/
    return (vcimage);
}   /*  End Function contour_create_viewable  */

/*EXPERIMENTAL_FUNCTION*/
KViewableContourImage *contour_create_viewable_sequence
    (KContourImage *cimages, unsigned int num, KWorldCanvas canvas,
     KViewableContourImageGroup group)
/*  [SUMMARY] Create a sequence of viewable contour images.
    [PURPOSE] This routine will create a sequence of viewable contour image
    objects from a sequence of contour images. At a later time, these viewable
    contour images may be made visible. This routine will not cause the canvas
    to be refreshed.
    <cimage> The contour images.
    <num> The number of contour images in the sequence.
    <canvas> The world canvas onto which the viewable contour images may be
    drawn. Many viewable contour images may be associated with a single canvas.
    <group> The KViewableContourImageGroup to add the new objects too. If NULL,
    it is not added to any group.
    [NOTE] The KViewableContourImage objects will be automatically destroyed if
    the canvas is destroyed.
    [RETURNS] A pointer to a dynamically allocated array of viewable contour
    image objects on success, else NULL.
*/
{
    unsigned int count;
    KViewableContourImage *vcimages;
    static char function_name[] = "contour_create_viewable_sequence";

    if ( ( vcimages = m_alloc (sizeof *vcimages * num) ) == NULL )
    {
	m_error_notify (function_name, "array of pointers");
	return (NULL);
    }
    for (count = 0; count < num; ++count)
    {
	if ( ( vcimages[count] =
	       contour_create_viewable (cimages[count], canvas, group) )
	     == NULL )
	{
	    while (count > 0) contour_destroy_viewable (vcimages[--count]);
	    m_free (vcimages);
	    return (NULL);
	}
    }
    return (vcimages);
}   /*  End Function contour_create_viewable_sequence  */

/*EXPERIMENTAL_FUNCTION*/
void contour_destroy_viewable (KViewableContourImage vcimage)
/*  [SUMMARY] Destroy a viewable contour image.
    [PURPOSE] This routine will destroy a viewable contour image.
    <vcimage> The viewable contour image.
    [RETURNS] Nothing.
*/
{
    KContourImage cimage;
    CanvasHolder holder;
    static char function_name[] = "contour_destroy_viewable";

    VERIFY_USED_VCIMAGE (vcimage);
    cimage = vcimage->cimage;
    holder = vcimage->canvas_holder;
    _contour_dealloc_viewable_levels (vcimage);
    KLIST_REMOVE3 (holder, holder_chain, vcimage);
    KLIST_REMOVE4 (cimage->first_vcimage, cimage->last_vcimage, cimage_chain,
		   vcimage);
    if (vcimage->group) KLIST_REMOVE3 (vcimage->group, group_chain, vcimage);
    vcimage->magic_number = VCIMAGE_FREE_MAGIC_NUMBER;
    vcimage->holder_chain.next = first_free_vcimage;
    first_free_vcimage = vcimage;
}   /*  End Function contour_destroy_viewable  */

/*EXPERIMENTAL_FUNCTION*/
flag contour_set_active (KViewableContourImage vcimage, flag active,
			 flag force_refresh, flag refresh_if_changed)
/*  [SUMMARY] Set active state for a viewable contour image.
    [PURPOSE] This routine will make a viewable contour image active or
    inactive.
    <vcimage> The viewable contour image.
    <active> If TRUE, the viewable contour image is made active, else it is
    made inactive.
    <force_refresh> If TRUE, the canvas is always refreshed.
    <refresh_if_changed> If TRUE, the canvas is refreshed if the viewable
    contour image active state changed.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    CanvasHolder holder;
    static char function_name[] = "contour_set_active";

    VERIFY_USED_VCIMAGE (vcimage);
    FLAG_VERIFY (active);
    FLAG_VERIFY (force_refresh);
    FLAG_VERIFY (refresh_if_changed);
    holder = vcimage->canvas_holder;
    if (active != vcimage->active)
    {
	vcimage->active = active;
	if (refresh_if_changed) force_refresh = TRUE;
    }
    if (force_refresh) return canvas_resize (holder->canvas, NULL, FALSE);
    return (TRUE);
}   /*  End Function contour_set_active  */

/*EXPERIMENTAL_FUNCTION*/
flag contour_set_all_inactive (KViewableContourImageGroup group,
			       flag refresh_if_changed)
/*  [SUMMARY] Clear active state for all viewable contour images in a group.
    [PURPOSE] This routine will make all viewable contour images in a group
    inactive.
    <group> The group of viewable contour images.
    <refresh_if_changed> If TRUE, canvases are refreshed if the active state
    of any viewable contour image changed.
    <key> The key to use for inactivation of viewable contour images. If this
    is 0 then all viewable contour images will be made inactive.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    KViewableContourImage curr;
    static char function_name[] = "contour_set_all_inactive";

    VERIFY_VCIMAGE_GROUP (group);
    FLAG_VERIFY (refresh_if_changed);
    /*  Make all inactive  */
    for (curr = group->first; curr; curr = curr->group_chain.next)
    {
	if (refresh_if_changed && curr->active)
	    curr->canvas_holder->needs_refresh = TRUE;
	curr->active = FALSE;
    }
    if (!refresh_if_changed) return (TRUE);
    for (curr = group->first; curr; curr = curr->group_chain.next)
    {
	if (curr->canvas_holder->needs_refresh)
	{
	    canvas_resize (curr->canvas_holder->canvas, NULL, FALSE);
	    curr->canvas_holder->needs_refresh = FALSE;
	}
    }
    return (TRUE);
}   /*  End Function contour_set_all_inactive  */

/*EXPERIMENTAL_FUNCTION*/
flag contour_register_data_change (KContourImage cimage)
/*  [SUMMARY] Notify data for contour image has changed.
    [PURPOSE] This routine will register a change in the Karma data structure
    associated with a contour image. If the contour image is active, it will
    be immediately redrawn on its canvas.
    <cimage> The contour image.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    KViewableContourImage vcimage;
    struct linearsegments_cache_struct *linsegs_cache;
    static char function_name[] = "contour_register_data_change";

    VERIFY_USED_CIMAGE (cimage);
    cimage->world_segs_valid = FALSE;
    for (linsegs_cache = cimage->first_linsegs_cache; linsegs_cache;
	 linsegs_cache = linsegs_cache->next)
	linsegs_cache->valid = FALSE;
    for (vcimage = cimage->first_vcimage; vcimage;
	 vcimage = vcimage->cimage_chain.next)
    {
	CanvasHolder holder = vcimage->canvas_holder;

	vcimage->pixel_segments_valid = FALSE;
	if (vcimage->active)  /*  Active image: refresh  */
	    canvas_resize (holder->canvas, NULL, FALSE);
    }
    return (TRUE);
}   /*  End Function contour_register_data_change  */

/*EXPERIMENTAL_FUNCTION*/
void contour_set_canvas_attributes (KWorldCanvas canvas, ...)
/*  [SUMMARY] Set the contour image attributes for a world canvas.
    <canvas> The world canvas.
    [VARARGS] The list of parameter attribute-key attribute-value-ptr pairs
    must follow. This list must be terminated with the CONTOUR_CANVAS_ATT_END.
    See [<contour_CANVAS_ATTRIBUTES>] for the list of attributes.
    [NOTE] The canvas is not refreshed by this operation.
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder;
    va_list argp;
    static char function_name[] = "contour_set_canvas_attributes";

    if (canvas == NULL)
    {
	fprintf (stderr, "NULL canvas passed\n");
	a_prog_bug (function_name);
    }
    /*  Dummy operation to ensure canvas is OK  */
    canvas_get_attributes (canvas, CANVAS_ATT_END);
    holder = _contour_get_canvas_holder (canvas, function_name);
    va_start (argp, canvas);
    _contour_process_canvas_attributes (holder, argp);
    va_end (argp);
}   /*  End Function contour_set_canvas_attributes  */

/*EXPERIMENTAL_FUNCTION*/
void contour_set_levels (KContourImage cimage, unsigned int num_levels,
			 CONST double *contour_levels)
/*  [SUMMARY] Set/update the contour levels for a KContourImage object.
    [PURPOSE] This routine will set/update the contour levels for a
    KContourImage object. Canvases are not refreshed by this operation.
    <cimage> The KContourImage object.
    <num_levels> The number of contour levels.
    <contour_levels> The array of contour levels. These are copied.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "contour_set_levels";

    VERIFY_USED_CIMAGE (cimage);
    if (num_levels == cimage->num_levels)
    {
	if ( (contour_levels == NULL) || (cimage->contour_levels == NULL) ||
	     !m_cmp (cimage->contour_levels, contour_levels,
		     sizeof *contour_levels * num_levels) )
	{
	    KViewableContourImage vcimage;
	    struct linearsegments_cache_struct *linsegs_cache;

	    cimage->world_segs_valid = FALSE;
	    for (linsegs_cache = cimage->first_linsegs_cache; linsegs_cache;
		 linsegs_cache = linsegs_cache->next)
		linsegs_cache->valid = FALSE;
	    for (vcimage = cimage->first_vcimage; vcimage;
		 vcimage = vcimage->cimage_chain.next)
		vcimage->pixel_segments_valid = FALSE;
	}
	m_copy (cimage->contour_levels, contour_levels,
		sizeof *contour_levels * num_levels);
    }
    else
    {
	_contour_dealloc_levels (cimage);
	if ( ( cimage->contour_levels =
	       m_dup (contour_levels, sizeof *contour_levels * num_levels) )
	     == NULL )
	    m_abort (function_name, "contour levels");
	if ( ( cimage->worldsegs =
	       m_calloc (sizeof *cimage->worldsegs * num_levels) ) == NULL )
	    m_abort (function_name, "contour level structures");
	cimage->num_levels = num_levels;
    }
}   /*  End Function contour_set_levels  */

/*EXPERIMENTAL_FUNCTION*/
void contour_set_group_levels (KContourImageGroup group,
			       unsigned int num_levels,
			       CONST double *contour_levels)
/*  [SUMMARY] Set/update contour levels for a group of KContourImage objects.
    [PURPOSE] This routine will set/update the contour levels for a group of
    KContourImage objects. Canvases are not refreshed by this operation.
    <group> The group of KContourImage objects.
    <num_levels> The number of contour levels.
    <contour_levels> The array of contour levels. These are copied.
    [RETURNS] Nothing.
*/
{
    KContourImage cimage;
    static char function_name[] = "contour_set_group_levels";

    VERIFY_CIMAGE_GROUP (group);
    for (cimage = group->first; cimage; cimage = cimage->group_chain.next)
	contour_set_levels (cimage, num_levels, contour_levels);
}   /*  End Function contour_set_group_levels  */

/*EXPERIMENTAL_FUNCTION*/
void contour_set_level_styles (KViewableContourImage vcimage,
			       CONST unsigned long *contour_pixels,
			       CONST flag *dash,
			       CONST double *linewidths)
/*  [SUMMARY] Set/update contour level styles for KViewableContourImage object.
    [PURPOSE] This routine will set/update the contour level tyles for a
    KViewableContourImage object. No canvas is refreshed by this operation.
    <vcimage> The KViewableContourImage object.
    <contour_pixels> The array of contour pixels. These are copied. This may be
    NULL.
    <dash> The array of dash flags. This may be NULL.
    <linewidths> The array of line widths. This may be NULL.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    static char function_name[] = "contour_set_level_styles";

    VERIFY_USED_VCIMAGE (vcimage);
    if (vcimage->cimage->num_levels < 1)
    {
	fprintf (stderr, "No contour levels set yet\n");
	a_prog_bug (function_name);
    }
    if (!vcimage->levels)
    {
	if ( ( vcimage->levels =
	       m_calloc (sizeof *vcimage->levels *
			 vcimage->cimage->num_levels) )
	     == NULL )
	    m_abort (function_name, "contour level structures");
    }
    for (count = 0; count < vcimage->cimage->num_levels; ++count)
	vcimage->levels[count].pix_limits.xmin[0] = TOOBIG;
    for (count = 0; count < vcimage->cimage->num_levels; ++count)
    {
	if (contour_pixels == NULL)
	    vcimage->levels[count].pixel_defined = FALSE;
	else
	{
	    vcimage->levels[count].pixel_value = contour_pixels[count];
	    vcimage->levels[count].pixel_defined = TRUE;
	}
	if (dash == NULL) vcimage->levels[count].dash = FALSE;
	else vcimage->levels[count].dash = dash[count];
	if (linewidths == NULL) vcimage->levels[count].linewidth = 0.0;
	else vcimage->levels[count].linewidth = linewidths[count];
    }
}   /*  End Function contour_set_level_styles  */

/*EXPERIMENTAL_FUNCTION*/
void contour_set_group_level_styles (KViewableContourImageGroup group,
				     CONST unsigned long *contour_pixels,
				     CONST flag *dash,
				     CONST double *linewidths)
/*  [SUMMARY] Set/update clevel styles for KViewableContourImage objects.
    [PURPOSE] This routine will set/update the contour level styles for a group
    of KViewableContourImage objects. No canvas is refreshed by this operation.
    <group> The group of KViewableContourImage objects.
    <contour_pixels> The array of contour pixels. These are copied. This may be
    NULL.
    <dash> The array of dash flags. This may be NULL.
    <linewidths> The array of line widths. This may be NULL.
    [RETURNS] Nothing.
*/
{
    KViewableContourImage vcimage;
    static char function_name[] = "contour_set_group_level_styles";

    VERIFY_VCIMAGE_GROUP (group);
    for (vcimage = group->first; vcimage; vcimage = vcimage->group_chain.next)
	contour_set_level_styles (vcimage, contour_pixels, dash, linewidths);
}   /*  End Function contour_set_group_level_styles  */

/*EXPERIMENTAL_FUNCTION*/
KWorldCanvas contour_get_worldcanvas (KViewableContourImage vcimage)
/*  [SUMMARY] Get the world canvas for a viewable contour image.
    <vcimage> The KViewableContourImage object.
    [RETURNS] The KWorldCanvas object.
*/
{
    static char function_name[] = "contour_get_worldcanvas";

    VERIFY_USED_VCIMAGE (vcimage);
    return (vcimage->canvas_holder->canvas);
}   /*  End Function contour_get_worldcanvas  */


/*  Private functions follow  */

static KContourImage _contour_create_restr (multi_array *multi_desc,
					    flag try_ap,
					    array_desc *arr_desc, char *slice,
					    unsigned int hdim,
					    unsigned int vdim,
					    unsigned int elem_index,
					    unsigned int num_restr,
					    char **restr_names,
					    double *restr_values,
					    KContourImageGroup group)
/*  [SUMMARY] Create contour image from 2D slice with restrictions.
    [PURPOSE] This routine will create a contour image object from a
    2-dimensional slice of a Karma data structure. At a later time, this
    contour image may be used to create a viewable contour image.
    <multi_desc> The  multi_array  descriptor which contains the Karma data
    structure. The routine increments the attachment count on the descriptor
    on successful completion. This may be NULL.
    <try_ap> If TRUE, a KwcsAstro object is created.
    <arr_desc> The array descriptor.
    <slice> The start of the slice data.
    <hdim> The dimension index of the horizontal dimension.
    <vdim> The dimension index of the vertical dimension.
    <elem_index> The element index in the data packets.
    <num_restr> The number of matched restrictions.
    <restr_names> The restriction names.
    <restr_values> The restriction values.
    <group> The KContourImageGroup to add the new object too. If NULL, it is
    not added to any group.
    [NOTE] Restriction information is automatically deallocated when
    [<contour_destroy>] is called.
    [RETURNS] A KContourImage object on success, NULL.
*/
{
    KContourImage cimage;
    unsigned int index;
    static char function_name[] = "_contour_create_restr";

    if (first_free_cimage == NULL)
    {
	if ( ( cimage = m_alloc (sizeof *cimage) ) == NULL ) return (NULL);
    }
    else
    {
	cimage = first_free_cimage;
	first_free_cimage = cimage->next_free;
	VERIFY_FREE_CIMAGE (cimage);
    }
    m_clear (cimage, sizeof *cimage);
    cimage->magic_number = CIMAGE_USED_MAGIC_NUMBER;
    cimage->multi_desc = multi_desc;
    cimage->group = group;
    cimage->arr_desc = arr_desc;
    cimage->slice = slice + ds_get_element_offset (arr_desc->packet,
						   elem_index);
    cimage->hdim = hdim;
    cimage->vdim = vdim;
    cimage->elem_index = elem_index;
    if (try_ap && multi_desc)
    {
	index = _contour_get_gen_index (multi_desc, arr_desc);
	cimage->astro_projection = wcs_astro_get_associated (multi_desc,index);
    }
    cimage->num_restrictions = num_restr;
    cimage->restriction_names = restr_names;
    cimage->restriction_values = restr_values;
    if (group) KLIST_APPEND3 (group, group_chain, cimage);
    /*  Attach  */
    if (multi_desc) ++multi_desc->attachments;
    return (cimage);
}   /*  End Function _contour_create_restr  */

static unsigned int _contour_get_gen_index (CONST multi_array *multi_desc,
					    CONST array_desc *arr_desc)
/*  [SUMMARY] Find general data structure which has an array descriptor.
    <multi_desc> The multi_array header.
    <arr_desc> The array descriptor to search for.
    [RETURNS] The index of the general data structure in the multi_array header
*/
{
    unsigned int count;
    char *encls_desc;
    static char function_name[] = "_contour_get_gen_index";

    for (count = 0; count < multi_desc->num_arrays; ++count)
    {
	switch ( ds_f_name_in_packet (multi_desc->headers[count],
				      arr_desc->dimensions[0]->name,
				      &encls_desc, NULL) )
	{
	  case IDENT_NOT_FOUND:
	    break;
	  case IDENT_DIMENSION:
	    if (arr_desc == (array_desc *) encls_desc) return (count);
	    break;
	  default:
	    fprintf (stderr, "Illegal return from <ds_f_name_in_packet>\n");
	    a_prog_bug (function_name);
	    break;
	}
    }
    fprintf (stderr, "Array descriptor not found\n");
    a_prog_bug (function_name);
    return (0);
}   /*  End Function get_gen_index  */

static CanvasHolder _contour_get_canvas_holder (KWorldCanvas canvas,
						char *func_name)
/*  [SUMMARY] Get the canvas holder for a world canvas.
    <canvas> The canvas.
    <func_name> The name of the calling function. If this is not NULL, the
    process will abort if a canvas holder is not found.
    [RETURNS] The canvas holder on success, else NULL.
*/
{
    CanvasHolder canvas_holder;

    /*  First search for existing canvas holder  */
    for (canvas_holder = first_canvas_holder; canvas_holder;
	 canvas_holder = canvas_holder->next)
    {
	if (canvas == canvas_holder->canvas) return (canvas_holder);
    }
    if (func_name)
    {
	fprintf (stderr, "CanvasHolder not found!\n");
	a_prog_bug (func_name);
    }
    return (NULL);
}   /*  End Function get_canvas_holder  */

static CanvasHolder _contour_alloc_canvas_holder (KWorldCanvas canvas)
/*  This routine will allocate a canvas holder for a world canvas.
    The canvas must be given by  canvas  .
    The routine returns the canvas holder on success, else it returns NULL.
*/
{
    CanvasHolder holder;
    static char function_name[] = "_contour_alloc_canvas_holder";

    /*  Must allocate holder  */
    if ( ( holder = m_calloc (sizeof *holder) ) == NULL )
    {
	m_error_notify (function_name, "canvas holder");
	return (NULL);
    }
    holder->magic_number = HOLDER_MAGIC_NUMBER;
    holder->canvas = canvas;
    KLIST_APPEND2 (first_canvas_holder, last_canvas_holder, holder);
    canvas_register_precompute_func (canvas,
				     _contour_worldcanvas_precompute_func,
				     holder);
    canvas_register_refresh_func (canvas, _contour_worldcanvas_refresh_func,
				  holder);
    canvas_register_destroy_func (canvas, FALSE, _contour_free_canvas_holder,
				  holder);
    return (holder);
}   /*  End Function alloc_canvas_holder  */

static void _contour_free_canvas_holder (CanvasHolder holder)
/*  [SUMMARY] Free a canvas holder (and any associated KContourImage objects).
    <holder> The canvas holder.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "_contour_free_canvas_holder";

    VERIFY_HOLDER (holder);
    while (holder->first) contour_destroy_viewable (holder->first);
    KLIST_REMOVE2 (first_canvas_holder, last_canvas_holder, holder);
    m_free2 (holder->colourname);
    holder->magic_number = 0;
    m_free (holder);
}   /*  End Function _contour_free_canvas_holder  */

static void _contour_process_canvas_attributes (CanvasHolder holder,
						va_list argp)
/*  [PURPOSE] This routine will set the attributes for a canvas.
    <holder> The canvas holder.
    <argp> The optional list of parameter attribute-key attribute-value
    pairs.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int att_key;
    char *colour;
    static char function_name[] = "_contour_process_canvas_attributes";

    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    CONTOUR_CANVAS_ATT_END )
    {
	switch (att_key)
	{
	  case CONTOUR_CANVAS_ATT_COLOURNAME:
	    colour = va_arg (argp, char *);
	    if (colour == NULL) colour = st_dup (DEFAULT_COLOUR);
	    else colour = st_dup (colour);
	    if (colour == NULL) m_abort (function_name, "colour name");
	    m_free2 (holder->colourname);
	    holder->colourname = colour;
	    break;
	  default:
	    fprintf (stderr, "Illegal attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	}
    }
}   /*  End Function _contour_process_canvas_attributes  */

static void _contour_worldcanvas_precompute_func
    (KWorldCanvas canvas, int width, int height,
     CONST struct win_scale_type *win_scale, Kcolourmap cmap, flag cmap_resize,
     void **info, PostScriptPage pspage,
     unsigned int num_areas, KPixCanvasRefreshArea *areas)
/*  [PURPOSE] This routine processes a precompute event for a world canvas.
    <canvas> The world canvas.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The canvas colourmap.
    <cmap_resize> If TRUE, the refresh is due to a colourmap resize.
    <info> A pointer to the arbitrary information pointer.
    <pspage> The PostScriptPage object. If this is NULL, the refresh is *not*
    destined for a PostScript page.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder = *info;
    KViewableContourImage vcimage;
    static char function_name[] = "_contour_worldcanvas_precompute_func";

    VERIFY_HOLDER (holder);
    if (canvas != holder->canvas)
    {
	fprintf (stderr, "Different canvas in canvas holder object\n");
	a_prog_bug (function_name);
    }
    /*  Show each active image  */
    for (vcimage = holder->first;vcimage; vcimage = vcimage->holder_chain.next)
    {
	if (vcimage->active) _contour_precompute_image (vcimage, win_scale,
							num_areas, areas);
    }
}   /*  End Function _contour_worldcanvas_precompute_func  */

static void _contour_precompute_image (KViewableContourImage vcimage,
				       CONST struct win_scale_type *win_scale,
				       unsigned int num_areas,
				       KPixCanvasRefreshArea *areas)
/*  [PURPOSE] This routine processes a precompute event for a world canvas.
    <vcimage> The KViewableContourImage object.
    <win_scale> A pointer to the window scaling information.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    [RETURNS] Nothing.
*/
{
    KContourImage cimage = vcimage->cimage;
    KwcsAstro ap;
    unsigned int count;
    CONST packet_desc *pack_desc;
    CONST uaddr *hoffsets, *voffsets;
    CONST dim_desc *hdim, *vdim;
    /*static char function_name[] = "_contour_precompute_image";*/

    if (cimage->num_levels < 1) return;
    pack_desc = cimage->arr_desc->packet;
    hdim = cimage->arr_desc->dimensions[cimage->hdim];
    vdim = cimage->arr_desc->dimensions[cimage->vdim];
    hoffsets = cimage->arr_desc->offsets[cimage->hdim];
    voffsets = cimage->arr_desc->offsets[cimage->vdim];
    ap = cimage->astro_projection;
    if (!cimage->world_segs_valid)
    {
	/*  Compute segments  */
	if ( !ds_contour (cimage->slice,
			  pack_desc->element_types[cimage->elem_index],
			  hdim, hoffsets, vdim, voffsets,
			  cimage->num_levels, cimage->contour_levels,
			  sizeof *cimage->worldsegs,
			  &cimage->worldsegs[0].coord_buf_size,
			  &cimage->worldsegs[0].x0, &cimage->worldsegs[0].y0,
			  &cimage->worldsegs[0].x1, &cimage->worldsegs[0].y1,
			  &cimage->worldsegs[0].num_segments) ) return;
	/*  Convert from data pixel to world co-ordinates  */
	for (count = 0; count < cimage->num_levels; ++count)
	{
	    if (ap == NULL) break;
	    wcs_astro_transform3 (ap, TRUE,
				  cimage->worldsegs[count].num_segments,
				  hdim->name,cimage->worldsegs[count].x0,FALSE,
				  vdim->name,cimage->worldsegs[count].y0,FALSE,
				  NULL, NULL, FALSE,
				  cimage->num_restrictions,
				  (CONST char **) cimage->restriction_names,
				  cimage->restriction_values);
	    wcs_astro_transform3 (ap, TRUE,
				  cimage->worldsegs[count].num_segments,
				  hdim->name,cimage->worldsegs[count].x1,FALSE,
				  vdim->name,cimage->worldsegs[count].y1,FALSE,
				  NULL, NULL, FALSE,
				  cimage->num_restrictions,
				  (CONST char **) cimage->restriction_names,
				  cimage->restriction_values);
	}
	cimage->world_segs_valid = TRUE;
    }
    if (vcimage->pixel_segments_valid)
    {
	if ( (vcimage->old_x_offset != win_scale->x_offset)
	     || (vcimage->old_y_offset != win_scale->y_offset)
	     || (vcimage->old_x_pixels != win_scale->x_pixels)
	     || (vcimage->old_y_pixels != win_scale->y_pixels)
	     || (vcimage->old_left_x != win_scale->left_x)
	     || (vcimage->old_right_x != win_scale->right_x)
	     || (vcimage->old_bottom_y != win_scale->bottom_y)
	     || (vcimage->old_top_y != win_scale->top_y) )
	{   /*  The linear->pixel transformation has changed  */
	    vcimage->pixel_segments_valid = FALSE;
	}
    }
    _contour_find_linsegs_cache (vcimage, win_scale);
    /*  Now go ahead and process each level separately  */
    for (count = 0; count < cimage->num_levels; ++count)
	_contour_precompute_level (vcimage, count, win_scale, num_areas,areas);
    vcimage->linsegs_cache->valid = TRUE;
    vcimage->pixel_segments_valid = TRUE;
    vcimage->old_x_offset = win_scale->x_offset;
    vcimage->old_y_offset = win_scale->y_offset;
    vcimage->old_x_pixels = win_scale->x_pixels;
    vcimage->old_y_pixels = win_scale->y_pixels;
    vcimage->old_left_x = win_scale->left_x;
    vcimage->old_right_x = win_scale->right_x;
    vcimage->old_bottom_y = win_scale->bottom_y;
    vcimage->old_top_y = win_scale->top_y;
}   /*  End Function _contour_precompute_image  */

static void _contour_precompute_level (KViewableContourImage vcimage,
				       unsigned int level_index,
				       CONST struct win_scale_type *win_scale,
				       unsigned int num_areas,
				       KPixCanvasRefreshArea *areas)
/*  [PURPOSE] This routine will precompute a single level.
    <vcimage> The KViewableContourImage object.
    <level_index> The contour level to draw.
    <win_scale> A pointer to the window scaling information.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <same_trans> If TRUE the world->linear co-ordinate transformation has not
    changed since the last refresh.
    <same_world> If TRUE the linear->pixel co-ordinate transformation has not
    changed since the last refresh.
    [RETURNS] Nothing.
*/
{
    KContourImage cimage = vcimage->cimage;
    CanvasHolder holder = vcimage->canvas_holder;
    KWorldCanvas canvas = holder->canvas;
    flag section_visible, segment_visible;
    unsigned int count, num_pix_segs, area_count;
    unsigned int section_count, subsection_count;
    unsigned int pix_section_size, num_segments;
    double win_x_min, win_y_min, win_x_max, win_y_max, tmp;
    double area_startx, area_endx, area_starty, area_endy;
    double x0, y0, x1, y1;
    double offset = 1.0;
    CONST struct worldsegments_level_struct *worldsegs_level =
	&cimage->worldsegs[level_index];
    struct linearsegments_level_struct *linsegs_level =
	&vcimage->linsegs_cache->levels[level_index];
    struct viewable_levelinfo_type *vc_level =
	&vcimage->levels[level_index];
    static char function_name[] = "_contour_precompute_level";

    num_segments = worldsegs_level->num_segments;
    if ( (num_segments > linsegs_level->coord_buf_size) ||
	 (linsegs_level->coord_buf_size > worldsegs_level->coord_buf_size) )
    {   /*  Reallocate linear co-ordinate buffers  */
	m_free2 (linsegs_level->x0);
	m_free2 (linsegs_level->y0);
	m_free2 (linsegs_level->x1);
	m_free2 (linsegs_level->y1);
	if ( ( linsegs_level->x0 =
	       m_alloc (sizeof *linsegs_level->x0 * num_segments) ) == NULL )
	    m_abort (function_name, "linear x0 array");
	if ( ( linsegs_level->y0 =
	       m_alloc (sizeof *linsegs_level->y0 * num_segments) ) == NULL )
	    m_abort (function_name, "linear y0 array");
	if ( ( linsegs_level->x1 =
	       m_alloc (sizeof *linsegs_level->x1 * num_segments) ) == NULL )
	    m_abort (function_name, "linear x1 array");
	if ( ( linsegs_level->y1 =
	       m_alloc (sizeof *linsegs_level->y1 * num_segments) ) == NULL )
	    m_abort (function_name, "linear y1 array");
	linsegs_level->coord_buf_size = num_segments;
    }
    if (!vcimage->linsegs_cache->valid)
    {   /*  Have to compute linear co-ordinates from world co-ordinates  */
	for (count = 0; count < num_segments; ++count)
	    linsegs_level->x0[count] = worldsegs_level->x0[count];
	for (count = 0; count < num_segments; ++count)
	    linsegs_level->y0[count] = worldsegs_level->y0[count];
	canvas_coords_transform (canvas, num_segments,
				 linsegs_level->x0, TRUE,
				 linsegs_level->y0, TRUE);
	for (count = 0; count < num_segments; ++count)
	    linsegs_level->x1[count] = worldsegs_level->x1[count];
	for (count = 0; count < num_segments; ++count)
	    linsegs_level->y1[count] = worldsegs_level->y1[count];
	canvas_coords_transform (canvas, num_segments,
				 linsegs_level->x1, TRUE,
				 linsegs_level->y1, TRUE);
    }
    vc_level->num_area_segments = 0;
    /*  Find edge of canvas in linear co-ordinates  */
    win_x_min = win_scale->x_offset;
    win_y_min = win_scale->y_offset;
    win_x_max = win_scale->x_offset + win_scale->x_pixels - 1;
    win_y_max = win_scale->y_offset + win_scale->y_pixels - 1;
    canvas_convert_to_canvas_coords (canvas, FALSE, 1,
				     &win_x_min, &win_y_min,
				     &win_x_min, &win_y_min,
				     NULL, NULL);
    canvas_convert_to_canvas_coords (canvas, FALSE, 1,
				     &win_x_max, &win_y_max,
				     &win_x_max, &win_y_max,
				     NULL, NULL);
    if (win_x_min > win_x_max) swap (win_x_min, win_x_max, tmp);
    if (win_y_min > win_y_max) swap (win_y_min, win_y_max, tmp);
    if ( (num_segments > vc_level->coord_buf_size) ||
	 (vc_level->coord_buf_size > worldsegs_level->coord_buf_size) )
    {
	/*  Reallocate pixel co-ordinate buffers  */
	m_free2 (vc_level->pix_x0);
	m_free2 (vc_level->pix_y0);
	m_free2 (vc_level->pix_x1);
	m_free2 (vc_level->pix_y1);
	if ( ( vc_level->pix_x0 =
	       m_alloc (sizeof (double) * num_segments) ) == NULL )
	    m_abort (function_name, "pixel x0 array");
	if ( ( vc_level->pix_y0 =
	       m_alloc (sizeof (double) * num_segments) ) == NULL )
	    m_abort (function_name, "pixel y0 array");
	if ( ( vc_level->pix_x1 =
	       m_alloc (sizeof (double) * num_segments) ) == NULL )
	    m_abort (function_name, "pixel x1 array");
	if ( ( vc_level->pix_y1 =
	       m_alloc (sizeof (double) * num_segments) ) == NULL )
	    m_abort (function_name, "pixel y1 array");
	vc_level->coord_buf_size = num_segments;
    }
    if (!vcimage->linsegs_cache->valid)
    {
	_contour_compute_lin_limits (vcimage, level_index, win_scale,
				     win_x_min, win_y_min,
				     win_x_max, win_y_max);
	/*  The pixel segments have now also been computed  */
    }
    else if (!vcimage->pixel_segments_valid)
    {
	_contour_compute_pix_segs (vcimage, level_index, win_scale,
				   win_x_min, win_y_min, win_x_max, win_y_max);
    }
    /* Special case for areas which are used for refreshing the entire canvas*/
    if ( (num_areas == 5)
	 && (areas[0].endy < win_scale->y_offset)
	 && ( (areas[1].starty >= win_scale->y_offset + win_scale->y_pixels)
	      || (areas[1].endy < areas[1].starty) )
	 && (areas[2].endx < win_scale->x_offset)
	 && ( (areas[3].startx >= win_scale->x_offset + win_scale->x_pixels)
	      || (areas[3].endx < areas[3].startx) )
	 && (areas[4].startx == win_scale->x_offset)
	 && (areas[4].endx == win_scale->x_offset + win_scale->x_pixels - 1)
	 && (areas[4].starty == win_scale->y_offset)
	 && (areas[4].endy == win_scale->y_offset + win_scale->y_pixels - 1) )
	num_areas = 0;
    if ( (num_areas < 1) || (vc_level->num_pixel_segments < NUM_SECTIONS) )
    {
	/*  Either whole canvas is to be refreshed (hence all cached pixel
	    segments should be drawn) or there are not enough segments to be
	    sectioned, so they are all drawn too  */
	return;
    }
    /*  Several refresh areas requested: draw only those areas  */
    if ( ( vc_level->pix_area_x0 =
	   m_alloc_no_checks (sizeof *vc_level->pix_area_x0 *
			      vc_level->num_pixel_segments * 4) )
	 == NULL )
	m_abort (function_name, "pixel area array");
    vc_level->pix_area_y0 = vc_level->pix_area_x0+vc_level->num_pixel_segments;
    vc_level->pix_area_x1 = vc_level->pix_area_y0+vc_level->num_pixel_segments;
    vc_level->pix_area_y1 = vc_level->pix_area_x1+vc_level->num_pixel_segments;
    pix_section_size = vc_level->num_pixel_segments / NUM_SECTIONS;
    for (count = 0, section_count = 0, subsection_count = 0, num_pix_segs = 0;
	 count < vc_level->num_pixel_segments; ++count, ++subsection_count)
    {
	if (subsection_count >= pix_section_size)
	{
	    if (++section_count < NUM_SECTIONS) subsection_count = 0;
	    else
	    {
		subsection_count = 0;  /* Pretend last section not processed */
		section_count = NUM_SECTIONS - 1;
	    }
	}
	if ( (subsection_count == 0) &&
	     (vc_level->num_pixel_segments >= NUM_SECTIONS) )
	{
	    /*  First time in this section: use the limits to quickly check
		if any segments should be drawn  */
	    section_visible = FALSE;
	    for (area_count = 0; area_count < num_areas; ++area_count)
	    {
		area_startx = (double) areas[area_count].startx - offset;
		area_endx = (double) areas[area_count].endx + offset;
		area_starty = (double) areas[area_count].starty - offset;
		area_endy = (double) areas[area_count].endy + offset;
		if ( (vc_level->pix_limits.xmin[section_count] < area_startx)
		     &&
		     (vc_level->pix_limits.xmax[section_count] < area_startx) )
		    continue;
		if ( (vc_level->pix_limits.xmin[section_count] > area_endx)
		     &&
		     (vc_level->pix_limits.xmax[section_count] > area_endx) )
		    continue;
		if ( (vc_level->pix_limits.ymin[section_count] < area_starty)
		     &&
		     (vc_level->pix_limits.ymax[section_count] < area_starty) )
		    continue;
		if ( (vc_level->pix_limits.ymin[section_count] > area_endy) &&
		     (vc_level->pix_limits.ymax[section_count] > area_endy) )
		    continue;
		/*  Section is visible in this area  */
		section_visible = TRUE;
		area_count = num_areas;
	    }
	    if (!section_visible)
	    {
		count += pix_section_size - 1;
		subsection_count += pix_section_size - 1;
		continue;
	    }
	}
	/*  Segment is in a visible section: test if segment itself is
	    visible  */
	x0 = vc_level->pix_x0[count];
	y0 = vc_level->pix_y0[count];
	x1 = vc_level->pix_x1[count];
	y1 = vc_level->pix_y1[count];
	segment_visible = FALSE;
	for (area_count = 0; area_count < num_areas; ++area_count)
	{
	    area_startx = (double) areas[area_count].startx - offset;
	    area_endx = (double) areas[area_count].endx + offset;
	    area_starty = (double) areas[area_count].starty - offset;
	    area_endy = (double) areas[area_count].endy + offset;
	    if ( (x0 < area_startx) && (x1 < area_startx) ) continue;
	    if ( (x0 > area_endx) && (x1 > area_endx) ) continue;
	    if ( (y0 < area_starty) && (y1 < area_starty) ) continue;
	    if ( (y0 > area_endy) && (y1 > area_endy) ) continue;
	    /*  Segment is visible in this area  */
	    segment_visible = TRUE;
	    area_count = num_areas;
	}
	if (segment_visible)
	{
	    vc_level->pix_area_x0[num_pix_segs] = x0;
	    vc_level->pix_area_y0[num_pix_segs] = y0;
	    vc_level->pix_area_x1[num_pix_segs] = x1;
	    vc_level->pix_area_y1[num_pix_segs] = y1;
	    ++num_pix_segs;
	}
    }
    vc_level->num_area_segments = num_pix_segs;
}   /*  End Function _contour_precompute_level  */

static void _contour_compute_lin_limits(KViewableContourImage vcimage,
					unsigned int level_index,
					CONST struct win_scale_type *win_scale,
					double win_x_min, double win_y_min,
					double win_x_max, double win_y_max)
/*  [SUMMARY] Compute linear world canvas segments.
    <vcimage> The KViewableContourImage object.
    <level_index> The contour level to draw.
    <win_x_min> The lower horizontal linear co-orinate.
    <win_x_max> The upper horizontal linear co-orinate.
    <win_y_min> The lower vertical linear co-orinate.
    <win_y_max> The upper vertical linear co-orinate.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas canvas = vcimage->canvas_holder->canvas;
    unsigned int count, num_pix_segs;
    unsigned int section_count, subsection_count, lin_section_size;
    unsigned int num_segments =
	vcimage->cimage->worldsegs[level_index].num_segments;
    double sec_xmin, sec_ymin, sec_xmax, sec_ymax, x0, y0, x1, y1;
    struct linearsegments_level_struct *linsegs_level =
	&vcimage->linsegs_cache->levels[level_index];
    struct viewable_levelinfo_type *vc_level = &vcimage->levels[level_index];

    if (num_segments < NUM_SECTIONS)
    {
	/*  Not enough segments: the sectioning algorithm will hang  */
	/*  Convert linear world segments to pixel segments  */
	for (count = 0, num_pix_segs = 0; count < num_segments; ++count)
	{
	    x0 = linsegs_level->x0[count];
	    y0 = linsegs_level->y0[count];
	    x1 = linsegs_level->x1[count];
	    y1 = linsegs_level->y1[count];
	    /*  Add segment to pixel list if inside world canvas  */
	    if ( (x0 < win_x_min) || (x0 > win_x_max) ||
		 (x1 < win_x_min) || (x1 > win_x_max) ||
		 (y0 < win_y_min) || (y0 > win_y_max) ||
		 (y1 < win_y_min) || (y1 > win_y_max) ) continue;
	    vc_level->pix_x0[num_pix_segs] = x0;
	    vc_level->pix_y0[num_pix_segs] = y0;
	    vc_level->pix_x1[num_pix_segs] = x1;
	    vc_level->pix_y1[num_pix_segs] = y1;
	    ++num_pix_segs;
	}
	vc_level->num_pixel_segments = num_pix_segs;
	canvas_convert_from_canvas_coords
	    (canvas, FALSE, TRUE, vc_level->num_pixel_segments,
	     vc_level->pix_x0, vc_level->pix_y0,
	     vc_level->pix_x0, vc_level->pix_y0);
	canvas_convert_from_canvas_coords
	    (canvas, FALSE, TRUE, vc_level->num_pixel_segments,
	     vc_level->pix_x1, vc_level->pix_y1,
	     vc_level->pix_x1, vc_level->pix_y1);
	/*  Pixel segments are now ready to be drawn  */
	return;
    }
    /*  Divide the segment arrays into a number of sections and compute the
	limits of each section. This can speed later operations. Since all
	the segments have to be processed anyway, may as well perform the
	world->pixel selection and conversion as well, even though it would
	have been done later anyway. It should be faster to combine the
	operations in one loop, rather than processing all the segments
	twice  */
    lin_section_size = num_segments / NUM_SECTIONS;
    section_count = 0;
    sec_xmin = TOOBIG;
    sec_xmax = -TOOBIG;
    sec_ymin = TOOBIG;
    sec_ymax = -TOOBIG;
    for (count = 0, num_pix_segs = 0, subsection_count = 0;
	 count < num_segments; ++count, ++subsection_count)
    {
	if (subsection_count >= lin_section_size)
	{
	    /*  End of section: save current section and go to next  */
	    linsegs_level->limits.xmin[section_count] = sec_xmin;
	    linsegs_level->limits.xmax[section_count] = sec_xmax;
	    linsegs_level->limits.ymin[section_count] = sec_ymin;
	    linsegs_level->limits.ymax[section_count] = sec_ymax;
	    if (++section_count < NUM_SECTIONS)
	    {
		sec_xmin = TOOBIG;
		sec_xmax = -TOOBIG;
		sec_ymin = TOOBIG;
		sec_ymax = -TOOBIG;
		subsection_count = 0;
	    }
	    else
	    {
		subsection_count = 1;  /*  Carry on with the last section  */
		section_count = NUM_SECTIONS - 1;
	    }
	}
	x0 = linsegs_level->x0[count];
	y0 = linsegs_level->y0[count];
	x1 = linsegs_level->x1[count];
	y1 = linsegs_level->y1[count];
	/*  Update section limits  */
	if (x0 < sec_xmin) sec_xmin = x0;
	if (x0 > sec_xmax) sec_xmax = x0;
	if (y0 < sec_ymin) sec_ymin = y0;
	if (y0 > sec_ymax) sec_ymax = y0;
	if (x1 < sec_xmin) sec_xmin = x1;
	if (x1 > sec_xmax) sec_xmax = x1;
	if (y1 < sec_ymin) sec_ymin = y1;
	if (y1 > sec_ymax) sec_ymax = y1;
	/*  Add segment to pixel list if inside world canvas  */
	if ( (x0 < win_x_min) || (x0 > win_x_max)
	     || (x1 < win_x_min) || (x1 > win_x_max)
	     || (y0 < win_y_min) || (y0 > win_y_max)
	     || (y1 < win_y_min) || (y1 > win_y_max) ) continue;
	vc_level->pix_x0[num_pix_segs] = x0;
	vc_level->pix_y0[num_pix_segs] = y0;
	vc_level->pix_x1[num_pix_segs] = x1;
	vc_level->pix_y1[num_pix_segs] = y1;
	++num_pix_segs;
    }
    /*  Save last section limits  */
    linsegs_level->limits.xmin[section_count] = sec_xmin;
    linsegs_level->limits.xmax[section_count] = sec_xmax;
    linsegs_level->limits.ymin[section_count] = sec_ymin;
    linsegs_level->limits.ymax[section_count] = sec_ymax;
    _contour_compute_pixel_coords (canvas, vc_level, linsegs_level, win_scale,
				   num_segments, num_pix_segs);
}   /*  End Function _contour_compute_lin_limits  */

static void _contour_compute_pix_segs (KViewableContourImage vcimage,
				       unsigned int level_index,
				       CONST struct win_scale_type *win_scale,
				       double win_x_min, double win_y_min,
				       double win_x_max, double win_y_max)
/*  [SUMMARY] Compute pixel canvas segments.
    <vcimage> The KViewableContourImage object.
    <level_index> The contour level to draw.
    <win_x_min> The lower horizontal linear co-orinate.
    <win_x_max> The upper horizontal linear co-orinate.
    <win_y_min> The lower vertical linear co-orinate.
    <win_y_max> The upper vertical linear co-orinate.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas canvas = vcimage->canvas_holder->canvas;
    unsigned int count, num_pix_segs;
    unsigned int section_count, subsection_count, lin_section_size;
    unsigned int num_segments =
	vcimage->cimage->worldsegs[level_index].num_segments;
    double x0, y0, x1, y1;
    struct linearsegments_level_struct *linsegs_level =
	&vcimage->linsegs_cache->levels[level_index];
    struct viewable_levelinfo_type *vc_level = &vcimage->levels[level_index];

    if (num_segments < NUM_SECTIONS)
    {
	/*  Not enough segments: the sectioning algorithm will hang  */
	/*  Convert linear world segments to pixel segments  */
	for (count = 0, num_pix_segs = 0; count < num_segments; ++count)
	{
	    x0 = linsegs_level->x0[count];
	    y0 = linsegs_level->y0[count];
	    x1 = linsegs_level->x1[count];
	    y1 = linsegs_level->y1[count];
	    /*  Add segment to pixel list if inside world canvas  */
	    if ( (x0 < win_x_min) || (x0 > win_x_max) ||
		 (x1 < win_x_min) || (x1 > win_x_max) ||
		 (y0 < win_y_min) || (y0 > win_y_max) ||
		 (y1 < win_y_min) || (y1 > win_y_max) ) continue;
	    vc_level->pix_x0[num_pix_segs] = x0;
	    vc_level->pix_y0[num_pix_segs] = y0;
	    vc_level->pix_x1[num_pix_segs] = x1;
	    vc_level->pix_y1[num_pix_segs] = y1;
	    ++num_pix_segs;
	}
	vc_level->num_pixel_segments = num_pix_segs;
	canvas_convert_from_canvas_coords
	    (canvas, FALSE, TRUE, vc_level->num_pixel_segments,
	     vc_level->pix_x0, vc_level->pix_y0,
	     vc_level->pix_x0, vc_level->pix_y0);
	canvas_convert_from_canvas_coords
	    (canvas, FALSE, TRUE, vc_level->num_pixel_segments,
	     vc_level->pix_x1, vc_level->pix_y1,
	     vc_level->pix_x1, vc_level->pix_y1);
	/*  Pixel segments are now ready to be drawn  */
	return;
    }
    /*  Process only those segments which lie inside the canvas  */
    lin_section_size = num_segments / NUM_SECTIONS;
    for (count = 0, num_pix_segs = 0, section_count = 0, subsection_count = 0;
	 count < num_segments; ++count, ++subsection_count)
    {
	if (subsection_count >= lin_section_size)
	{
	    if (++section_count < NUM_SECTIONS) subsection_count = 0;
	    else
	    {
		subsection_count = 0; /* Pretend last section not processed */
		section_count = NUM_SECTIONS - 1;
	    }
	}
	if (subsection_count == 0)
	{
	    /*  First time in this section: use the limits to quickly check
		if any segments should be drawn  */
	    if ( (linsegs_level->limits.xmin[section_count] < win_x_min) &&
		 (linsegs_level->limits.xmax[section_count] < win_x_min) )
	    {
		count += lin_section_size - 1;
		subsection_count += lin_section_size - 1;
		continue;
	    }
	    if ( (linsegs_level->limits.xmin[section_count] > win_x_max) &&
		 (linsegs_level->limits.xmax[section_count] > win_x_max) )
	    {
		count += lin_section_size - 1;
		subsection_count += lin_section_size - 1;
		continue;
	    }
	    if ( (linsegs_level->limits.ymin[section_count] < win_y_min) &&
		 (linsegs_level->limits.ymax[section_count] < win_y_min) )
	    {
		count += lin_section_size - 1;
		subsection_count += lin_section_size - 1;
		continue;
	    }
	    if ( (linsegs_level->limits.ymin[section_count] > win_y_max) &&
		 (linsegs_level->limits.ymax[section_count] > win_y_max) )
	    {
		count += lin_section_size - 1;
		subsection_count += lin_section_size - 1;
		continue;
	    }
	}
	x0 = linsegs_level->x0[count];
	y0 = linsegs_level->y0[count];
	x1 = linsegs_level->x1[count];
	y1 = linsegs_level->y1[count];
	if ( (x0 < win_x_min) || (x0 > win_x_max) ||
	     (x1 < win_x_min) || (x1 > win_x_max) ||
	     (y0 < win_y_min) || (y0 > win_y_max) ||
	     (y1 < win_y_min) || (y1 > win_y_max) ) continue;
	vc_level->pix_x0[num_pix_segs] = x0;
	vc_level->pix_y0[num_pix_segs] = y0;
	vc_level->pix_x1[num_pix_segs] = x1;
	vc_level->pix_y1[num_pix_segs] = y1;
	++num_pix_segs;
    }
    _contour_compute_pixel_coords (canvas, vc_level, linsegs_level, win_scale,
				   num_segments, num_pix_segs);
}   /*  End Function _contour_compute_pix_segs  */

static void _contour_compute_pixel_coords
    (KWorldCanvas canvas, struct viewable_levelinfo_type *vc_level,
     CONST struct linearsegments_level_struct *linsegs_level,
     CONST struct win_scale_type *win_scale,
     unsigned int num_world_segs, unsigned int num_pix_segs)
/*  [SUMMARY] Compute pixel co-ordinates.
    <canvas> The world canvas.
    <level> The contour level.
    <linsegs_level> The linear segments level information.
    <win_scale> The window scaling information.
    <num_pix_segs> The number of pixel segments.
    [RETURNS] Nothing.
*/
{
    unsigned int count, section_count, subsection_count, pix_section_size;
    double sec_xmin, sec_ymin, sec_xmax, sec_ymax;
    double x0, y0, x1, y1;
    /*static char function_name[] = "_contour_compute_pixel_coords";*/

    vc_level->num_pixel_segments = num_pix_segs;
    canvas_convert_from_canvas_coords (canvas, FALSE, TRUE,
				       vc_level->num_pixel_segments,
				       vc_level->pix_x0, vc_level->pix_y0,
				       vc_level->pix_x0, vc_level->pix_y0);
    canvas_convert_from_canvas_coords (canvas, FALSE, TRUE,
				       vc_level->num_pixel_segments,
				       vc_level->pix_x1, vc_level->pix_y1,
				       vc_level->pix_x1, vc_level->pix_y1);
    pix_section_size = vc_level->num_pixel_segments / NUM_SECTIONS;
    if (num_pix_segs == num_world_segs)
    {
	/*  Since all segments are visible, assume that the order of the
	    segments in sections will be the same for linear world and
	    pixel co-ordinates  */
	canvas_convert_from_canvas_coords (canvas, FALSE, TRUE,
					   NUM_SECTIONS,
					   linsegs_level->limits.xmin,
					   linsegs_level->limits.ymin,
					   vc_level->pix_limits.xmin,
					   vc_level->pix_limits.ymin);
	canvas_convert_from_canvas_coords (canvas, FALSE, TRUE,
					   NUM_SECTIONS,
					   linsegs_level->limits.xmax,
					   linsegs_level->limits.ymax,
					   vc_level->pix_limits.xmax,
					   vc_level->pix_limits.ymax);
    }
    else if (vc_level->num_pixel_segments >= NUM_SECTIONS)
    {
	/*  Have to compute limits on pixel segments  */
	sec_xmin = TOOBIG;
	sec_xmax = -TOOBIG;
	sec_ymin = TOOBIG;
	sec_ymax = -TOOBIG;
	for (count = 0, section_count = 0, subsection_count = 0;
	     count < num_pix_segs; ++count, ++subsection_count)
	{
	    if (subsection_count >= pix_section_size)
	    {
		/*  End of section: save current section and go to next  */
		vc_level->pix_limits.xmin[section_count] = sec_xmin;
		vc_level->pix_limits.xmax[section_count] = sec_xmax;
		vc_level->pix_limits.ymin[section_count] = sec_ymin;
		vc_level->pix_limits.ymax[section_count] = sec_ymax;
		if (++section_count < NUM_SECTIONS)
		{
		    sec_xmin = TOOBIG;
		    sec_xmax = -TOOBIG;
		    sec_ymin = TOOBIG;
		    sec_ymax = -TOOBIG;
		    subsection_count = 0;
		}
		else
		{
		    subsection_count = 1; /* Carry one with the last section */
		    section_count = NUM_SECTIONS - 1;
		}
	    }
	    x0 = vc_level->pix_x0[count];
	    y0 = vc_level->pix_y0[count];
	    x1 = vc_level->pix_x1[count];
	    y1 = vc_level->pix_y1[count];
	    /*  Update section limits  */
	    if (x0 < sec_xmin) sec_xmin = x0;
	    if (x0 > sec_xmax) sec_xmax = x0;
	    if (y0 < sec_ymin) sec_ymin = y0;
	    if (y0 > sec_ymax) sec_ymax = y0;
	    if (x1 < sec_xmin) sec_xmin = x1;
	    if (x1 > sec_xmax) sec_xmax = x1;
	    if (y1 < sec_ymin) sec_ymin = y1;
	    if (y1 > sec_ymax) sec_ymax = y1;
	}
	/*  Save last section limits  */
	vc_level->pix_limits.xmin[section_count] = sec_xmin;
	vc_level->pix_limits.xmax[section_count] = sec_xmax;
	vc_level->pix_limits.ymin[section_count] = sec_ymin;
	vc_level->pix_limits.ymax[section_count] = sec_ymax;
    }
}   /*  End Function _contour_compute_pixel_coords  */

static void _contour_worldcanvas_refresh_func
    (KWorldCanvas canvas, int width, int height,
     CONST struct win_scale_type *win_scale,
     Kcolourmap cmap, flag cmap_resize, void **info, PostScriptPage pspage,
     unsigned int num_areas, KPixCanvasRefreshArea *areas,flag *honoured_areas)
/*  [PURPOSE] This routine processes a refresh event for a world canvas.
    <canvas> The world canvas.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The canvas colourmap.
    <cmap_resize> If TRUE, the refresh is due to a colourmap resize.
    <info> A pointer to the arbitrary information pointer.
    <pspage> The PostScriptPage object. If this is NULL, the refresh is *not*
    destined for a PostScript page.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder = *info;
    KViewableContourImage vcimage;
    static char function_name[] = "_contour_worldcanvas_refresh_func";

    VERIFY_HOLDER (holder);
    if (canvas != holder->canvas)
    {
	fprintf (stderr, "Different canvas in canvas holder object\n");
	a_prog_bug (function_name);
    }
    /*  Show each active image  */
    for (vcimage = holder->first;vcimage; vcimage = vcimage->holder_chain.next)
	if (vcimage->active)
	    _contour_refresh_image (vcimage, win_scale, num_areas, areas);
    *honoured_areas = TRUE;
}   /*  End Function _contour_worldcanvas_refresh_func  */

static void _contour_refresh_image (KViewableContourImage vcimage,
				    CONST struct win_scale_type *win_scale,
				    unsigned int num_areas,
				    KPixCanvasRefreshArea *areas)
/*  [PURPOSE] This routine processes a refresh event for a world canvas.
    <vcimage> The KViewableContourImage object.
    <win_scale> A pointer to the window scaling information.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    [RETURNS] Nothing.
*/
{
    flag dash;
    KPixCanvas pixcanvas;
    unsigned int count;
    double linewidth;
    /*static char function_name[] = "_contour_refresh_image";*/

    if (vcimage->cimage->num_levels < 1) return;
    pixcanvas = canvas_get_pixcanvas (vcimage->canvas_holder->canvas);
    kwin_get_attributes (pixcanvas,
			 KWIN_ATT_LINE_DASH, &dash,
			 KWIN_ATT_LINEWIDTH, &linewidth,
			 KWIN_ATT_END);
    for (count = 0; count < vcimage->cimage->num_levels; ++count)
	_contour_draw_level (vcimage, count, win_scale, num_areas, areas);
    kwin_set_attributes (pixcanvas,
			 KWIN_ATT_LINE_DASH, dash,
			 KWIN_ATT_LINEWIDTH, linewidth,
			 KWIN_ATT_END);
}   /*  End Function _contour_refresh_image  */

static void _contour_draw_level (KViewableContourImage vcimage,
				 unsigned int level_index,
				 CONST struct win_scale_type *win_scale,
				 unsigned int num_areas,
				 KPixCanvasRefreshArea *areas)
/*  [PURPOSE] This routine will draw a single level.
    <vcimage> The KViewableContourImage object.
    <level> The contour level to draw.
    <win_scale> A pointer to the window scaling information.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <same_trans> If TRUE the world->linear co-ordinate transformation has not
    changed since the last refresh.
    <same_world> If TRUE the linear->pixel co-ordinate transformation has not
    changed since the last refresh.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas canvas;
    KPixCanvas pixcanvas;
    CanvasHolder holder;
    unsigned long pixel_value;
    struct viewable_levelinfo_type *vc_level = &vcimage->levels[level_index];
    static char function_name[] = "_contour_draw_level";

    holder = vcimage->canvas_holder;
    canvas = holder->canvas;
    if (vc_level->pixel_defined) pixel_value = vc_level->pixel_value;
    else
    {
	/*  Get colour, with a hack to use grey rather than real dashing  */
	if ( !canvas_get_colour (canvas,
				 vc_level->dash ?
				 "grey70" : holder->colourname,
				 &pixel_value, NULL, NULL, NULL) )
	{
	    fprintf (stderr, "%s: error getting colour\n", function_name);
	    return;
	}
    }
    pixcanvas = canvas_get_pixcanvas (canvas);
    kwin_set_attributes (pixcanvas,
#ifdef DISABLED
			 KWIN_ATT_LINE_DASH, vc_level->dash,
#endif
			 KWIN_ATT_LINEWIDTH, vc_level->linewidth,
			 KWIN_ATT_END);
    if ( (num_areas < 1) || (vc_level->num_pixel_segments < NUM_SECTIONS) ||
	 (vc_level->num_area_segments < 1) )
    {
	/*  Either whole canvas is to be refreshed (hence all cached pixel
	    segments should be drawn) or there are not enough segments to be
	    sectioned, so they are all drawn too  */
	kwin_draw_segments_TRANSITION (pixcanvas,
				       vc_level->pix_x0, vc_level->pix_y0,
				       vc_level->pix_x1, vc_level->pix_y1,
				       vc_level->num_pixel_segments,
				       pixel_value);
	return;
    }
    if (vc_level->num_area_segments < 1) return;
    kwin_draw_segments_TRANSITION (canvas_get_pixcanvas (canvas),
				   vc_level->pix_area_x0,vc_level->pix_area_y0,
				   vc_level->pix_area_x1,vc_level->pix_area_y1,
				   vc_level->num_area_segments, pixel_value);
    m_free_no_checks (vc_level->pix_area_x0);
    vc_level->pix_area_x0 = NULL;
    vc_level->pix_area_y0 = NULL;
    vc_level->pix_area_x1 = NULL;
    vc_level->pix_area_y1 = NULL;
}   /*  End Function _contour_draw_level  */

static void _contour_dealloc_levels (KContourImage cimage)
/*  [SUMMARY] Dealloc the contour levels information.
    <cimage> The KContourImage object.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    static char function_name[] = "_contour_dealloc_levels";

    if (cimage->num_levels > 0)
    {
	KViewableContourImage vcimage;

	for (vcimage = cimage->first_vcimage; vcimage;
	     vcimage = vcimage->cimage_chain.next)
	    _contour_dealloc_viewable_levels (vcimage);
	if (cimage->first_linsegs_cache)
	{
	    fprintf (stderr, "Unreferenced linear segments cache\n");
	    a_prog_bug (function_name);
	}
	m_free (cimage->contour_levels);
	for (count = 0; count < cimage->num_levels; ++count)
	{
	    m_free2 (cimage->worldsegs[count].x0);
	    m_free2 (cimage->worldsegs[count].y0);
	    m_free2 (cimage->worldsegs[count].x1);
	    m_free2 (cimage->worldsegs[count].y1);
	}
	m_free (cimage->worldsegs);
    }
    cimage->num_levels = 0;
    cimage->world_segs_valid = FALSE;
    cimage->worldsegs = NULL;
    cimage->contour_levels = NULL;
}   /*  End Function _contour_dealloc_levels  */

static void _contour_dealloc_viewable_levels (KViewableContourImage vcimage)
/*  [SUMMARY] Dealloc the contour levels information.
    <vcimage> The KViewableContourImage object.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    /*static char function_name[] = "_contour_dealloc_viewable_levels";*/

    _contour_release_linsegs_cache (vcimage->linsegs_cache);
    vcimage->linsegs_cache = NULL;
    vcimage->pixel_segments_valid = FALSE;
    if (!vcimage->levels) return;
    for (count = 0; count < vcimage->cimage->num_levels; ++count)
    {
	m_free2 (vcimage->levels[count].pix_x0);
	m_free2 (vcimage->levels[count].pix_y0);
	m_free2 (vcimage->levels[count].pix_x1);
	m_free2 (vcimage->levels[count].pix_y1);
    }
    m_free (vcimage->levels);
    vcimage->levels = NULL;
}   /*  End Function _contour_dealloc_viewable_levels  */

void _contour_find_linsegs_cache (KViewableContourImage vcimage,
				  CONST struct win_scale_type *win_scale)
/*  [SUMMARY] Find the best linear segments cache for a KViewableContourImage.
    <vcimage> The KViewableContourImage.
    <win_scale> A pointer to the window scaling information.
    [RETURNS] Nothing.
*/
{
    KContourImage cimage = vcimage->cimage;
    unsigned int count;
    double wx_thres, wy_thres;
    struct linearsegments_cache_struct *linsegs_cache = vcimage->linsegs_cache;
    CONST dim_desc *hdim, *vdim;
    double test_x[5], test_y[5];
    static char function_name[] = "_contour_find_linsegs_cache";

    /*  At this point the contour segments have been computed in world
	co-ordinates. To speed the redisplay, try to find a linear segments
	cache where the non-linear to linear world co-ordinate transformation
	has not changed  */
    test_x[0] = win_scale->x_offset;
    test_y[0] = win_scale->y_offset;
    test_x[1] = test_x[0] + 1e-2;
    test_y[1] = test_y[0] + 1e-2;
    canvas_convert_to_canvas_coords (vcimage->canvas_holder->canvas, FALSE,
				     2, test_x, test_y, NULL, NULL,
				     test_x, test_y);
    wx_thres = fabs (test_x[0] - test_x[1]);
    wy_thres = fabs (test_y[0] - test_y[1]);
    hdim = cimage->arr_desc->dimensions[cimage->hdim];
    vdim = cimage->arr_desc->dimensions[cimage->vdim];
    test_x[0] = hdim->first_coord;
    test_y[0] = vdim->first_coord;
    test_x[1] = hdim->last_coord;
    test_y[1] = vdim->first_coord;
    test_x[2] = (hdim->first_coord + hdim->last_coord) / 2.0;
    test_y[2] = (vdim->first_coord + vdim->last_coord) / 2.0;
    test_x[3] = hdim->first_coord;
    test_y[3] = vdim->last_coord;
    test_x[4] = hdim->last_coord;
    test_y[4] = vdim->last_coord;
    canvas_coords_transform (vcimage->canvas_holder->canvas, 5,
			     test_x, FALSE, test_y, FALSE);
    if (linsegs_cache)
    {   /*  I already have a cache. Use if valid, else update if it doesn't
	    hurt another user  */
	if ( m_cmp (linsegs_cache->test_x, test_x, sizeof *test_x * 5)
	     && m_cmp (linsegs_cache->test_y, test_y, sizeof *test_y * 5) )
	    return;  /*  Same world<->linear transformation  */
	for (count = 0; count < 5; ++count)
	{
	    if (fabs (test_x[count] - linsegs_cache->test_x[count]) > wx_thres)
		break;
	    if (fabs (test_y[count] - linsegs_cache->test_y[count]) > wy_thres)
		break;
	}
	if (count >= 5) return;  /*  Almost same world<->linear transform  */
	/*  If I am the only user of this cache, invalidate it so it will be
	    updated  */
	if (linsegs_cache->refcount == 1) linsegs_cache->valid = FALSE;
	if (!linsegs_cache->valid)
	{   /*  If current cache is invalid, may as well update it  */
	    m_copy (linsegs_cache->test_x, test_x, sizeof *test_x * 5);
	    m_copy (linsegs_cache->test_y, test_y, sizeof *test_y * 5);
	    return;
	}
	/*  Don't want to update this cache: find/create another  */
	_contour_release_linsegs_cache (linsegs_cache);
	vcimage->linsegs_cache = NULL;
    }
    /*  Find any cache which is not valid or has the same world<->linear
	transformation  */
    for (linsegs_cache = cimage->first_linsegs_cache; linsegs_cache;
	 linsegs_cache = linsegs_cache->next)
    {
	if (!linsegs_cache->valid) break;  /*  OK to update this one  */
	if ( m_cmp (linsegs_cache->test_x, test_x, sizeof *test_x * 5)
	     && m_cmp (linsegs_cache->test_y, test_y, sizeof *test_y * 5) )
	    break;  /*  This one is valid for me: use it  */
	for (count = 0; count < 5; ++count)
	{
	    if (fabs (test_x[count] - linsegs_cache->test_x[count]) > wx_thres)
		break;
	    if (fabs (test_y[count] - linsegs_cache->test_y[count]) > wy_thres)
		break;
	}
	if (count >= 5) break;  /*  Close enough to be valid for me: use it  */
    }
    if (!linsegs_cache)
    {   /*  Need to create a new one  */
	if ( ( linsegs_cache = m_calloc (sizeof *linsegs_cache) ) == NULL )
	    m_abort (function_name, "linear segments cache");
	if ( ( linsegs_cache->levels = m_calloc (sizeof *linsegs_cache->levels
						 * cimage->num_levels) )
	     == NULL )
	    m_abort (function_name, "linear segments cache levels");
	linsegs_cache->cimage = cimage;
	KLIST_APPEND2 (cimage->first_linsegs_cache, cimage->last_linsegs_cache,
		       linsegs_cache);
    }
    ++linsegs_cache->refcount;
    vcimage->linsegs_cache = linsegs_cache;
    m_copy (linsegs_cache->test_x, test_x, sizeof *test_x * 5);
    m_copy (linsegs_cache->test_y, test_y, sizeof *test_y * 5);
}   /*  End Function _contour_find_linsegs_cache  */

void _contour_release_linsegs_cache (struct linearsegments_cache_struct *cache)
/*  [SUMMARY] Release linear segments cache.
    <cache> The cache. This may be NULL.
    [RETURNS] Nothing.
*/
{
    unsigned int count;

    if (!cache) return;
    if (--cache->refcount > 0) return;
    KLIST_REMOVE2 (cache->cimage->first_linsegs_cache,
		   cache->cimage->last_linsegs_cache, cache);
    for (count = 0; count < cache->cimage->num_levels; ++count)
    {
	m_free2 (cache->levels[count].x0);
	m_free2 (cache->levels[count].y0);
	m_free2 (cache->levels[count].x1);
	m_free2 (cache->levels[count].y1);
    }
    m_free (cache->levels);
    m_free (cache);
}   /*  End Function _contour_release_linsegs_cache  */
