/*LINTLIBRARY*/
/*  misc.c

    This code provides KContourImage objects.

    Copyright (C) 1998-2004  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for manipulating contour images.


    Written by      Richard Gooch   17-SEP-1998

    Updated by      Richard Gooch   20-NOV-2002: Documentation fix.

    Updated by      Richard Gooch   21-FEB-2003: Supported "min" and "max"
  values.

    Updated by      Richard Gooch   5-OCT-2003: Added crude thickening for
  "major" contour levels.

    Updated by      Richard Gooch   10-OCT-2004: Disable thickening of "major"
  contour levels when there is only one contour level.

    Last updated by Richard Gooch   25-OCT-2004: Fixed bug in function
  <contour_unscale_levels>: don't crash when <<num_levels>> is zero.


*/
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <karma.h>
#include <karma_contour.h>
#include <karma_ex.h>
#include <karma_r.h>


/*  Private functions  */
STATIC_FUNCTION (double get_decimal_step, (double range) );


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
unsigned int contour_parse_levels (double *contour_levels,
				   unsigned int max_levels, CONST char *string,
				   double min, double max)
/*  [SUMMARY] Parse a textual contour level specification.
    <contour_levels> The raw contour levels are written here.
    <max_levels> The maximum number of levels.
    <string> The string to parse.
    <min> The minimum value in the data range.
    <max> The maximum value in the data range.
    [RETURNS] The number of levels. On error 0 is returned.
*/
{
    flag relative, geometric;
    unsigned int num_levels;
    double value, start, stop, increment, first_inc;
    CONST char *p = string;

    while ( isspace (*p) ) ++p;
    for (num_levels = 0; p; )
    {
	if (*p == '%')
	{
	    relative = TRUE;
	    ++p;
	}
	else relative = FALSE;
	/*  Get a value  */
	if (strncmp (p, "min", 3) == 0)
	{
	    value = min;
	    p += 3;
	    if (*p == '\0') p = NULL;
	}
	else if (strncmp (p, "max", 3) == 0)
	{
	    value = max;
	    p += 3;
	    if (*p == '\0') p = NULL;
	}
	else value = ex_float (p, &p);
	if ( (p == NULL) || isdigit (*p) || (*p == '.') || (*p == '%') ||
	     isspace (p[-1]) )
	{
	    /*  End of the line or separate value follows  */
	    if (num_levels >= max_levels)
	    {
		fprintf (stderr, "Too many contour levels: maximum is %u\n",
			 max_levels);
		return (0);
	    }
	    if (relative) value = min + value * 1e-2 * (max - min);
	    contour_levels[num_levels++] = value;
	    continue;
	}
	/*  Special character  */
	if (*p != ':')
	{
	    fprintf (stderr, "Illegal control character: '%c'\n", *p);
	    return (0);
	}
	start = value;
	++p;
	/*  Get stop and increment values  */
	if (strncmp (p, "min", 3) == 0)
	{
	    stop = min;
	    p += 3;
	    if (*p == '\0') p = NULL;
	}
	else if (strncmp (p, "max", 3) == 0)
	{
	    stop = max;
	    p += 3;
	    if (*p == '\0') p = NULL;
	}
	else stop = ex_float (p, &p);
	if (p == NULL)
	{
	    fprintf (stderr, "Increment character, found end of line\n");
	    return (0);
	}
	switch (*p)
	{
	  case '+':
	    geometric = FALSE;
	    break;
	  case '*':
	    geometric = TRUE;
	    break;
	  default:
	    fprintf (stderr,
		     "Missing increment character, found: '%c' instead\n", *p);
	    return (0);
	    /*break;*/
	}
	++p;
	increment = ex_float (p, &p);
	first_inc = TOOBIG;
	for (value = start; value <= stop; )
	{
	    if (num_levels >= max_levels)
	    {
		fprintf (stderr, "Too many contour levels: maximum is %u\n",
			 max_levels);
		return (0);
	    }
	    contour_levels[num_levels++] = relative ?
		( min + value * 1e-2 * (max - min) ) : value;
	    if (geometric) value *= increment;
	    else value += increment;
	    if ( DOUBLE_IS_BLANK (first_inc) ) first_inc = value - start;
	    if (fabs (stop - value) < first_inc * 1e-5) value = stop;
	}
    }
    return (num_levels);
}   /*  End Function contour_parse_levels  */

/*EXPERIMENTAL_FUNCTION*/
void contour_unscale_levels (double *contour_levels, flag *dash,
			     double *linewidths, unsigned int num_levels,
			     double scale, double offset,
			     double neg_width, double pos_width)
/*  [SUMMARY] Unscale contour levels.
    <contour_levels> The scaled contour levels. This is updated with the raw
    (unscaled) contour levels.
    <dash> An array of booleans will be written here. TRUE is written if the
    corresponding scaled contour level is negative, else FALSE is written.
    <linewidths> An array of linewidths will be written here.
    <num_levels> The number of levels.
    <scale> The scale factor.
    <offset> The offset.
    [NOTE] scaled = raw * <<scale>> + <<offset>>.
    <neg_width> The linewidth for negative contours.
    <pos_width> The linewidth for positive contours.
    [RETURNS] Nothing.
*/
{
    flag want_majors;
    unsigned int count;
    double major_width = pos_width * 2.0;
    double range = 0.0;
    double threshold = TOOBIG;
    double maj_step;
    static flag first_time = TRUE;
    static flag disable_majors = FALSE;

    if (first_time)
    {
	first_time = FALSE;
	if ( r_getenv ("CONTOUR_DISABLE_MAJORS") )
	{
	    fprintf (stderr, "Disabling major contour thickening\n");
	    disable_majors = TRUE;
	}
    }
    if (num_levels < 1) return;
    want_majors = ( !disable_majors && (num_levels > 1) ) ? TRUE : FALSE;
    if (major_width < 2.0) major_width = 2.0;
    /*  Find range major step and comparison threshold  */
    for (count = 0; count < num_levels; ++count)
    {
	double tmp = fabs (contour_levels[count]);

	if (tmp > range) range = tmp;
    }
    maj_step = get_decimal_step (range);
    for (count = 0; count < num_levels - 1; ++count)
    {
	double tmp = fabs (contour_levels[count + 1] - contour_levels[count]);

	if (tmp < threshold) threshold = tmp;
    }
    threshold *= 1e-3;
    for (count = 0; count < num_levels; ++count)
    {
	double tmp;

	if (contour_levels[count] < 0.0)
	{
	    dash[count] = TRUE;
	    linewidths[count] = neg_width;
	}
	else
	{
	    dash[count] = FALSE;
	    linewidths[count] = pos_width;
	}
	if (want_majors)
	{
	    tmp = floor (fabs (contour_levels[count]) / maj_step) * maj_step;
	    if (fabs (contour_levels[count] - tmp) < threshold)
		linewidths[count] = major_width;
	}
	contour_levels[count] = (contour_levels[count] - offset) / scale;
    }
}   /*  End Function contour_unscale_levels  */


/*  Private functions follow  */

static double get_decimal_step (double range)
/*  [SUMMARY] Get a decimal step.
    <range> The range of the data.
    [RETURNS] The step value.
*/
{
    double factor, step;

    factor = pow ( 10.0, -floor ( log10 (range) ) );
    range *= factor;
    if (range >= 6.0)      step = 2.0;
    else if (range >= 3.0) step = 1.0;
    else if (range >= 1.5) step = 0.5;
    else                   step = 0.2;
    return (step / factor);
}   /*  End Function get_decimal_step  */
