/*LINTLIBRARY*/
/*  main.c

    This code provides generic DisplayData objects.

    Copyright (C) 2002-2004  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for manipulating generic display
    data.


    Written by      Richard Gooch   10-NOV-2002: Serious coding begins.

    Updated by      Richard Gooch   12-NOV-2002: First cut of code. Compiles.

    Updated by      Richard Gooch   23-NOV-2002: Second cut of code. Compiles.

    Updated by      Richard Gooch   11-DEC-2002: Third cut of code. Tested with
  new <kvis> application.

    Updated by      Richard Gooch   12-DEC-2002: Added <<domain>> parameter and
  directory handling to <dispdata_create_dataobject> function.

    Updated by      Richard Gooch   17-DEC-2002: Fixed allocation bug in
  <dispdata_regenerate_list>: only one extra block would be allocated, not the
  total allocation requirements. Fixed allocation bug for dataclass namelist.

    Updated by      Richard Gooch   24-JAN-2003: Added data attribute:
  KDISPLAYDATA_DATA_ATT_REPLACE. Added KDisplayDataFuncCopyDataAttributes
  hook.

    Updated by      Richard Gooch   25-JAN-2003: Created experimental function
  <dispdata_register_data_create_func>.

    Updated by      Richard Gooch   27-JAN-2003: Created experimental function
  <dispdata_get_blinkstate_index>.

    Updated by      Richard Gooch   28-JAN-2003: Added load-order sorting.

    Updated by      Richard Gooch   17-FEB-2003: Added <<before_blinkentries>>
  parameter to KDisplayDataFuncActivateBlinkState hook.

    Updated by      Richard Gooch   18-FEB-2003: Fixed bug in function
  <dispdata_destroy_dataobject>: was calling blinkstate refresh function before
  viewable was destroyed. This screwed up <_dispdata_array_set_title>.

    Updated by      Richard Gooch   21-FEB-2003: Fixed memory leak in function
  <dispdata_associate_data_object>: <<destroy_func>> would not be called by
  <j> package, so register a callback.

    Updated by      Richard Gooch   11-JUN-2003: Created experimental functions
  <dispdata_get_window_index> and <dispdata_register_window_renumber_func>.

    Updated by      Richard Gooch   13-JUN-2003: Created experimental function
  <dispdata_get_num_windows>.

    Updated by      Richard Gooch   2-AUG-2003: Created experimental function
  <dispdata_set_class_attributes> and changed API to function
  <dispdata_create_dataclass>.

    Updated by      Richard Gooch   3-AUG-2003: Created experimental functions
  <dispdata_get_first_dataobject> and <dispdata_get_next_dataobject>.

    Updated by      Richard Gooch   5-AUG-2003: Created experimental function
  <dispdata_register_data_destroy_func>.

    Updated by      Richard Gooch   6-AUG-2003: Made use of new experimental
  function <c_unregister_callback2>.

    Updated by      Richard Gooch   10-SEP-2003: Refresh windows in
  <dispdata_destroy_dataobject> if active blinkentry showed data.

    Updated by      Richard Gooch   26-SEP-2003: Fixed bug caused by previous
  change: data replacement no longer caused window refresh. Fixed by not
  clearing the flag except in <dispdata_refresh_window>.

    Updated by      Richard Gooch   14-MAR-2004: Created experimental function
  <dispdata_get_objectlist>.

    Updated by      Richard Gooch   17-MAR-2004: Added <<dataobjects>> and
  <<filter_func>> parameters to <dispdata_regenerate_namelist>.

    Updated by      Richard Gooch   14-JUL-2004: Added dataclass attribute:
  KDISPLAYDATA_CLASS_ATT_CURRENTLY_CREATING.

    Last updated by Richard Gooch   18-JUL-2004: Added <<dataclass>> argument
  for <dispdata_get_dataobject_subclass>.


*/
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <sys/param.h>
#include <limits.h>
#include <errno.h>
#include <karma.h>
#include <k_list.h>
#include <karma_dispdata.h>
#include <karma_canvas.h>
#include <karma_kwin.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_c.h>
#include <karma_j.h>
#include <karma_r.h>

#include <karma_dispdata_hooks.h>


#define ALLOCATED_LINES_STEP           16
#define ALLOCATED_CHARS_STEP         1024

#ifndef PATH_MAX  /*  SunOS doesn't have this  */
#  define PATH_MAX MAXPATHLEN
#endif

#define DATACLASS_MAGIC_NUMBER         (unsigned int) 1238433284
#define DATAOBJECT_MAGIC_NUMBER        (unsigned int)  419283402
#define WINDOW_MAGIC_NUMBER            (unsigned int)  598437339
#define BLINKSTATE_MAGIC_NUMBER        (unsigned int) 1342908238
#define BLINKSTATE_CLASS_MAGIC_NUMBER  (unsigned int) 2089234873
#define BLINKENTRY_MAGIC_NUMBER        (unsigned int) 1393843202
#define NAMELIST_MAGIC_NUMBER          (unsigned int)  517352283

#define VERIFY_DATACLASS(dataclass) {if (dataclass == NULL) \
{fprintf (stderr, "NULL dataclass passed\n"); \
 a_prog_bug (function_name); } \
if (dataclass->magic_number != DATACLASS_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid dataclass object\n"); \
 a_prog_bug (function_name); } }

#define VERIFY_DATAOBJECT(dataobject) {if (dataobject == NULL) \
{fprintf (stderr, "NULL dataobject passed\n"); \
 a_prog_bug (function_name); } \
if (dataobject->magic_number != DATAOBJECT_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid dataobject object\n"); \
 a_prog_bug (function_name); } }

#define VERIFY_WINDOW(window) {if (window == NULL) \
{fprintf (stderr, "NULL window passed\n"); \
 a_prog_bug (function_name); } \
if (window->magic_number != WINDOW_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid window object\n"); \
 a_prog_bug (function_name); } }

#define VERIFY_BLINKSTATE(blinkstate) {if (blinkstate == NULL) \
{fprintf (stderr, "NULL blinkstate passed\n"); \
 a_prog_bug (function_name); } \
if (blinkstate->magic_number != BLINKSTATE_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid blinkstate object\n"); \
 a_prog_bug (function_name); } }

#define VERIFY_BLINKSTATE_CLASS(blinkclass) {if (blinkclass == NULL) \
{fprintf (stderr, "NULL blinkclass passed\n"); \
 a_prog_bug (function_name); } \
if (blinkclass->magic_number != BLINKSTATE_CLASS_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid blinkclass object\n"); \
 a_prog_bug (function_name); } }

#define VERIFY_BLINKENTRY(blinkentry) {if (blinkentry == NULL) \
{fprintf (stderr, "NULL blinkentry passed\n"); \
 a_prog_bug (function_name); } \
if (blinkentry->magic_number != BLINKENTRY_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid blinkentry object\n"); \
 a_prog_bug (function_name); } }

#define VERIFY_NAMELIST(namelist) { \
if (namelist->magic_number != NAMELIST_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid namelist object\n"); \
 a_prog_bug (function_name); } }


struct displaydata_dataclass_type        /*  Describes a dataclass  */
{
    unsigned int magic_number;
    char *name;
    KDisplayDataFuncSetClassAttributes            set_class_attributes;
    KDisplayDataFuncDestroyDataObject             destroy_dataobject;
    KDisplayDataFuncCopyDataAttributes            copy_data_attributes;
    KDisplayDataFuncSetDataAttributes             set_data_attributes;
    KDisplayDataFuncGetDataAttributes             get_data_attributes;
    KDisplayDataFuncCreateWindow                  create_window;
    KDisplayDataFuncDestroyWindow                 destroy_window;
    KDisplayDataFuncSetWindowAttributes           set_window_attributes;
    KDisplayDataFuncGetWindowAttributes           get_window_attributes;
    KDisplayDataFuncCreateViewable                create_viewable;
    KDisplayDataFuncDestroyViewable               destroy_viewable;
    KDisplayDataFuncCreateBlinkState              create_blinkstate;
    KDisplayDataFuncDestroyBlinkState             destroy_blinkstate;
    KDisplayDataFuncActivateBlinkState            activate_blinkstate;
    KDisplayDataFuncCreateBlinkEntry              create_blinkentry;
    KDisplayDataFuncDestroyBlinkEntry             destroy_blinkentry;
    KDisplayDataFuncCopyBlinkEntry                copy_blinkentry;
    KDisplayDataFuncActivateBlinkEntry            activate_blinkentry;
    KDisplayDataFuncSetBlinkEntryAttributes       set_blinkentry_attributes;
    KDisplayDataFuncGetBlinkEntryAttributes       get_blinkentry_attributes;
    char **names;                          /*  Sorted                        */
    KDisplayDataObject *objects;           /*  Sorted                        */
    unsigned int num_sets;
    unsigned int num_chars;
    unsigned int namebuf_length;
    flag namelist_valid;
    unsigned int num_replace_flags;
    KCallbackList create_list;
    KCallbackList destroy_list;
    KDisplayDataObject first;
    KDisplayDataObject last;
    KDisplayDataClass prev;
    KDisplayDataClass next;
    KDisplayDataObject currently_creating;
};

struct displaydata_dataobject_type         /*  Describes a data object       */
{
    unsigned int magic_number;
    char *domain;
    char *name;
    char *directory;
    flag replace;
    KDisplayDataClass dataclass;           /*  KDisplayDataClass we are in   */
    void *subclass;
    KJoinedPairList associations;
    KCallbackList associations_destroy_list;
    KCallbackList destroy_list;
    KDisplayDataObject prev;               /*  Chained in KDisplayDataClass  */
    KDisplayDataObject next;               /*  Chained in KDisplayDataClass  */
};

struct canvas_holder
{
    KDisplayDataCanvas canvas;
    KCallbackFunc destroy_func;
};

struct displaydata_window_type             /*  Describes a window  */
{
    unsigned int magic_number;
    struct canvas_holder pc;
    struct canvas_holder rgb;
    void (*show_canvas) (void *info, KWorldCanvas canvas);
    void *info;                            /*  For <<show_canvas>> */
    flag refresh;                       /*  Cleared when window is refreshed */
    KCallbackList destroy_list;
    struct displaydata_window_class_entry *first;
    struct displaydata_window_class_entry *last;
    struct
    {
	KDisplayDataBlinkState first;
	KDisplayDataBlinkState last;
    }
    blinkstates;
    unsigned int num_blinkstates;
    flag indices_valid;
    KDisplayDataBlinkState active;
    unsigned int index;                    /*  Index number of this window  */
    KCallbackList renumber_list;           /*  Callbacks for index changed  */
    struct                                 /*  Unsorted, all data classes   */
    {
	struct displaydata_window_dataobject_type *first;
	struct displaydata_window_dataobject_type *last ;
    }
    dataobjects;
    KDisplayDataWindow prev;
    KDisplayDataWindow next;
};

struct displaydata_window_class_entry      /*  Window data for this class  */
{
    KDisplayDataClass dataclass;
    void *subclass;
    struct displaydata_window_class_entry *prev;
    struct displaydata_window_class_entry *next;
};

struct displaydata_blinkstate_type         /*  Describes one blink state  */
{
    unsigned int magic_number;
    KDisplayDataWindow window;
    void (*active_change) (void *info, flag active);
    void *info;
    unsigned int index;
    KDisplayDataClassBlinkState first;
    KDisplayDataClassBlinkState last;
    struct displaydata_blinkstate_type *prev;
    struct displaydata_blinkstate_type *next;
};

struct displaydata_private_namelist
{
    unsigned int magic_number;
    unsigned int num_allocated_entries;
    unsigned int num_allocated_chars;
    char *block;
    char **strings;
    KDisplayDataObject *dataobjects;
};

struct displaydata_class_blinkstate_type       /*  Lists data in this class  */
{
    unsigned int magic_number;
    KDisplayDataClass dataclass;
    void *subclass;
    void (*refresh_func) (void *info, KDisplayDataClass dataclass);
    void *(*create_blinkentry) (void *info, KDisplayDataBlinkEntry be);
    void (*destroy_blinkentry) (void *info);
    void (*copy_blinkentry) (void *dest, void *source);
    void *info;
    KDisplayDataPrivateNameList namelist;
    unsigned int num_allocated_entries;
    KDisplayDataBlinkEntry *blinkentries;
    KDisplayDataBlinkEntry first;
    KDisplayDataBlinkEntry last;
    KDisplayDataClassBlinkState prev;
    KDisplayDataClassBlinkState next;
};

struct displaydata_window_dataobject_type
{
    KDisplayDataObject dataobject;
    void *viewable;
    struct displaydata_window_dataobject_type *prev;
    struct displaydata_window_dataobject_type *next;
};

struct displaydata_blinkentry_type         /*  Per blink state and dataclass */
{
    unsigned int magic_number;
    struct displaydata_window_dataobject_type *windataobject;
    void *subclass;
    void *info;                            /*  For the application           */
    KDisplayDataBlinkEntry prev;
    KDisplayDataBlinkEntry next;
};


/*  Private data  */
static KDisplayDataClass first_dataclass = NULL;
static KDisplayDataClass last_dataclass = NULL;
static unsigned int num_dataclasses = 0;
static KDisplayDataWindow first_window = NULL;
static KDisplayDataWindow last_window = NULL;
static flag window_was_created = FALSE;


/*  Private functions  */
STATIC_FUNCTION (void _dispdata_destroy_window,
		 (KDisplayDataWindow window, flag canvas_destroy) );
STATIC_FUNCTION (void _dispdata_canvas_destroy_callback, (void *info) );
STATIC_FUNCTION (flag _dispdata_create_destroy_data_event_func,
		 (void *object, void *client1_data, void *call_data,
		  void *client2_data) );


/* Public functions follow */

/*EXPERIMENTAL_FUNCTION*/
void dispdata_init (void)
/*  [SUMMARY] Initialise the package.
    [RETURNS] Nothing.
*/
{
    static flag first_time = TRUE;

    if (first_time)
    {
	first_time = FALSE;
	dispdata_array_create_dataclass ();
	dispdata_overlay_create_dataclass ();
    }
}   /*  End Function dispdata_init  */

/*UNPUBLISHED_FUNCTION*/
KDisplayDataClass dispdata_create_dataclass
    (CONST char *name,
     KDisplayDataFuncSetClassAttributes set_class_attributes,
     KDisplayDataFuncDestroyDataObject destroy_dataobject,
     KDisplayDataFuncCopyDataAttributes copy_data_attributes,
     KDisplayDataFuncSetDataAttributes set_data_attributes,
     KDisplayDataFuncGetDataAttributes get_data_attributes,
     KDisplayDataFuncCreateWindow create_window,
     KDisplayDataFuncDestroyWindow destroy_window,
     KDisplayDataFuncSetWindowAttributes set_window_attributes,
     KDisplayDataFuncGetWindowAttributes get_window_attributes,
     KDisplayDataFuncCreateViewable create_viewable,
     KDisplayDataFuncDestroyViewable destroy_viewable,
     KDisplayDataFuncCreateBlinkState create_blinkstate,
     KDisplayDataFuncDestroyBlinkState destroy_blinkstate,
     KDisplayDataFuncActivateBlinkState activate_blinkstate,
     KDisplayDataFuncCreateBlinkEntry create_blinkentry,
     KDisplayDataFuncDestroyBlinkEntry destroy_blinkentry,
     KDisplayDataFuncCopyBlinkEntry copy_blinkentry,
     KDisplayDataFuncActivateBlinkEntry activate_blinkentry,
     KDisplayDataFuncSetBlinkEntryAttributes set_blinkentry_attributes,
     KDisplayDataFuncGetBlinkEntryAttributes get_blinkentry_attributes)
/*  [SUMMARY] Create a display data class (i.e. array, overlay, etc.).
    <name> The type of data this dataclass manages.
    [RETURNS] A KDisplayDataClass object on success, else NULL.
*/
{
    KDisplayDataClass dataclass;
    static char function_name[] = "dispdata_create_dataclass";

    if (window_was_created)
    {
	fprintf (stderr, "Cannot create dataclass after windows created\n");
	a_prog_bug (function_name);
    }
    if ( ( dataclass = m_calloc (sizeof *dataclass) ) == NULL )
    {
	m_error_notify (function_name, "KDisplayDataClass object");
	return (NULL);
    }
    if ( ( dataclass->name = st_dup (name) ) == NULL )
    {
	m_error_notify (function_name, "KDisplayDataClass name");
	m_free (dataclass);
	return (NULL);
    }
    if ( ( dataclass->names = m_alloc (sizeof *dataclass->names * 16) )
	 == NULL )
    {
	m_error_notify (function_name, "name list");
	m_free (dataclass->name);
	m_free (dataclass);
	return (NULL);
    }
    if ( ( dataclass->objects = m_alloc (sizeof *dataclass->objects * 16) )
	 == NULL )
    {
	m_error_notify (function_name, "object list");
	m_free (dataclass->names);
	m_free (dataclass->name);
	m_free (dataclass);
	return (NULL);
    }
    dataclass->namebuf_length = 16;
    dataclass->magic_number = DATACLASS_MAGIC_NUMBER;
    KLIST_APPEND2 (first_dataclass, last_dataclass, dataclass);
    dataclass->set_class_attributes = set_class_attributes;
    dataclass->destroy_dataobject = destroy_dataobject;
    dataclass->copy_data_attributes = copy_data_attributes;
    dataclass->set_data_attributes = set_data_attributes;
    dataclass->get_data_attributes = get_data_attributes;
    dataclass->create_window = create_window;
    dataclass->destroy_window = destroy_window;
    dataclass->set_window_attributes = set_window_attributes;
    dataclass->get_window_attributes = get_window_attributes;
    dataclass->create_viewable = create_viewable;
    dataclass->destroy_viewable = destroy_viewable;
    dataclass->create_blinkstate = create_blinkstate;
    dataclass->destroy_blinkstate = destroy_blinkstate;
    dataclass->activate_blinkstate = activate_blinkstate;
    dataclass->create_blinkentry = create_blinkentry;
    dataclass->destroy_blinkentry = destroy_blinkentry;
    dataclass->copy_blinkentry = copy_blinkentry;
    dataclass->activate_blinkentry = activate_blinkentry;
    dataclass->set_blinkentry_attributes = set_blinkentry_attributes;
    dataclass->get_blinkentry_attributes = get_blinkentry_attributes;
    ++num_dataclasses;
    return (dataclass);
}   /*  End Function dispdata_create_dataclass  */

/*EXPERIMENTAL_FUNCTION*/
unsigned int dispdata_get_classlist (KDisplayDataClass **classes,char ***names)
/*  [SUMMARY] Get the list of supported dataclasses.
    <classes> The list of KDisplayDataClass objects is written here. This may
    be subsequently deallocated.
    <names> The list of name pointers is written here. This may be subsequently
    deallocated.
    [RETURNS] The number of dataclasses. On failure the process aborts.
*/
{
    KDisplayDataClass dataclass;
    unsigned int count;
    static char function_name[] = "dispdata_get_classlist";

    if ( ( *classes = m_alloc (sizeof **classes * num_dataclasses) ) == NULL )
	m_abort (function_name, "dataclass array");
    if ( ( *names = m_alloc (sizeof *names * num_dataclasses) ) == NULL )
	m_abort (function_name, "name array");
    for (dataclass = first_dataclass, count = 0; dataclass;
	 dataclass = dataclass->next, ++count)
    {
	(*classes)[count] = dataclass;
	(*names)[count] = dataclass->name;
    }
    return (count);
}   /*  End Function dispdata_get_classlist  */

/*EXPERIMENTAL_FUNCTION*/
KCallbackFunc dispdata_register_data_create_func (KDisplayDataClass dataclass,
						  void (*func)
						  (void *info,
						   KDisplayDataObject data),
						  void *info)
/*  [SUMMARY] Register a function to be called when a dataobject is created.
    <dataclass> The KDisplayDataClass the data belong to.
    <func> The function that is called when a new dataobject is created.
    <info> An arbitrary information pointer passed to <<func>>.
    [RETURNS] A KCallbackFunc object. On failure, the process aborts.
*/
{
    static char function_name[] = "dispdata_register_data_create_func";

    VERIFY_DATACLASS (dataclass);
    return c_register_callback (&dataclass->create_list,
				_dispdata_create_destroy_data_event_func, info,
				(void *) func, FALSE, NULL, FALSE, FALSE);
}   /*  End Function dispdata_register_data_create_func  */

/*EXPERIMENTAL_FUNCTION*/
KCallbackFunc dispdata_register_data_destroy_func (KDisplayDataClass dataclass,
						   void (*func)
						   (void *info),
						   void *info)
/*  [SUMMARY] Register a function to be called when a dataobject is destroyed.
    <dataclass> The KDisplayDataClass the data belong to.
    <func> The function that is called when a dataobject is destroyed.
    <info> An arbitrary information pointer passed to <<func>>.
    [RETURNS] A KCallbackFunc object. On failure, the process aborts.
*/
{
    static char function_name[] = "dispdata_register_data_destroy_func";

    VERIFY_DATACLASS (dataclass);
    return c_register_callback (&dataclass->destroy_list,
				_dispdata_create_destroy_data_event_func, info,
				(void *) func, FALSE, NULL, FALSE, FALSE);
}   /*  End Function dispdata_register_data_destroy_func  */

/*EXPERIMENTAL_FUNCTION*/
KCallbackFunc dispdata_register_dataobject_destroy_func
    (KDisplayDataObject dataobject, void (*func) (void *info), void *info)
/*  [SUMMARY] Register a function to be called when a dataobject is destroyed.
    <dataobject> The dataobject for which the callback function is registered.
    <func> The function that is called when the dataobject is destroyed.
    <info> An arbitrary information pointer passed to <<func>>.
    [RETURNS] A KCallbackFunc object. On failure, the process aborts.
*/
{
    static char function_name[] = "dispdata_register_dataobject_destroy_func";

    VERIFY_DATAOBJECT (dataobject);
    return c_register_callback (&dataobject->destroy_list,
				( flag (*) () ) func, info,
				NULL, FALSE, NULL, FALSE, FALSE);
}   /*  End Function dispdata_register_dataobject_destroy_func  */

/*EXPERIMENTAL_FUNCTION*/
void dispdata_set_dataclass_attributes (KDisplayDataClass dataclass, ...)
/*  [SUMMARY] Set global attributes for the a dataclass.
    <dataclass> The KDisplayDataClass the attributes are destined for.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    KDISPLAYDATA_CLASS_ATT_END.
    [NOTE] Some windows may be refreshed.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    static char function_name[] = "dispdata_set_dataclass_attributes";

    VERIFY_DATACLASS (dataclass);
    /*  This interface function is needed because the stdarg interface is
	broken and doesn't allow functions with no variable arguments  */
    va_start (argp, dataclass);
    (*dataclass->set_class_attributes) (argp);
    va_end (argp);
}   /*  End Function dispdata_set_dataclass_attributes  */

/*EXPERIMENTAL_FUNCTION*/
void dispdata_get_dataclass_attributes (KDisplayDataClass dataclass, ...)
/*  [SUMMARY] Get global attributes for the a dataclass.
    <dataclass> The KDisplayDataClass the attributes are destined for.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    KDISPLAYDATA_CLASS_ATT_END.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    unsigned int att_key;
    static char function_name[] = "dispdata_get_dataclass_attributes";

    VERIFY_DATACLASS (dataclass);
    /*  First process the generic attributes  */
    va_start (argp, dataclass);
    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    KDISPLAYDATA_CLASS_ATT_END )
    {
	switch (att_key)
	{
	  case KDISPLAYDATA_CLASS_ATT_CURRENTLY_CREATING:
	    *va_arg (argp, KDisplayDataObject *) =
		dataclass->currently_creating;
	    break;
	}
    }
    va_end (argp);
}   /*  End Function dispdata_get_dataclass_attributes  */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataWindow dispdata_create_window (KDisplayDataCanvas pc,
					   KDisplayDataCanvas rgb,
					   void (*show_canvas)
					   (void *info, KWorldCanvas canvas),
					   void *info )
/*  [SUMMARY] Create a display data window.
    <pc> The PseudoColour canvas. This may contain NULL KWorldCanvas objects.
    <rgb> The TrueColour/DirectColour canvas. This may contain NULL
    KWorldCanvas objects.
    <show_canvas> The function that is called when a canvas needs to be shown.
    <info> An arbitrary information pointer passed to <<show_canvas>>.
    [NOTE] If either of the main canvases is destroyed, the KDisplayDataWindow
    object is automatically destroyed.
    [RETURNS] A KDisplayDataWindow object on success, else NULL.
*/
{
    KDisplayDataWindow window;
    KDisplayDataClass dataclass;
    static char function_name[] = "dispdata_create_window";

    if (!pc.main && !rgb.main)
    {
	fprintf (stderr, "NULL canvases passes\n");
	a_prog_bug (function_name);
    }
    if ( ( window = m_calloc (sizeof *window) ) == NULL )
    {
	m_error_notify (function_name, "KDisplayWindow object");
	return (NULL);
    }
    window_was_created = TRUE;
    window->magic_number = WINDOW_MAGIC_NUMBER;
    if (last_window) window->index = last_window->index + 1;
    KLIST_APPEND2 (first_window, last_window, window);
    window->pc.canvas = pc;
    window->rgb.canvas = rgb;
    window->show_canvas = show_canvas;
    window->info = info;
    /*  Create window data objects for all data objects  */
    for (dataclass = first_dataclass; dataclass; dataclass = dataclass->next)
    {
	KDisplayDataObject dataobject;
	struct displaydata_window_class_entry *winclass;

	if ( ( winclass = m_calloc (sizeof *winclass) ) == NULL )
	{
	    m_error_notify (function_name, "window object");
	    dispdata_destroy_window (window);
	}
	winclass->dataclass = dataclass;
	if ( ( winclass->subclass = (*dataclass->create_window) (window,
							       &pc, &rgb) )
	     == NULL )
	{
	    m_free (winclass);
	    dispdata_destroy_window (window);
	}
	KLIST_APPEND (window, winclass);
	for (dataobject = dataclass->first; dataobject;
	     dataobject = dataobject->next)
	{
	    struct displaydata_window_dataobject_type *windataobject;

	    if ( ( windataobject = m_calloc (sizeof *windataobject) ) == NULL )
	    {
		m_error_notify (function_name, "window data object");
		dispdata_destroy_window (window);
	    }
	    windataobject->dataobject = dataobject;
	    KLIST_APPEND (&window->dataobjects, windataobject);
	    if ( ( windataobject->viewable =
		   (*dataclass->create_viewable) (dataobject->subclass,
						  winclass->subclass) )
		 == NULL )
	    {
		m_free (windataobject);
		dispdata_destroy_window (window);
	    }
	}
    }
    if (pc.main)
	window->pc.destroy_func =
	    canvas_register_destroy_func (pc.main, FALSE,
					  _dispdata_canvas_destroy_callback,
					  window);
    if (rgb.main)
	window->rgb.destroy_func =
	    canvas_register_destroy_func (rgb.main, FALSE,
					  _dispdata_canvas_destroy_callback,
					  window);
    return (window);
}   /*  End Function dispdata_create_window  */

/*EXPERIMENTAL_FUNCTION*/
void dispdata_destroy_window (KDisplayDataWindow window)
/*  [SUMMARY] Destroy a display data window.
    <window> The KDisplayDataWindow to destroy.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "dispdata_destroy_window";

    VERIFY_WINDOW (window);
    _dispdata_destroy_window (window, FALSE);
}   /*  End Function dispdata_destroy_window  */

/*EXPERIMENTAL_FUNCTION*/
KCallbackFunc dispdata_register_window_destroy_func (KDisplayDataWindow window,
						     void (*func) (),
						     void *info)
/*  [SUMMARY] Register destroy callback.
    [PURPOSE] This routine will register a destroy function for a display data
    window. The function will be called when the window is destroyed. Many
    destroy functions may be registered per window. The first function
    registered is the first function called upon destruction.
    <window> The KDisplayDataWindow.
    <func> The function that is called when the window is destroyed. The
    prototype function is [<dispdata_PROTO_destroy_func>].
    <info> The arbitrary function information pointer.
    [RETURNS] A KCallbackFunc object. On failure, the process aborts.
*/
{
    static char function_name[] = "dispdata_register_window_destroy_func";

    VERIFY_WINDOW (window);
    return c_register_callback (&window->destroy_list,
				( flag (*) () ) func, info,
				NULL, FALSE, NULL, FALSE, FALSE);
}   /*  End Function dispdata_register_window_destroy_func  */

/*EXPERIMENTAL_FUNCTION*/
void dispdata_set_window_attributes (KDisplayDataWindow window, ...)
/*  [SUMMARY] Set the attributes for a window.
    <window> The KDisplayDataWindow.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    KDISPLAYDATA_WINDOW_ATT_END.
    [NOTE] Some windows may be refreshed.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    unsigned int att_key;
    struct displaydata_window_class_entry *winclass;
    static char function_name[] = "dispdata_set_window_attributes";

    VERIFY_WINDOW (window);
    /*  First process the generic attributes  */
    va_start (argp, window);
    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    KDISPLAYDATA_WINDOW_ATT_END )
    {
	switch (att_key)
	{
	}
    }
    va_end (argp);
    /*  Now process the class-specific attributes  */
    for (winclass = window->first; winclass; winclass = winclass->next)
    {
	va_start (argp, window);
	(*winclass->dataclass->set_window_attributes) (winclass->subclass,
						       argp);
	va_end (argp);
    }
}   /*  End Function dispdata_set_window_attributes  */

/*EXPERIMENTAL_FUNCTION*/
void dispdata_get_window_attributes (KDisplayDataWindow window, ...)
/*  [SUMMARY] Get the attributes for a window object.
    <window> The KDisplayDataWindow.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    KDISPLAYDATA_WINDOW_ATT_END.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    unsigned int att_key;
    struct displaydata_window_class_entry *winclass;
    static char function_name[] = "dispdata_get_window_attributes";

    VERIFY_WINDOW (window);
    /*  First process the generic attributes  */
    va_start (argp, window);
    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    KDISPLAYDATA_WINDOW_ATT_END )
    {
	switch (att_key)
	{
	}
    }
    va_end (argp);
    /*  Now process the class-specific attributes  */
    for (winclass = window->first; winclass; winclass = winclass->next)
    {
	va_start (argp, window);
	(*winclass->dataclass->get_window_attributes) (winclass->subclass,
						       argp);
	va_end (argp);
    }
}   /*  End Function dispdata_get_window_attributes  */

/*EXPERIMENTAL_FUNCTION*/
unsigned int dispdata_get_num_windows ()
/*  [SUMMARY] Get the number of windows.
    [RETURNS] The number of windows.
*/
{
    return last_window ? (last_window->index + 1) : 0;
}   /*  End Function dispdata_get_num_windows  */

/*EXPERIMENTAL_FUNCTION*/
void dispdata_refresh_window (KDisplayDataWindow window, KWorldCanvas canvas)
/*  [SUMMARY] Refresh a window.
    <window> The KDisplayDataWindow object to refresh.
    <canvas> The preferred KWorldCanvas to display. This may be NULL.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "dispdata_refresh_window";

    VERIFY_WINDOW (window);
    if (canvas) (*window->show_canvas) (window->info, canvas);
    else
    {  /*  Panner canvas is refreshed when main canvas is refreshed  */
	canvas_refresh_if_visible (window->pc.canvas.main, FALSE);
	canvas_refresh_if_visible (window->pc.canvas.magnifier, FALSE);
	if (window->pc.canvas.main == window->rgb.canvas.main) return;
	canvas_refresh_if_visible (window->rgb.canvas.main, FALSE);
	canvas_refresh_if_visible (window->rgb.canvas.magnifier, FALSE);
    }
    window->refresh = FALSE;
}   /*  End Function dispdata_refresh_window  */

/*EXPERIMENTAL_FUNCTION*/
unsigned int dispdata_get_num_blinkstates (KDisplayDataWindow window)
/*  [SUMMARY] Get the number of blinkstates for a window.
    <window> The KDisplayDataWindow object.
    [RETURNS] The number of blinkstates.
*/
{
    static char function_name[] = "dispdata_get_num_blinkstates";

    VERIFY_WINDOW (window);
    return (window->num_blinkstates);
}   /*  End Function dispdata_get_num_blinkstates  */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataBlinkState dispdata_get_active_blinkstate
    (KDisplayDataWindow window, void **info)
/*  [SUMMARY] Get the active KDisplayDataBlinkState for a window.
    <window> The window.
    <info> The arbitrary information pointer is written here. If this is NULL,
    nothing is written here.
    [RETURNS] The active blink state. NULL is returned if there is no active
    blink state.
*/
{
    static char function_name[] = "dispdata_get_active_blinkstate";

    VERIFY_WINDOW (window);
    if (info) *info = window->active ? window->active->info : NULL;
    return (window->active);
}   /*  End Function dispdata_get_active_blinkstate  */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataWindow dispdata_get_first_window (void **info)
/*  [SUMMARY] Get first window.
    <info> The arbitrary information pointer is written here. If this is NULL,
    nothing is written here.
    [RETURNS] A KDisplayDataWindow object. NULL is returned if there are no
    windows.
*/
{
    /*static char function_name[] = "dispdata_get_first_window";*/

    if (info) *info = first_window ? first_window->info : NULL;
    return (first_window);
}   /*  End Function dispdata_get_first_window  */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataWindow dispdata_get_prev_window (KDisplayDataWindow window,
					     flag circular, void **info)
/*  [SUMMARY] Get previous window.
    <window> The current window.
    <circular> If TRUE, the list of windows is treated as circular.
    <info> The arbitrary information pointer is written here. If this is NULL,
    nothing is written here.
    [RETURNS] A KDisplayDataWindow object. NULL is returned if there are no
    more windows.
*/
{
    static char function_name[] = "dispdata_get_prev_window";

    VERIFY_WINDOW (window);
    window = window->prev;
    if (!window && circular) window = last_window;
    if (info && window) *info = window->info;
    return (window);
}   /*  End Function dispdata_get_prev_window  */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataWindow dispdata_get_next_window (KDisplayDataWindow window,
					     flag circular, void **info)
/*  [SUMMARY] Get next window.
    <window> The current window.
    <circular> If TRUE, the list of windows is treated as circular.
    <info> The arbitrary information pointer is written here. If this is NULL,
    nothing is written here.
    [RETURNS] A KDisplayDataWindow object. NULL is returned if there are no
    more windows.
*/
{
    static char function_name[] = "dispdata_get_next_window";

    VERIFY_WINDOW (window);
    window = window->next;
    if (!window && circular) window = first_window;
    if (info) *info = window->info;
    return (window);
}   /*  End Function dispdata_get_next_window  */

/*EXPERIMENTAL_FUNCTION*/
unsigned int dispdata_get_window_index (KDisplayDataWindow window)
/*  [SUMMARY] Get the index for a window.
    <window> The KDisplayDataWindow object.
    [RETURNS] The index of the window.
*/
{
    static char function_name[] = "dispdata_get_window_index";

    VERIFY_WINDOW (window);
    return (window->index);
}   /*  End Function dispdata_get_window_index  */

/*EXPERIMENTAL_FUNCTION*/
KCallbackFunc dispdata_register_window_renumber_func(KDisplayDataWindow window,
						     void (*func) (),
						     void *info)
/*  [SUMMARY] Register renumber callback.
    [PURPOSE] This routine will register a renumber function for a display data
    window. The function will be called when the window index changes (due to
    renumbering). Many renumber functions may be registered per window. The
    first function registered is the first function called upon renumbering.
    <window> The KDisplayDataWindow.
    <func> The function that is called when the window is renumbered. The
    prototype function is [<dispdata_PROTO_renumber_func>].
    <info> The arbitrary function information pointer.
    [RETURNS] A KCallbackFunc object. On failure, the process aborts.
*/
{
    static char function_name[] = "dispdata_register_window_renumber_func";

    VERIFY_WINDOW (window);
    return c_register_callback (&window->renumber_list,
				( flag (*) () ) func, info,
				NULL, FALSE, NULL, FALSE, FALSE);
}   /*  End Function dispdata_register_window_renumber_func  */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataObject dispdata_create_dataobject (KDisplayDataClass dataclass,
					       CONST char *domain,
					       CONST char *name,
					       void *data)
/*  [SUMMARY] Create a display data object (i.e. array, overlay, etc.).
    <dataclass> The KDisplayDataClass the data belong to.
    <domain> The domain from where data came from. The string contents are
    copied.
    <name> The name of the object. The string contents are copied.
    <data> The data for the object.
    [RETURNS] A KDisplayDataObject object on success, else NULL.
*/
{
    KDisplayDataWindow window;
    KDisplayDataObject dataobject, next_dataobject;
    KDisplayDataObject olddataobject = NULL;
    static flag first_time = TRUE;
    static flag alphabetic_sorting = TRUE;
    static char function_name[] = "dispdata_create_dataobject";

    VERIFY_DATACLASS (dataclass);
    if (!domain || !name || !data)
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    if (first_time)
    {
	first_time = FALSE;
	if ( r_getenv ("DISPDATA_SORT_BY_LOAD_ORDER") )
	    alphabetic_sorting = FALSE;
    }
    dataclass->namelist_valid = FALSE;
    /*  Ensure namelist has enough room  */
    if (dataclass->num_sets >= dataclass->namebuf_length)
    {
	KDisplayDataObject *objects;
	char **names;

	if ( ( names =
	       m_alloc (sizeof *names * dataclass->namebuf_length * 2) )
	     == NULL )
	{
	    m_error_notify (function_name, "name list");
	    return (NULL);
	}
	if ( ( objects =
	       m_alloc (sizeof *objects * dataclass->namebuf_length * 2) )
	     == NULL )
	{
	    m_error_notify (function_name, "object list");
	    m_free (names);
	    return (NULL);
	}
	if (dataclass->names) m_free (dataclass->names);
	if (dataclass->objects) m_free (dataclass->objects);
	dataclass->names = names;
	dataclass->objects = objects;
	dataclass->namebuf_length *= 2;
    }
    if ( ( dataobject = m_calloc (sizeof *dataobject) ) == NULL )
    {
	m_error_notify (function_name, "KDisplayDataObject object");
	return (NULL);
    }
    dataobject->dataclass = dataclass;
    dataobject->subclass = data;
    if ( ( dataobject->domain = st_dup (domain) ) == NULL )
    {
	m_error_notify (function_name, "domain name");
	m_free (dataobject);
	return (NULL);
    }
    if (strstr (domain, "FILE") != NULL)
    {   /*  Extract directory name and strip from name  */
	CONST char *ptr;
	char *directory;
	char dirpath[PATH_MAX], realdir[PATH_MAX];

	if ( ( ptr = strrchr (name, '/') ) != NULL )
	{
	    unsigned int length;

	    length = ptr - name;
	    if (length >= PATH_MAX - 1) length = PATH_MAX - 2;
	    strncpy (dirpath, name, length);
	    dirpath[length] = '\0';
	    name = ptr + 1;
	}
	else strcpy (dirpath, ".");
	if ( ( directory = realpath (dirpath, realdir) ) == NULL )
	{
	    fprintf (stderr, "%s: error converting path: \"%s\"\t%s\n",
		     function_name, dirpath, ERRSTRING);
	    directory = dirpath;
	}
	if ( ( dataobject->directory = st_dup (directory) ) == NULL )
	{
	    m_error_notify (function_name, "directory name");
	    m_free (dataobject->domain);
	    m_free (dataobject);
	    return (NULL);
	}
    }
    if ( ( dataobject->name = st_dup (name) ) == NULL )
    {
	m_error_notify (function_name, "dataobject name");
	if (dataobject->directory) m_free (dataobject->directory);
	m_free (dataobject->domain);
	m_free (dataobject);
	return (NULL);
    }
    if (dataclass->num_replace_flags > 0)
    {   /*  Some datasets want to be replaced: search for a match  */
	for (olddataobject = dataclass->first; olddataobject;
	     olddataobject = olddataobject->next)
	    if (olddataobject->replace &&
		(strcmp (olddataobject->domain, domain) == 0) ) break;
    }
    if (olddataobject)
    {
	if ( !(*dataclass->copy_data_attributes) (dataobject->subclass,
						  olddataobject->subclass) )
	{
	    m_free (dataobject->name);
	    if (dataobject->directory) m_free (dataobject->directory);
	    m_free (dataobject->domain);
	    m_free (dataobject);
	    return (NULL);
	}
	dataobject->associations = olddataobject->associations;
	dataobject->associations_destroy_list =
	    olddataobject->associations_destroy_list;
	olddataobject->associations = NULL;
	olddataobject->associations_destroy_list = NULL;
	dataobject->replace = TRUE;
	++dataclass->num_replace_flags;
    }
    else if ( ( dataobject->associations = j_create (NULL, strcmp, NULL,NULL) )
	      == NULL )
    {
	m_error_notify (function_name, "dataobject association list");
	m_free (dataobject->name);
	if (dataobject->directory) m_free (dataobject->directory);
	m_free (dataobject->domain);
	m_free (dataobject);
	return (NULL);
    }
    dataobject->magic_number = DATAOBJECT_MAGIC_NUMBER;
    ++dataclass->num_sets;
    dataclass->currently_creating = dataobject;
    /*  Insert dataobject into dataclass list at sorted position  */
    if (alphabetic_sorting)
    {
	if (dataclass->first)
	{
	    for (next_dataobject = dataclass->first; next_dataobject;
		 next_dataobject = next_dataobject->next)
		if (strcmp (name, next_dataobject->name) < 0) break;
	}
	else next_dataobject = NULL;
    }
    else next_dataobject = olddataobject;
    if (next_dataobject)
    {   /*  Need braces  */
	KLIST_INSERT (dataclass, next_dataobject, dataobject);
    }
    else KLIST_APPEND (dataclass, dataobject);
    /*  For each window and blink state, create and insert blink entry  */
    for (window = first_window; window; window = window->next)
    {
	KDisplayDataBlinkState blinkstate;
	struct displaydata_window_dataobject_type *windataobject;
	struct displaydata_window_class_entry *winclass = NULL;

	/*  Find class entry for the dataclass in this window  */
	for (winclass = window->first; winclass;
	     winclass = winclass->next)
	    if (winclass->dataclass == dataclass) break;
	if (winclass == NULL)
	{
	    fprintf (stderr, "Classentry for window not found!\n");
	    a_prog_bug (function_name);
	}
	if ( ( windataobject = m_calloc (sizeof *windataobject) ) == NULL )
	    m_abort (function_name, "window data object");
	windataobject->dataobject = dataobject;
	KLIST_APPEND (&window->dataobjects, windataobject);
	if ( ( windataobject->viewable =
	       (*dataclass->create_viewable) (dataobject->subclass,
					      winclass->subclass) )
	     == NULL )
	    m_abort (function_name, "specific window data object");
	for (blinkstate = window->blinkstates.first; blinkstate;
	     blinkstate = blinkstate->next)
	{
	    KDisplayDataClassBlinkState classentry;
	    KDisplayDataBlinkEntry blinkentry, newblinkentry;
	    flag visible = FALSE;

	    classentry = dispdata_get_blinkstate_class (blinkstate, dataclass);
	    if ( ( newblinkentry = m_calloc (sizeof *newblinkentry) ) == NULL )
		m_abort (function_name, "blink entry");
	    newblinkentry->windataobject = windataobject;
	    newblinkentry->magic_number = BLINKENTRY_MAGIC_NUMBER;
	    if ( ( newblinkentry->subclass =
		   (*dataclass->create_blinkentry) (classentry->subclass,
						    windataobject->viewable) )
		 == NULL )
		m_abort (function_name, "specific blink entry state");
	    if (classentry->create_blinkentry)
	    {
		if ( ( newblinkentry->info =
		       (*classentry->create_blinkentry) (classentry->info,
							 newblinkentry) )
		     == NULL )
		    m_abort (function_name, "application blink entry state");
	    }
	    /*  Search for location to insert blink entry into  */
	    for (blinkentry = classentry->first; blinkentry;
		 blinkentry = blinkentry->next)
		if (blinkentry->windataobject->dataobject == next_dataobject)
		    break;
	    if (blinkentry)
	    {   /*  Need braces  */
		KLIST_INSERT (classentry, blinkentry, newblinkentry);
	    }
	    else KLIST_APPEND (classentry, newblinkentry);
	    if (olddataobject)
	    {   /*  Search for old blinkentry to copy  */
		for (blinkentry = classentry->first; blinkentry;
		     blinkentry = blinkentry->next)
		    if (blinkentry->windataobject->dataobject == olddataobject)
			break;
		(*dataclass->copy_blinkentry) (newblinkentry->subclass,
					       blinkentry->subclass);
		(*classentry->copy_blinkentry) (newblinkentry->info,
						blinkentry->info);
	    }
	    if (!olddataobject && classentry->refresh_func)
		(*classentry->refresh_func) (classentry->info,
					     classentry->dataclass);
	    if (blinkstate == window->active)
	    {
		(*dataclass->activate_blinkentry) (newblinkentry->subclass,
						   &visible);
		if (visible) window->refresh = TRUE;
	    }
	}
    }
    if (olddataobject)
    {
	dispdata_destroy_dataobject (olddataobject);
	for (window = first_window; window; window = window->next)
	{
	    if (window->refresh) dispdata_refresh_window (window, NULL);
	}
    }
    c_call_callbacks (dataclass->create_list, dataobject);
    dataclass->currently_creating = NULL;
    return (dataobject);
}   /*  End Function dispdata_create_dataobject  */

/*EXPERIMENTAL_FUNCTION*/
void dispdata_destroy_dataobject (KDisplayDataObject dataobject)
/*  [SUMMARY] Destroy a display data object (i.e. array, overlay, etc.).
    <dataobject> The KDisplayDataObject to destroy.
    [RETURNS] Nothing.
*/
{
    KDisplayDataWindow window;
    static char function_name[] = "dispdata_destroy_dataobject";

    VERIFY_DATAOBJECT (dataobject);
    dataobject->dataclass->namelist_valid = FALSE;
    --dataobject->dataclass->num_sets;
    if (dataobject->replace) --dataobject->dataclass->num_replace_flags;
    KLIST_REMOVE (dataobject->dataclass, dataobject);
    if (dataobject->associations) j_destroy (dataobject->associations);
    c_call_callbacks (dataobject->associations_destroy_list, dataobject);
    c_destroy_list (dataobject->associations_destroy_list);
    c_call_callbacks (dataobject->destroy_list, dataobject);
    c_call_callbacks (dataobject->dataclass->destroy_list, NULL);
    c_destroy_list (dataobject->destroy_list);
    /*  For each window and blink state, remove and free blink entry and
	data objects  */
    for (window = first_window; window; window = window->next)
    {
	KDisplayDataBlinkState blinkstate;
	struct displaydata_window_dataobject_type *windataobject;

	/*  First destroy the blinkentry for each blinkstate  */
	for (blinkstate = window->blinkstates.first; blinkstate;
	     blinkstate = blinkstate->next)
	{
	    KDisplayDataClassBlinkState classentry = NULL;
	    KDisplayDataBlinkEntry blinkentry;
	    flag was_active;

	    /*  Find class entry for the dataclass in this blink state  */
	    for (classentry = blinkstate->first; classentry;
		 classentry = classentry->next)
		if (classentry->dataclass == dataobject->dataclass) break;
	    if (classentry == NULL)
	    {
		fprintf (stderr, "Classentry not found!\n");
		a_prog_bug (function_name);
	    }
	    /*  Search for blink entry with corresponding data object  */
	    for (blinkentry = classentry->first; blinkentry;
		 blinkentry = blinkentry->next)
		if (blinkentry->windataobject->dataobject == dataobject) break;
	    if (classentry->destroy_blinkentry)
		(*classentry->destroy_blinkentry) (blinkentry->info);
	    was_active = (*dataobject->dataclass->destroy_blinkentry)
		(blinkentry->subclass);
	    if ( was_active && (blinkstate == window->active) )
		window->refresh = TRUE;
	    KLIST_REMOVE (classentry, blinkentry);
	    blinkentry->magic_number = 0;
	    m_free (blinkentry);
	}
	/*  Now destroy the viewable  */
	for (windataobject = window->dataobjects.first; windataobject;
	     windataobject = windataobject->next)
	    if (windataobject->dataobject == dataobject) break;
	(*dataobject->dataclass->destroy_viewable) (windataobject->viewable,
						    FALSE);
	KLIST_REMOVE (&window->dataobjects, windataobject);
	m_free (windataobject);
	/*  Finally call the refresh functions for each blinkstate  */
	for (blinkstate = window->blinkstates.first; blinkstate;
	     blinkstate = blinkstate->next)
	{
	    KDisplayDataClassBlinkState classentry = NULL;

	    /*  Find class entry for the dataclass in this blink state  */
	    for (classentry = blinkstate->first; classentry;
		 classentry = classentry->next)
		if (classentry->dataclass == dataobject->dataclass) break;
	    if (classentry == NULL)
	    {
		fprintf (stderr, "Classentry not found!\n");
		a_prog_bug (function_name);
	    }
	    if (classentry->refresh_func)
		(*classentry->refresh_func) (classentry->info,
					     classentry->dataclass);
	}
	if (window->refresh) dispdata_refresh_window (window, NULL);
    }
    (*dataobject->dataclass->destroy_dataobject) (dataobject->subclass);
    if (dataobject->domain) m_free (dataobject->domain);
    if (dataobject->directory) m_free (dataobject->directory);
    m_free (dataobject->name);
    dataobject->magic_number = 0;
    m_free (dataobject);
}   /*  End Function dispdata_destroy_dataobject  */

/*EXPERIMENTAL_FUNCTION*/
void *dispdata_get_dataobject_subclass (KDisplayDataObject dataobject,
					KDisplayDataClass dataclass)
/*  [SUMMARY] Get the subclass for a KDisplayDataObject.
    <dataobject> The KDisplayDataObject.
    <dataclass> The KDisplayDataClass the data belong to.
    [RETURNS] A pointer to the subclass.
*/
{
    static char function_name[] = "dispdata_get_dataobject_subclass";

    VERIFY_DATAOBJECT (dataobject);
    VERIFY_DATACLASS (dataclass);
    if (dataobject->dataclass != dataclass)
    {
	fprintf (stderr, "Dataclass mis-match\n");
	a_prog_bug (function_name);
    }
    return (dataobject->subclass);
}   /*  End Function dispdata_get_dataobject_subclass  */

/*EXPERIMENTAL_FUNCTION*/
void dispdata_set_data_attributes (KDisplayDataObject dataobject, ...)
/*  [SUMMARY] Set the attributes for a data object.
    <dataobject> The KDisplayDataObject.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    KDISPLAYDATA_DATA_ATT_END.
    [NOTE] Some windows may be refreshed.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    KDisplayDataWindow window;
    flag keep_going = TRUE;
    flag bool;
    flag refresh = FALSE;
    static char function_name[] = "dispdata_set_data_attributes";

    /*  First process the generic attributes  */
    va_start (argp, dataobject);
    while (keep_going)
    {
	unsigned int att_key = va_arg (argp, unsigned int);

	switch (att_key)
	{
	  case KDISPLAYDATA_DATA_ATT_REPLACE:
	    bool = va_arg (argp, flag);
	    FLAG_VERIFY (bool);
	    if (bool != dataobject->replace) refresh = TRUE;
	    if (bool && !dataobject->replace)
		++dataobject->dataclass->num_replace_flags;
	    else if (!bool && dataobject->replace)
		--dataobject->dataclass->num_replace_flags;
	    dataobject->replace = bool;
	    break;
	  default:
	    keep_going = FALSE;
	    break;
	}
    }
    va_end (argp);
    /*  Now process the class-specific attributes  */
    va_start (argp, dataobject);
    VERIFY_DATAOBJECT (dataobject);
    (*dataobject->dataclass->set_data_attributes) (dataobject->subclass, argp);
    va_end (argp);
    if (!refresh) return;
    /*  Refresh all blinkstate interfaces  */
    for (window = first_window; window; window = window->next)
    {
	KDisplayDataBlinkState blinkstate;

	for (blinkstate = window->blinkstates.first; blinkstate;
	     blinkstate = blinkstate->next)
	{
	    KDisplayDataClassBlinkState classentry;

	    classentry = dispdata_get_blinkstate_class (blinkstate,
							dataobject->dataclass);
	    (*classentry->refresh_func) (classentry->info,
					 classentry->dataclass);
	}
    }
}   /*  End Function dispdata_set_data_attributes  */

/*EXPERIMENTAL_FUNCTION*/
void dispdata_get_data_attributes (KDisplayDataObject dataobject, ...)
/*  [SUMMARY] Get the attributes for a data object.
    <dataobject> The KDisplayDataObject.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    KDISPLAYDATA_DATA_ATT_END.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    unsigned int att_key;
    void *dummy;
    static char function_name[] = "dispdata_get_data_attributes";

    VERIFY_DATAOBJECT (dataobject);
    /*  First process the generic attributes  */
    va_start (argp, dataobject);
    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    KDISPLAYDATA_DATA_ATT_END )
    {
	switch (att_key)
	{
	  case KDISPLAYDATA_DATA_ATT_DOMAIN:
	    *va_arg (argp, char **) = dataobject->domain;
	    break;
	  case KDISPLAYDATA_DATA_ATT_NAME:
	    *va_arg (argp, char **) = dataobject->name;
	    break;
	  case KDISPLAYDATA_DATA_ATT_DIRECTORY:
	    *va_arg (argp, char **) = dataobject->directory;
	    break;
	  case KDISPLAYDATA_DATA_ATT_REPLACE:
	    *va_arg (argp, flag *) = dataobject->replace;
	    break;
	  default:
	    dummy = va_arg (argp, void *);
	    break;
	}
    }
    va_end (argp);
    /*  Now process the class-specific attributes  */
    va_start (argp, dataobject);
    (*dataobject->dataclass->get_data_attributes) (dataobject->subclass, argp);
    va_end (argp);
}   /*  End Function dispdata_get_data_attributes  */

/*EXPERIMENTAL_FUNCTION*/
void dispdata_associate_data_object (KDisplayDataObject dataobject,
				     CONST char *name,
				     void (*destroy_func) (void *object),
				     void *object)
/*  [SUMMARY] Associate a named object with a data object.
    <dataobject> The KDisplayDataObject.
    <name> The name of the associated data object. The contents are copied.
    <destroy_func> The function that is called when [<dataobect>] is destroyed.
    <object> The object pointer.
    [RETURNS] Nothing. On failure, the process aborts.
*/
{
    static char function_name[] = "dispdata_associate_data_object";

    VERIFY_DATAOBJECT (dataobject);
    if (j_put_pair (dataobject->associations, name, strlen (name) + 1,
		    object, 0, NULL, NULL, KJ_REPLACEMENT_POLICY_NEW, FALSE)
	== NULL)
	m_abort (function_name, "association");
    if (destroy_func)
	c_register_callback (&dataobject->associations_destroy_list,
			     ( flag (*) () ) destroy_func, object,
			     NULL, FALSE, NULL, FALSE, FALSE);
}   /*  End Function dispdata_associate_data_object  */

/*EXPERIMENTAL_FUNCTION*/
void *dispdata_get_associated_data_object (KDisplayDataObject dataobject,
					   CONST char *name)
/*  [SUMMARY] Get an associated named object for a data object.
    <dataobject> The KDisplayDataObject.
    <name> The name of the associated data object.
    [RETURNS] A pointer to the associated object.
*/
{
    void *object;
    static char function_name[] = "dispdata_get_associated_data_object";

    VERIFY_DATAOBJECT (dataobject);
    if (j_get_pair (dataobject->associations, name, &object) == NULL)
	return (NULL);
    return (object);
}   /*  End Function dispdata_get_associated_data_object  */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataObject dispdata_get_first_dataobject (KDisplayDataClass dataclass)
/*  [SUMMARY] Get first data object for a dataclass.
    <dataclass> The KDisplayDataClass the data belong to.
    [RETURNS] A KDisplayDataObject object. NULL is returned if there are no
    dataobjects.
*/
{
    static char function_name[] = "dispdata_get_first_dataobject";

    VERIFY_DATACLASS (dataclass);
    return (dataclass->first);
}   /*  End Function dispdata_get_first_dataobject  */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataObject dispdata_get_next_dataobject (KDisplayDataObject dataobject)
/*  [SUMMARY] Get next data object.
    <dataobject> The current data object.
    [RETURNS] A KDisplayDataObject object. NULL is returned if there are no
    more data objects.
*/
{
    static char function_name[] = "dispdata_get_next_dataobject";

    VERIFY_DATAOBJECT (dataobject);
    return (dataobject->next);
}   /*  End Function dispdata_get_next_dataobject  */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataBlinkState dispdata_create_blinkstate
    (KDisplayDataWindow window,
     void (*active_change) (void *info, flag active),
     void *info)
/*  [SUMMARY] Create a blink state object.
    <window> The window the blink state is associated with.
    <active_change> The function that is called when the blinkstate is made
    active or inactive.
    <info> An arbitrary information pointer.
    [RETURNS] A KDisplayDataBlinkState object on success, else NULL.
*/
{
    struct displaydata_window_class_entry *winclass;
    KDisplayDataBlinkState blinkstate;
    static char function_name[] = "dispdata_create_blinkstate";

    VERIFY_WINDOW (window);
    if ( ( blinkstate = m_calloc (sizeof *blinkstate) ) == NULL )
    {
	m_error_notify (function_name, "KDisplayDataBlinkState object");
	return (NULL);
    }
    blinkstate->window = window;
    blinkstate->active_change = active_change;
    blinkstate->info = info;
    blinkstate->magic_number = BLINKSTATE_MAGIC_NUMBER;
    KLIST_APPEND (&window->blinkstates, blinkstate);
    if (window->indices_valid) blinkstate->index = window->num_blinkstates;
    ++window->num_blinkstates;
    /*  Create data class list for this blink state and create blink entries
     */
    for (winclass = window->first; winclass; winclass = winclass->next)
    {
	KDisplayDataObject dataobject;
	KDisplayDataClass dataclass = winclass->dataclass;
	KDisplayDataClassBlinkState classentry;

	if ( ( classentry = m_calloc (sizeof *classentry) ) == NULL )
	    m_error_notify (function_name, "class entry");
	else
	{
	    if ( ( classentry->subclass =
		   (*dataclass->create_blinkstate) (classentry,
						    winclass->subclass) )
		 == NULL )
	    {
		m_error_notify (function_name, "blinkstate");
		m_free (classentry);
		classentry = NULL;
	    }
	}
	if (classentry == NULL)  /*  Cleanup  */
	{
	    dispdata_destroy_blinkstate (blinkstate);
	    return (NULL);
	}
	classentry->magic_number = BLINKSTATE_CLASS_MAGIC_NUMBER;
	classentry->dataclass = dataclass;
	KLIST_APPEND (blinkstate, classentry);
	for (dataobject = dataclass->first; dataobject;
	     dataobject = dataobject->next)
	{
	    struct displaydata_window_dataobject_type *windataobject;
	    KDisplayDataBlinkEntry blinkentry;

	    for (windataobject = window->dataobjects.first; windataobject;
		 windataobject = windataobject->next)
		if (windataobject->dataobject == dataobject) break;
	    if ( ( blinkentry = m_calloc (sizeof *blinkentry) ) == NULL )
	    {
		m_error_notify (function_name, "blink entry");
		dispdata_destroy_blinkstate (blinkstate);
		return (NULL);
	    }
	    blinkentry->windataobject = windataobject;
	    if ( ( blinkentry->subclass =
		   (*dataclass->create_blinkentry) (classentry->subclass,
						    windataobject->viewable) )
		 == NULL )
	    {
		m_free (blinkentry);
		dispdata_destroy_blinkstate (blinkstate);
		return (NULL);
	    }
	    if (classentry->create_blinkentry)
	    {
		if ( ( blinkentry->info =
		       (*classentry->create_blinkentry) (classentry->info,
							 blinkentry) )
		     == NULL )
		{
		    (*dataclass->destroy_blinkentry) (blinkentry->subclass);
		    m_free (blinkentry);
		    dispdata_destroy_blinkstate (blinkstate);
		    return (NULL);
		}
	    }
	    blinkentry->magic_number = BLINKENTRY_MAGIC_NUMBER;
	    KLIST_APPEND (classentry, blinkentry);
	}
    }
    return (blinkstate);
}   /*  End Function dispdata_create_blinkstate  */

/*EXPERIMENTAL_FUNCTION*/
void dispdata_destroy_blinkstate (KDisplayDataBlinkState blinkstate)
/*  [SUMMARY] Destroy a blink state object.
    <blinkstate> The KDisplayBlinkState to destroy.
    [RETURNS] Nothing.
*/
{
    KDisplayDataClassBlinkState classentry, next_class;
    KDisplayDataWindow window;
    static char function_name[] = "dispdata_destroy_blinkstate";

    VERIFY_BLINKSTATE (blinkstate);
    window = blinkstate->window;
    window->indices_valid = FALSE;
    --window->num_blinkstates;
    KLIST_REMOVE (&window->blinkstates, blinkstate);
    if (window->active == blinkstate) window->active = NULL;
    for (classentry = blinkstate->first; classentry; classentry = next_class)
    {
	KDisplayDataBlinkEntry blinkentry, next_blinkentry;

	next_class = classentry->next;
	for (blinkentry = classentry->first; blinkentry;
	     blinkentry = next_blinkentry)
	{
	    next_blinkentry = blinkentry->next;
	    if (classentry->destroy_blinkentry)
		(*classentry->destroy_blinkentry) (blinkentry->info);
	    (*classentry->dataclass->destroy_blinkentry)
		(blinkentry->subclass);
	    blinkentry->magic_number = 0;
	    m_free (blinkentry);
	}
	(*classentry->dataclass->destroy_blinkstate) (classentry->subclass);
	if (classentry->blinkentries) m_free (classentry->blinkentries);
	classentry->magic_number = 0;
	m_free (classentry);
    }
    blinkstate->magic_number = 0;
    m_free (blinkstate);
}   /*  End Function dispdata_destroy_blinkstate  */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataClassBlinkState dispdata_get_blinkstate_class
    (KDisplayDataBlinkState blinkstate, KDisplayDataClass dataclass)
/*  [SUMMARY] Get the dataclass-specific object for a blink state.
    <blinkstate> The KDisplayDataBlinkState object.
    <dataclass> The KDisplayDataClass object.
    [RETURNS] A KDisplayDataClassBlinkState object.
*/
{
    KDisplayDataClassBlinkState entry;
    static char function_name[] = "dispdata_get_blinkstate_class";

    VERIFY_BLINKSTATE (blinkstate);
    VERIFY_DATACLASS (dataclass);
    for (entry = blinkstate->first; entry; entry = entry->next)
	if (entry->dataclass == dataclass) break;
    if (entry == NULL)
    {
	fprintf (stderr, "Classentry for blinkstate not found!\n");
	a_prog_bug (function_name);
    }
    return (entry);
}   /*  End Function dispdata_get_blinkstate_class  */

/*EXPERIMENTAL_FUNCTION*/
void dispdata_activate_blinkstate (KDisplayDataBlinkState blinkstate)
/*  [SUMMARY] Make a blinkstate active (change window contents).
    <blinkstate> The KDisplayDataBlinkState object to make active.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas wc = NULL;
    KDisplayDataWindow window;
    KDisplayDataClassBlinkState classentry;
    static char function_name[] = "dispdata_activate_blinkstate";

    VERIFY_BLINKSTATE (blinkstate);
    window = blinkstate->window;
    if (window->active == blinkstate) return;
    if (window->active && window->active->active_change)
	(*window->active->active_change) (window->active->info, FALSE);
    blinkstate->window->active = blinkstate;
    for (classentry = blinkstate->first; classentry;
	 classentry = classentry->next)
    {
	KDisplayDataBlinkEntry blinkentry;

	(*classentry->dataclass->activate_blinkstate) (classentry->subclass,
						       TRUE);
	for (blinkentry = classentry->first; blinkentry;
	     blinkentry = blinkentry->next)
	{
	    KWorldCanvas tmp_wc;

	    tmp_wc = (*classentry->dataclass->activate_blinkentry)
		(blinkentry->subclass, NULL);
	    if (tmp_wc) wc = tmp_wc;
	}
	(*classentry->dataclass->activate_blinkstate) (classentry->subclass,
						       FALSE);
    }
    dispdata_refresh_window (blinkstate->window, wc);
    if (window->active->active_change)
	(*window->active->active_change) (window->active->info, TRUE);
}   /*  End Function dispdata_activate_blinkstate  */

/*EXPERIMENTAL_FUNCTION*/
flag dispdata_test_blinkstate_active (KDisplayDataBlinkState blinkstate)
/*  [SUMMARY] Test if blinkstate is the active one for its window.
    <blinkstate> The KDisplayDataBlinkState object.
    [RETURNS] TRUE if active, else FALSE.
*/
{
    static char function_name[] = "dispdata_test_blinkstate_active";

    VERIFY_BLINKSTATE (blinkstate);
    return (blinkstate->window->active == blinkstate) ? TRUE : FALSE;
}   /*  End Function dispdata_test_blinkstate_active  */

/*EXPERIMENTAL_FUNCTION*/
unsigned int dispdata_get_blinkstate_index (KDisplayDataBlinkState blinkstate)
/*  [SUMMARY] Get the index for a blinkstate.
    <blinkstate> The KDisplayDataBlinkState object.
    [RETURNS] The index of the blinkstate.
*/
{
    static char function_name[] = "dispdata_get_blinkstate_index";

    VERIFY_BLINKSTATE (blinkstate);
    if (!blinkstate->window->indices_valid)
    {
	KDisplayDataBlinkState curr;
	unsigned int count;

	for (curr = blinkstate->window->blinkstates.first, count = 0; curr;
	     curr = curr->next, ++count) curr->index = count;
	blinkstate->window->indices_valid = TRUE;
    }
    return (blinkstate->index);
}   /*  End Function dispdata_get_blinkstate_index  */

/*EXPERIMENTAL_FUNCTION*/
void dispdata_copy_blinkstate (KDisplayDataBlinkState dest,
			       KDisplayDataBlinkState source)
/*  [SUMMARY] Copy blink state.
    <dest> The destination KDisplayDataBlinkState object to copy to.
    <source> The source KDisplayDataBlinkState object to copy from.
    [NOTE] This function does not refresh the blink state interface.
    [RETURNS] Nothing.
*/
{
    KDisplayDataClassBlinkState destclass, sourceclass;
    static char function_name[] = "dispdata_copy_blinkstate";

    VERIFY_BLINKSTATE (dest);
    VERIFY_BLINKSTATE (source);
    for (destclass = dest->first, sourceclass = source->first;
	 destclass;
	 destclass = destclass->next, sourceclass = sourceclass->next)
    {
	KDisplayDataClass dataclass = destclass->dataclass;
	KDisplayDataBlinkEntry destentry, sourceentry;

	for (destentry = destclass->first, sourceentry = sourceclass->first;
	     destentry;
	     destentry = destentry->next, sourceentry = sourceentry->next)
	{
	    (*dataclass->copy_blinkentry) (destentry->subclass,
					   sourceentry->subclass);
	    if (destclass->copy_blinkentry == sourceclass->copy_blinkentry)
		(*destclass->copy_blinkentry) (destentry->info,
					       sourceentry->info);
	}
    }
}   /*  End Function dispdata_copy_blinkstate  */

/*EXPERIMENTAL_FUNCTION*/
void dispdata_refresh_blinkstate (KDisplayDataBlinkState blinkstate)
/*  [SUMMARY] Refresh the user interface for a blink state.
    <blinkstate> The blink state to refresh.
    [RETURNS] Nothing.
*/
{
    struct displaydata_class_blinkstate_type *classentry;
    static char function_name[] = "dispdata_refresh_blinkstate";

    VERIFY_BLINKSTATE (blinkstate);
    for (classentry = blinkstate->first; classentry;
	 classentry = classentry->next)
    {
	if (classentry->refresh_func)
	    (*classentry->refresh_func) (classentry->info,
					 classentry->dataclass);
    }
}   /*  End Function dispdata_refresh_blinkstate  */

/*EXPERIMENTAL_FUNCTION*/
void *dispdata_get_blinkstate_info (KDisplayDataBlinkState blinkstate)
/*  [SUMMARY] Get the arbitrary pointer for a blinkstate.
    <blinkstate> The blinkstate.
    [RETURNS] The arbitrary pointer.
*/
{
    static char function_name[] = "dispdata_get_blinkstate_info";

    VERIFY_BLINKSTATE (blinkstate);
    return (blinkstate->info);
}   /*  End Function dispdata_get_blinkstate_info  */

/*EXPERIMENTAL_FUNCTION*/
void dispdata_register_blinkstate_funcs
    (KDisplayDataClassBlinkState blinkstate,
     void (*refresh_func) (void *info, KDisplayDataClass dataclass),
     void *(*create_blinkentry) (void *info, KDisplayDataBlinkEntry be),
     void (*destroy_blinkentry) (void *info),
     void (*copy_blinkentry) (void *dest, void *source),
     void *info)
/*  [SUMMARY] Register callback functions for a blinkstate.
    <blinkstate> The KDisplayDataClassBlinkState object.
    <dataclass> The dataclass for refresh events.
    <refresh_func> The function to be called to refresh the blinkstate
    interface.
    <info> An arbitrary information pointer to be passed to <<refresh_func>>
    and <<create_blinkentry>>.
    [RETURNS] Nothing.
*/
{
    KDisplayDataBlinkEntry blinkentry;
    static char function_name[] = "dispdata_register_blinkstate_funcs";

    VERIFY_BLINKSTATE_CLASS (blinkstate);
    blinkstate->refresh_func = refresh_func;
    blinkstate->create_blinkentry = create_blinkentry;
    blinkstate->destroy_blinkentry = destroy_blinkentry;
    blinkstate->copy_blinkentry = copy_blinkentry;
    blinkstate->info = info;
    for (blinkentry = blinkstate->first; blinkentry;
	 blinkentry = blinkentry->next)
    {
	if ( ( blinkentry->info = (*create_blinkentry) (info, blinkentry) )
	     == NULL )
	    m_abort (function_name, "application blink entry");
    }
}   /*  End Function dispdata_register_blinkstate_funcs  */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataBlinkState dispdata_get_first_blinkstate(KDisplayDataWindow window,
						     void **info)
/*  [SUMMARY] Get first blinkstate for a window.
    <window> The window.
    <info> The arbitrary information pointer is written here. If this is NULL,
    nothing is written here.
    [RETURNS] A KDisplayDataBlinkState object. NULL is returned if there are no
    blinkstates associated with this window.
*/
{
    static char function_name[] = "dispdata_get_first_blinkstate";

    VERIFY_WINDOW (window);
    if (info)
	*info =
	    window->blinkstates.first ? window->blinkstates.first->info : NULL;
    return (window->blinkstates.first);
}   /*  End Function dispdata_get_first_blinkstate  */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataBlinkState dispdata_get_last_blinkstate (KDisplayDataWindow window,
						     void **info)
/*  [SUMMARY] Get last blinkstate for a window.
    <window> The window.
    <info> The arbitrary information pointer is written here. If this is NULL,
    nothing is written here.
    [RETURNS] A KDisplayDataBlinkState object. NULL is returned if there are no
    blinkstates associated with this window.
*/
{
    static char function_name[] = "dispdata_get_last_blinkstate";

    VERIFY_WINDOW (window);
    if (info)
	*info =
	    window->blinkstates.last ? window->blinkstates.last->info : NULL;
    return (window->blinkstates.last);
}   /*  End Function dispdata_get_last_blinkstate  */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataBlinkState dispdata_get_prev_blinkstate
    (KDisplayDataBlinkState blinkstate, flag circular, void **info)
/*  [SUMMARY] Get previous blinkstate for a window.
    <blinkstate> The current blinkstate.
    <circular> If TRUE, the list of windows is treated as circular.
    <info> The arbitrary information pointer is written here. If this is NULL,
    nothing is written here.
    [RETURNS] A KDisplayDataBlinkState object. NULL is returned if there are no
    more blinkstates associated with this window.
*/
{
    KDisplayDataWindow window;
    static char function_name[] = "dispdata_get_prev_blinkstate";

    VERIFY_BLINKSTATE (blinkstate);
    window = blinkstate->window;
    blinkstate = blinkstate->prev;
    if (!blinkstate && circular) blinkstate = window->blinkstates.last;
    if (info && blinkstate) *info = blinkstate->info;
    return (blinkstate);
}   /*  End Function dispdata_get_prev_blinkstate  */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataBlinkState dispdata_get_next_blinkstate
    (KDisplayDataBlinkState blinkstate, flag circular, void **info)
/*  [SUMMARY] Get next blinkstate for a window.
    <blinkstate> The current blinkstate.
    <circular> If TRUE, the list of blinkstates is treated as circular.
    <info> The arbitrary information pointer is written here. If this is NULL,
    nothing is written here.
    [RETURNS] A KDisplayDataBlinkState object. NULL is returned if there are no
    more blinkstates associated with this window.
*/
{
    KDisplayDataWindow window;
    static char function_name[] = "dispdata_get_next_blinkstate";

    VERIFY_BLINKSTATE (blinkstate);
    window = blinkstate->window;
    blinkstate = blinkstate->next;
    if (!blinkstate && circular) blinkstate = window->blinkstates.first;
    if (info && blinkstate) *info = blinkstate->info;
    return (blinkstate);
}   /*  End Function dispdata_get_next_blinkstate  */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataWindow dispdata_get_window_from_blinkstate
    (KDisplayDataBlinkState blinkstate)
/*  [SUMMARY] Get the windor for a blinkstate.
    <blinkstate> The blinkstate.
    [RETURNS] The window.
*/
{
    static char function_name[] = "dispdata_get_window_from_blinkstate";

    VERIFY_BLINKSTATE (blinkstate);
    return (blinkstate->window);
}   /*  End Function dispdata_get_window_from_blinkstate  */

/*EXPERIMENTAL_FUNCTION*/
CONST char **dispdata_get_namelist (KDisplayDataClass dataclass,
				    unsigned int *num_entries,
				    unsigned int *num_chars)
/*  [SUMMARY] Get the list of (sorted) data names for a dataclass.
    <dataclass> The KDisplayDataClass object.
    <num_entries> The number of entries in the list is written here.
    <num_chars> The sum of the lengths of the names is written here. This does
    not include terminator characters.
    [RETURNS] The array of names.
*/
{
    static char function_name[] = "dispdata_get_namelist";

    VERIFY_DATACLASS (dataclass);
    if (!dataclass->namelist_valid)
    {
	KDisplayDataObject dataobject;
	unsigned int count;

	dataclass->num_chars = 0;
	for (dataobject = dataclass->first, count = 0;
	     dataobject;
	     dataobject = dataobject->next, ++count)
	{
	    dataclass->names[count] = dataobject->name;
	    dataclass->num_chars += strlen (dataobject->name);
	    dataclass->objects[count] = dataobject;
	}
	dataclass->namelist_valid = TRUE;
    }
    if (num_entries) *num_entries = dataclass->num_sets;
    if (num_chars) *num_chars = dataclass->num_chars;
    return ( (CONST char **) dataclass->names );
}   /*  End Function dispdata_get_namelist  */

/*EXPERIMENTAL_FUNCTION*/
CONST KDisplayDataObject *dispdata_get_objectlist (KDisplayDataClass dataclass,
						   unsigned int *num_entries)
/*  [SUMMARY] Get the list of (sorted) data objects for a dataclass.
    <dataclass> The KDisplayDataClass object.
    <num_entries> The number of entries in the list is written here.
    [RETURNS] The array of data objects.
*/
{
    static char function_name[] = "dispdata_get_objectlist";

    VERIFY_DATACLASS (dataclass);
    if (!dataclass->namelist_valid)
	dispdata_get_namelist (dataclass, NULL, NULL);
    if (num_entries) *num_entries = dataclass->num_sets;
    return (dataclass->objects);
}   /*  End Function dispdata_get_objectlist  */

/*EXPERIMENTAL_FUNCTION*/
char **dispdata_regenerate_namelist (KDisplayDataClass dataclass,
				     KDisplayDataPrivateNameList *namelist_ptr,
				     unsigned int num_reserved,
				     unsigned int *num_entries,
				     KDisplayDataObject **dataobjects,
				     flag (*filter_func) (void *info,
							  KDisplayDataObject dataobject),
				     void *info)
/*  [SUMMARY] Regenerate the list of (sorted) data names for a dataclass.
    <dataclass> The KDisplayDataClass object.
    <namelist_ptr> A pointer to a special buffer. This may be updated.
    <num_reserved> The number of characters to reserve for each entry. The
    reserved characters will be filled with ' ' (space).
    <num_entries> The number of entries in the list is written here.
    <dataobjects> The array of dataobjects is written here. This may be NULL.
    <filter_func> A function to filter dataobjects. If it returns TRUE, a
    dataobject will be included in the list, otherwise it won't. If this is
    NULL, all dataobjects are included.
    <info> An arbitrary pointer passed to <<filter_func>>.
    [RETURNS] The array of string pointers. The string contents may be modified
    and will persist until the next call to this function.
*/
{
    KDisplayDataPrivateNameList namelist = *namelist_ptr;
    CONST KDisplayDataObject *real_dataobjects;
    unsigned int real_num_entries, icount, ocount, num_chars, pad;
    char *ptr;
    CONST char **names;
    static char function_name[] = "dispdata_regenerate_namelist";

    VERIFY_DATACLASS (dataclass);
    if (namelist)
    {
	VERIFY_NAMELIST (namelist);
    }
    else
    {
	if ( ( namelist = m_calloc (sizeof *namelist) ) == NULL )
	    m_abort (function_name, "namelist object");
	*namelist_ptr = namelist;
	namelist->magic_number = NAMELIST_MAGIC_NUMBER;
    }
    names = dispdata_get_namelist (dataclass, &real_num_entries, &num_chars);
    num_chars += real_num_entries + num_reserved * real_num_entries;
    /*  Ensure there is enough storage  */
    if (real_num_entries >= namelist->num_allocated_entries)
    {
	icount = real_num_entries - namelist->num_allocated_entries;
	pad = icount % ALLOCATED_LINES_STEP;
	icount += pad ? (ALLOCATED_LINES_STEP - pad) : ALLOCATED_LINES_STEP;
	namelist->num_allocated_entries += icount;
	if (namelist->strings) m_free (namelist->strings);
	if ( ( namelist->strings =
	       m_alloc (namelist->num_allocated_entries *
			sizeof *namelist->strings) ) == NULL )
	    m_abort (function_name, "array of string pointers");
	if (namelist->dataobjects) m_free (namelist->dataobjects);
	if ( ( namelist->dataobjects =
	       m_alloc (namelist->num_allocated_entries *
			sizeof *namelist->dataobjects) ) == NULL )
	    m_abort (function_name, "array of dataobject pointers");
    }
    if (num_chars >= namelist->num_allocated_chars)
    {
	icount = num_chars - namelist->num_allocated_chars;
	pad = icount % ALLOCATED_CHARS_STEP;
	icount += pad ? (ALLOCATED_CHARS_STEP - pad) : ALLOCATED_CHARS_STEP;
	namelist->num_allocated_chars += icount;
	if (namelist->block) m_free (namelist->block);
	if ( ( namelist->block = m_alloc (namelist->num_allocated_chars) )
	     == NULL )
	    m_abort (function_name, "block of characters");
    }
    /*  Now copy names and object pointers  */
    real_dataobjects = dispdata_get_objectlist (dataclass, NULL);
    for (icount = 0, ocount = 0, ptr = namelist->block;
	 icount < real_num_entries; ++icount)
    {
	if ( filter_func && !(*filter_func) (info, real_dataobjects[icount]) )
	    continue;
	namelist->strings[ocount] = ptr;
	memset (ptr, ' ', num_reserved);
	ptr += num_reserved;
	strcpy (ptr, names[icount]);
	ptr += strlen (names[icount]) + 1;
	namelist->dataobjects[ocount] = real_dataobjects[icount];
	++ocount;
    }
    *num_entries = ocount;
    if (dataobjects) *dataobjects = namelist->dataobjects;
    return (namelist->strings);
}   /*  End Function dispdata_regenerate_namelist  */

/*EXPERIMENTAL_FUNCTION*/
void dispdata_free_namelist (KDisplayDataPrivateNameList namelist)
/*  [SUMMARY] Free the list of (sorted) data names for a dataclass.
    <namelist> The namelist buffer.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "dispdata_free_namelist";

    if (!namelist) return;
    VERIFY_NAMELIST (namelist);
    if (namelist->block) m_free (namelist->block);
    if (namelist->strings) m_free (namelist->strings);
}   /*  End Function dispdata_free_namelist  */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataBlinkEntry dispdata_get_first_blinkentry
    (KDisplayDataClassBlinkState class_blinkstate)
/*  [SUMMARY] Get the first blink entry for a blink state.
    <class_blinkstate> The KDisplayDataClassBlinkState object.
    [RETURNS] A KDisplayDataBlinkEntry object. NULL may be returned if there
    are no blink state entries (i.e. no data sets loaded for this class).
*/
{
    static char function_name[] = "dispdata_get_first_blinkentry";

    VERIFY_BLINKSTATE_CLASS (class_blinkstate);
    return (class_blinkstate->first);
}   /*  End Function dispdata_get_first_blinkentry  */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataBlinkEntry dispdata_get_next_blinkentry
    (KDisplayDataBlinkEntry blinkentry)
/*  [SUMMARY] Get the next blink entry after a specified entry.
    <blinkentry> The KDisplayDataBlinkEntry object.
    [RETURNS] A KDisplayDataBlinkEntry object. NULL may be returned if there
    are no more blink state entries.
*/
{
    static char function_name[] = "dispdata_get_next_blinkentry";

    VERIFY_BLINKENTRY (blinkentry);
    return (blinkentry->next);
}   /*  End Function dispdata_get_next_blinkentry  */

/*EXPERIMENTAL_FUNCTION*/
void *dispdata_get_blinkentry_info (KDisplayDataBlinkEntry blinkentry)
/*  [SUMMARY] Get the application-specific information for a blink entry.
    <blinkentry> The KDisplayDataBlinkEntry object.
    [RETURNS] The application-specific information pointer.
*/
{
    static char function_name[] = "dispdata_get_blinkentry_info";

    VERIFY_BLINKENTRY (blinkentry);
    return (blinkentry->info);
}   /*  End Function dispdata_get_blinkentry_info  */

/*EXPERIMENTAL_FUNCTION*/
void dispdata_set_blinkentry_attributes (KDisplayDataBlinkEntry blinkentry,
					 ...)
/*  [SUMMARY] Set attributes for a blink entry.
    <blinkentry> The KDisplayDataBlinkEntry object.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    KDISPLAYDATA_BLINKENTRY_ATT_END.
    [NOTE] Some windows may be refreshed.
    [RETURNS] Nothing.
*/
{
    KDisplayDataClass dataclass;
    va_list argp;
    static char function_name[] = "dispdata_set_blinkentry_attributes";

    va_start (argp, blinkentry);
    VERIFY_BLINKENTRY (blinkentry);
    dataclass = blinkentry->windataobject->dataobject->dataclass;
    (*dataclass->set_blinkentry_attributes) (blinkentry->subclass, argp);
    va_end (argp);
}   /*  End Function dispdata_set_blinkentry_attributes  */

/*EXPERIMENTAL_FUNCTION*/
void dispdata_get_blinkentry_attributes (KDisplayDataBlinkEntry blinkentry,
					 ...)
/*  [SUMMARY] Get attributes for a blink entry.
    <blinkentry> The KDisplayDataBlinkEntry object.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    KDISPLAYDATA_BLINKENTRY_ATT_END.
    [RETURNS] Nothing.
*/
{
    KDisplayDataClass dataclass;
    va_list argp;
    static char function_name[] = "dispdata_get_blinkentry_attributes";

    va_start (argp, blinkentry);
    VERIFY_BLINKENTRY (blinkentry);
    dataclass = blinkentry->windataobject->dataobject->dataclass;
    (*dataclass->get_blinkentry_attributes) (blinkentry->subclass, argp);
    va_end (argp);
}   /*  End Function dispdata_get_blinkentry_attributes  */

/*EXPERIMENTAL_FUNCTION*/
CONST KDisplayDataBlinkEntry *dispdata_regenerate_blinkentry_list
    (KDisplayDataClassBlinkState blinkstate,
     unsigned int *num_entries)
/*  [SUMMARY] Regenerate a list of blink entries for a blink state.
    <blinkstate> The KDisplayDataClassBlinkState object.
    <num_entries> The number of entries is written here.
    [RETURNS] The array of blink entries.
*/
{
    KDisplayDataBlinkEntry blinkentry;
    unsigned int count, pad;
    static char function_name[] = "dispdata_regenerate_blinkentry_list";

    VERIFY_BLINKSTATE_CLASS (blinkstate);
    dispdata_get_namelist (blinkstate->dataclass, num_entries, NULL);
    /*  Ensure there is enough storage  */
    if (*num_entries >= blinkstate->num_allocated_entries)
    {
	count = *num_entries - blinkstate->num_allocated_entries;
	pad = count % ALLOCATED_LINES_STEP;
	count += pad ? (ALLOCATED_LINES_STEP - pad) : ALLOCATED_LINES_STEP;
	blinkstate->num_allocated_entries += count;
	if (blinkstate->blinkentries) m_free (blinkstate->blinkentries);
	if ( ( blinkstate->blinkentries =
	       m_alloc (blinkstate->num_allocated_entries *
			sizeof *blinkstate->blinkentries) ) == NULL )
	    m_abort (function_name, "array of blinkentries");
    }
    /*  Now copy blinkentries  */
    blinkentry = blinkstate->first;
    for (count = 0; count < *num_entries; ++count)
    {
	blinkstate->blinkentries[count] = blinkentry;
	blinkentry = blinkentry->next;
    }
    return (blinkstate->blinkentries);
}   /*  End Function dispdata_regenerate_blinkentry_list  */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataObject dispdata_get_dataobject_from_blinkentry
    (KDisplayDataBlinkEntry blinkentry)
/*  [SUMMARY] Get the dataobject for a blinkentry.
    <blinkentry> The blinkentry.
    [RETURNS] The dataobject.
*/
{
    static char function_name[] = "dispdata_get_dataobject_from_blinkentry";

    VERIFY_BLINKENTRY (blinkentry);
    return (blinkentry->windataobject->dataobject);
}   /*  End Function dispdata_get_dataobject_from_blinkentry  */


/*  Private functions follow  */

static void _dispdata_destroy_window (KDisplayDataWindow window,
				      flag canvas_destroy)
/*  [SUMMARY] Destroy a display data window.
    <window> The KDisplayDataWindow to destroy.
    <canvas_destroy> If TRUE, one or more of the canvases is being destroyed.
    [RETURNS] Nothing.
*/
{
    KDisplayDataWindow curr;
    unsigned int count;
    struct displaydata_window_dataobject_type *windataobject, *next_wd;
    struct displaydata_window_class_entry *winclass, *next_wc;

    c_destroy_list (window->renumber_list);
    curr = window->next;
    KLIST_REMOVE2 (first_window, last_window, window);
    /*  Renumber the affected windows now, as destroy callbacks may need to
	reference the new indices  */
    for (; curr; curr = curr->next, ++count)
    {
	--curr->index;
	c_call_callbacks (curr->renumber_list, NULL);
    }
    c_unregister_callback2 (window->pc.destroy_func);
    c_unregister_callback2 (window->rgb.destroy_func);
    c_call_callbacks (window->destroy_list, NULL);
    c_destroy_list (window->destroy_list);
    while (window->blinkstates.first)
	dispdata_destroy_blinkstate (window->blinkstates.first);
    for (windataobject = window->dataobjects.first; windataobject;
	 windataobject = next_wd)
    {
	next_wd = windataobject->next;
	(*windataobject->dataobject->dataclass->destroy_viewable)
	    (windataobject->viewable, canvas_destroy);
	m_free (windataobject);
    }
    for (winclass = window->first; winclass; winclass = next_wc)
    {
	next_wc = winclass->next;
	(*winclass->dataclass->destroy_window) (winclass->subclass);
	m_free (winclass);
    }
    window->magic_number = 0;
    m_free (window);
}   /*  End Function _dispdata_destroy_window  */

static void _dispdata_canvas_destroy_callback (void *info)
/*  [SUMMARY] Process a destroy event on a world canvas.
    <info> An arbitrary function information pointer.
    [RETURNS] Nothing.
*/
{
    _dispdata_destroy_window (info, TRUE);
}   /*  End Function _dispdata_canvas_destroy_callback  */

static flag _dispdata_create_destroy_data_event_func (void *object,
						      void *client1_data,
						      void *call_data,
						      void *client2_data)
/*  [SUMMARY] Generic callback.
    [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    void (*func) (void *info, KDisplayDataObject data);

    func = ( void (*) (void *info, KDisplayDataObject data) ) client1_data;
    (*func) (object, call_data);
    return (FALSE);
}   /*  End Function _dispdata_create_destroy_data_event_func  */
