/*LINTLIBRARY*/
/*  overlay.c

    This code provides overlay-type DisplayData objects.

    Copyright (C) 2002-2003  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for manipulating overlay-type
    display data.


    Written by      Richard Gooch   10-DEC-2002: Serious coding begins.

    Updated by      Richard Gooch   12-DEC-2002: First cut of code. Works.

    Updated by      Richard Gooch   24-JAN-2003: Took account of API change to
  KDisplayDataFuncActivateBlinkEntry. Added KDisplayDataFuncCopyDataAttributes
  hook.

    Updated by      Richard Gooch   27-JAN-2003: Added return flag for all
  get/set hooks.

    Updated by      Richard Gooch   15-FEB-2003: Disabled explicit window
  refresh in <set_blinkentry_attributes> until <overlay> package is improved.

    Updated by      Richard Gooch   21-MAR-2003: Made use of new class
  KViewableOverlayList.

    Updated by      Richard Gooch   2-AUG-2003: Took account of API change to
  <dispdata_create_dataclass>: pass NULL <<set_class_attributes>>.

    Last updated by Richard Gooch   10-SEP-2003: Added return value to
  <destroy_blinkentry>.


*/
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdarg.h>
#include <karma.h>
#define NEW_WIN_SCALE
#include <karma_dispdata.h>
#include <karma_overlay.h>
#include <karma_canvas.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>

#include <karma_dispdata_hooks.h>


struct window
{
    KDisplayDataWindow superclass;
    KWorldCanvas pc;
    KWorldCanvas rgb;
    struct blinkstate *active_blinkstate;
};

struct viewable
{
    KViewableOverlayList pc_vlist;
    KViewableOverlayList rgb_vlist;
    struct window *window;
};

struct blinkstate
{
    struct window *window;
};

struct blinkentry
{
    struct viewable *viewable;
    struct blinkstate *blinkstate;
    unsigned char active;
};


/*  Private data  */
static KDisplayDataClass dataclass = NULL;


/*  Hook functions  */
STATIC_FUNCTION (void _dispdata_ov_destroy_dataobject, (void *dataobject) );
STATIC_FUNCTION (flag _dispdata_ov_copy_data_attributes,
		 (void *dest_p, void *source_p) );
STATIC_FUNCTION (flag _dispdata_ov_set_data_attributes,
		 (void *data, va_list argp) );
STATIC_FUNCTION (flag _dispdata_ov_get_data_attributes,
		 (void *data, va_list argp) );
STATIC_FUNCTION (void *_dispdata_ov_create_window,
		 (KDisplayDataWindow superclass,
		  KDisplayDataCanvas *pc, KDisplayDataCanvas *rgb) );
STATIC_FUNCTION (void _dispdata_ov_destroy_window, (void *window_p) );
STATIC_FUNCTION (flag _dispdata_ov_set_window_attributes,
		 (void *win, va_list argp) );
STATIC_FUNCTION (flag _dispdata_ov_get_window_attributes,
		 (void *win, va_list argp) );
STATIC_FUNCTION (void *_dispdata_ov_create_viewable,
		 (void *data, void *window_p) );
STATIC_FUNCTION (void _dispdata_ov_destroy_viewable,
		 (void *viewable_p, flag canvas_destroy) );
STATIC_FUNCTION (void *_dispdata_ov_create_blinkstate,
		 (KDisplayDataClassBlinkState superclass, void *window) );
STATIC_FUNCTION (void _dispdata_ov_destroy_blinkstate,
		 (void *blinkstate_p) );
STATIC_FUNCTION (void _dispdata_ov_activate_blinkstate,
		 (void *blinkstate_p, flag before_blinkentries) );
STATIC_FUNCTION (void *_dispdata_ov_create_blinkentry,
		 (void *blinkstate, void *viewable) );
STATIC_FUNCTION (flag _dispdata_ov_destroy_blinkentry,
		 (void *blinkentry_p) );
STATIC_FUNCTION (void _dispdata_ov_copy_blinkentry,
		 (void *dest_p, void *source_p) );
STATIC_FUNCTION (KWorldCanvas _dispdata_ov_activate_blinkentry,
		 (void *blinkentry_p, flag *visible) );
STATIC_FUNCTION (flag _dispdata_ov_set_blinkentry_attributes,
		 (void *blinkentry_p, va_list argp) );
STATIC_FUNCTION (flag _dispdata_ov_get_blinkentry_attributes,
		 (void *blinkentry_p, va_list argp) );


/*  Private functions  */


/* Public functions follow */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataClass dispdata_overlay_create_dataclass (void)
/*  [SUMMARY] Initialise overlay data support for the package.
    [RETURNS] A KDisplayDataClass object on success, else NULL.
*/
{
    KDisplayDataClass dc;

    if (dataclass) return (dataclass);
    dc = dispdata_create_dataclass ("Annotations",
				    NULL,
				    _dispdata_ov_destroy_dataobject,
				    _dispdata_ov_copy_data_attributes,
				    _dispdata_ov_set_data_attributes,
				    _dispdata_ov_get_data_attributes,
				    _dispdata_ov_create_window,
				    _dispdata_ov_destroy_window,
				    _dispdata_ov_set_window_attributes,
				    _dispdata_ov_get_window_attributes,
				    _dispdata_ov_create_viewable,
				    _dispdata_ov_destroy_viewable,
				    _dispdata_ov_create_blinkstate,
				    _dispdata_ov_destroy_blinkstate,
				    _dispdata_ov_activate_blinkstate,
				    _dispdata_ov_create_blinkentry,
				    _dispdata_ov_destroy_blinkentry,
				    _dispdata_ov_copy_blinkentry,
				    _dispdata_ov_activate_blinkentry,
				    _dispdata_ov_set_blinkentry_attributes,
				    _dispdata_ov_get_blinkentry_attributes);
    dataclass = dc;
    return (dc);
}   /*  End Function dispdata_overlay_create_dataclass  */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataObject dispdata_overlay_create_dataobject (KOverlayList olist,
						       CONST char *domain,
						       CONST char *name)
/*  [SUMMARY] Create an overlay data object.
    <olist> The overlay list.
    <domain> The domain from where list came from. The string contents are
    copied.
    <name> The name of the object.
    [RETURNS] A KDisplayDataObject on success, else NULL.
*/
{
    return dispdata_create_dataobject (dataclass, domain, name, olist);
}   /*  End Function dispdata_overlay_create_dataobject  */


/*  Hook functions follow  */

static void _dispdata_ov_destroy_dataobject (void *dataobject)
/*  [SUMMARY] Destroy a data object.
    <dataobject> The dataobject to destroy.
    [RETURNS] Nothing.
*/
{
    overlay_destroy_list (dataobject);
}   /*  End Function _dispdata_ov_destroy_dataobject  */

static flag _dispdata_ov_copy_data_attributes (void *dest_p, void *source_p)
/*  [SUMMARY] Copy data attributes.
    <dest_p> The destination data object.
    <source_p> The source data object.
    [RETURNS] TRUE on success, else FALSE..
*/
{
    return (TRUE);
}   /*  End Function _dispdata_ov_copy_data_attributes  */

static flag _dispdata_ov_set_data_attributes (void *data, va_list argp)
{
    return (FALSE);
}   /*  End Function _dispdata_ov_set_data_attributes  */

static flag _dispdata_ov_get_data_attributes (void *data, va_list argp)
{
    return (FALSE);
}   /*  End Function _dispdata_ov_get_data_attributes  */

static void *_dispdata_ov_create_window (KDisplayDataWindow superclass,
					 KDisplayDataCanvas *pc,
					 KDisplayDataCanvas *rgb)
/*  [SUMMARY] Create a window.
    <superclass> The superclass KDisplayDataWindow object.
    <pc> The PseudoColour canvas. This may contain NULL KWorldCanvas objects.
    <rgb> The TrueColour/DirectColour canvas. This may contain NULL
    KWorldCanvas objects.
    [RETURNS] The window object.
*/
{
    struct window *window;
    static char function_name[] = "_dispdata_ov_create_window";

    if ( ( window = m_calloc (sizeof *window) ) == NULL )
    {
	m_error_notify (function_name, "window information");
	return (NULL);
    }
    window->superclass = superclass;
    window->pc = pc->main;
    window->rgb = rgb->main;
    return (window);
}   /*  End Function _dispdata_ov_create_window  */

static void _dispdata_ov_destroy_window (void *window_p)
/*  [SUMMARY] Deallocate a window.
    [RETURNS] Nothing.
*/
{
    m_free (window_p);
}   /*  End Function _dispdata_ov_destroy_window  */

static flag _dispdata_ov_set_window_attributes (void *win, va_list argp)
{
    return (FALSE);
}   /*  End Function _dispdata_ov_set_window_attributes  */

static flag _dispdata_ov_get_window_attributes (void *win, va_list argp)
{
    return (FALSE);
}   /*  End Function _dispdata_ov_get_window_attributes  */

static void *_dispdata_ov_create_viewable (void *data, void *window_p)
{
    KOverlayList olist = data;
    struct viewable *viewable;
    struct window *window = window_p;
    static char function_name[] = "_dispdata_ov_create_viewable";

    if ( ( viewable = m_calloc (sizeof *viewable) ) == NULL )
    {
	m_error_notify (function_name, "window/data information");
	return (NULL);
    }
    if (window->pc)
    {
	if ( ( viewable->pc_vlist =
	       overlay_create_viewable (olist, window->pc) ) == NULL )
	{
	    m_free (viewable);
	    return (NULL);
	}
    }
    if (window->rgb)
    {
	if ( ( viewable->rgb_vlist =
	       overlay_create_viewable (olist, window->rgb) ) == NULL )
	{
	    if (viewable->pc_vlist)
		overlay_destroy_viewable (viewable->pc_vlist);
	    m_free (viewable);
	    return (NULL);
	}
    }
    viewable->window = window;
    return (viewable);
}   /*  End Function _dispdata_ov_create_viewable  */

static void _dispdata_ov_destroy_viewable (void *viewable_p,
					   flag canvas_destroy)
/*  [SUMMARY] Deallocate the cache for data on a window.
    [RETURNS] Nothing.
*/
{
    struct viewable *viewable = viewable_p;

    if (!canvas_destroy)
    {
	if (viewable->pc_vlist) overlay_destroy_viewable (viewable->pc_vlist);
	if (viewable->rgb_vlist) overlay_destroy_viewable(viewable->rgb_vlist);
    }
    m_free (viewable);
}   /*  End Function _dispdata_ov_destroy_viewable  */

static void *_dispdata_ov_create_blinkstate
    (KDisplayDataClassBlinkState superclass, void *window)
{
    struct blinkstate *blinkstate;

    if ( ( blinkstate = m_calloc (sizeof *blinkstate) ) == NULL ) return NULL;
    blinkstate->window = window;
    return (blinkstate);
}   /*  End Function _dispdata_ov_create_blinkstate  */

static void _dispdata_ov_destroy_blinkstate (void *blinkstate_p)
{
    m_free (blinkstate_p);
}   /*  End Function _dispdata_ov_destroy_blinkstate  */

static void _dispdata_ov_activate_blinkstate (void *blinkstate_p,
					      flag before_blinkentries)
{
    struct blinkstate *blinkstate = blinkstate_p;

    blinkstate->window->active_blinkstate = blinkstate;
}   /*  End Function _dispdata_ov_activate_blinkstate  */

static void *_dispdata_ov_create_blinkentry (void *blinkstate, void *viewable)
{
    struct blinkentry *blinkentry;

    if ( ( blinkentry = m_calloc (sizeof *blinkentry) ) == NULL )
	return (NULL);
    blinkentry->viewable = viewable;
    blinkentry->blinkstate = blinkstate;
    return (blinkentry);
}   /*  End Function _dispdata_ov_create_blinkentry  */

static flag _dispdata_ov_destroy_blinkentry (void *blinkentry_p)
/*  [SUMMARY] Destroy a blinkentry.
    <blinkentry_p> The blinkentry subclass data.
    [RETURNS] TRUE if the viewable was set to visible in this blinkentry.
*/
{
    flag visible;
    struct blinkentry *blinkentry = blinkentry_p;

    visible = blinkentry->active;
    m_free (blinkentry_p);
    return (visible);
}   /*  End Function _dispdata_ov_destroy_blinkentry  */

static void _dispdata_ov_copy_blinkentry (void *dest_p, void *source_p)
{
    struct blinkentry *dest = dest_p;
    struct blinkentry *source = source_p;

    dest->active = source->active;
}   /*  End Function _dispdata_ov_copy_blinkentry  */

static KWorldCanvas _dispdata_ov_activate_blinkentry (void *blinkentry_p,
						      flag *visible)
{
    struct blinkentry *blinkentry = blinkentry_p;
    struct viewable *viewable = blinkentry->viewable;

    if (visible && blinkentry->active) *visible = TRUE;
    if (viewable->pc_vlist)
	overlay_set_active (viewable->pc_vlist, blinkentry->active,
			    FALSE, FALSE, FALSE, 0);
    if (viewable->rgb_vlist)
	overlay_set_active (viewable->rgb_vlist, blinkentry->active,
			    FALSE, FALSE, FALSE, 0);
    return (NULL);
}   /*  End Function _dispdata_ov_activate_blinkentry  */

static flag _dispdata_ov_set_blinkentry_attributes (void *blinkentry_p,
						    va_list argp)
{
    flag bool;
    flag changed = FALSE;
    flag all_known = TRUE;
    unsigned int att_key, dummy;
    struct blinkentry *blinkentry = blinkentry_p;
    struct viewable *viewable = blinkentry->viewable;

    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    KDISPLAYDATA_BLINKENTRY_ATT_END )
    {
	switch (att_key)
	{
	  case KDISPLAYDATA_BLINKENTRY_ATT_ACTIVE:
	    bool = va_arg (argp, flag);
	    if (blinkentry->active != bool)
	    {
		blinkentry->active = bool;
		changed = TRUE;
	    }
	    break;
	  default:
	    dummy = va_arg (argp, unsigned int);
	    all_known = FALSE;
	    break;
	}
    }
    if (!changed) return (all_known);
    if (viewable->window->active_blinkstate == blinkentry->blinkstate)
    {
	_dispdata_ov_activate_blinkentry (blinkentry, NULL);
	dispdata_refresh_window (viewable->window->superclass, NULL);
    }
    return (all_known);
}   /*  End Function _dispdata_ov_set_blinkentry_attributes  */

static flag _dispdata_ov_get_blinkentry_attributes (void *blinkentry_p,
						    va_list argp)
{
    flag all_known = TRUE;
    unsigned int att_key;
    struct blinkentry *blinkentry = blinkentry_p;
    void *dummy;

    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    KDISPLAYDATA_BLINKENTRY_ATT_END )
    {
	switch (att_key)
	{
	  case KDISPLAYDATA_BLINKENTRY_ATT_ACTIVE:
	    *va_arg (argp, flag *) = blinkentry->active;
	    break;
	  default:
	    dummy = va_arg (argp, void *);
	    all_known = FALSE;
	    break;
	}
    }
    return (all_known);
}   /*  End Function _dispdata_ov_get_blinkentry_attributes  */


/*  Private functions follow  */
