/*LINTLIBRARY*/
/*  main.c

    This code provides 1-D plotting facilities

    Copyright (C) 2000-2004  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for plotting 1-D arrays.


    Written by      Richard Gooch   15-JUN-2000

    Updated by      Richard Gooch   17-JUN-2000: Support NULL <<x>> arrays:
  functions assume equally spaced co-ordinates.

    Updated by      Richard Gooch   20-JUN-2000: Renamed to <graph1d> package.

    Updated by      Richard Gooch   17-MAY-2001: Created <graph1d_points>
  function.

    Updated by      Richard Gooch   3-JAN-2002: Bug fix in function
  <graph1d_lines>: trap for <<num_points>> less than 2.

    Updated by      Richard Gooch   24-DEC-2002: Created KViewableGraph class.

    Updated by      Richard Gooch   26-DEC-2002: Added colourmap support.

    Updated by      Richard Gooch   28-DEC-2002: Added scale and data change
  interfaces.

    Updated by      Richard Gooch   29-DEC-2002: Allow empty list of points
  in <graph1d_create>.

    Updated by      Richard Gooch   30-DEC-2002: Added check for <<num_points>>
  equal to zero, and avoid redrawing.

    Updated by      Richard Gooch   3-JAN-2003: Changed interface and
  implementation so that co-ordinate arrays are allocated internally. Created
  <graph1d_save_ascii>.

    Updated by      Richard Gooch   11-JAN-2003: Created experimental function
  <graph1d_test_if_zoomed>.

    Updated by      Richard Gooch   12-JAN-2003: Created experimental function
  <graph1d_unzoom>.

    Updated by      Richard Gooch   22-JAN-2003: Renamed function
  <graph1d_test_if_zoomed> to <graph1d_test_if_zoomed_horizontally> and changed
  interface to <graph1d_unzoom>.

    Updated by     Richard Gooch    3-FEB-2003: Fixed <_graph1d_steps> to plot
  left horizontal line to data point and not left boundary. No longer read past
  end of Y array.

    Updated by      Richard Gooch   5-FEB-2003: Optimised <_graph1d_lines>.

    Updated by      Richard Gooch   6-FEB-2003: Optimised <_graph1d_steps>.

    Updated by      Richard Gooch   20-FEB-2003: Bugfix in <_graph1d_lines> and
  <_graph1d_steps>: was not drawing all lines when using colourmap.

    Updated by      Richard Gooch   11-APR-2003: Added horizontal and vertical
  line support.

    Updated by      Richard Gooch   30-APR-2003: Created experimental function
  <graph1d_refresh_if_needed>.

    Updated by      Richard Gooch   28-MAY-2003: Generalised smoothing code to
  support 5, 7 and 9 point Hanning windows and 3, 5, 7, and 9 point uniform.

    Updated by      Richard Gooch   11-JUL-2003: Added dashing of hline and
  vline.

    Updated by      Richard Gooch   12-JUL-2003: Fixed bug in <_graph1d_steps>
  where zero-length segments were being drawn. This led to points being drawn
  in PostScript output and a 1.5x slowdown. Also fixed bug where vertical line
  for last input value was not drawn.

    Updated by      Richard Gooch   2-OCT-2003: Handle turnover/wraparound in X
  axis.

    Updated by      Richard Gooch   6-OCT-2003: Fixed bug in <_graph1d_points>:
  if all points lie outside boundary, would loop forever.

    Updated by      Richard Gooch   1-JAN-2004: Minor code migration from
  <graph1d_orthogonal_save> to <graph1d_save_ascii>.

    Updated by      Richard Gooch   2-JAN-2004: Created experimental function
  <graph1d_get_last>.

    Last updated by Richard Gooch   6-MAY-2004: Improved function
  <graph1d_change_draw_order> to not set <<need_refresh>> flag if not needed.


*/
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <stdarg.h>
#include <karma.h>
#include <k_list.h>
#define NEW_WIN_SCALE
#include <karma_graph1d.h>
#include <karma_canvas.h>
#include <karma_kcmap.h>
#include <karma_kwin.h>
#include <karma_ch.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_c.h>


#define VGRAPH_USED_MAGIC_NUMBER  (unsigned int)  393843029
#define VGRAPH_FREE_MAGIC_NUMBER  (unsigned int)  398128732
#define HOLDER_MAGIC_NUMBER       (unsigned int) 2143457982
#define DEFAULT_DARK_COLOUR       "black"
#define DEFAULT_LIGHT_COLOUR      "white"
#define BUF_SIZE                  16384
#define MAX_WEIGHTS                   9

#define VERIFY_HOLDER(holder) {if (holder == NULL) \
{fprintf (stderr, "NULL canvas holder passed\n"); \
 a_prog_bug (function_name); } \
if (holder->magic_number != HOLDER_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid canvas holder object\n"); \
 a_prog_bug (function_name); } }

#define VERIFY_USED_VGRAPH(vgraph) if (vgraph == NULL) \
{fprintf (stderr, "NULL viewable graph passed\n"); \
 a_prog_bug (function_name); } \
if (vgraph->magic_number != VGRAPH_USED_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid viewable graph object\n"); \
 a_prog_bug (function_name); }

#define VERIFY_FREE_VGRAPH(vgraph) if (vgraph == NULL) \
{fprintf (stderr, "NULL viewable graph passed\n"); \
 a_prog_bug (function_name); } \
if (vgraph->magic_number != VGRAPH_FREE_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid freed viewable graph object\n"); \
 a_prog_bug (function_name); }

#define X_IN_LIMITS(x,info) ( ( (x >= info->min1_wx) && (x <= info->max1_wx) )\
			      || ( info->have_2nd_range                       \
				   && (x >= info->min2_wx)                    \
				   && (x <= info->max2_wx) ) )

typedef struct canvas_holder_type * CanvasHolder;

struct canvas_info_struct
{
    double left_wx;
    double right_wx;
    double min1_wx;
    double max1_wx;
    double min2_wx;
    double max2_wx;
    double min_wy;
    double max_wy;
    flag have_2nd_range;
    CONST struct win_scale_type *win_scale;
};

struct canvas_holder_type
{
    unsigned int        magic_number;
    KWorldCanvas        canvas;
    KPixCanvas          pixcanvas;
    /*  Attributes  */
    unsigned int        draw_style;
    unsigned int        smooth_mode;
    KWorldCanvasLimits  limits;
    double              minimum_y;
    double              maximum_y;
    flag                auto_y;
    flag                show_hline;
    flag                show_vline;
    /*  Internal management  */
    flag                need_refresh;
    unsigned long       previous_background;
    float               previous_intensity;
    flag                light;
    unsigned int        num_weights;
    double              weights[MAX_WEIGHTS];
    struct canvas_info_struct canvas_info;  /*  Only valid during refresh  */
    KViewableGraph      first;
    KViewableGraph      last;
};

struct viewable_graph_type
{
    unsigned int        magic_number;
    CanvasHolder        canvas_holder;
    flag                active;
    /*  Attributes  */
    unsigned int        draw_style;
    unsigned long       dark_pixel;  /*  Pixel to use for light backgrounds  */
    unsigned long       light_pixel; /*  Pixel to use for dark backgrounds   */
    flag                use_kcmap;
    double              scale;
    double              offset;
    double              hline_y;     /*  Y position of horizontal marker     */
    double              vline_x;     /*  X position of vertical marker       */
    /*  Input data  */
    unsigned int        num_points;
    /*  Other information  */
    unsigned int        num_coord_pairs_allocated;
    double              *coords;              /*  The co-ordinate buffer     */
    uaddr               key;                  /*  Key for grouping           */
    /*  Internal management  */
    KCallbackList       destroy_list;
    KViewableGraph      prev;
    KViewableGraph      next;
};


/*  Public data  */
char *graph1d_draw_style_choices[GRAPH1D_NUM_DRAW_STYLES] =
{
    "join", "hist", "cross",
};

char *graph1d_smooth_mode_choices[GRAPH1D_NUM_SMOOTH_MODES] =
{
    "none",
    "hanning-3", "hanning-5", "hanning-7", "hanning-9",
    "uniform-3", "uniform-5", "uniform-7", "uniform-9",
};


/*  Private data  */
static char _graph1d_canvas_holder_key;           /*  Unique key  */
static KViewableGraph first_free = NULL;


/*  Private functions  */
STATIC_FUNCTION (CanvasHolder _graph1d_get_canvas_holder,
		 (KWorldCanvas canvas, char *func_name) );
STATIC_FUNCTION (CanvasHolder _graph1d_alloc_canvas_holder,
		 (KWorldCanvas canvas) );
STATIC_FUNCTION (void _graph1d_free_canvas_holder, (CanvasHolder holder) );
STATIC_FUNCTION (flag _graph1d_process_vgraph_attributes,
		 (KViewableGraph vgraph, va_list argp) );
STATIC_FUNCTION (void _graph1d_worldcanvas_precompute_func,
		 (KWorldCanvas canvas, int width, int height,
		  CONST struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize,
		  void **info, PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void _graph1d_precompute_vgraph,
		 (KViewableGraph vgraph,CONST struct win_scale_type *win_scale,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void _graph1d_worldcanvas_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  CONST struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize,
		  void **info, PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (void _graph1d_refresh_vgraph,
		 (KViewableGraph vgraph, Kcolourmap cmap,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void _graph1d_lines,
		 (CONST struct canvas_holder_type *holder,
		  double scale, double offset,
		  unsigned int num_points, CONST double *x, CONST double *y,
		  unsigned long pixel_value,
		  unsigned int num_pixels, unsigned long *pixel_values) );
STATIC_FUNCTION (void _graph1d_steps,
		 (CONST struct canvas_holder_type *holder,
		  double scale, double offset,
		  unsigned int num_points, CONST double *x, CONST double *y,
		  unsigned long pixel_value,
		  unsigned int num_pixels, unsigned long *pixel_values) );
STATIC_FUNCTION (void _graph1d_points,
		 (CONST struct canvas_holder_type *holder,
		  double scale, double offset,
		  unsigned int num_points, CONST double *x, CONST double *y,
		  unsigned long pixel_value) );


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void graph1d_init (KWorldCanvas canvas)
/*  [SUMMARY] Initialise the package for a particular canvas.
    [PURPOSE] This routine will initialise the [<graph1d>] package for a
    particular world canvas. Calling this routine causes a number of callback
    routines internal to the package to be registered with the canvas (such
    as refresh and position event callbacks). This routine must be called
    before creating viewable graphs.
    <canvas> The world canvas object.
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder;
    static char function_name[] = "graph1d_init";

    if (canvas == NULL)
    {
	fprintf (stderr, "NULL world canvas passed\n");
	a_prog_bug (function_name);
    }
    if ( canvas_get_associated_object (canvas, &_graph1d_canvas_holder_key) )
    {
	fprintf (stderr, "Canvas already initialised\n");
	a_prog_bug (function_name);
    }
    if ( ( holder = _graph1d_alloc_canvas_holder (canvas) ) == NULL )
	m_abort (function_name, "canvas holder");
    canvas_zoom_handle_mouse (canvas, &holder->limits);
    canvas_zoom_handle_keys (canvas, &holder->limits);
}   /*  End Function graph1d_init  */

/*EXPERIMENTAL_FUNCTION*/
void graph1d_set_canvas_attributes (KWorldCanvas canvas, ...)
/*  [SUMMARY] Set attributes for a canvas.
    <canvas> The world canvas.
    [VARARGS] The list of parameter attribute-key attribute-value-ptr pairs
    must follow. This list must be terminated with the GRAPH1D_CANVAS_ATT_END.
    See [<graph1d_CANVAS_ATTRIBUTES>] for the list of attributes.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    flag bool;
    unsigned int att_key, uival;
    double dval;
    CanvasHolder holder;
    static char function_name[] = "graph1d_set_canvas_attributes";

    va_start (argp, canvas);
    if (canvas == NULL)
    {
	fprintf (stderr, "NULL world canvas passed\n");
	a_prog_bug (function_name);
    }
    holder = _graph1d_get_canvas_holder (canvas, function_name);
    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    GRAPH1D_CANVAS_ATT_END )
    {
	switch (att_key)
	{
	  case GRAPH1D_CANVAS_ATT_DRAW_STYLE:
	    uival = va_arg (argp, unsigned int);
	    if (holder->draw_style == uival) break;
	    switch (uival)
	    {
	      case GRAPH1D_DRAW_STYLE_LINES:
	      case GRAPH1D_DRAW_STYLE_STEPS:
	      case GRAPH1D_DRAW_STYLE_CROSSES:
		break;
	      case GRAPH1D_DRAW_STYLE_DEFAULT:
		holder->draw_style = GRAPH1D_DRAW_STYLE_LINES;
		break;
	      default:
		fprintf (stderr, "Illegal draw style: %u\n", uival);
		a_prog_bug (function_name);
		break;
	    }
	    holder->draw_style = uival;
	    holder->need_refresh = TRUE;
	    break;
	  case GRAPH1D_CANVAS_ATT_SMOOTH_MODE:
	    uival = va_arg (argp, unsigned int);
	    if (holder->smooth_mode == uival) break;
	    if (uival >= GRAPH1D_NUM_SMOOTH_MODES)
	    {
		fprintf (stderr, "Illegal smooth mode: %u\n",
			 holder->smooth_mode);
		a_prog_bug (function_name);
	    }
	    holder->smooth_mode = uival;
	    holder->need_refresh = TRUE;
	    holder->weights[0] = TOOBIG;
	    break;
	  case GRAPH1D_CANVAS_ATT_LEFT_X:
	    dval = va_arg (argp, double);
	    if (holder->limits.soft.left_x == dval) break;
	    holder->limits.soft.left_x = dval;
	    holder->need_refresh = TRUE;
	    break;
	  case GRAPH1D_CANVAS_ATT_RIGHT_X:
	    dval = va_arg (argp, double);
	    if (holder->limits.soft.right_x == dval) break;
	    holder->limits.soft.right_x = dval;
	    holder->need_refresh = TRUE;
	    break;
	  case GRAPH1D_CANVAS_ATT_BOTTOM_Y:
	    dval = va_arg (argp, double);
	    if (holder->limits.soft.bottom_y == dval) break;
	    holder->limits.soft.bottom_y = dval;
	    holder->need_refresh = TRUE;
	    break;
	  case GRAPH1D_CANVAS_ATT_TOP_Y:
	    dval = va_arg (argp, double);
	    if (holder->limits.soft.top_y == dval) break;
	    holder->limits.soft.top_y = dval;
	    holder->need_refresh = TRUE;
	    break;
	  case GRAPH1D_CANVAS_ATT_MINIMUM_Y:
	    dval = va_arg (argp, double);
	    if (holder->minimum_y == dval) break;
	    holder->minimum_y = dval;
	    holder->need_refresh = TRUE;
	    break;
	  case GRAPH1D_CANVAS_ATT_MAXIMUM_Y:
	    dval = va_arg (argp, double);
	    if (holder->maximum_y == dval) break;
	    holder->maximum_y = dval;
	    holder->need_refresh = TRUE;
	    break;
	  case GRAPH1D_CANVAS_ATT_AUTO_Y:
	    bool = va_arg (argp, flag);
	    if (holder->auto_y == bool) break;
	    holder->auto_y = bool;
	    holder->need_refresh = TRUE;
	    break;
	  case GRAPH1D_CANVAS_ATT_SHOW_HLINE:
	    bool = va_arg (argp, flag);
	    if (holder->show_hline == bool) break;
	    holder->show_hline = bool;
	    holder->need_refresh = TRUE;
	    break;
	  case GRAPH1D_CANVAS_ATT_SHOW_VLINE:
	    bool = va_arg (argp, flag);
	    if (holder->show_vline == bool) break;
	    holder->show_vline = bool;
	    holder->need_refresh = TRUE;
	    break;
	  default:
	    fprintf (stderr, "Illegal attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	}
    }
    va_end (argp);
    if (holder->auto_y)
    {
	canvas_set_attributes (holder->canvas,
			       CANVAS_ATT_BOTTOM_Y, holder->minimum_y,
			       CANVAS_ATT_TOP_Y, holder->maximum_y,
			       CANVAS_ATT_END);
	holder->need_refresh = TRUE;
    }
}   /*  End Function graph1d_set_canvas_attributes  */

/*EXPERIMENTAL_FUNCTION*/
flag graph1d_test_if_zoomed_horizontally (KWorldCanvas canvas)
/*  [SUMMARY] Test if a canvas is zoomed horizontally.
    <canvas> The world canvas.
    [RETURNS] TRUE if the canvas is zoomed, else FALSE.
*/
{
    CanvasHolder holder;
    double left_x, right_x;
    static char function_name[] = "graph1d_test_if_zoomed_horizontally";

    holder = _graph1d_get_canvas_holder (canvas, function_name);
    canvas_get_attributes (canvas,
			   CANVAS_ATT_LEFT_X, &left_x,
			   CANVAS_ATT_RIGHT_X, &right_x,
			   CANVAS_ATT_END);
    if (holder->limits.soft.left_x != left_x) return (TRUE);
    if (holder->limits.soft.right_x != right_x) return (TRUE);
    return (FALSE);
}   /*  End Function graph1d_test_if_zoomed_horizontally  */

/*EXPERIMENTAL_FUNCTION*/
void graph1d_unzoom (KWorldCanvas canvas, flag horizontal, flag vertical,
		     flag auto_v)
/*  [SUMMARY] Unzoom a canvas.
    <canvas> The world canvas.
    <horizontal> If TRUE, the canvas is unzoomed horizontally.
    <vertical> If TRUE, the canvas is unzoomed vertically.
    <auto_v> If TRUE, the canvas is auto-scaled vertically.
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder;
    static char function_name[] = "graph1d_unzoom";

    holder = _graph1d_get_canvas_holder (canvas, function_name);
    if (horizontal)
	canvas_set_attributes (canvas,
			       CANVAS_ATT_LEFT_X, holder->limits.soft.left_x,
			       CANVAS_ATT_RIGHT_X, holder->limits.soft.right_x,
			       CANVAS_ATT_END);
    if (vertical)
	canvas_set_attributes(canvas,
			      CANVAS_ATT_BOTTOM_Y,holder->limits.soft.bottom_y,
			      CANVAS_ATT_TOP_Y, holder->limits.soft.top_y,
			      CANVAS_ATT_END);
    if (auto_v)
	canvas_set_attributes (canvas,
			       CANVAS_ATT_BOTTOM_Y, holder->minimum_y,
			       CANVAS_ATT_TOP_Y, holder->maximum_y,
			       CANVAS_ATT_END);
    holder->need_refresh = TRUE;
}   /*  End Function graph1d_unzoom  */

/*EXPERIMENTAL_FUNCTION*/
void graph1d_refresh_if_needed (KWorldCanvas canvas)
/*  [SUMMARY] Refresh a canvas if needed.
    <canvas> The world canvas to be refreshed.
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder;
    static char function_name[] = "graph1d_refresh_if_needed";

    if (canvas == NULL)
    {
	fprintf (stderr, "NULL world canvas passed\n");
	a_prog_bug (function_name);
    }
    holder = _graph1d_get_canvas_holder (canvas, function_name);
    if (!holder->need_refresh) return;
    kwin_refresh_if_visible (holder->pixcanvas, TRUE);
}   /*  End Function graph1d_refresh_if_needed  */

/*EXPERIMENTAL_FUNCTION*/
KViewableGraph graph1d_create (KWorldCanvas canvas, unsigned int num_points,
			       double **x, double **y, uaddr key, ...)
/*  [SUMMARY] Create viewable graph.
    [PURPOSE] This routine will create a viewable graph object from two arrays
    of values. At a later time, this viewable graph may be made visible. This
    routine will not cause the canvas to be refreshed.
    <canvas> The world canvas onto which the viewable graph may be drawn. Many
    viewable graphs may be associated with a single canvas.
    <num_points> The number of points in this viewable graph.
    <x> The pointer to the array of horizontal values is written here. The
    space is automatically dealloated when the graph is destroyed.
    <y> The pointer to the array of vertical values is written here. The space
    is automatically dealloated when the graph is destroyed.
    <key> The key to associate with this viewable graph.
    [NOTE] The KViewableGraph will be automatically destroyed if the canvas is
    destroyed.
    [VARARGS] The list of parameter attribute-key attribute-value-ptr pairs
    must follow. This list must be terminated with the GRAPH1D_VGRAPH_ATT_END.
    See [<graph1d_VGRAPH_ATTRIBUTES>] for the list of attributes.
    [RETURNS] A KViewableGraph object on success, NULL.
*/
{
    va_list argp;
    CanvasHolder holder;
    KViewableGraph vgraph;
    static char function_name[] = "graph1d_create";

    va_start (argp, key);
    if (canvas == NULL)
    {
	fprintf (stderr, "NULL world canvas passed\n");
	a_prog_bug (function_name);
    }
    holder = _graph1d_get_canvas_holder (canvas, function_name);
    if (first_free == NULL)
    {
	if ( ( vgraph = m_alloc (sizeof *vgraph) ) == NULL ) return (NULL);
    }
    else
    {
	vgraph = first_free;
	first_free = vgraph->next;
	VERIFY_FREE_VGRAPH (vgraph);
    }
    m_clear (vgraph, sizeof *vgraph);
    vgraph->magic_number = VGRAPH_USED_MAGIC_NUMBER;
    vgraph->canvas_holder = holder;
    vgraph->draw_style = GRAPH1D_DRAW_STYLE_DEFAULT;
    vgraph->key = key;
    vgraph->scale = 1.0;
    vgraph->offset = 0.0;
    vgraph->hline_y = TOOBIG;
    vgraph->vline_x = TOOBIG;
    KLIST_APPEND (holder, vgraph);
    if ( !graph1d_allocate_coords (vgraph, num_points, x, y) )
    {
	graph1d_destroy (vgraph);
	return (NULL);
    }
    graph1d_set_vgraph_attributes (vgraph,
				   GRAPH1D_VGRAPH_ATT_DARK_COLOUR, NULL,
				   GRAPH1D_VGRAPH_ATT_LIGHT_COLOUR, NULL,
				   GRAPH1D_VGRAPH_ATT_END);
    _graph1d_process_vgraph_attributes (vgraph, argp);
    va_end (argp);
    return (vgraph);
}   /*  End Function graph1d_create  */

/*EXPERIMENTAL_FUNCTION*/
void graph1d_set_vgraph_attributes (KViewableGraph vgraph, ...)
/*  [SUMMARY] Set attributes for a viewable graph.
    <vgraph> The viewable graph.
    [VARARGS] The list of parameter attribute-key attribute-value-ptr pairs
    must follow. This list must be terminated with the GRAPH1D_VGRAPH_ATT_END.
    See [<graph1d_VGRAPH_ATTRIBUTES>] for the list of attributes.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    flag changed;
    static char function_name[] = "graph1d_set_vgraph_attributes";

    va_start (argp, vgraph);
    VERIFY_USED_VGRAPH (vgraph);
    changed = _graph1d_process_vgraph_attributes (vgraph, argp);
    if (vgraph->active && vgraph->num_points && changed)
	vgraph->canvas_holder->need_refresh = TRUE;
    va_end (argp);
}   /*  End Function graph1d_set_vgraph_attributes  */

/*EXPERIMENTAL_FUNCTION*/
flag graph1d_set_active (KViewableGraph vgraph, flag active,flag force_refresh,
			 flag refresh_if_changed, flag exclusive, uaddr key)
/*  [SUMMARY] Set active state for a viewable graph controlled refresh.
    [PURPOSE] This routine will make a viewable graph active or inactive.
    <vgraph> The viewable graph.
    <active> If TRUE, the viewable graph is made active, else it is made
    inactive.
    <force_refresh> If TRUE, the canvas is always refreshed.
    <refresh_if_changed> If TRUE, the canvas is refreshed if the viewable graph
    active state changed.
    <exclusive> If TRUE and <<active>> is TRUE, make this viewable graph the
    only active one for this canvas. Other viewable graphs for this canvas will
    be made inactive.
    <key> The key to use for inactivation of other viewable graphs. If this is
    0 then all other viewable graphs will be made inactive.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    CanvasHolder holder;
    KViewableGraph curr;
    static char function_name[] = "graph1d_set_active";

    VERIFY_USED_VGRAPH (vgraph);
    FLAG_VERIFY (active);
    FLAG_VERIFY (force_refresh);
    FLAG_VERIFY (refresh_if_changed);
    holder = vgraph->canvas_holder;
    if (active != vgraph->active)
    {
	vgraph->active = active;
	if (vgraph->num_points)
	{
	    if (refresh_if_changed) force_refresh = TRUE;
	    holder->need_refresh = TRUE;
	}
    }
    if (exclusive && active)
    {
	/*  Make others inactive  */
	for (curr = holder->first; curr; curr = curr->next)
	{
	    if (curr == vgraph) continue;
	    if ( (key != 0) && (key != curr->key) ) continue;
	    if (vgraph->num_points)
	    {
		if (curr->active) holder->need_refresh = TRUE;
		if (curr->active && refresh_if_changed) force_refresh = TRUE;
	    }
	    curr->active = FALSE;
	}
    }
    if (force_refresh) return kwin_refresh_if_visible (holder->pixcanvas,TRUE);
    return (TRUE);
}   /*  End Function graph1d_set_active  */

/*EXPERIMENTAL_FUNCTION*/
flag graph1d_set_all_inactive (KWorldCanvas canvas, flag force_refresh,
			       flag refresh_if_changed, uaddr key)
/*  [SUMMARY] Clear active state all viewable graphs controlled refresh.
    [PURPOSE] This routine will make all viewable graphs associated with a
    world canvas inactive.
    <canvas> The world canvas.
    <force_refresh> If TRUE, the canvas is always refreshed.
    <refresh_if_changed> If TRUE, the canvas is refreshed if the active state
    of any viewable graph changed.
    <key> The key to use for inactivation of viewable graphs. If this is 0 then
    all viewable graphs will be made inactive.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    CanvasHolder holder;
    KViewableGraph curr;
    static char function_name[] = "graph1d_set_all_inactive";

    holder = _graph1d_get_canvas_holder (canvas, function_name);
    FLAG_VERIFY (force_refresh);
    FLAG_VERIFY (refresh_if_changed);
    /*  Make all inactive  */
    for (curr = holder->first; curr; curr = curr->next)
    {
	if ( (key != 0) && (key != curr->key) ) continue;
	if (curr->num_points)
	{
	    if (curr->active) holder->need_refresh = TRUE;
	    if (curr->active && refresh_if_changed) force_refresh = TRUE;
	}
	curr->active = FALSE;
    }
    if (force_refresh) return canvas_resize (holder->canvas, NULL, FALSE);
    return (TRUE);
}   /*  End Function graph1d_set_all_inactive  */

/*EXPERIMENTAL_FUNCTION*/
void graph1d_destroy (KViewableGraph vgraph)
/*  [SUMMARY] Destroy a viewable graph.
    <vgraph> The viewable graph.
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder;
    static char function_name[] = "graph1d_destroy";

    VERIFY_USED_VGRAPH (vgraph);
    holder = vgraph->canvas_holder;
    if (vgraph->active && vgraph->num_points) holder->need_refresh = TRUE;
    if (vgraph->coords) m_free (vgraph->coords);
    c_call_callbacks (vgraph->destroy_list, NULL);
    c_destroy_list (vgraph->destroy_list);
    KLIST_REMOVE (holder, vgraph);
    vgraph->magic_number = VGRAPH_FREE_MAGIC_NUMBER;
    vgraph->next = first_free;
    first_free = vgraph;
}   /*  End Function graph1d_destroy  */

/*EXPERIMENTAL_FUNCTION*/
KCallbackFunc graph1d_register_destroy_func (KViewableGraph vgraph,
					     void (*func) (), void *info)
/*  [SUMMARY] Register destroy callback.
    [PURPOSE] This routine will register a destroy function for a viewable
    graph. The function will be called when the canvas is destroyed. Many
    destroy functions may be registered per graph. The first function
    registered is the first function called upon destruction.
    <vgraph> The viewable graph.
    <func> The function that is called when the graph is destroyed. The
    prototype function is [<graph1d_PROTO_destroy_func>].
    <info> The arbitrary function information pointer.
    [RETURNS] A KCallbackFunc object. On failure, the process aborts.
*/
{
    static char function_name[] = "graph1d_register_destroy_func";

    VERIFY_USED_VGRAPH (vgraph);
    return c_register_callback (&vgraph->destroy_list, ( flag (*) () ) func,
				info, NULL, FALSE, NULL, FALSE, FALSE);
}   /*  End Function graph1d_register_destroy_func  */

/*EXPERIMENTAL_FUNCTION*/
KWorldCanvas graph1d_get_worldcanvas (KViewableGraph vgraph)
/*  [SUMMARY] Get the world canvas for a viewable graph.
    <vgraph> The viewable graph.
    [RETURNS] The KWorldCanvas object.
*/
{
    static char function_name[] = "graph1d_get_worldcanvas";

    VERIFY_USED_VGRAPH (vgraph);
    return (vgraph->canvas_holder->canvas);
}   /*  End Function graph1d_get_worldcanvas  */

/*EXPERIMENTAL_FUNCTION*/
void graph1d_change_draw_order (KViewableGraph vgraph, KViewableGraph position,
				flag before)
/*  [SUMMARY] Change the draw order for a viewable graph.
    <vgraph> The viewable graph to move in the draw order.
    <position> The viewable graph to use as a reference point.
    <before> If TRUE, <<vgraph>> will be moved just before <<position>> (or to
    the start of the list if <<position>> is NULL). If FALSE, <<vgraph>> will
    be moved just after <<position>> (or to the end of the list if <<position>>
    is NULL).
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder;
    static char function_name[] = "graph1d_change_draw_order";

    VERIFY_USED_VGRAPH (vgraph);
    holder = vgraph->canvas_holder;
    if (position)
    {
	VERIFY_USED_VGRAPH (position);
	if (position->canvas_holder != holder)
	{
	    fprintf (stderr, "Inconsistent canvas holders: %p and %p\n",
		     holder, position->canvas_holder);
	    a_prog_bug (function_name);
	}
    }
    else position = before ? holder->first : holder->last;
    if (vgraph == position) return;
    if (before)
    {   /*  Insert before a specified position  */
	if (vgraph->next == position) return;       /*  Already just before  */
	KLIST_REMOVE (holder, vgraph);
	KLIST_INSERT (holder, position, vgraph);
    }
    else
    {   /*  Insert after a specified position  */
	if (vgraph->prev == position) return;       /*  Already just after   */
	position = position->next;
	KLIST_REMOVE (holder, vgraph);
	if (position)
	{
	    KLIST_INSERT (holder, position, vgraph);
	}
	else
	{
	    KLIST_APPEND (holder, vgraph);
	}
    }
    if (vgraph->active) holder->need_refresh = TRUE;
}   /*  End Function graph1d_change_draw_order  */

/*EXPERIMENTAL_FUNCTION*/
KViewableGraph graph1d_get_last (KWorldCanvas canvas)
/*  [SUMMARY] Get last viewable graph for a world canvas.
    <canvas> The world canvas.
    [RETURNS] The viewable graph if there is one, else NULL.
*/
{
    CanvasHolder holder;
    static char function_name[] = "graph1d_get_last";

    holder = _graph1d_get_canvas_holder (canvas, function_name);
    return (holder->last);
}   /*  End Function graph1d_get_last  */

/*EXPERIMENTAL_FUNCTION*/
flag graph1d_allocate_coords (KViewableGraph vgraph, unsigned int num_coords,
			      double **x, double **y)
/*  [SUMMARY] Allocate co-ordinates for a viewable graph.
    <vgraph> The viewable graph.
    <num_points> The number of co-ordinate pairs to allocate.
    <x> The pointer to the array of horizontal values is written here. The
    space is automatically dealloated when the graph is destroyed.
    <y> The pointer to the array of vertical values is written here. The space
    is automatically dealloated when the graph is destroyed.
    [RETURNS] TRUE on success, else FALSE, indicating an allocation error.
*/
{
    static char function_name[] = "graph1d_allocate_coords";

    VERIFY_USED_VGRAPH (vgraph);
    if (vgraph->active && vgraph->num_points)
	vgraph->canvas_holder->need_refresh = TRUE;
    vgraph->num_points = 0;
    if (num_coords > vgraph->num_coord_pairs_allocated)
    {
	if (vgraph->coords) m_free (vgraph->coords);
	vgraph->num_coord_pairs_allocated = 0;
	vgraph->coords = NULL;
	if ( ( vgraph->coords =
	       m_alloc (sizeof *vgraph->coords * num_coords * 2) ) == NULL )
	{
	    m_error_notify (function_name, "co-ordinate array");
	    return (FALSE);
	}
	vgraph->num_coord_pairs_allocated = num_coords;
    }
    if (x) *x = vgraph->coords;
    if (y) *y = vgraph->coords + vgraph->num_coord_pairs_allocated;
    return (TRUE);
}   /*  End Function graph1d_allocate_coords  */

/*EXPERIMENTAL_FUNCTION*/
void graph1d_set_size (KViewableGraph vgraph, unsigned int num_points)
/*  [SUMMARY] Set the size of a viewable graph.
    <vgraph> The viewable graph.
    <num_points> The number of valid points in the graph.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "graph1d_set_size";

    VERIFY_USED_VGRAPH (vgraph);
    if (num_points > vgraph->num_coord_pairs_allocated)
    {
	fprintf (stderr, "Number of points: %u exceeds allocation: %u\n",
		 num_points, vgraph->num_coord_pairs_allocated);
	a_prog_bug (function_name);
    }
    if ( vgraph->active && (vgraph->num_points != num_points) )
	vgraph->canvas_holder->need_refresh = TRUE;
    vgraph->num_points = num_points;
}   /*  End Function graph1d_set_size  */

/*EXPERIMENTAL_FUNCTION*/
unsigned int graph1d_get_size (KViewableGraph vgraph, double **x, double **y)
/*  [SUMMARY] Get the size of a viewable graph.
    <vgraph> The viewable graph.
    <num_points> The number of points in the graph is written here.
    <x> The pointer to the array of horizontal values is written here. This may
    be NULL.
    <y> The pointer to the array of vertical values is written here. This may
    be NULL.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "graph1d_get_size";

    VERIFY_USED_VGRAPH (vgraph);
    if (x) *x = vgraph->coords;
    if (y) *y = vgraph->coords + vgraph->num_coord_pairs_allocated;
    return (vgraph->num_points);
}   /*  End Function graph1d_get_size  */

/*EXPERIMENTAL_FUNCTION*/
flag graph1d_save_ascii (KViewableGraph vgraph, Channel channel)
/*  [SUMMARY] Save the data in a viewable graph in ASCII format.
    <vgraph> The viewable graph.
    <channel> The Channel object to write to.
    [RETURNS] TRUE on success, else FALSE indicating error.
*/
{
    unsigned int count;
    char *xlabel, *ylabel;
    double *x, *y;
    static char function_name[] = "graph1d_save_ascii";

    VERIFY_USED_VGRAPH (vgraph);
    canvas_get_specification (vgraph->canvas_holder->canvas,
			      &xlabel, &ylabel, NULL, NULL, NULL);
    ch_printf (channel, "# X axis name: \"%s\"  Y axis name: \"%s\"\n",
	       xlabel, ylabel);
    x = vgraph->coords;
    y = vgraph->coords + vgraph->num_coord_pairs_allocated;
    ch_printf (channel, "# %u values follow\n", vgraph->num_points);
    for (count = 0; count < vgraph->num_points; ++count, ++x, ++y)
	ch_printf (channel, "%-20.13e   %.13e\n", *x, *y);
    return (TRUE);
}   /*  End Function graph1d_save_ascii  */


/*  Private functions follow  */

static CanvasHolder _graph1d_get_canvas_holder (KWorldCanvas canvas,
						char *func_name)
/*  [SUMMARY] Get the canvas holder for a world canvas.
    <canvas> The canvas.
    <func_name> The name of the calling function. If this is not NULL, the
    process will abort if a canvas holder is not found.
    [RETURNS] The canvas holder on success, else NULL.
*/
{
    CanvasHolder holder;

    holder = canvas_get_associated_object (canvas,&_graph1d_canvas_holder_key);
    if (!holder)
    {
	fprintf (stderr, "CanvasHolder not found!\n");
	a_prog_bug (func_name);
    }
    return (holder);
}   /*  End Function get_canvas_holder  */

static CanvasHolder _graph1d_alloc_canvas_holder (KWorldCanvas canvas)
/*  This routine will allocate a canvas holder for a world canvas.
    The canvas must be given by  canvas  .
    The routine returns the canvas holder on success, else it returns NULL.
*/
{
    CanvasHolder holder;
    unsigned short red, green, blue;
    double r, g, b;
    static char function_name[] = "_graph1d_alloc_canvas_holder";

    /*  Must allocate holder  */
    if ( ( holder = m_calloc (sizeof *holder) ) == NULL )
    {
	m_error_notify (function_name, "canvas holder");
	return (NULL);
    }
    holder->magic_number = HOLDER_MAGIC_NUMBER;
    holder->canvas = canvas;
    holder->pixcanvas = canvas_get_pixcanvas (canvas);
    canvas_associate_object (canvas, &_graph1d_canvas_holder_key,holder,FALSE);
    canvas_register_precompute_func (canvas,
				     _graph1d_worldcanvas_precompute_func,
				     holder);
    canvas_register_refresh_func (canvas, _graph1d_worldcanvas_refresh_func,
				  holder);
    canvas_register_destroy_func (canvas, FALSE, _graph1d_free_canvas_holder,
				  holder);
    holder->limits.soft.left_x = 0.0;
    holder->limits.soft.right_x = 1.0;
    holder->limits.soft.bottom_y = 0.0;
    holder->limits.soft.top_y = 1.0;
    holder->limits.hard.left_x = TOOBIG;
    holder->limits.hard.right_x = TOOBIG;
    holder->limits.hard.bottom_y = TOOBIG;
    holder->limits.hard.top_y = TOOBIG;
    kwin_get_attributes (holder->pixcanvas,
			 KWIN_ATT_BACKGROUND_PIXEL,
			 &holder->previous_background,
			 KWIN_ATT_END);
    kwin_get_pixel_RGB_values (holder->pixcanvas,
			       &holder->previous_background,
			       &red, &green, &blue, 1);
    r = red;
    g = green;
    b = blue;
    holder->previous_intensity = sqrt (r * r + g * g + b * b) / 65535.0;
    return (holder);
}   /*  End Function alloc_canvas_holder  */

static void _graph1d_free_canvas_holder (CanvasHolder holder)
/*  [SUMMARY] Free a canvas holder (and any associated KViewableGraph objects).
    <holder> The canvas holder.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "_graph1d_free_canvas_holder";

    VERIFY_HOLDER (holder);
    while (holder->first) graph1d_destroy (holder->first);
    holder->magic_number = 0;
    m_free (holder);
}   /*  End Function _graph1d_free_canvas_holder  */

static flag _graph1d_process_vgraph_attributes (KViewableGraph vgraph,
						va_list argp)
/*  [PURPOSE] This routine will set the attributes for a viewable graph.
    <vgraph> The viewable graph.
    <argp> The optional list of parameter attribute-key attribute-value
    pairs.
    [RETURNS] TRUE if an attribute value was changed, else FALSE.
*/
{
    flag changed = FALSE;
    flag bool;
    unsigned int att_key, uival;
    unsigned long pixel;
    double dval;
    char *colourname;
    static char function_name[] = "_graph1d_process_vgraph_attributes";

    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    GRAPH1D_VGRAPH_ATT_END )
    {
	switch (att_key)
	{
	  case GRAPH1D_VGRAPH_ATT_DRAW_STYLE:
	    uival = va_arg (argp, unsigned int);
	    if (vgraph->draw_style == uival) break;
	    switch (uival)
	    {
	      case GRAPH1D_DRAW_STYLE_LINES:
	      case GRAPH1D_DRAW_STYLE_STEPS:
	      case GRAPH1D_DRAW_STYLE_CROSSES:
	      case GRAPH1D_DRAW_STYLE_DEFAULT:
		break;
	      default:
		fprintf (stderr,"Illegal draw style: %u\n", uival);
		a_prog_bug (function_name);
		break;
	    }
	    vgraph->draw_style = uival;
	    changed = TRUE;
	    break;
	  case GRAPH1D_VGRAPH_ATT_DARK_COLOUR:
	    colourname = va_arg (argp, char *);
	    if (!colourname) colourname = DEFAULT_DARK_COLOUR;
	    kwin_get_colour_harder (vgraph->canvas_holder->pixcanvas,
				    colourname, &pixel);
	    if (vgraph->dark_pixel == pixel) break;
	    vgraph->dark_pixel = pixel;
	    changed = TRUE;
	    break;
	  case GRAPH1D_VGRAPH_ATT_LIGHT_COLOUR:
	    colourname = va_arg (argp, char *);
	    if (!colourname) colourname = DEFAULT_LIGHT_COLOUR;
	    kwin_get_colour_harder (vgraph->canvas_holder->pixcanvas,
				    colourname, &pixel);
	    if (vgraph->light_pixel == pixel) break;
	    vgraph->light_pixel = pixel;
	    changed = TRUE;
	    break;
	  case GRAPH1D_VGRAPH_ATT_USE_COLOURMAP:
	    bool = va_arg (argp, flag);
	    if (vgraph->use_kcmap == bool) break;
	    vgraph->use_kcmap = bool;
	    changed = TRUE;
	    break;
	  case GRAPH1D_VGRAPH_ATT_SCALE:
	    dval = va_arg (argp, double);
	    if (vgraph->scale == dval) break;
	    vgraph->scale = dval;
	    changed = TRUE;
	    break;
	  case GRAPH1D_VGRAPH_ATT_OFFSET:
	    dval = va_arg (argp, double);
	    if (vgraph->offset == dval) break;
	    vgraph->offset = dval;
	    changed = TRUE;
	    break;
	  case GRAPH1D_VGRAPH_ATT_HLINE_Y:
	    dval = va_arg (argp, double);
	    if (vgraph->hline_y == dval) break;
	    vgraph->hline_y = dval;
	    changed = TRUE;
	    break;
	  case GRAPH1D_VGRAPH_ATT_VLINE_X:
	    dval = va_arg (argp, double);
	    if (vgraph->vline_x == dval) break;
	    vgraph->vline_x = dval;
	    changed = TRUE;
	    break;
	  default:
	    fprintf (stderr, "Illegal attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	}
    }
    return (changed);
}   /*  End Function process_vgraph_attributes  */

static void _graph1d_worldcanvas_precompute_func
    (KWorldCanvas canvas, int width, int height,
     CONST struct win_scale_type *win_scale, Kcolourmap cmap, flag cmap_resize,
     void **info, PostScriptPage pspage,
     unsigned int num_areas, KPixCanvasRefreshArea *areas)
/*  [PURPOSE] This routine processes a precompute event for a world canvas.
    <canvas> The world canvas.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information. The data are only
    valid during this callback.
    <cmap> The canvas colourmap.
    <cmap_resize> If TRUE, the refresh is due to a colourmap resize.
    <info> A pointer to the arbitrary information pointer.
    <pspage> The PostScriptPage object. If this is NULL, the refresh is *not*
    destined for a PostScript page.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder = *info;
    KViewableGraph vgraph;
    KPixCanvas pixcanvas;
    unsigned long background_pixel;
    static char function_name[] = "_graph1d_worldcanvas_precompute_func";

    VERIFY_HOLDER (holder);
    if (canvas != holder->canvas)
    {
	fprintf (stderr, "Different canvas in canvas holder object\n");
	a_prog_bug (function_name);
    }
    pixcanvas = canvas_get_pixcanvas (canvas);
    kwin_get_attributes (pixcanvas,
			 KWIN_ATT_BACKGROUND_PIXEL, &background_pixel,
			 KWIN_ATT_END);
    if (holder->previous_background != background_pixel)
    {
	unsigned short red, green, blue;
	double r, g, b;

	holder->previous_background = background_pixel;
	kwin_get_pixel_RGB_values (pixcanvas, &background_pixel,
				   &red, &green, &blue, 1);
	r = red;
	g = green;
	b = blue;
	holder->previous_intensity = sqrt (r * r + g * g + b * b) / 65535.0;
    }
    holder->light = ( (holder->previous_intensity >= 0.5) ||
		      pspage ) ? TRUE : FALSE;
    if ( (holder->smooth_mode != GRAPH1D_SMOOTH_MODE_NONE) &&
	 DOUBLE_IS_BLANK (holder->weights[0]) )
    {
	flag hanning = FALSE;

	switch (holder->smooth_mode)
	{
	  case GRAPH1D_SMOOTH_MODE_HANNING_3:
	    holder->num_weights = 3;
	    hanning = TRUE;
	    break;
	  case GRAPH1D_SMOOTH_MODE_HANNING_5:
	    holder->num_weights = 5;
	    hanning = TRUE;
	    break;
	  case GRAPH1D_SMOOTH_MODE_HANNING_7:
	    holder->num_weights = 7;
	    hanning = TRUE;
	    break;
	  case GRAPH1D_SMOOTH_MODE_HANNING_9:
	    holder->num_weights = 9;
	    hanning = TRUE;
	    break;
	  case GRAPH1D_SMOOTH_MODE_UNIFORM_3:
	    holder->num_weights = 3;
	    break;
	  case GRAPH1D_SMOOTH_MODE_UNIFORM_5:
	    holder->num_weights = 5;
	    break;
	  case GRAPH1D_SMOOTH_MODE_UNIFORM_7:
	    holder->num_weights = 7;
	    break;
	  case GRAPH1D_SMOOTH_MODE_UNIFORM_9:
	    holder->num_weights = 9;
	    break;
	}
	if (hanning)
	{
	    signed int count;
	    signed int midpoint = holder->num_weights >> 1;
	    double np1 = holder->num_weights + 1;

	    for (count = 0; count < holder->num_weights; ++count)
	    {
		double distance = abs (midpoint - count);

		holder->weights[count] = 2.0 * (np1 - 2.0 *distance) /np1 /np1;
	    }
	}
	else
	{
	    unsigned int count;
	    double weight = 1.0 / (double) holder->num_weights;

	    for (count = 0; count < holder->num_weights; ++count)
		holder->weights[count] = weight;
	}
    }
    /*  Show each active vgraph  */
    for (vgraph = holder->first; vgraph; vgraph = vgraph->next)
    {
	if (vgraph->active && vgraph->num_points)
	    _graph1d_precompute_vgraph (vgraph, win_scale, num_areas, areas);
    }
}   /*  End Function _graph1d_worldcanvas_precompute_func  */

static void _graph1d_precompute_vgraph (KViewableGraph vgraph,
					CONST struct win_scale_type *win_scale,
					unsigned int num_areas,
					KPixCanvasRefreshArea *areas)
/*  [PURPOSE] This routine processes a precompute event for a world canvas.
    <vgraph> The KViewableGraph object.
    <win_scale> A pointer to the window scaling information.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    [RETURNS] Nothing.
*/
{
    /*static char function_name[] = "_graph1d_precompute_vgraph";*/

}   /*  End Function _graph1d_precompute_vgraph  */

static void _graph1d_worldcanvas_refresh_func(KWorldCanvas canvas,
					      int width, int height,
					      CONST struct win_scale_type *win_scale,
					      Kcolourmap cmap,
					      flag cmap_resize,
					      void **info,
					      PostScriptPage pspage,
					      unsigned int num_areas,
					      KPixCanvasRefreshArea *areas,
					      flag *honoured_areas)
/*  [PURPOSE] This routine processes a refresh event for a world canvas.
    <canvas> The world canvas.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information. The data are only
    valid during this callback.
    <cmap> The canvas colourmap.
    <cmap_resize> If TRUE, the refresh is due to a colourmap resize.
    <info> A pointer to the arbitrary information pointer.
    <pspage> The PostScriptPage object. If this is NULL, the refresh is *not*
    destined for a PostScript page.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder = *info;
    KViewableGraph vgraph;
    double xstep, tmp;
    double xb[4], yb[4];
    static char function_name[] = "_graph1d_worldcanvas_refresh_func";

    VERIFY_HOLDER (holder);
    if (canvas != holder->canvas)
    {
	fprintf (stderr, "Different canvas in canvas holder object\n");
	a_prog_bug (function_name);
    }
    /*  Compute the boundaries, and also detect if there is X turnover  */
    xb[0] = win_scale->left_x;
    yb[0] = win_scale->bottom_y;
    xb[1] = win_scale->left_x + (win_scale->right_x - win_scale->left_x) * 0.1;
    yb[1] = win_scale->bottom_y;
    xb[2] = win_scale->left_x + (win_scale->right_x - win_scale->left_x) * 0.2;
    yb[2] = win_scale->bottom_y;
    xb[3] = win_scale->right_x;
    yb[3] = win_scale->top_y;
    canvas_coords_transform (canvas, 4, xb, FALSE, yb, FALSE);
    holder->canvas_info.left_wx = xb[0];
    holder->canvas_info.right_wx = xb[3];
    /*  Test 2 intervals in case X turnover is within one interval  */
    xstep = xb[1] - xb[0];
    tmp = xb[2] - xb[1];
    if ( fabs (xstep) > fabs (tmp) ) xstep = tmp;
    if (xstep < 0.0)
    {   /*  X co-ordinates decrease left-to-right  */
	if (xb[0] < xb[3])
	{   /*  Left < right, so we must have turnover  */
	    /* TODO: fix assumption that turnover only happens for longitude */
	    holder->canvas_info.min1_wx = 0.0;
	    holder->canvas_info.max1_wx = xb[0];
	    holder->canvas_info.min2_wx = xb[3];
	    holder->canvas_info.max2_wx = 360.0;
	    holder->canvas_info.have_2nd_range = TRUE;
	}
	else
	{
	    holder->canvas_info.min1_wx = xb[3];
	    holder->canvas_info.max1_wx = xb[0];
	    holder->canvas_info.have_2nd_range = FALSE;
	}
    }
    else
    {   /*  X co-ordinates increase left-to-right  */
	if (xb[0] > xb[3])
	{   /*  Left > right, so we must have turnover  */
	    /* TODO: fix assumption that turnover only happens for longitude */
	    holder->canvas_info.min1_wx = xb[0];
	    holder->canvas_info.max1_wx = 360.0;
	    holder->canvas_info.min2_wx = 0.0;
	    holder->canvas_info.max2_wx = xb[3];
	    holder->canvas_info.have_2nd_range = TRUE;
	}
	else
	{
	    holder->canvas_info.min1_wx = xb[0];
	    holder->canvas_info.max1_wx = xb[3];
	    holder->canvas_info.have_2nd_range = FALSE;
	}
    }
    if (yb[0] < yb[3])
    {
	holder->canvas_info.min_wy = yb[0];
	holder->canvas_info.max_wy = yb[3];
    }
    else
    {
	holder->canvas_info.min_wy = yb[3];
	holder->canvas_info.max_wy = yb[0];
    }
    holder->canvas_info.win_scale = win_scale;
    /*  Show each active vgraph  */
    for (vgraph = holder->first; vgraph; vgraph = vgraph->next)
    {
	if (vgraph->active && vgraph->num_points)
	    _graph1d_refresh_vgraph (vgraph, cmap, num_areas, areas);
    }
    /**honoured_areas = TRUE;*/
    holder->need_refresh = FALSE;
}   /*  End Function _graph1d_worldcanvas_refresh_func  */

static void _graph1d_refresh_vgraph (KViewableGraph vgraph,
				     Kcolourmap cmap,
				     unsigned int num_areas,
				     KPixCanvasRefreshArea *areas)
/*  [PURPOSE] This routine processes a refresh event for a world canvas.
    <vgraph> The KViewableGraph object.
    <cmap> The canvas colourmap.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas canvas = vgraph->canvas_holder->canvas;
    flag free_scratch = FALSE;
    flag dash_on = FALSE;
    unsigned int draw_style = vgraph->draw_style;
    unsigned int num_points = vgraph->num_points;
    unsigned int num_pixels;
    unsigned long pixel;
    unsigned long *pixels;
    struct canvas_holder_type *holder = vgraph->canvas_holder;
    struct canvas_info_struct *canvas_info = &holder->canvas_info;
    double *x, *y;
    static char function_name[] = "_graph1d_refresh_vgraph";

    pixel = vgraph->canvas_holder->light ?
	vgraph->dark_pixel : vgraph->light_pixel;
    if (vgraph->use_kcmap && cmap)
	num_pixels = kcmap_get_pixels (cmap, &pixels);
    else
    {
	num_pixels = 0;
	pixels = NULL;
    }
    if (draw_style == GRAPH1D_DRAW_STYLE_DEFAULT)
	draw_style = holder->draw_style;
    if (holder->smooth_mode != GRAPH1D_SMOOTH_MODE_NONE)
    {
	unsigned int count;
	unsigned int midpoint = holder->num_weights >> 1;
	CONST double *ox = vgraph->coords;
	CONST double *oy = ox + vgraph->num_coord_pairs_allocated;

	if (holder->num_weights > num_points) return;
	num_points -= (holder->num_weights - 1);
	x = m_alloc_scratch (sizeof *x * num_points * 2, function_name);
	y = x + num_points;
	free_scratch = TRUE;
	for (count = 0; count < num_points; ++count)
	{
	    unsigned int wcount;
	    double sum = 0.0;

	    for (wcount = 0; wcount < holder->num_weights; ++wcount)
	    {
		if ( DOUBLE_IS_BLANK (oy[count + wcount]) ) continue;
		sum += oy[count + wcount] * holder->weights[wcount];
	    }
	    x[count] = ox[count + midpoint];
	    y[count] = sum;
	}
    }
    else
    {
	x = vgraph->coords;
	y = x + vgraph->num_coord_pairs_allocated;
    }
    switch (draw_style)
    {
      case GRAPH1D_DRAW_STYLE_LINES:
	_graph1d_lines (holder, vgraph->scale, vgraph->offset,
			num_points, x, y, pixel, num_pixels, pixels);
	break;
      case GRAPH1D_DRAW_STYLE_STEPS:
	_graph1d_steps (holder, vgraph->scale, vgraph->offset,
			num_points, x, y, pixel, num_pixels, pixels);
	break;
      case GRAPH1D_DRAW_STYLE_CROSSES:
	_graph1d_points (holder, vgraph->scale, vgraph->offset,
			 num_points, x, y, pixel);
	break;
    }
    if ( holder->show_hline && DOUBLE_IS_VALID (vgraph->hline_y) &&
	 (vgraph->hline_y >= canvas_info->min_wy) &&
	 (vgraph->hline_y <= canvas_info->max_wy) )
    {
	double xt, yt;

	xt = canvas_info->left_wx;
	yt = vgraph->hline_y;
	canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1, &xt, &yt,
					   &xt, &yt);
	kwin_set_attributes (holder->pixcanvas,
			     KWIN_ATT_LINE_DASH, TRUE,
			     KWIN_ATT_END);
	dash_on = TRUE;
	kwin_draw_line (holder->pixcanvas,
			canvas_info->win_scale->x_offset, yt,
			canvas_info->win_scale->x_offset +
			canvas_info->win_scale->x_pixels - 1, yt,
			pixel);
    }
    if ( holder->show_vline && X_IN_LIMITS (vgraph->vline_x, canvas_info) )
    {
	double xt, yt;

	xt = vgraph->vline_x;
	yt = canvas_info->min_wy;
	canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1, &xt, &yt,
					   &xt, &yt);
	kwin_set_attributes (holder->pixcanvas,
			     KWIN_ATT_LINE_DASH, TRUE,
			     KWIN_ATT_END);
	dash_on = TRUE;
	kwin_draw_line (holder->pixcanvas,
			xt, canvas_info->win_scale->y_offset,
			xt, canvas_info->win_scale->y_offset +
			canvas_info->win_scale->y_pixels - 1,
			pixel);
    }
    if (dash_on) kwin_set_attributes (holder->pixcanvas,
				      KWIN_ATT_LINE_DASH, FALSE,
				      KWIN_ATT_END);
    if (free_scratch) m_free_scratch (x);
}   /*  End Function _graph1d_refresh_vgraph  */

static void _graph1d_lines (CONST struct canvas_holder_type *holder,
			    double scale, double offset,
			    unsigned int num_points,
			    CONST double *x, CONST double *y,
			    unsigned long pixel_value, unsigned int num_pixels,
			    unsigned long *pixel_values)
/*  [SUMMARY] Plot a 1-dimensional graph by connecting the vertices.
    <holder> A pointer to the canvas information.
    <scale> The scale value to apply to the data before drawing.
    <offset> The offset value to apply to the scaled data before drawing.
    <num_points> The number of data points to plot.
    <x> The array of horizontal world co-ordinates. These must be in left to
    right order.
    <y> The array of vertical world co-ordinates.
    <pixel_value> The pixel value to draw with.
    <num_pixels> The number of pixel values. If this is 0, <<pixel_value>> is
    used instead.
    <pixel_values> The array of pixel values.
    [RETURNS] Nothing.
*/
{
    unsigned int icount;
    double min_wy, max_wy;
    CONST struct canvas_info_struct *canvas_info = &holder->canvas_info;

    if (num_points < 2) return;
    min_wy = canvas_info->min_wy;
    max_wy = canvas_info->max_wy;
    for (icount = 0; icount < num_points; )
    {   /*  Process in blocks  */
	unsigned int ocount, npoints;
	unsigned int istart = icount;
	double xbuf[BUF_SIZE], ybuf[BUF_SIZE];

	/*  Extract a list of points which lie inside the boundary  */
	for (ocount = 0; (ocount < BUF_SIZE) && (icount < num_points); )
	{
	    double xt, yt;

	    ++icount;   /* Must ensure increments happen even if loop broken */
	    xt = *x++;
	    yt = *y++;
	    if ( !X_IN_LIMITS (xt, canvas_info) ) break;
	    if ( DOUBLE_IS_BLANK (yt) ) break;
	    yt = yt * scale + offset;
	    if ( (yt < min_wy) || (yt > max_wy) ) break;
	    xbuf[ocount] = xt;
	    ybuf[ocount] = yt;
	    ++ocount;
	}
	npoints = ocount;
	if (npoints < 2) continue;
	canvas_convert_from_canvas_coords (holder->canvas, FALSE, FALSE,
					   npoints, xbuf, ybuf, xbuf, ybuf);
	if (num_pixels < 1)
	{
	    kwin_draw_lines_TRANSITION (holder->pixcanvas,
					xbuf, ybuf, npoints,
					pixel_value);
	    continue;
	}
	/*  Individual pixel values: do it the hard way  */
	for (ocount = 0; ocount < npoints - 1; ++ocount)
	{
	    pixel_value = pixel_values[(ocount + istart) * (num_pixels - 1) /
				      (num_points - 1)];
	    kwin_draw_line (holder->pixcanvas, xbuf[ocount], ybuf[ocount],
			    xbuf[ocount + 1], ybuf[ocount + 1], pixel_value);
	}
    }
}   /*  End Function _graph1d_lines  */

static void _graph1d_steps (CONST struct canvas_holder_type *holder,
			    double scale, double offset,
			    unsigned int num_points,
			    CONST double *x, CONST double *y,
			    unsigned long pixel_value, unsigned int num_pixels,
			    unsigned long *pixel_values)
/*  [SUMMARY] Plot a 1-dimensional graph by drawing steps.
    <holder> A pointer to the canvas information.
    <scale> The scale value to apply to the data before drawing.
    <offset> The offset value to apply to the scaled data before drawing.
    <num_points> The number of data points to plot.
    <x> The array of horizontal world co-ordinates. These must be in left to
    right order.
    <y> The array of vertical world co-ordinates.
    <pixel_value> The pixel value to draw with.
    <num_pixels> The number of pixel values. If this is 0, <<pixel_value>> is
    used instead.
    <pixel_values> The array of pixel values.
    [RETURNS] Nothing.
*/
{
    unsigned int icount;
    double min_wy, max_wy, last_x, next_x;
    CONST struct canvas_info_struct *canvas_info = &holder->canvas_info;

    min_wy = canvas_info->min_wy;
    max_wy = canvas_info->max_wy;
    last_x = x[0];
    for (icount = 0; icount < num_points; )
    {   /*  Process in blocks  */
	unsigned int ocount, npoints;
	unsigned int istart = icount;
	double xbuf[BUF_SIZE], ybuf[BUF_SIZE];

	/*  Extract a list of points which lie inside the boundary  */
	for (ocount = 0; (ocount + 2 < BUF_SIZE) && (icount < num_points); )
	{
	    flag draw_h = TRUE;
	    flag bad_v = FALSE;
	    double x0, y0, y1;

	    icount++;   /* Must ensure increments happen even if loop broken */
	    x0 = *x++;
	    y0 = *y++;
	    xbuf[ocount] = last_x;
	    if (icount >= num_points) next_x = x0;
	    else next_x = (x0 + *x) * 0.5;
	    last_x = next_x;
	    if ( !X_IN_LIMITS (x0, canvas_info) ) break;
	    if ( DOUBLE_IS_BLANK (y0) ) break;
	    y0 = y0 * scale + offset;
	    if (y0 < min_wy) y0 = min_wy, draw_h = FALSE;
	    if (y0 > max_wy) y0 = max_wy, draw_h = FALSE;
	    if (draw_h) ybuf[ocount++] = y0;
	    xbuf[ocount] = next_x;
	    ybuf[ocount++] = y0;
	    if (icount >= num_points) break;
	    y1 = *y;
	    if ( DOUBLE_IS_BLANK (y1) ) break;
	    y1 = y1 * scale + offset;
	    if (y1 < min_wy)
	    {
		if (!draw_h) break;
		y1 = min_wy;
		bad_v = TRUE;
	    }
	    if (y1 > max_wy)
	    {
		if (!draw_h) break;
		y1 = max_wy;
		bad_v = TRUE;
	    }
	    if (bad_v)
	    {
		xbuf[ocount] = next_x;
		ybuf[ocount++] = y1;
		break;
	    }  /*  Else vertex will be added in the next iteration  */
	}
	npoints = ocount;
	if (npoints < 2) continue;
	canvas_convert_from_canvas_coords (holder->canvas, FALSE, FALSE,
					   npoints, xbuf, ybuf, xbuf, ybuf);
	if (num_pixels < 1)
	{
	    kwin_draw_lines_TRANSITION (holder->pixcanvas,
					xbuf, ybuf, npoints, pixel_value);
	    continue;
	}
	/*  Individual pixel values: do it the hard way  */
	for (ocount = 0; ocount < npoints - 1; ++ocount)
	{
	    pixel_value = pixel_values[(ocount /3 + istart) * (num_pixels -1) /
				      (num_points - 1)];
	    kwin_draw_line (holder->pixcanvas, xbuf[ocount], ybuf[ocount],
			    xbuf[ocount + 1], ybuf[ocount + 1], pixel_value);
	}
    }
}   /*  End Function _graph1d_steps  */

static void _graph1d_points (CONST struct canvas_holder_type *holder,
			     double scale,double offset,
			     unsigned int num_points,
			     CONST double *x, CONST double *y,
			     unsigned long pixel_value)
/*  [SUMMARY] Plot a 1-dimensional graph by drawing crosses.
    <holder> A pointer to the canvas information.
    <scale> The scale value to apply to the data before drawing.
    <offset> The offset value to apply to the scaled data before drawing.
    <num_points> The number of data points to plot.
    <x> The array of horizontal world co-ordinates. These must be in left to
    right order.
    <y> The array of vertical world co-ordinates.
    <pixel_value> The pixel value to draw with.
    [RETURNS] Nothing.
*/
{
    double min_wy, max_wy;
    CONST struct canvas_info_struct *canvas_info = &holder->canvas_info;

    min_wy = canvas_info->min_wy;
    max_wy = canvas_info->max_wy;
    while (num_points > 0)
    {   /*  Process in blocks  */
	unsigned int count, npoints;
	double xbuf[BUF_SIZE], ybuf[BUF_SIZE];
	double x0[BUF_SIZE], y0[BUF_SIZE], x1[BUF_SIZE], y1[BUF_SIZE];

	/*  Extract a list of points which lie inside the boundary  */
	for (count = 0; (count < BUF_SIZE) && (num_points > 0); )
	{
	    double xt, yt;

	    --num_points; /*Must ensure increments happen even if loop broken*/
	    xt = *x++;
	    yt = *y++;
	    if ( !X_IN_LIMITS (xt, canvas_info) ) continue;
	    if ( DOUBLE_IS_BLANK (yt) ) continue;
	    yt = yt * scale + offset;
	    if ( (yt < min_wy) || (yt > max_wy) ) continue;
	    xbuf[count] = xt;
	    ybuf[count] = yt;
	    ++count;
	}
	npoints = count;
	canvas_convert_from_canvas_coords (holder->canvas, FALSE, FALSE,
					   npoints, xbuf, ybuf, xbuf, ybuf);
	/*  Generate cross shapes  */
	for (count = 0; count < npoints; ++count)
	{
	    double two = 2.0;

	    x0[count] = xbuf[count] - two;
	    y0[count] = ybuf[count] - two;
	    x1[count] = xbuf[count] + two;
	    y1[count] = ybuf[count] + two;
	}
	kwin_draw_segments_TRANSITION (holder->pixcanvas,
				       x0, ybuf, x1, ybuf, npoints,
				       pixel_value);
	kwin_draw_segments_TRANSITION (holder->pixcanvas,
				       xbuf, y0, xbuf, y1, npoints,
				       pixel_value);
    }
}   /*  End Function _graph1d_points  */
