/*LINTLIBRARY*/
/*  queue_event.c

    This code provides event queueing routines for pixel canvases.

    Copyright (C) 2003  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains event queueing routines for pixel canvases.


    Written by      Richard Gooch   17-SEP-2003

    Last updated by Richard Gooch   17-SEP-2003


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <karma_kwin.h>
#include <karma_w.h>
#include <karma_m.h>
#include <k_event_codes.h>


#define CANVAS_KEY    &_kwin_queue_event_canvas_key


/*  Structure declarations  */

struct queue_struct
{
    KPixCanvas canvas;
    int x;
    int y;
    flag clip;
    unsigned int event_code;
    KWorkFunc work_func;
};


/*  Private data  */
static char _kwin_queue_event_canvas_key;           /*  Unique key  */


/*  Private functions  */
STATIC_FUNCTION (flag _kwin_queue_event_work_func, (void **info) );


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
flag kwin_queue_position_event (KPixCanvas canvas, int x, int y, flag clip,
				unsigned int event_code, void *event_info)
/*  [SUMMARY] Queue a position event on a pixel canvas for possible processing.
    [PURPOSE] This function is similar to the [<kwin_process_position_event>]
    function, except that non-critical events may be dropped if the system is
    too busy. This avoids application response lagging behind user input.
    <canvas> The pixel canvas.
    <x> The horizontal position of the event, relative to the parent window.
    <y> The vertical position of the event, relative to the parent window.
    <clip> If this is TRUE, events outside of the canvas boundary are clipped
    to the nearest boundary and passed on to the registered position event
    consumers. If this is FALSE, events outside of the canvas boundary are not
    consumed (i.e. they are ignored).
    <event_code> The arbitrary event code.
    <event_info> The arbitrary event information.
    [NOTE] Only mouse move and mouse drag events are considered non-critical.
    [NOTE] Older events are dropped, ensuring that the most recent event is
    always processed.
    [NOTE] Events may be processed in the context of a work function (see the
    [<w>] package for information on work functions).
    [RETURNS] TRUE if the event was consumed or queued, else FALSE.
*/
{
    flag disposable_event;
    struct queue_struct *queue;
    static char function_name[] = "kwin_queue_position_event";

    if ( !w_test_supported () )
	return kwin_process_position_event (canvas, x, y, clip,
					    event_code, event_info);
    queue = kwin_get_associated_object (canvas, CANVAS_KEY);
    if (!queue)
    {
	if ( ( queue = m_calloc (sizeof *queue) ) == NULL )
	    m_abort (function_name, "event queue");
	kwin_associate_object (canvas, CANVAS_KEY, queue, FALSE);
	queue->canvas = canvas;
	queue->event_code = K_CANVAS_EVENT_UNDEFINED;
    }
    switch (event_code)
    {
      case K_CANVAS_EVENT_LEFT_MOUSE_DRAG:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_DRAG:
      case K_CANVAS_EVENT_RIGHT_MOUSE_DRAG:
      case K_CANVAS_EVENT_POINTER_MOVE:
      case K_CANVAS_EVENT_MOUSE_BUTTON4_DRAG:
      case K_CANVAS_EVENT_MOUSE_BUTTON5_DRAG:
	disposable_event = event_info ? FALSE : TRUE;
	break;
      default:
	disposable_event = FALSE;
	break;
    }
    if (!disposable_event)
    {   /*  Should be processed now. First process any queued events  */
	if (queue->work_func) w_do_work ();
	return kwin_process_position_event (canvas, x, y, clip,
					    event_code, event_info);
    }
    /*  If pending event is different: process it now  */
    if ( queue->work_func && (event_code != queue->event_code) ) w_do_work ();
    queue->event_code = event_code;
    queue->x = x;
    queue->y = y;
    queue->clip = clip;
    if (!queue->work_func)
	queue->work_func = w_register_func (_kwin_queue_event_work_func, queue,
					    KWF_PRIORITY_HIGHEST);
    return (TRUE);
}   /*  End Function kwin_queue_position_event  */


/*  Private functions follow  */

static flag _kwin_queue_event_work_func (void **info)
/*  [SUMMARY] Work function.
    [PURPOSE] This routine is called to perform some work.
    <info> A pointer to the arbitrary work function information pointer.
    [RETURNS] TRUE if the work function should be called again, else FALSE
    indicating that the work function is to be unregistered.
*/
{
    struct queue_struct *queue = *info;

    kwin_process_position_event (queue->canvas,
				 queue->x, queue->y, queue->clip,
				 queue->event_code, NULL);
    queue->work_func = NULL;
    return (FALSE);
}   /*  End Function work_func  */
