/*LINTLIBRARY*/
/*  Canvas.c

    This code provides a canvas widget for Xt.

    Copyright (C) 1994-2005  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains all routines needed for a canvas widget for Xt.


    Written by      Richard Gooch   17-JUL-1994

    Updated by      Richard Gooch   19-JUL-1994

    Updated by      Richard Gooch   30-JUL-1994: Create graphics context on
  realisation.

    Updated by      Richard Gooch   9-AUG-1994: Added MapNotify and UnmapNotify
  event handlers.

    Updated by      Richard Gooch   12-OCT1994: Added ButtonRelease translation
  and used button release event codes.

    Updated by      Richard Gooch   25-OCT-1994: Moved some #include's around.

    Updated by      Richard Gooch   5-NOV-1994: Added VisibilityNotify event
  handler.

    Updated by      Richard Gooch   23-DEC-1994: Added realiseCallback resource

    Updated by      Richard Gooch   4-JAN-1995: Cosmetic changes.

    Updated by      Richard Gooch   28-JAN-1995: Added stereo support.

    Updated by      Richard Gooch   17-APR-1995: Fixed mix of XtRBool and
  sizeof Boolean in resource fields. Use Bool because it is of int size.

    Updated by      Richard Gooch   26-JUL-1995: Removed buggy setting of
  border pixmap in <Realise>.

    Updated by      Richard Gooch   23-SEP-1995: Made use of XGL.

    Updated by      Richard Gooch   13-OCT-1995: Ensure XGL stereo canvases are
  not visible upon creation.

    Updated by      Richard Gooch   20-OCT-1995: Removed call to special XGL
  resize function.

    Updated by      Richard Gooch   22-OCT-1995: Force redisplay upon Expose
  even if canvas state not visible: print a warning message.

    Updated by      Richard Gooch   29-OCT-1995: Made use of OpenGL.

    Updated by      Richard Gooch   23-DEC-1995: Supported split (fake) stereo
  mode.

    Updated by      Richard Gooch   28-DEC-1995: Added verbose resource.

    Updated by      Richard Gooch   30-APR-1996: Added Initialise method.

    Updated by      Richard Gooch   4-MAY-1996: Subclassed from SimpleWidget.

    Updated by      Richard Gooch   9-MAY-1996: Took account of changes to
  OpenGL interface.

    Updated by      Richard Gooch   11-MAY-1996: Added verticalSplit and
  splitSeparation resources.

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   15-SEP-1996: Made use of new <xv_*>
  routines.

    Updated by      Richard Gooch   23-OCT-1996: Freeze/unfreeze pointer move
  events upon pressing of spacebar.

    Updated by      Richard Gooch   26-OCT-1996: Refresh canvas when 'r' key is
  pressed.

    Updated by      Richard Gooch   27-OCT-1996: Supported key presses with
  modifiers.

    Updated by      Richard Gooch   6-NOV-1996: Cosmetic changes.

    Updated by      Richard Gooch   23-NOV-1996: Clear canvas when 'r' is
  pressed.

    Updated by      Richard Gooch   27-JAN-1997: Added popup and popdown
  callbacks to Shell widget ancestor.

    Updated by      Richard Gooch   19-FEB-1997: Set background pixel for
  pixel canvas.

    Updated by      Richard Gooch   3-MAR-1997: Quash notice message for zero
  length strings.

    Updated by      Richard Gooch   21-MAR-1997: Don't set canvas visible in
  shell popup callback if widget was not mapped.

    Updated by      Richard Gooch   13-MAY-1997: Allocate fresh copy of
  fontName string.

    Updated by      Richard Gooch   14-AUG-1997: Supported cursor resources
  defined in Simple widget class.

    Updated by      Richard Gooch   4-SEP-1997: Added timing controls.

    Updated by      Richard Gooch   20-OCT-1997: Added fallback to "fixed" font
  if selected font not available.

    Updated by      Richard Gooch   8-DEC-1997: Added cursor key bindings.

    Updated by      Richard Gooch   22-FEB-1998: Disabled popup callback.

    Updated by      Richard Gooch   16-JUN-1998: Disable messages about
  installing colourmap.

    Updated by      Richard Gooch   29-JUL-1998: Added control-G binding.

    Updated by      Richard Gooch   12-AUG-1998: Added generation of event
  K_CANVAS_EVENT_POINTER_WARP.

    Updated by      Richard Gooch   15-OCT-1998: Made use of new function
  <kwin_warp_pointer>.

    Updated by      Richard Gooch   2-SEP-1999: Took account of interface
  change to function <kwin_warp_pointer> and discard events.

    Updated by      Richard Gooch   14-SEP-1999: Added generation of events
  K_CANVAS_EVENT_POINTER_ENTER and K_CANVAS_EVENT_POINTER_LEAVE.

    Updated by      Richard Gooch   23-SEP-1999: No longer discard events when
  warping pointer due to arrow key press (was not necessary).

    Updated by      Richard Gooch   8-DEC-1999: Added generation of event
  K_CANVAS_EVENT_KEYSYM.

    Updated by      Richard Gooch   9-JUN-2000: Put Xmultibuffer-specific code
  inside #ifdef HAS_Xmbuf.

    Updated by      Richard Gooch   28-JUN-2000: Created Destroy method.

    Updated by      Richard Gooch   2-JUL-2000: Free fontName resource in
  Destroy method.

    Updated by      Richard Gooch   13-JUL-2000: Convert control characters to
  lowercase characters so that upper levels only need to check the modifier
  mask.

    Updated by      Richard Gooch   8-DEC-2001: Fixed built-in control key
  bindings broken by previous change. Split 'r'-triggered refresh to clear,
  flush and then redisplay, so that precompute/buffering effects can be seen.

    Updated by      Richard Gooch   17-DEC-2001: Force clipping for LeaveNotify
  event.

    Updated by      Richard Gooch   23-DEC-2001: Drop LeaveNotify events when
  pointer moves are disabled.

    Updated by      Richard Gooch   1-JAN-2002: Added generation of events for
  mouse buttons 4 and 5.

    Updated by      Richard Gooch   23-JAN-2002: Added marker for frozen cursor

    Updated by      Richard Gooch   23-MAY-2002: Added verbose fprintf to
  <ConvertCursor>.

    Updated by      Richard Gooch   14-AUG-2002: Added #include <stdlib.h>

    Updated by      Richard Gooch   17-AUG-2002: Added generation of events:
  K_CANVAS_EVENT_PLAIN_KEY_RELEASE, K_CANVAS_EVENT_MODIFIED_KEY_RELEASE and
  K_CANVAS_EVENT_KEYSYM_RELEASE.

    Updated by      Richard Gooch   9-DEC-2002: Registered binding list.

    Updated by      Richard Gooch   10-DEC-2002: Added showBindings resource.

    Updated by      Richard Gooch   11-DEC-2002: Made 'r' key binding use
  <kwin_resize> to clear canvas, so that double buffer is cleared if needed.
  Moved timing functionality to <kwin> package.

    Updated by      Richard Gooch   22-DEC-2002: Stupid compiler warning fix.

    Updated by      Richard Gooch   10-FEB-2003: Added suggestion line when
  bad key modifier is found.

    Updated by      Richard Gooch   17-SEP-2003: Switched to function
  <kwin_queue_position_event>.

    Updated by      Richard Gooch   11-JAN-2004: Show colourmap install events
  when verbose flag is set.

    Updated by      Richard Gooch   10-MAR-2004: Added processSpecialKeys
  resource.

    Updated by      Richard Gooch   11-MAR-2004: Now also generate
  K_CANVAS_EVENT_SHIFTED_MASK.

    Updated by      Richard Gooch   26-APR-2004: Changed to use fputs() where
  possible.

    Updated by      Richard Gooch   7-MAY-2004: Fixed bug so that event mask
  K_CANVAS_EVENT_SHIFTED_MASK is actually generated when needed (11-MAR-2004
  change had a bug that prevented this from working).

    Updated by      Richard Gooch   15-AUG-2004: Drop EnterNotify events when
  pointer moves are disabled.

    Updated by      Richard Gooch   10-OCT-2004: Handle NULL return from
  XKeysymToString().

    Last updated by Richard Gooch   24-MAY-2005: Added control-spacebar binding
  to freeze button and keyboard events as well.


*/
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#ifndef X11
#  define X11
#endif
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/XawInit.h>
#define XK_MISCELLANY
#include <X11/keysymdef.h>
#include <karma.h>
#include <karma_kwin.h>
#include <karma_xv.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>
#include <k_event_codes.h>
#include <Xkw/CanvasP.h>
#include <X11/Xmu/Drawing.h>


#define CURSOR_RADIUS 4
#define USE_WM_CMAP_INSTALLATION

#define offset(field) XtOffsetOf(CanvasRec, canvas.field)

static XtResource resources[] =
{
    {XtNvisual, XtCVisual, XtRVisual, sizeof (Visual *),
     offset (visual), XtRImmediate, CopyFromParent},
    {XkwNstereoMode, XkwCStereoMode, XtRInt, sizeof (int),
     offset (stereoMode), XtRImmediate, XkwSTEREO_MODE_MONO},
    {XkwNverticalSplit, XkwCVerticalSplit, XtRBool, sizeof (Bool),
     offset (verticalSplit), XtRImmediate, False},
    {XkwNsplitSeparation, XkwCSplitSeparation, XtRInt, sizeof (int),
     offset (splitSeparation), XtRImmediate, 0},
    {XkwNmonoPixCanvas, XkwCPixCanvas, XtRPointer, sizeof (XtPointer),
     offset (monoPixCanvas), XtRImmediate, NULL},
    {XkwNleftPixCanvas, XkwCPixCanvas, XtRPointer, sizeof (XtPointer),
     offset (leftPixCanvas), XtRImmediate, NULL},
    {XkwNrightPixCanvas, XkwCPixCanvas, XtRPointer, sizeof (XtPointer),
     offset (rightPixCanvas), XtRImmediate, NULL},
    {XkwNclipEvents, XkwCClipEvents, XtRBool, sizeof (Bool),
     offset (clipEvents), XtRImmediate, False},
    {XkwNsilenceUnconsumed, XkwCSilenceUnconsumed, XtRBool, sizeof (Bool),
     offset (silence_unconsumed_events), XtRImmediate, False},
    {XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
     offset (foreground_pixel), XtRString, XtDefaultForeground},
    {XkwNforceNewCmap, XkwCForceNewCmap, XtRBool, sizeof (Bool),
     offset (force_new_cmap), XtRImmediate, False},
    {XkwNretainFgBg, XkwCRetainFgBg, XtRBool, sizeof (Bool),
     offset (retain_fgbg), XtRImmediate, False},
    {XkwNrealiseCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
     offset (realiseCallback), XtRCallback, (caddr_t) NULL},
    {XtNfont, XtCFont, XtRString, sizeof (String),
     offset (fontName), XtRImmediate, NULL},
    {XkwNshowBindings, XkwCShowBindings, XtRBool, sizeof (Bool),
     offset (showBindings), XtRImmediate, False},
    {XkwNprocessSpecialKeys, XkwCProcessSpecialKeys, XtRBool, sizeof (Bool),
     offset (processSpecialKeys), XtRImmediate, (XtPointer) True},
    {XkwNverbose, XkwCVerbose, XtRBool, sizeof (Bool),
     offset (verbose), XtRImmediate, False},
};
#undef offset

STATIC_FUNCTION (void _Canvas_Initialise, (Widget request, Widget new) );
STATIC_FUNCTION (void _Canvas_Realise, (Widget w, Mask *valueMask,
					XSetWindowAttributes *attributes) );
STATIC_FUNCTION (void _Canvas_Destroy, (Widget W) );
STATIC_FUNCTION (Boolean _Canvas_SetValues, () );
STATIC_FUNCTION (void _Canvas_Redisplay,
		 (Widget W, XEvent *event, Region region) );
STATIC_FUNCTION (void _Canvas_refresh, (CanvasWidget cnv, flag clear) );
STATIC_FUNCTION (void CanvasColourmapChange,
		 (CanvasWidget cnv, XEvent *event) );
STATIC_FUNCTION (void CanvasEnterNotify, (CanvasWidget cnv, XEvent *event) );
STATIC_FUNCTION (void CanvasLeaveNotify, (CanvasWidget cnv, XEvent *event) );
STATIC_FUNCTION (void CanvasMotionNotify, (CanvasWidget cnv, XEvent *event) );
STATIC_FUNCTION (void CanvasButtonNotify, (CanvasWidget cnv, XEvent *event) );
STATIC_FUNCTION (void CanvasMapNotify, (CanvasWidget cnv, XEvent *event) );
STATIC_FUNCTION (void CanvasUnmapNotify, (CanvasWidget cnv, XEvent *event) );
STATIC_FUNCTION (void CanvasVisibilityNotify,
		 (CanvasWidget cnv, XEvent *event) );
STATIC_FUNCTION (void CanvasKeyPress, (CanvasWidget cnv, XEvent *event) );
STATIC_FUNCTION (void CanvasKeyRelease, (CanvasWidget cnv, XEvent *event) );
STATIC_FUNCTION (void CanvasKeyEvent, (CanvasWidget cnv, XEvent *event,
				       flag press) );
#ifdef HAS_Xmbuf
STATIC_FUNCTION (void create_stereo_window,
		 (Widget widget, unsigned int window_class, Visual *visual,
		  XtValueMask value_mask, XSetWindowAttributes *attributes,
		  Multibuffer *left, Multibuffer *right) );
#else
#  define create_stereo_window(widget,wc,v,vm,a,l,r) {}
#endif
STATIC_FUNCTION (void set_visibility, (CanvasWidget cnv, flag visible) );
STATIC_FUNCTION (void inject_event,
		 (CanvasWidget cnv, int x, int y, unsigned int k_event,
		  flag clip, CONST char *string) );
STATIC_FUNCTION (void popup_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void popdown_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _Canvas_ConvertCursor, (Widget W) );
STATIC_FUNCTION (void _Canvas_draw_marker, (CanvasWidget w) );


static char defaultTranslations[] =
  "#override \n\
   <ColormapNotify>:   CanvasColourmapChange()  \n\
   <EnterNotify>:      CanvasEnterNotify() \n\
   <LeaveNotify>:      CanvasLeaveNotify() \n\
   <MotionNotify>:     CanvasMotionNotify() \n\
   <ButtonPress>:      CanvasButtonNotify() \n\
   <ButtonRelease>:    CanvasButtonNotify() \n\
   <MapNotify>:        CanvasMapNotify() \n\
   <UnmapNotify>:      CanvasUnmapNotify() \n\
   <VisibilityNotify>: CanvasVisibilityNotify() \n\
   <KeyPress>:         CanvasKeyPress() \n\
   <KeyRelease>:       CanvasKeyRelease()";

static XtActionsRec actions[] = {
  {"CanvasColourmapChange",      (XtActionProc) CanvasColourmapChange},
  {"CanvasEnterNotify",          (XtActionProc) CanvasEnterNotify},
  {"CanvasLeaveNotify",          (XtActionProc) CanvasLeaveNotify},
  {"CanvasMotionNotify",         (XtActionProc) CanvasMotionNotify},
  {"CanvasButtonNotify",         (XtActionProc) CanvasButtonNotify},
  {"CanvasMapNotify",            (XtActionProc) CanvasMapNotify},
  {"CanvasUnmapNotify",          (XtActionProc) CanvasUnmapNotify},
  {"CanvasVisibilityNotify",     (XtActionProc) CanvasVisibilityNotify},
  {"CanvasKeyPress",             (XtActionProc) CanvasKeyPress},
  {"CanvasKeyRelease",           (XtActionProc) CanvasKeyRelease},
};

CanvasClassRec canvasClassRec =
{
    {
	/* core fields  */
	(WidgetClass) &simpleClassRec,  /*  superclass             */
	"Canvas",                       /*  class_name             */
	sizeof (CanvasRec),             /*  widget_size            */
	NULL,                           /*  class_initialise       */
	NULL,                           /*  class_part_initialise  */
	FALSE,                          /*  class_inited           */
	(XtInitProc) _Canvas_Initialise,/*  initialise             */
	NULL,                           /*  initialise_hook        */
	_Canvas_Realise,                /*  realise                */
	actions,                        /*  actions                */
	XtNumber (actions),             /*  num_actions            */
	resources,                      /*  resources              */
	XtNumber (resources),           /*  num_resources          */
	NULLQUARK,                      /*  xrm_class              */
	TRUE,                           /*  compress_motion        */
	TRUE,                           /*  compress_exposure      */
	TRUE,                           /*  compress_enterleave    */
	TRUE,                           /*  visible_interest       */
	_Canvas_Destroy,                /*  destroy                */
	NULL,                           /*  resize                 */
	_Canvas_Redisplay,              /*  expose                 */
	_Canvas_SetValues,              /*  set_values             */
	NULL,                           /*  set_values_hook        */
	XtInheritSetValuesAlmost,       /*  set_values_almost      */
	NULL,                           /*  get_values_hook        */
	NULL,                           /*  accept_focus           */
	XtVersion,                      /*  version                */
	NULL,                           /*  callback_private       */
	defaultTranslations,            /*  tm_table               */
	XtInheritQueryGeometry,         /*  query_geometry         */
	XtInheritDisplayAccelerator,    /*  display_accelerator    */
	NULL                            /*  extension              */
    },
    {   /* simple fields */
	XtInheritChangeSensitive        /* change_sensitive        */
    },
    {
	/*  canvas fields */
	0                               /*  ignore                 */
    }
};

WidgetClass canvasWidgetClass = (WidgetClass) &canvasClassRec;

static KPixCanvasBinding bindings[] =
{
  {"arrow, left",  "move pointer one screen pixel left (control: 10 pixels)"},
  {"arrow, right", "move pointer one screen pixel right (control: 10 pixels)"},
  {"arrow, up",     "move pointer one screen pixel up (control: 10 pixels)"},
  {"arrow, down",   "move pointer one screen pixel down (control: 10 pixels)"},
  {"spacebar",           "toggle freeze/unfreeze of pointer move events"},
  {"spacebar, control",  "remember frozen position for button/key events"},
  {"r",                  "clear and refresh window"},
  {"t, control",         "toggle timing"},
  {"g, control",         "ring bell"},
  {NULL,                 NULL}
};


static void _Canvas_Initialise (Widget Request, Widget New)
{
    int nstereo = 0;
    Window root_window;
    Widget shell = NULL;
    Widget parent;
    /*CanvasWidget request = (CanvasWidget) Request;*/
    CanvasWidget new = (CanvasWidget) New;
    Display *dpy;
    Screen *screen;
    Visual *visual;
    XVisualInfo *vinfo;
#ifdef HAS_Xmbuf
    int nmono, dummy;
    XmbufBufferInfo *dummy_info1, *dummy_info2;
#endif
    static char function_name[] = "CanvasWidget::Initialise";

    new->canvas.frozen_x = -1;
    new->canvas.frozen_y = -1;
    new->canvas.remember_freeze_pos = FALSE;
#ifdef HAS_Xmbuf
    new->canvas.left_window = 0;
    new->canvas.right_window = 0;
#endif
    if (new->canvas.fontName)
    {
	if ( ( new->canvas.fontName = st_dup (new->canvas.fontName) ) == NULL )
	    m_abort (function_name, "fontName");
    }
    dpy = XtDisplay (New);
    screen = XtScreen (New);
    root_window = RootWindowOfScreen (screen);
    for ( parent = XtParent (New); parent; parent = XtParent (parent) )
    {
	if ( XtIsShell (parent) )
	{
	    shell = parent;
	    break;
	}
    }
    if (shell == NULL)
    {
	fputs ("No Shell (sub)class ancestor!\n", stderr);
	a_prog_bug (function_name);
    }
    XtAddCallback (shell, XtNpopupCallback, popup_cbk, New);
    XtAddCallback (shell, XtNpopdownCallback, popdown_cbk, New);
    /*  Get visual information  */
    if ( (visual = new->canvas.visual) == CopyFromParent )
	visual = XDefaultVisualOfScreen (screen);
    vinfo = xv_get_visinfo_for_visual (dpy, visual);
    /*  If stereo mode requested, check if supported  */
    switch (new->canvas.stereoMode)
    {
      case XkwSTEREO_MODE_XMBUF:
#if HAS_Xmbuf
	if ( !XmbufQueryExtension (dpy, &dummy, &dummy) )
	{
	    fputs ("Multi-Buffering extension not available\n", stderr);
	    exit (RV_UNDEF_ERROR);
	}
	if (XmbufGetScreenInfo (dpy, root_window, &nmono, &dummy_info1,
				&nstereo, &dummy_info2) == 0)
	{
	    fputs ("Error getting Multi-Buffering screen info\n", stderr);
	    exit (RV_UNDEF_ERROR);
	}
	if (nmono > 0) XFree (dummy_info1);
	if (nstereo > 0) XFree (dummy_info2);
#endif
	if (nstereo < 1)
	{
	    fputs ("No stereo multibuffers available\n", stderr);
	    exit (RV_UNDEF_ERROR);
	}
	break;
      case XkwSTEREO_MODE_XGL:
	if (kwin_xgl_test_stereo (dpy, root_window)
	    != KWIN_XGL_STEREO_AVAILABLE)
	{
	    fputs ("XGL stereo not available\n", stderr);
	    exit (RV_UNDEF_ERROR);
	}
	break;
      case XkwSTEREO_MODE_OpenGL:
	if ( !kwin_open_gl_test_stereo (dpy, vinfo) )
	{
	    fputs ("OpenGL stereo not available\n", stderr);
	    exit (RV_UNDEF_ERROR);
	}
	break;
      case XkwSTEREO_MODE_SPLIT:
      case XkwSTEREO_MODE_MONO:
	break;
      default:
	fprintf (stderr, "Illegal stereo mode: %d\n", new->canvas.stereoMode);
	a_prog_bug (function_name);
	return;
    }
    XFree (vinfo);
}   /*  End Function Initialise  */

static void _Canvas_Realise (Widget w, Mask *valueMask,
			     XSetWindowAttributes *attributes)
{
    KPixCanvasFont font = NULL;  /*  Initialised to keep compiler happy  */
    KPixCanvas pixcanvas, stereopixcanvases[2];
    flag verbose, show_bindings;
    int count;
    int split_width, split_height, split_xoff, split_yoff;
    GC gc;
    int num_vinfos, num_cmaps;
    XWindowAttributes parent_att;
    Widget parent;
    Colormap cmap, *cmap_list;
    Display *dpy;
    Visual *visual;
    Screen *screen;
    CanvasWidget cnv = (CanvasWidget) w;
    XVisualInfo vinfo_template;
    XGCValues gcvalues;
    XVisualInfo *vinfos;
    unsigned long pixels[2];
    XColor colour_defs[2];
    static char function_name[] = "CanvasWidget::Realise";

    show_bindings = cnv->canvas.showBindings;
    verbose = cnv->canvas.verbose;
    /*  Create the X window  */
    visual = cnv->canvas.visual;
    dpy = XtDisplay (cnv);
    parent = XtParent (w);
    screen = XtScreen (parent);
    /*  Determine parent information  */
    if (XGetWindowAttributes (dpy, XtWindow (parent), &parent_att) == 0)
    {
	fputs ("Error getting parent window attributes\n", stderr);
	a_prog_bug (function_name);
    }
    if (visual == CopyFromParent) visual = parent_att.visual;
    if (visual == parent_att.visual)
    {
	if (verbose)
	{
	    fprintf (stderr, "%s(%p): creating window with default visual\n",
		     function_name, w);
	}
    }
    else
    {
	/*  Canvas visual different from parent  */
	vinfo_template.visualid = XVisualIDFromVisual (visual);
	vinfos = XGetVisualInfo (dpy, VisualIDMask, &vinfo_template,
				 &num_vinfos);
	if (num_vinfos < 1)
	{
	    fprintf (stderr, "Error getting visual info for visual: %p\n",
		     (void *) visual);
	    a_prog_bug (function_name);
	}
	if (num_vinfos > 1)
	{
	    fprintf (stderr,
		     "%s: WARNING: number of visuals for visual: %p is: %d\n",
		     function_name, (void *) visual, num_vinfos);
	}
	if (verbose)
	{
	    fprintf (stderr,
		     "%s(%p): creating window with visualID: 0x%lx\tdepth: %d\n",
		     function_name, w, vinfos[0].visualid, vinfos[0].depth);
	}
	cnv->core.depth = vinfos[0].depth;
	XFree ( (void *) vinfos );
	if (parent_att.colormap == cnv->core.colormap)
	{
	    /*  Invalidate original colourmap  */
	    cnv->canvas.force_new_cmap = True;
	}
    }
    if (cnv->canvas.force_new_cmap)
    {
	/*  Need to create a colourmap  */
	if ( ( cnv->core.colormap =
	       XCreateColormap (dpy, XRootWindowOfScreen (screen),
				visual, AllocNone) ) == (Colormap) None )
	{
	    fputs ("Could not create colourmap\n", stderr);
	    exit (1);
	}
	if (verbose) fprintf (stderr, "%s: new cmap: 0x%lx\n",
			      function_name, cnv->core.colormap);
	/* Copy over foreground and background colours from parent colourmap */
	colour_defs[0].pixel = cnv->canvas.foreground_pixel;
	colour_defs[1].pixel = cnv->core.background_pixel;
	XQueryColors (dpy, parent_att.colormap, colour_defs, 2);
	if (XAllocColor (dpy, cnv->core.colormap, colour_defs) == 0)
	{
	    fputs ("Could not allocate colour\n", stderr);
	    exit (1);
	}
	if (XAllocColor (dpy, cnv->core.colormap, colour_defs + 1) == 0)
	{
	    fputs ("Could not allocate colour\n", stderr);
	    exit (1);
	}
	pixels[0] = colour_defs[0].pixel;
	pixels[1] = colour_defs[1].pixel;
	if (!cnv->canvas.retain_fgbg) 
	{
	    XFreeColors (dpy, cnv->core.colormap, pixels, 2, 0);
	}
	cnv->canvas.foreground_pixel = pixels[0];
	cnv->core.background_pixel = pixels[1];
    }
    attributes->colormap = cnv->core.colormap;
    attributes->background_pixel = cnv->core.background_pixel;
    *valueMask |= CWColormap | CWBackPixel;
    _Canvas_ConvertCursor (w);
    if ( ( attributes->cursor = cnv->simple.cursor ) != None )
        *valueMask |= CWCursor;
    switch (cnv->canvas.stereoMode)
    {
      case XkwSTEREO_MODE_XMBUF:
	create_stereo_window (w, (unsigned int) InputOutput, visual,
			      *valueMask, attributes,
			      &cnv->canvas.left_window,
			      &cnv->canvas.right_window);
	break;
      case XkwSTEREO_MODE_XGL:
      case XkwSTEREO_MODE_OpenGL:
	if (cnv->core.depth == 8) fprintf (stderr, "bg: %lu\n",
					   cnv->core.background_pixel);
	/*  Fall through  */
      case XkwSTEREO_MODE_SPLIT:
      case XkwSTEREO_MODE_MONO:
	XtCreateWindow (w, (unsigned int) InputOutput, visual,
			*valueMask, attributes);
	break;
      default:
	fprintf (stderr, "Illegal stereo mode: %d\n", cnv->canvas.stereoMode);
	a_prog_bug (function_name);
	return;
    }
    cnv->canvas.cmap_installed = FALSE;
    cnv->canvas.mapped = FALSE;
    cnv->canvas.partly_unobscured = FALSE;
    cmap_list = XListInstalledColormaps (dpy, cnv->core.window, &num_cmaps);
    cmap = cnv->core.colormap;
    for (count = 0; count < num_cmaps; ++count)
    {
	if (cmap == cmap_list[count]) cnv->canvas.cmap_installed = TRUE;
    }
    XFree ( (void *) cmap_list );
    cnv->canvas.cmap_installed_internally = FALSE;
    /*  Create graphics context  */
    gcvalues.function = GXcopy;
    gcvalues.foreground = cnv->canvas.foreground_pixel;
    gcvalues.background = cnv->core.background_pixel;
    gc = XCreateGC (dpy, cnv->core.window,
		    GCForeground | GCBackground | GCFunction,
		    &gcvalues);
    /*  Create the KPixCanvas  */
    switch (cnv->canvas.stereoMode)
    {
#ifdef HAS_Xmbuf
      case XkwSTEREO_MODE_XMBUF:
	if ( ( stereopixcanvases[0] =
	       kwin_create_x (dpy, cnv->canvas.left_window, gc,
			      0, 0, cnv->core.width, cnv->core.height) )
	    == NULL )
	    m_abort (function_name, "left pixel canvas");
	kwin_register_bindings (stereopixcanvases[0], bindings, FALSE);
	if (cnv->canvas.fontName)
	{
	    if ( ( font = kwin_load_font (stereopixcanvases[0],
					  cnv->canvas.fontName) ) == NULL )
	    {
		m_error_notify (function_name, "font");
		if ( ( font = kwin_load_font (stereopixcanvases[0],
					      "fixed") ) == NULL )
		    m_abort (function_name, "font");
	    }
	    kwin_set_attributes (stereopixcanvases[0],
				 KWIN_ATT_FONT, font,
				 KWIN_ATT_END);
	}
	kwin_set_attributes (stereopixcanvases[0],
			     KWIN_ATT_VISIBLE, FALSE,
			     KWIN_ATT_BACKGROUND_PIXEL,
			     cnv->core.background_pixel,
			     KWIN_ATT_SHOW_BINDINGS, show_bindings,
			     KWIN_ATT_END);
	cnv->canvas.leftPixCanvas = stereopixcanvases[0];
	if ( ( stereopixcanvases[1] =
	      kwin_create_x (dpy, cnv->canvas.right_window, gc,
			     0, 0, cnv->core.width, cnv->core.height) )
	    == NULL )
	    m_abort (function_name, "right pixel canvas");
	kwin_register_bindings (stereopixcanvases[1], bindings, FALSE);
	if (cnv->canvas.fontName)
	    kwin_set_attributes (stereopixcanvases[1],
				 KWIN_ATT_FONT, font,
				 KWIN_ATT_END);
	kwin_set_attributes (stereopixcanvases[1],
			     KWIN_ATT_VISIBLE, FALSE,
			     KWIN_ATT_BACKGROUND_PIXEL,
			     cnv->core.background_pixel,
			     KWIN_ATT_SHOW_BINDINGS, show_bindings,
			     KWIN_ATT_END);
	cnv->canvas.rightPixCanvas = stereopixcanvases[1];
	XtCallCallbacks (w, XkwNrealiseCallback,(XtPointer) stereopixcanvases);
	break;
#endif
      case XkwSTEREO_MODE_XGL:
	if ( !kwin_xgl_create_stereo (dpy, cnv->core.window,
				      0, 0, cnv->core.width, cnv->core.height,
				      &pixcanvas, stereopixcanvases,
				      stereopixcanvases + 1) )
	{
	    fprintf (stderr, "%s: error creating stereo canvas\n",
		     function_name);
	    exit (RV_UNDEF_ERROR);
	}
#ifdef DISABLED
	if (cnv->canvas.fontName)
	{
	    if ( ( font = kwin_load_font (stereopixcanvases[0],
					  cnv->canvas.fontName) ) == NULL )
	    {
		m_error_notify (function_name, "font");
		if ( ( font = kwin_load_font (stereopixcanvases[0],
					      "fixed") ) == NULL )
		    m_abort (function_name, "font");
	    }
	    kwin_set_attributes (stereopixcanvases[0],
				 KWIN_ATT_FONT, font,
				 KWIN_ATT_END);
	    kwin_set_attributes (stereopixcanvases[1],
				 KWIN_ATT_FONT, font,
				 KWIN_ATT_END);
	}
#endif
	kwin_set_attributes (stereopixcanvases[0],
			     KWIN_ATT_VISIBLE, FALSE,
			     KWIN_ATT_BACKGROUND_PIXEL,
			     cnv->core.background_pixel,
			     KWIN_ATT_END);
	kwin_set_attributes (stereopixcanvases[1],
			     KWIN_ATT_VISIBLE, FALSE,
			     KWIN_ATT_BACKGROUND_PIXEL,
			     cnv->core.background_pixel,
			     KWIN_ATT_END);
	cnv->canvas.leftPixCanvas = stereopixcanvases[0];
	cnv->canvas.rightPixCanvas = stereopixcanvases[1];
	XtCallCallbacks (w, XkwNrealiseCallback,(XtPointer) stereopixcanvases);
	break;
      case XkwSTEREO_MODE_OpenGL:
	if ( !kwin_open_gl_create_stereo (dpy, cnv->core.window, 0, 0,
					  cnv->core.width, cnv->core.height,
					  stereopixcanvases,
					  stereopixcanvases + 1) )
	{
	    fprintf (stderr, "%s: error creating stereo canvas\n",
		     function_name);
	    exit (RV_UNDEF_ERROR);
	}
	if (cnv->canvas.fontName)
	{
	    if ( ( font = kwin_load_font (stereopixcanvases[0],
					  cnv->canvas.fontName) ) == NULL )
	    {
		m_error_notify (function_name, "font");
		if ( ( font = kwin_load_font (stereopixcanvases[0],
					      "fixed") ) == NULL )
		    m_abort (function_name, "font");
	    }
	    kwin_set_attributes (stereopixcanvases[0],
				 KWIN_ATT_FONT, font,
				 KWIN_ATT_END);
	    kwin_set_attributes (stereopixcanvases[1],
				 KWIN_ATT_FONT, font,
				 KWIN_ATT_END);
	}
	kwin_set_attributes (stereopixcanvases[0],
			     KWIN_ATT_VISIBLE, FALSE,
			     KWIN_ATT_BACKGROUND_PIXEL,
			     cnv->core.background_pixel,
			     KWIN_ATT_END);
	kwin_set_attributes (stereopixcanvases[1],
			     KWIN_ATT_VISIBLE, FALSE,
			     KWIN_ATT_BACKGROUND_PIXEL,
			     cnv->core.background_pixel,
			     KWIN_ATT_END);
	cnv->canvas.leftPixCanvas = stereopixcanvases[0];
	cnv->canvas.rightPixCanvas = stereopixcanvases[1];
	XtCallCallbacks (w, XkwNrealiseCallback,(XtPointer) stereopixcanvases);
	break;
      case XkwSTEREO_MODE_SPLIT:
	if (cnv->canvas.verticalSplit)
	{
	    split_width = cnv->core.width;
	    split_height = (cnv->core.height -cnv->canvas.splitSeparation) / 2;
	    split_xoff = 0;
	    split_yoff = split_height + cnv->canvas.splitSeparation;
	}
	else
	{
	    split_width = (cnv->core.width - cnv->canvas.splitSeparation) / 2;
	    split_height = cnv->core.height;
	    split_xoff = split_width + cnv->canvas.splitSeparation;
	    split_yoff = 0;
	}
	if ( ( stereopixcanvases[0] =
	       kwin_create_x (dpy, cnv->core.window, gc,
			      0, 0,
			      split_width, split_height) )
	    == NULL )
	    m_abort (function_name, "pixel canvas");
	kwin_register_bindings (stereopixcanvases[0], bindings, FALSE);
	kwin_set_attributes (stereopixcanvases[0],
			     KWIN_ATT_VISIBLE, FALSE,
			     KWIN_ATT_BACKGROUND_PIXEL,
			     cnv->core.background_pixel,
			     KWIN_ATT_SHOW_BINDINGS, show_bindings,
			     KWIN_ATT_END);
	if ( ( stereopixcanvases[1] =
	       kwin_create_x (dpy, cnv->core.window, gc,
			      split_xoff, split_yoff,
			      split_width, split_height) )
	    == NULL )
	    m_abort (function_name, "pixel canvas");
	kwin_register_bindings (stereopixcanvases[1], bindings, FALSE);
	kwin_set_attributes (stereopixcanvases[1],
			     KWIN_ATT_VISIBLE, FALSE,
			     KWIN_ATT_BACKGROUND_PIXEL,
			     cnv->core.background_pixel,
			     KWIN_ATT_SHOW_BINDINGS, show_bindings,
			     KWIN_ATT_END);
	if (cnv->canvas.fontName)
	{
	    if ( ( font = kwin_load_font (stereopixcanvases[0],
					  cnv->canvas.fontName) ) == NULL )
	    {
		m_error_notify (function_name, "font");
		if ( ( font = kwin_load_font (stereopixcanvases[0],
					      "fixed") ) == NULL )
		    m_abort (function_name, "font");
	    }
	    kwin_set_attributes (stereopixcanvases[0],
				 KWIN_ATT_FONT, font,
				 KWIN_ATT_END);
	    kwin_set_attributes (stereopixcanvases[1],
				 KWIN_ATT_FONT, font,
				 KWIN_ATT_END);
	}
	cnv->canvas.leftPixCanvas = stereopixcanvases[0];
	cnv->canvas.rightPixCanvas = stereopixcanvases[1];
	XtCallCallbacks (w, XkwNrealiseCallback,(XtPointer) stereopixcanvases);
	break;
      case XkwSTEREO_MODE_MONO:
	if ( ( pixcanvas =
	       kwin_create_x (dpy, cnv->core.window, gc,
			      0, 0, cnv->core.width, cnv->core.height) )
	     == NULL )
	    m_abort (function_name, "pixel canvas");
	kwin_register_bindings (pixcanvas, bindings, FALSE);
	if (cnv->canvas.fontName)
	{
	    if ( ( font = kwin_load_font (pixcanvas,
					  cnv->canvas.fontName) ) == NULL )
	    {
		m_error_notify (function_name, "font");
		if ( ( font = kwin_load_font (pixcanvas, "fixed") ) == NULL )
		    m_abort (function_name, "font");
	    }
	    kwin_set_attributes (pixcanvas,
				 KWIN_ATT_FONT, font,
				 KWIN_ATT_END);
	}
	kwin_set_attributes (pixcanvas,
			     KWIN_ATT_VISIBLE, FALSE,
			     KWIN_ATT_BACKGROUND_PIXEL,
			     cnv->core.background_pixel,
			     KWIN_ATT_SHOW_BINDINGS, show_bindings,
			     KWIN_ATT_END);
	cnv->canvas.monoPixCanvas = pixcanvas;
	XtCallCallbacks (w, XkwNrealiseCallback, (XtPointer) pixcanvas);
	break;
    }
#ifdef USE_WM_CMAP_INSTALLATION
    if (parent_att.colormap != cnv->core.colormap)
    {
	Widget shell;

	for (shell = parent; shell; shell = XtParent (shell))
	    if ( XtIsShell (shell) ) break;
	if (shell == NULL)
	{
	    fprintf (stderr, "Canvas does not have parent shell!\n");
	    a_prog_bug (function_name);
	}
	/*  A quick hack  */
	XtSetWMColormapWindows (shell, &w, 1);
    }
#endif
}   /*  End Function Realise  */

static void _Canvas_Destroy (Widget W)
{
    CanvasWidget w = (CanvasWidget) W;

    if (w->canvas.rightPixCanvas) kwin_destroy (w->canvas.rightPixCanvas);
    if (w->canvas.leftPixCanvas) kwin_destroy (w->canvas.leftPixCanvas);
#ifdef HAS_Xmbuf
    if (w->canvas.left_window != 0)
    {
	XmbufDestroyBuffers (XtDisplay (W), w->core.window);
    }
#endif
    if (w->canvas.monoPixCanvas) kwin_destroy (w->canvas.monoPixCanvas);
    if (w->canvas.fontName) m_free (w->canvas.fontName);
}   /*  End Function Destroy  */

/* ARGSUSED */
static Boolean _Canvas_SetValues (Widget Current, Widget Request, Widget New)
{
    CanvasWidget old = (CanvasWidget) Current;
    CanvasWidget new = (CanvasWidget) New;
    static char function_name[] = "CanvasWidget::SetValues";

    new->canvas.monoPixCanvas = old->canvas.monoPixCanvas;
    new->canvas.leftPixCanvas = old->canvas.leftPixCanvas;
    new->canvas.rightPixCanvas = old->canvas.rightPixCanvas;
    new->canvas.visual = old->canvas.visual;
    new->canvas.stereoMode = old->canvas.stereoMode;
    if (new->canvas.fontName != old->canvas.fontName)
    {
	if (old->canvas.fontName) m_free (old->canvas.fontName);
	if ( ( new->canvas.fontName = st_dup (new->canvas.fontName) ) == NULL )
	    m_abort (function_name, "fontName");
    }
    return False;   
}   /*  End Function SetValues  */

static void _Canvas_Redisplay (Widget W, XEvent *event, Region region)
{
    CanvasWidget w = (CanvasWidget) W;
    static char function_name[] = "CanvasWidget::Redisplay";

    if ( !XtIsRealized (W) )
    {
	fputs ("Expose on non-realised canvas widget!\n", stderr);
	return;
    }
    if (event->xexpose.window != w->core.window)
    {
	fputs ("Expose event not on canvas widget window!\n", stderr);
	return;
    }
    if (!w->canvas.cmap_installed)
    {
	if (w->canvas.verbose) fprintf (stderr, "%s: installing cmap: 0x%lx\n",
					function_name, w->core.colormap);
	XInstallColormap (XtDisplay (W), w->core.colormap);
	w->canvas.cmap_installed_internally = TRUE;
    }
    if (!w->core.visible)
    {
	fprintf (stderr, "Expose on: %p which is not visible\n", (void *) w);
    }
    if (!w->canvas.partly_unobscured)
    {
	fprintf (stderr,"Expose on: %p which is fully obscured: unobscuring\n",
		 (void *) w);
	w->canvas.partly_unobscured = TRUE;
	set_visibility (w, w->canvas.mapped);
    }
    _Canvas_refresh (w, FALSE);
}   /*  End Function Redisplay  */

static void _Canvas_refresh (CanvasWidget cnv, flag clear)
/*  [SUMMARY] Refresh a Canvas widget.
    <cnv> The CanvasWidget object.
    <clear> If TRUE, the canvas(es) are cleared.
    [RETURNS] Nothing.
*/
{
    int split_width, split_height, split_xoff, split_yoff;
    static char function_name[] = "CanvasWidget::refresh";

    if (cnv->canvas.monoPixCanvas)
    {
	if ( !kwin_resize (cnv->canvas.monoPixCanvas, clear,
			   0, 0, cnv->core.width, cnv->core.height) )
	{
	    fprintf (stderr, "%s: error redisplaying mono KPixCanvas\n",
		     function_name);
	    return;
	}
	_Canvas_draw_marker (cnv);
    }
    if (cnv->canvas.stereoMode == XkwSTEREO_MODE_SPLIT)
    {
	if (cnv->canvas.verticalSplit)
	{
	    split_width = cnv->core.width;
	    split_height = (cnv->core.height -cnv->canvas.splitSeparation) / 2;
	    split_xoff = 0;
	    split_yoff = split_height + cnv->canvas.splitSeparation;
	}
	else
	{
	    split_width = (cnv->core.width - cnv->canvas.splitSeparation) / 2;
	    split_height = cnv->core.height;
	    split_xoff = split_width + cnv->canvas.splitSeparation;
	    split_yoff = 0;
	}
	if ( !kwin_resize (cnv->canvas.leftPixCanvas, clear,
			   0, 0, split_width, split_height) )
	{
	    fprintf (stderr, "%s: error redisplaying left KPixCanvas\n",
		     function_name);
	    return;
	}
	if ( !kwin_resize (cnv->canvas.rightPixCanvas, clear,
			   split_xoff, split_yoff, split_width, split_height) )
	{
	    fprintf (stderr, "%s: error redisplaying right KPixCanvas\n",
		     function_name);
	    return;
	}
    }
    else
    {
	/*  Conventional stereo canvas  */
	if (cnv->canvas.leftPixCanvas)
	{
	    if ( !kwin_resize (cnv->canvas.leftPixCanvas, clear,
			       0, 0, cnv->core.width, cnv->core.height) )
	    {
		fprintf (stderr, "%s: error redisplaying left KPixCanvas\n",
			 function_name);
		return;
	    }
	}
	if (cnv->canvas.rightPixCanvas)
	{
	    if ( !kwin_resize (cnv->canvas.rightPixCanvas, clear,
			       0, 0, cnv->core.width, cnv->core.height) )
	    {
		fprintf (stderr, "%s: error redisplaying right KPixCanvas\n",
			 function_name);
		return;
	    }
	}
    }
}   /*  End Function refresh  */

static void CanvasColourmapChange (CanvasWidget cnv, XEvent *event)
{
    static char function_name[] = "CanvasWidget::CanvasColourmapChange";

    if (event->xcolormap.window != cnv->core.window)
    {
	fputs ("Colourmap event not on canvas widget window!\n", stderr);
	return;
    }
    switch (event->xcolormap.state)
    {
      case ColormapInstalled:
	if (cnv->canvas.verbose)
	    fprintf (stderr, "%s: colourmap 0x%lx installed\n",
		     function_name, event->xcolormap.colormap);
	if (event->xcolormap.colormap == cnv->core.colormap)
	    cnv->canvas.cmap_installed = TRUE;
	else cnv->canvas.cmap_installed = FALSE;
	break;
      case ColormapUninstalled:
	if (cnv->canvas.verbose)
	    fprintf (stderr,"%s: colourmap 0x%lx uninstalled\n",
		     function_name, event->xcolormap.colormap);
	if (event->xcolormap.colormap == cnv->core.colormap)
	    cnv->canvas.cmap_installed = FALSE;
	break;
      default:
	fprintf (stderr, "Illegal state: %d\n", event->xcolormap.state);
	a_prog_bug (function_name);
	break;
    }
}   /*  End Function CanvasColourmapChange   */

static void CanvasEnterNotify (CanvasWidget cnv, XEvent *event)
{
    if (event->xcrossing.window != cnv->core.window)
    {
	fputs ("EnterNotify event not on canvas widget window!\n", stderr);
	return;
    }
    if (!cnv->canvas.cmap_installed)
    {
#ifndef USE_WM_CMAP_INSTALLATION
	XInstallColormap (XtDisplay (cnv), cnv->core.colormap);
	cnv->canvas.cmap_installed_internally = TRUE;
#endif
    }
    if (cnv->canvas.frozen_x > -1)
    {   /*  Give the user a reminder that pointer moves are still disabled  */
	XBell (XtDisplay (cnv), -90);
	return;
    }
    inject_event (cnv, event->xcrossing.x, event->xcrossing.y,
		  K_CANVAS_EVENT_POINTER_ENTER, FALSE, NULL);
}   /*  End Function CanvasEnterNotify   */

static void CanvasLeaveNotify (CanvasWidget cnv, XEvent *event)
{
    if (event->xcrossing.window != cnv->core.window)
    {
	fputs ("LeaveNotify event not on canvas widget window!\n", stderr);
	return;
    }
    if (cnv->canvas.frozen_x > -1) return;
    inject_event (cnv, event->xcrossing.x, event->xcrossing.y,
		  K_CANVAS_EVENT_POINTER_LEAVE, TRUE, NULL);
}   /*  End Function CanvasLeaveNotify   */

static void CanvasMotionNotify (CanvasWidget cnv, XEvent *event)
{
    unsigned int k_event;
    static char function_name[] = "CanvasWidget::CanvasMotionNotify";

    if (event->xmotion.window != cnv->core.window)
    {
	fputs ("MotionNotify event not on canvas widget window!\n", stderr);
	return;
    }
    if (!cnv->canvas.cmap_installed)
    {
	if (cnv->canvas.verbose)
	    fprintf (stderr, "%s: installing cmap: 0x%lx\n",
		     function_name, cnv->core.colormap);
	XInstallColormap (XtDisplay (cnv), cnv->core.colormap);
	cnv->canvas.cmap_installed_internally = TRUE;
    }
    if (event->xmotion.state & Button1Mask)
	k_event = K_CANVAS_EVENT_LEFT_MOUSE_DRAG;
    else if (event->xmotion.state & Button2Mask)
	k_event = K_CANVAS_EVENT_MIDDLE_MOUSE_DRAG;
    else if (event->xmotion.state & Button3Mask)
	k_event = K_CANVAS_EVENT_RIGHT_MOUSE_DRAG;
    else if (event->xmotion.state & Button4Mask)
	k_event = K_CANVAS_EVENT_MOUSE_BUTTON4_DRAG;
    else if (event->xmotion.state & Button5Mask)
	k_event = K_CANVAS_EVENT_MOUSE_BUTTON5_DRAG;
    else
    {
	k_event = K_CANVAS_EVENT_POINTER_MOVE;
	if (cnv->canvas.frozen_x > -1) return;
    }
    inject_event (cnv, event->xmotion.x, event->xmotion.y, k_event, FALSE,
		  NULL);
}   /*  End Function CanvasMotionNotify   */

static void CanvasButtonNotify (CanvasWidget cnv, XEvent *event)
{
    unsigned int k_event;
    static char function_name[] = "CanvasWidget::CanvasButtonNotify";

    if (event->xbutton.window != cnv->core.window)
    {
	fputs ("ButtonNotify event not on canvas widget window!\n", stderr);
	return;
    }
    if (!cnv->canvas.cmap_installed)
    {
	if (cnv->canvas.verbose)
	    fprintf (stderr, "%s: installing cmap: 0x%lx\n",
		     function_name, cnv->core.colormap);
	XInstallColormap (XtDisplay (cnv), cnv->core.colormap);
	cnv->canvas.cmap_installed_internally = TRUE;
    }
    switch (event->xbutton.button)
    {
      case Button1:
	if (event->xbutton.state & Button1Mask)
	    k_event = K_CANVAS_EVENT_LEFT_MOUSE_RELEASE;
	else k_event = K_CANVAS_EVENT_LEFT_MOUSE_CLICK;
	break;
      case Button2:
	if (event->xbutton.state & Button2Mask)
	    k_event = K_CANVAS_EVENT_MIDDLE_MOUSE_RELEASE;
	else k_event = K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK;
	break;
      case Button3:
	if (event->xbutton.state & Button3Mask)
	    k_event = K_CANVAS_EVENT_RIGHT_MOUSE_RELEASE;
	else k_event = K_CANVAS_EVENT_RIGHT_MOUSE_CLICK;
	break;
      case Button4:
	if (event->xbutton.state & Button4Mask)
	    k_event = K_CANVAS_EVENT_MOUSE_BUTTON4_RELEASE;
	else k_event = K_CANVAS_EVENT_MOUSE_BUTTON4_CLICK;
	break;
      case Button5:
	if (event->xbutton.state & Button5Mask)
	    k_event = K_CANVAS_EVENT_MOUSE_BUTTON5_RELEASE;
	else k_event = K_CANVAS_EVENT_MOUSE_BUTTON5_CLICK;
	break;
      default:
	k_event = K_CANVAS_EVENT_UNDEFINED;
	fprintf (stderr, "%s: undefined button event, id: 0x%x\n",
		 function_name, event->xbutton.button);
	break;
    }
    inject_event (cnv, event->xbutton.x, event->xbutton.y, k_event, FALSE,
		  NULL);
}   /*  End Function CanvasButtonNotify   */

static void CanvasMapNotify (CanvasWidget cnv, XEvent *event)
/*  [SUMMARY] The MapNotify event handler.
    [PURPOSE] This function is called when the canvas is mapped.
    [RETURNS] Nothing.
    [NOTE] X does not generate MapNotify events on children of a window which
    is being mapped. Of course, the first time each child is mapped, they each
    receive MapNotify events.
*/
{
    if (event->xmap.window != cnv->core.window)
    {
	fputs ("MapNotify event not on canvas widget window!\n", stderr);
	return;
    }
    cnv->canvas.mapped = TRUE;
    set_visibility (cnv, cnv->canvas.partly_unobscured);
}   /*  End Function CanvasMapNotify   */

static void CanvasUnmapNotify (CanvasWidget cnv, XEvent *event)
/*  [SUMMARY] The UnmapNotify event handler.
    [PURPOSE] This function is called when the canvas is unmapped.
    [RETURNS] Nothing.
    [NOTE] X does not generate UnmapNotify events on children of a window which
    is being unmapped.
*/
{
    if (event->xunmap.window != cnv->core.window)
    {
	fputs ("UnmapNotify event not on canvas widget window!\n", stderr);
	return;
    }
    cnv->canvas.mapped = FALSE;
    set_visibility (cnv, FALSE);
}   /*  End Function CanvasUnmapNotify   */

static void CanvasVisibilityNotify (CanvasWidget cnv, XEvent *event)
{
    if (event->xvisibility.window != cnv->core.window)
    {
	fputs ("VisibilityNotify event not on canvas widget window!\n",stderr);
	return;
    }
    switch (event->xvisibility.state)
    {
      case VisibilityFullyObscured:
	/*  This is only received when the window is obscured behind other
	    windows, not when it is iconified. There seem to be no events
	    dispatched when the window is iconified. A pity.  */
	cnv->canvas.partly_unobscured = FALSE;
	set_visibility (cnv, FALSE);
	break;
      case VisibilityPartiallyObscured:
      case VisibilityUnobscured:
	cnv->canvas.partly_unobscured = TRUE;
	set_visibility (cnv, cnv->canvas.mapped);
	break;
      default:
	fprintf (stderr, "Unknown visibility event state: %d\n",
		 event->xvisibility.state);
	break;
    }
}   /*  End Function CanvasVisibilityNotify   */

static void CanvasKeyPress (CanvasWidget cnv, XEvent *event)
{
    CanvasKeyEvent (cnv, event, TRUE);
}   /*  End Function CanvasKeyPress  */

static void CanvasKeyRelease (CanvasWidget cnv, XEvent *event)
{
    /*  Note that not all hardware can generate key release events  */
    CanvasKeyEvent (cnv, event, FALSE);
}   /*  End Function CanvasKeyRelease  */

static void CanvasKeyEvent (CanvasWidget cnv, XEvent *event, flag press)
{
    flag bad_modifier = FALSE;
    
    int len;
    unsigned int event_code;
    unsigned int shift_bits = 0;
    unsigned long mask = 0;
    KeySym keysym;
    char buffer[STRING_LENGTH + 1];
    static char function_name[] = "CanvasWidget::CanvasKeyEvent";

    if (event->xkey.window != cnv->core.window)
    {
	fputs ("KeyPress event not on canvas widget window!\n", stderr);
	return;
    }
    shift_bits = event->xkey.state & (ShiftMask | LockMask);
    if (event->xkey.state & Button1Mask)
	mask |= K_CANVAS_EVENT_LEFT_MOUSE_MASK;
    if (event->xkey.state & Button2Mask)
	mask |= K_CANVAS_EVENT_MIDDLE_MOUSE_MASK;
    if (event->xkey.state & Button3Mask)
	mask |= K_CANVAS_EVENT_RIGHT_MOUSE_MASK;
    if (event->xkey.state & Button4Mask)
	mask |= K_CANVAS_EVENT_MOUSE_BUTTON4_MASK;
    if (event->xkey.state & Button5Mask)
	mask |= K_CANVAS_EVENT_MOUSE_BUTTON5_MASK;
    if (event->xkey.state & ControlMask)
	mask |= K_CANVAS_EVENT_CONTROL_KEY_MASK;
    if (event->xkey.state & Mod1Mask)
    {
	fprintf (stderr, "%s: modifier 1 not supported\n", function_name);
	bad_modifier = TRUE;
    }
    if (event->xkey.state & Mod2Mask)
    {
	fprintf (stderr, "%s: modifier 2 not supported\n", function_name);
	bad_modifier = TRUE;
    }
    if (event->xkey.state & Mod3Mask)
    {
	fprintf (stderr, "%s: modifier 3 not supported\n", function_name);
	bad_modifier = TRUE;
    }
    if (event->xkey.state & Mod4Mask)
    {
	fprintf (stderr, "%s: modifier 4 not supported\n", function_name);
	bad_modifier = TRUE;
    }
    if (event->xkey.state & Mod5Mask)
    {
	fprintf (stderr, "%s: modifier 5 not supported\n", function_name);
	bad_modifier = TRUE;
    }
    if (bad_modifier)
    {
	fputs ("  Make sure keys like CapsLock, NumLock, ScrollLock and AltLock are turned off\n", stderr);
	return;
    }
    /*  Clear the shift bits so that I can get shifted keysyms  */
    event->xkey.state &= ~shift_bits & ~ControlMask;
    len = XLookupString (&event->xkey, buffer,
			 STRING_LENGTH - sizeof mask - 1, &keysym, NULL);
    if (len < 1)
    {
	/*  Possibly some special key like an arrow key  */
	int hshift = 0;
	int vshift = 0;

	switch (keysym)
	{
	  case XK_Left:
	    if (cnv->canvas.processSpecialKeys) hshift = -1;
	    break;
	  case XK_Right:
	    if (cnv->canvas.processSpecialKeys) hshift = 1;
	    break;
	  case XK_Down:
	    if (cnv->canvas.processSpecialKeys) vshift = 1;
	    break;
	  case XK_Up:
	    if (cnv->canvas.processSpecialKeys) vshift = -1;
	    break;
	}
	if (!hshift && !vshift)
	{
	    char *string = XKeysymToString (keysym);

	    if (!string) return;
	    strcpy (buffer, string);
	    event_code = press ? K_CANVAS_EVENT_KEYSYM_PRESS :
		                 K_CANVAS_EVENT_KEYSYM_RELEASE;
	    len = strlen (buffer);
	    if (shift_bits) mask |= K_CANVAS_EVENT_SHIFTED_MASK;
	    m_copy (buffer + len + 1, &mask, sizeof mask);
	    inject_event (cnv, event->xkey.x, event->xkey.y, event_code, FALSE,
			  buffer);
	    return;
	}
	if (!press) return;
	if (mask == K_CANVAS_EVENT_CONTROL_KEY_MASK)
	{
	    hshift *= 10;
	    vshift *= 10;
	}
	kwin_warp_pointer (cnv->canvas.monoPixCanvas, hshift, vshift, TRUE,
			   FALSE);
	return;
    }
    /*  Put the shift bits back and lookup the string again  */
    event->xkey.state |= shift_bits;
    len = XLookupString (&event->xkey, buffer,
			 STRING_LENGTH - sizeof mask - 1, &keysym, NULL);
    if (mask == 0) event_code = press ? K_CANVAS_EVENT_PLAIN_KEY_PRESS :
	                                K_CANVAS_EVENT_PLAIN_KEY_RELEASE;
    else event_code = press ? K_CANVAS_EVENT_MODIFIED_KEY_PRESS :
	                      K_CANVAS_EVENT_MODIFIED_KEY_RELEASE;
    buffer[len] = '\0';
    m_copy (buffer + len + 1, &mask, sizeof mask);
    if ( (event->xkey.state & ShiftMask) || (event->xkey.state & LockMask) )
	st_upr (buffer);
    if (!cnv->canvas.processSpecialKeys)
    {
	inject_event (cnv, event->xkey.x, event->xkey.y, event_code, FALSE,
		      buffer);
	return;
    }
    if ( (strcmp (buffer, " ") == 0) && !event->xkey.state )
    {   /*  Toggle the pointer freeze state  */
	if (!press) return;
	if (cnv->canvas.frozen_x < 0)
	{   /*  Not frozen: freeze  */
	    cnv->canvas.frozen_x = event->xkey.x;
	    cnv->canvas.frozen_y = event->xkey.y;
	    cnv->canvas.remember_freeze_pos =
		(mask == K_CANVAS_EVENT_CONTROL_KEY_MASK) ? TRUE : FALSE;
	    _Canvas_draw_marker (cnv);
	}
	else
	{   /*  Not frozen: unfreeze  */
	    cnv->canvas.frozen_x = -1;
	    cnv->canvas.frozen_y = -1;
	    kwin_resize (cnv->canvas.monoPixCanvas, TRUE,
			 0, 0, cnv->core.width, cnv->core.height);
	}
	return;
    }
    if ( (strcmp (buffer, "r") == 0) && (mask == 0) )
    {   /*  The canvas needs to be refreshed  */
	if (!press) return;
	_Canvas_refresh (cnv, TRUE);
	return;
    }
    if ( (strcmp (buffer, "t") == 0) &&
	 (mask == K_CANVAS_EVENT_CONTROL_KEY_MASK) )
    {   /*  Toggle the timing state  */
	flag bool;

	if (!press) return;
	kwin_get_attributes (cnv->canvas.monoPixCanvas,
			     KWIN_ATT_TIME_REFRESH, &bool,
			     KWIN_ATT_END);
	bool = bool ? FALSE : TRUE;
	kwin_set_attributes (cnv->canvas.monoPixCanvas,
			     KWIN_ATT_TIME_REFRESH, bool,
			     KWIN_ATT_END);
	return;
    }
    if ( (strcmp (buffer, "g") == 0) &&
	 (mask == K_CANVAS_EVENT_CONTROL_KEY_MASK) )
    {
	if (!press) return;
	XBell (XtDisplay (cnv), 100);
	return;
    }
    inject_event (cnv, event->xkey.x, event->xkey.y, event_code, FALSE,buffer);
}   /*  End Function CanvasKeyEvent  */

#ifdef HAS_Xmbuf
static void create_stereo_window (Widget widget, unsigned int window_class,
				  Visual *visual, XtValueMask value_mask,
				  XSetWindowAttributes *attributes,
				  Multibuffer *left, Multibuffer *right)
/*  [PURPOSE] This routine will create a stereo window.
*/
{
    Window win, parentwin;

    parentwin = (widget->core.parent ? widget->core.parent->core.window :
		 widget->core.screen->root);
    if (widget->core.window == None)
    {
	if (widget->core.width == 0 || widget->core.height == 0)
	{
	    fprintf (stderr, "Widget %s has zero width and/or height\n",
		     (char *) &widget->core.name);
	}
	win = XmbufCreateStereoWindow (XtDisplay (widget), parentwin,
				       (int) widget->core.x,
				       (int) widget->core.y,
				       (unsigned int) widget->core.width,
				       (unsigned int) widget->core.height,
				       (unsigned) widget->core.border_width,
				       (int) widget->core.depth,
				       window_class, visual, value_mask,
				       attributes, left, right);
	if (win == (Window) None)
	{
	    fputs ("Error creating stereo window\n", stderr);
	    exit (RV_UNDEF_ERROR);
	}
	widget->core.window = win;
    }
}   /*  End Function create_stereo_window  */
#endif

static void set_visibility (CanvasWidget cnv, flag visible)
/*  [PURPOSE] This routine will set the visibility state of any pixel canvases.
    <cnv> The Canvas widget.
    <visible> If TRUE, the pixel canvases are set to be visible, else they are
    invisible.
    [RETURNS] Nothing.
*/
{
    if (cnv->canvas.monoPixCanvas)
    {
	kwin_set_attributes (cnv->canvas.monoPixCanvas,
			     KWIN_ATT_VISIBLE, visible,
			     KWIN_ATT_END);
    }
    if (cnv->canvas.leftPixCanvas)
    {
	kwin_set_attributes (cnv->canvas.leftPixCanvas,
			     KWIN_ATT_VISIBLE, visible,
			     KWIN_ATT_END);
    }
    if (cnv->canvas.rightPixCanvas)
    {
	kwin_set_attributes (cnv->canvas.rightPixCanvas,
			     KWIN_ATT_VISIBLE, visible,
			     KWIN_ATT_END);
    }
}   /*  End Function set_visibility  */

static void inject_event (CanvasWidget cnv, int x, int y, unsigned int k_event,
			  flag clip, CONST char *string)
/*  [PURPOSE] This routine will inject an event into an appropriate pixel
    canvas.
    <cnv> The Canvas widget.
    <x> The horizontal co-ordinate of the event.
    <y> The vertical co-ordinate of the event.
    <k_event> The Karma event code.
    <clip> If TRUE, the event will always be clipped.
    <string> The keystring. This is NULL if a key was not pressed.
    [RETURNS] Nothing.
*/
{
    flag consumed = FALSE;

    if ( (cnv->canvas.frozen_x > -1) && cnv->canvas.remember_freeze_pos )
    {
	x = cnv->canvas.frozen_x;
	y = cnv->canvas.frozen_y;
    }
    if (cnv->canvas.clipEvents) clip = TRUE;
    if (cnv->canvas.monoPixCanvas)
	consumed = kwin_queue_position_event (cnv->canvas.monoPixCanvas,
					      x, y, clip, k_event,
					      (void *) string);
    else if (cnv->canvas.leftPixCanvas)
	consumed = kwin_queue_position_event (cnv->canvas.leftPixCanvas,
					      x, y, clip, k_event,
					      (void *) string);
    if (consumed || cnv->canvas.silence_unconsumed_events) return;
    fprintf (stderr, "Karma event: %u not consumed\n", k_event);
}   /*  End Function inject_event  */

static void popup_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the popup callback.
*/
{
#ifdef DISABLED
    CanvasWidget top = (CanvasWidget) client_data;

    set_visibility (top, top->canvas.mapped);
#endif
}   /*  End Function popup_cbk   */

static void popdown_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the popdown callback.
*/
{
    CanvasWidget top = (CanvasWidget) client_data;

    set_visibility (top, FALSE);
}   /*  End Function popdown_cbk   */

static void _Canvas_ConvertCursor (Widget W)
{
    CanvasWidget w = (CanvasWidget) W;
    XrmValue from, to;
    Cursor cursor;
    static char function_name[] = "CanvasWidget::ConvertCursor";

    if (w->simple.cursor_name == NULL) return;
    if (w->canvas.verbose)
	fprintf (stderr, "%s: converting cursor: \"%s\"\n",
		 function_name, w->simple.cursor_name);
    from.addr = (XPointer) w->simple.cursor_name;
    from.size = strlen ( (char *) from.addr ) + 1;
    to.size = sizeof (Cursor);
    to.addr = (XPointer) &cursor;
    if ( XtConvertAndStore (W, XtRString, &from, XtRColorCursor, &to) )
    {
        if (cursor != None) w->simple.cursor = cursor;
    }
    else
    {
        XtAppErrorMsg (XtWidgetToApplicationContext (W),
		       "convertFailed", "ConvertCursor", "XawError",
		       "Simple: ConvertCursor failed.",
		       (String *) NULL, (Cardinal *) NULL);
    }
}   /*  End Function _Canvas_ConvertCursor  */

static void _Canvas_draw_marker (CanvasWidget w)
{
    unsigned int visual;
    unsigned long pixel_value;

    if ( (w->canvas.frozen_x < 0) || (w->canvas.frozen_y < 0) ) return;
    kwin_get_attributes (w->canvas.monoPixCanvas,
			 KWIN_ATT_VISUAL, &visual,
			 KWIN_ATT_END);
    switch (visual)
    {
      case KWIN_VISUAL_DIRECTCOLOUR:
      case KWIN_VISUAL_TRUECOLOUR:
	kwin_get_attributes (w->canvas.monoPixCanvas,
			     KWIN_ATT_PIX_GREEN_MASK, &pixel_value,
			     KWIN_ATT_END);
	break;
      default:
	kwin_get_colour (w->canvas.monoPixCanvas, "green", &pixel_value,
			 NULL, NULL, NULL);
	break;
    }
    kwin_draw_ellipse (w->canvas.monoPixCanvas,
		       w->canvas.frozen_x, w->canvas.frozen_y,
		       CURSOR_RADIUS, CURSOR_RADIUS, pixel_value);
}   /*  End Function _Canvas_draw_marker  */
