/*LINTLIBRARY*/
/*  Postscript.c

    This code provides a postscript control widget for Xt.

    PostScript code Copyright (C) 1994-2005  Richard Gooch

    Original   widget code Copyright (C) 1994    Patrick Jordan & Richard Gooch
    Subsequent widget code Copyright (C) 1994-2005  Richard Gooch
    Original widget code incorporated into Karma by permission.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Patrick Jordan may be reached by email at  patrick@ariel.com.au
    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for a postscript control
    widget for Xt.


    Written by      Patrick Jordan  26-JUN-1994: Copied PostScript code from
  print_xview.c  and hard_copy.c  in  source/modules/kview_2d (XView version of
  kview_2d  ).

    Updated by      Patrick Jordan  27-JUL-1994: Removed print_mono button
   as code not implemented yet.

    Updated by      Richard Gooch   23-OCT-1994: Changed some resource names.

    Updated by      Richard Gooch   7-DEC-1994: Stripped declaration of  errno
  and added #include <errno.h>

    Updated by      Richard Gooch   9-APR-1995: Initialised  arrayfile_name
  field to NULL to avoid freeing a non-existant block.

    Updated by      Richard Gooch   13-JUL-1995: Added resources to change
  orientation, page offset and size.

    Updated by      Richard Gooch   30-JUL-1995: Made use of XkwNvaluePtr
  resource for ValueWidgetClass.

    Updated by      Richard Gooch   23-SEP-1995: Coped with NULL name pointer
  to <XkwPostscriptRegisterImageAndName>

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   7-JUN-1996: Added dialog widget for output
  filename.

    Updated by      Richard Gooch   28-JUN-1996: Created XkwNautoIncrement
  resource.

    Updated by      Richard Gooch   1-JUL-1996: Switched to <psw_va_create> and
  <psw_close> routines.

    Updated by      Richard Gooch   30-OCT-1996: Added #include <sys/types.h>

    Updated by      Richard Gooch   20-DEC-1996: Added dialog for queue name.

    Updated by      Richard Gooch   16-JAN-1997: Added tray selection.

    Updated by      Richard Gooch   5-FEB-1997: Fixed bug where tray member was
  not initialised.

    Updated by      Richard Gooch   20-MAY-1997: Ensure square pixels on screen
  are square on paper.

    Updated by      Richard Gooch   4-JUN-1997: Support paper tray.

    Updated by      Richard Gooch   9-JUN-1997: Ensure square pixels on screen
  are square on paper for landscape orientation as well.

    Updated by      Richard Gooch   2-AUG-1997: Made use of new function
  <r_fork_and_execvp>.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <r> package.

    Updated by      Richard Gooch   23-OCT-1998: Added event
  "PostscriptWidget::WritePS".

    Updated by      Richard Gooch   6-AUG-1999: Added linewidth slider.

    Updated by      Richard Gooch   9-SEP-1999: Added "Lock Filename" toggle.

    Updated by      Richard Gooch   22-MAY-2000: Increased size limit to 1m.

    Updated by      Richard Gooch   9-JUN-2000: Removed unused variables.

    Updated by      Richard Gooch   13-MAR-2001: Allow negative offsets.

    Updated by      Richard Gooch   21-MAR-2001: Created Destroy method.

    Updated by      Richard Gooch   25-SEP-2001: Removed #include for var-args.

    Updated by      Richard Gooch   2-JAN-2004: Made use of new experimental
  function <kwin_ps_generate>.

    Updated by      Richard Gooch   11-JAN-2004: Added media size menu, made
  use of exernally defined orientation menu.

    Updated by      Richard Gooch   16-AUG-2004: Switched to SimpleTextEntry
  widget class.

    Updated by      Richard Gooch   12-OCT-2004: Switched to Form widget class
  for mail layout, so that user resizes can change text entry field sizes.

    Updated by      Richard Gooch   20-OCT-2004: Changed name of queueDialog
  widget.

    Last updated by Richard Gooch   26-APR-2005: Made use of caseInsensitive
  resource for size menu.


*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <sys/wait.h>
#include <signal.h>
#include <Xkw/PostscriptP.h>
#include <X11/Xos.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Box.h>
#include <karma.h>
#include <karma_xtmisc.h>
#include <karma_kwin.h>
#include <karma_psw.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_ch.h>
#include <karma_r.h>
#include <karma_m.h>
#include <karma_c.h>
#include <Xkw/Value.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/ExclusiveMenu.h>
#include <Xkw/SimpleSlider.h>
#include <Xkw/SimpleTextEntry.h>



/*  Methods  */
STATIC_FUNCTION (void _Postscript_Initialise, (Widget request, Widget new) );
STATIC_FUNCTION (void _Postscript_Destroy, (Widget W) );
STATIC_FUNCTION (void auto_increment_cbk, (Widget w, XtPointer client_data,
					   XtPointer call_data) );
STATIC_FUNCTION (flag _Postscript_ds_event_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );


#define TheOffset(field) XtOffset(PostscriptWidget, postscript.field)

static XtResource PostscriptResources[] = 
{
    {XkwNpageHorizontalOffset, XkwCPageHorizontalOffset, XtRInt, sizeof (int),
     TheOffset (hoffset), XtRImmediate, (XtPointer) 10},
    {XkwNpageVerticalOffset, XkwCPageVerticalOffset, XtRInt, sizeof (int),
     TheOffset (voffset), XtRImmediate, (XtPointer) 10},
    {XkwNpageHorizontalSize, XkwCPageHorizontalSize, XtRInt, sizeof (int),
     TheOffset (hsize), XtRImmediate, (XtPointer) 180},
    {XkwNpageVerticalSize, XkwCPageVerticalSize, XtRInt, sizeof (int),
     TheOffset (vsize), XtRImmediate, (XtPointer) 180},
    {XkwNautoIncrement, XkwCAutoIncrement, XtRBool, sizeof (Bool),
     TheOffset (autoIncrement), XtRImmediate, (XtPointer) FALSE},
    {XtNcallback, XtCCallback, XtRCallback, sizeof (caddr_t),
     TheOffset (callback), XtRCallback, (caddr_t) NULL},
};

#undef TheOffset

PostscriptClassRec postscriptClassRec = 
{
  {        /* CoreClassPart */
    (WidgetClass)&topLevelShellClassRec,  /* superclass */
    "Postscript",                     /* class_name */
    sizeof(PostscriptRec),            /* widget_size */
    NULL,                          /* class_initialise */
    NULL,                          /* class_part_initialise */
    FALSE,                         /* class_init */
    (XtInitProc)_Postscript_Initialise,        /* initialise */
    NULL,                          /* initialise_hook */
    XtInheritRealize,              /* realize */
    NULL,                          /* actions */
    0,                             /* num_actions */
    PostscriptResources,              /* resources */
    XtNumber(PostscriptResources),    /* num_resources */
    NULLQUARK,                     /* xrm_class */
    TRUE,                          /* compress_motion */
    TRUE,                          /* compress_exposure */
    TRUE,                          /* compress_enterleave */
    TRUE,                          /* visible_interest */
    _Postscript_Destroy,           /* destroy */
    NULL,                          /* resize */
    NULL,                          /* expose */
    NULL,                          /* set_values */
    NULL,                          /* set_values_hook */
    XtInheritSetValuesAlmost,      /* set_values_almost */
    NULL,                          /* get_values_hook */
    NULL,                          /* accept_focus */
    XtVersion,                     /* version */
    NULL,                          /* callback_private */
    NULL,                          /* tm_translations */
    NULL,
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
    /* geometry_manager	  */	XtInheritGeometryManager,
    /* change_managed	  */	XtInheritChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension	  */	NULL
  },{ /* Shell */
    /* extension	  */	NULL
  },{ /* WMShell */
    /* extension	  */	NULL
  },{ /* VendorShell */
    /* extension	  */	NULL
  },{ /* TopLevelShell */
    /* extension	  */	NULL
  },
  {  /* PostscriptClassPart */
    0 /* empty */
  }
};

WidgetClass postscriptWidgetClass = (WidgetClass) &postscriptClassRec;


#define NUM_SIZE_CHOICES 8
static char *size_choices[NUM_SIZE_CHOICES] =
{
    "A3", "A4", "A5", "B4", "B5", "letter", "A4/letter", "user defined"
};


/*----------------------------------------------------------------------*/
/* check that a widget is actually a PostscriptWidget                      */
/*----------------------------------------------------------------------*/

static int check_type (Widget w, char *function_name)
{
    int fl;
    /*static char func_name[] = "Postscript.check_type";*/

    fl = XtIsSubclass (w, postscriptWidgetClass);
    if (!fl)
    fprintf (stderr, "ERROR: Widget passed to %s is not a PostscriptWidget\n",
	     function_name);
    return fl;
}

/*----------------------------------------------------------------------*/
/*  This routine will print the image to the queue named by the queue   */
/*  dialog.                                                             */
/*  If the value of client_data is TRUE, colour PostScript is written.  */
/*  The case where client_data is FALSE is not implemented yet.         */
/*  It is setup as a callback for the "queue" buttons.                  */
/*  The routine returns nothing.                                        */
/*----------------------------------------------------------------------*/

static void send_to_queue (Widget w, XtPointer client_data,XtPointer call_data)
{
    flag colour = (uaddr) client_data;
    Bool fix_aspect;
    PostscriptWidget top = (PostscriptWidget)XtParent(XtParent(XtParent(w)));
    String queue;
    static char function_name[] = "send_to_queue";

    XtCallCallbacks ( (Widget) top, XtNcallback, NULL );
    FLAG_VERIFY (colour);
    XtVaGetValues (top->postscript.queue_txt, XtNstring, &queue, NULL);
    if ( (queue == NULL) || (*queue == '\0') )
	fprintf (stderr, "No printer queue specified\n");
    XtVaGetValues (top->postscript.aspect_tgl, XtNstate, &fix_aspect, NULL);
    kwin_ps_generate (top->postscript.pixcanvas, queue, TRUE,
		      top->postscript.hoffset, top->postscript.voffset,
		      top->postscript.hsize, top->postscript.vsize,
		      top->postscript.orientation, FALSE,
		      top->postscript.linewidth, top->postscript.tray,
		      fix_aspect, size_choices[top->postscript.media_size]);
}   /*  End Function send_to_queue  */

/*----------------------------------------------------------------------*/
/*  This routine will print the image to a file.                        */
/*  If the value of client_data is TRUE, encapsulated PS is written.    */
/*  It is setup as a callback for the "write" buttons.                  */
/*----------------------------------------------------------------------*/

static void write_to_file (Widget w, XtPointer client_data,XtPointer call_data)
{
    flag check = TRUE;
    flag eps = (uaddr) client_data;
    unsigned int sequence_number = 0;
    struct stat statbuf;
    char filename[STRING_LENGTH];
    char basefile[STRING_LENGTH];
    Bool fix_aspect;
    String dialog_name;
    PostscriptWidget top = (PostscriptWidget) XtParent(XtParent(XtParent(w)));
    static char function_name[] = "write_to_file";

    XtCallCallbacks ( (Widget) top, XtNcallback, NULL );
    FLAG_VERIFY (eps);
    XtVaGetValues (top->postscript.name_txt, XtNstring, &dialog_name, NULL);
    if ( (dialog_name == NULL) || (*dialog_name == '\n') )
	strcpy (basefile, "fred");
    else strcpy (basefile, dialog_name);
    while (check)
    {
	if ( top->postscript.autoIncrement && (sequence_number > 0) )
	    sprintf (filename, "%s.%u.%s",
		     basefile, sequence_number, eps ? "eps" : "ps");
	else sprintf (filename, "%s.%s", basefile, eps ? "eps" : "ps");
	if (top->postscript.autoIncrement)
	{
	    ++sequence_number;
	    if (stat (filename, &statbuf) != 0) check = FALSE;
	}
	else check = FALSE;
    }
    XtVaGetValues (top->postscript.aspect_tgl, XtNstate, &fix_aspect, NULL);
    kwin_ps_generate (top->postscript.pixcanvas, filename, FALSE,
		      top->postscript.hoffset, top->postscript.voffset,
		      top->postscript.hsize, top->postscript.vsize,
		      top->postscript.orientation, eps,
		      top->postscript.linewidth, top->postscript.tray,
		      fix_aspect, size_choices[top->postscript.media_size]);
}   /*  End Function write_to_file  */

static Widget create_value (Widget parent, Widget upper_widget, char *name,
			    int min, int *valuepointer)
{
    Widget w;

    w = XtVaCreateManagedWidget (name, valueWidgetClass, parent,
				 XtNfromVert, upper_widget,
				 XtNright, XtChainRight,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XkwNmodifier, 1,
				 XkwNminimum, min,
				 XkwNmaximum, 1000,
				 XtNvalue, *valuepointer,
				 XtNborderWidth, 0,
				 XtNlabel, name,
				 XkwNvaluePtr, valuepointer,
				 NULL);
    return w;
}   /*  End Function create_value  */

static void create_text_btn (Widget *btn, char *name,char *cbk_val,
			     XtCallbackProc proc, Widget parent)
{
    *btn = XtVaCreateManagedWidget ("button", commandWidgetClass, parent,
				    XtNlabel, name,
				    NULL);
    XtAddCallback (*btn, XtNcallback, proc, cbk_val);
}  /*  End Function create_text_btn  */

static void _Postscript_Initialise (Widget Request, Widget New)
{
    /*PostscriptWidget request = (PostscriptWidget) Request;*/
    PostscriptWidget new = (PostscriptWidget) New;
    Widget form, box, upper_widget;
    Widget psbtn, epsbtn, qcolourbtn, w, closebtn;
    CONST char *queue, *default_media_size;
    static unsigned int instance_number = 0;

    new->postscript.instance_number = instance_number++;
    default_media_size = r_getenv ("PAPER_SIZE");
    if ( !psw_get_media_size (default_media_size, NULL, NULL) )
	default_media_size = NULL;
    if (!default_media_size) default_media_size = "A4";
    form = XtVaCreateManagedWidget ("form", formWidgetClass, New,
				    XtNborderWidth, 0,
				    NULL);
    box = XtVaCreateManagedWidget ("box", boxWidgetClass, form,
				   XtNleft, XtChainLeft,
				   XtNright, XtChainRight,
				   XtNtop, XtChainTop,
				   XtNbottom, XtChainTop,
				   XtNhorizDistance, 0,
				   XtNvertDistance, 0,
				   XtNborderWidth, 0,
				   XtNorientation, XtorientHorizontal,
				   NULL);
    upper_widget = box;
    closebtn = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, box,
					XtNlabel, "close",
					NULL);
    XtAddCallback (closebtn, XtNcallback, xtmisc_popdown_cbk, New);
    create_text_btn (&psbtn, "save .ps", (char *) FALSE, write_to_file, box);
    create_text_btn (&epsbtn, "save .eps", (char *) TRUE, write_to_file, box);
    create_text_btn (&qcolourbtn, "print", (char *) TRUE, send_to_queue, box);
    w = XtVaCreateManagedWidget ("menuButton", exclusiveMenuWidgetClass, form,
				 XtNleft, XtChainLeft,
				 XtNright, XtChainLeft,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, upper_widget,
				 XtNvertDistance, 0,
				 XtNmenuName, "trayMenu",
				 XkwNchoiceName, "Tray",
				 XkwNnumItems, PSW_NUM_TRAY_CHOICES,
				 XkwNitemStrings, psw_tray_choices,
				 XkwNvaluePtr, &new->postscript.tray,
				 NULL);
    upper_widget = w;
    w = XtVaCreateManagedWidget ("menuButton", exclusiveMenuWidgetClass, form,
				 XtNleft, XtChainLeft,
				 XtNright, XtChainLeft,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, upper_widget,
				 XtNmenuName, "sizeMenu",
				 XkwNchoiceName, "Size",
				 XkwNnumItems, NUM_SIZE_CHOICES,
				 XkwNitemStrings, size_choices,
				 XkwNsetChoice, default_media_size,
				 XkwNvaluePtr, &new->postscript.media_size,
				 XkwNcaseInsensitive, True,
				 NULL);
    w = XtVaCreateManagedWidget ("aspectToggle", ktoggleWidgetClass, form,
				 XtNleft, XtChainLeft,
				 XtNright, XtChainLeft,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNlabel, "Keep Aspect",
				 XkwNcrosses, FALSE,
				 XtNstate, True,
				 NULL);
    upper_widget = w;
    new->postscript.aspect_tgl = w;
    w = XtVaCreateManagedWidget ("menuButton", exclusiveMenuWidgetClass, form,
				 XtNleft, XtChainLeft,
				 XtNright, XtChainLeft,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, upper_widget,
				 XtNmenuName, "orientationMenu",
				 XkwNchoiceName, "Orientation",
				 XkwNnumItems, KWIN_PS_NUM_ORIENTATION_CHOICES,
				 XkwNitemStrings, kwin_ps_orientation_choices,
				 XkwNvaluePtr, &new->postscript.orientation,
				 NULL);
    w = create_value (form, w, "hoffset (mm)", -100, &new->postscript.hoffset);
    w = create_value (form, w, "voffset (mm)", -100, &new->postscript.voffset);
    w = create_value (form, w, "hsize (mm)", 0, &new->postscript.hsize);
    w = create_value (form, w, "vsize (mm)", 0, &new->postscript.vsize);
    upper_widget = w;
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				 XtNleft, XtChainLeft,
				 XtNright, XtChainRight,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, upper_widget,
				 XtNlabel, "Thin Line Width (mm)",
				 XtNwidth, 350,
				 XkwNminimum, 0,
				 XkwNmaximum, 100,
				 XkwNmodifier, 1,
				 XtNvalue, 10,
				 XkwNscaledValuePtr,&new->postscript.linewidth,
				 XkwNscaledFormat, "%.2f",
				 NULL);
    upper_widget = w;
    XkwSimpleSliderSetScale (w, 1e-2, 0.0, FALSE, TRUE);
    w = XtVaCreateManagedWidget ("incToggle", ktoggleWidgetClass, form,
				 XtNleft, XtChainLeft,
				 XtNright, XtChainLeft,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, upper_widget,
				 XtNlabel, "Auto Increment",
				 XtNstate, new->postscript.autoIncrement,
				 XkwNcrosses, FALSE,
				 NULL);
    XtAddCallback (w, XtNcallback, auto_increment_cbk, New);
    w = XtVaCreateManagedWidget ("lockToggle", ktoggleWidgetClass, form,
				 XtNleft, XtChainLeft,
				 XtNright, XtChainLeft,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNlabel, "Lock Filename",
				 XkwNcrosses, FALSE,
				 NULL);
    upper_widget = w;
    new->postscript.lock_tgl = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNleft, XtChainLeft,
				 XtNright, XtChainLeft,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, upper_widget,
				 XtNlabel, "Output file:  ",
				 XtNborderWidth, 0,
				 NULL);
    w = XtVaCreateManagedWidget ("nameDialog", simpleTextEntryWidgetClass,form,
				 XtNleft, XtChainLeft,
				 XtNright, XtChainRight,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNstring, "fred",
				 XtNwidth, 205,
				 XtNeditType, XawtextEdit,
				 NULL);
    upper_widget = w;
    new->postscript.name_txt = w;
    if ( ( queue = r_getenv ("PRINTER") ) == NULL ) queue = "lp";
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNleft, XtChainLeft,
				 XtNright, XtChainLeft,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, upper_widget,
				 XtNlabel, "Printer queue:",
				 XtNborderWidth, 0,
				 NULL);
    w = XtVaCreateManagedWidget ("queueDialog",simpleTextEntryWidgetClass,form,
				 XtNleft, XtChainLeft,
				 XtNright, XtChainRight,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNstring, queue,
				 XtNwidth, 205,
				 XtNeditType, XawtextEdit,
				 NULL);
    new->postscript.queue_txt = w;
    new->postscript.ds_event_callback =
	ds_event_register_func (_Postscript_ds_event_func, New);
}  /*  End Function Initialise  */

static void _Postscript_Destroy (Widget W)
{
    PostscriptWidget w = (PostscriptWidget) W;

    c_unregister_callback (w->postscript.ds_event_callback);
}   /*  End Function Destroy  */

static void auto_increment_cbk (Widget w, XtPointer client_data,
				XtPointer call_data)
{
    flag bool = (uaddr) call_data;
    PostscriptWidget top = (PostscriptWidget) client_data;

    top->postscript.autoIncrement = bool;
}   /*  End Function auto_increment_cbk  */

static flag _Postscript_ds_event_func (void *object, multi_array *multi_desc,
				       CONST char *domain, CONST char *name)
/*  [SUMMARY] Process an event.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    unsigned int instance_number;
    Bool fix_aspect;
    PostscriptWidget top = (PostscriptWidget) object;
    packet_desc *pack_desc;
    char *packet, *filename;
    /*static char function_name[] = "Postscript::ds_event_func";*/

    if (strcmp (name, "PostscriptWidget::WritePS") != 0) return (FALSE);
    if (multi_desc->num_arrays != 1) return (FALSE);
    pack_desc = multi_desc->headers[0];
    packet = multi_desc->data[0];
    if (pack_desc->num_elements != 2) return (FALSE);
    if (pack_desc->element_types[0] != K_UINT) return (FALSE);
    if (pack_desc->element_types[1] != K_VSTRING) return (FALSE);
    if (strcmp (pack_desc->element_desc[0], "instance_number") != 0)
	return (FALSE);
    if (strcmp (pack_desc->element_desc[1], "filename") != 0) return (FALSE);
    m_copy (&instance_number, packet, sizeof instance_number);
    if (instance_number != top->postscript.instance_number) return (FALSE);
    m_copy (&filename, packet + ds_get_element_offset (pack_desc, 1),
	    sizeof filename);
    XtVaGetValues (top->postscript.aspect_tgl, XtNstate, &fix_aspect, NULL);
    kwin_ps_generate (top->postscript.pixcanvas, filename, FALSE,
		      top->postscript.hoffset, top->postscript.voffset,
		      top->postscript.hsize, top->postscript.vsize,
		      top->postscript.orientation, FALSE,
		      top->postscript.linewidth, top->postscript.tray,
		      fix_aspect, size_choices[top->postscript.media_size]);
    return (TRUE);
}   /*  End Function _Postscript_ds_event_func  */


/*  Public functions follow  */

void XkwPostscriptRegisterImageAndName (Widget w, KPixCanvas image, char *name)
{
    PostscriptWidget top = (PostscriptWidget) w;
    Bool locked;
    static char fn_name[] = "XkwPostscriptRegisterImageAndName";

    check_type (w, fn_name);
    top->postscript.pixcanvas = image;
    XtVaGetValues (top->postscript.lock_tgl, XtNstate, &locked, NULL);
    if (locked) return;
    if ( (name == NULL) || (*name == '\0') ) return;
    XtVaSetValues (top->postscript.name_txt, XtNstring, name, NULL);
}   /*  End Function XkwPostscriptRegisterImageAndName  */
