/*LINTLIBRARY*/
/*  ViewDatasets.c

    This code provides a multiple dataset view control widget for Xt.

    Copyright (C) 2002-2004  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for a multiple dataset view control
    widget for Xt.


    Written by      Richard Gooch   5-DEC-2002: Copied parts from
  View2Datasets.c.

    Updated by      Richard Gooch   11-DEC-2002: First cut of code. All
  functionality of <kview> now available in <kvis>.

    Updated by      Richard Gooch   13-DEC-2002: Fixed bug with scatter plot
  computation and tracking: restriction information was not being extracted and
  passed along. Old bug, present in <kview> too.

    Updated by      Richard Gooch   21-DEC-2002: Partial support for different
  grids for image and profile arrays.

    Updated by      Richard Gooch   23-DEC-2002: Added "Square" button for
  scatterplot.

    Updated by      Richard Gooch   25-DEC-2002: Switched to KViewableGraph for
  radial profile points display.

    Updated by      Richard Gooch   27-DEC-2002: Fixed bug in line profile
  display: frame number was converted to world co-ordinate.

    Updated by      Richard Gooch   20-JAN-2003: Removed remaining profile
  implementation code as <dispdata> package now implements this.

    Updated by      Richard Gooch   25-JAN-2003: Fixed bug in <<Initialise>>
  function: now stop searching for ancestor widgets once desired one is found.

    Updated by      Richard Gooch   28-JAN-2003: Added cursor broadcast.

    Updated by      Richard Gooch   4-FEB-2003: Removed create AtInit resource
  because profile widget must be created after ImageDisplay, magnifier and
  panner widgets are all realised, and there is no way of finding that out.

    Updated by      Richard Gooch   17-FEB-2003: Removed auto title code, since
  <dispdata> package does it now.

    Updated by      Richard Gooch   18-FEB-2003: Fixed bug in function
  <trace_position_func>: check for NULL image_xlabel (data could be unloaded).

    Updated by      Richard Gooch   28-FEB-2003: Took account of API change for
  Hershey font rendering.

    Updated by      Richard Gooch   10-APR-2003: Fixed bug in function
  <scatter_cursor_func>: main canvas with different grid than scatter plot X
  image was not handled properly. Added yellow circle to marker.

    Updated by      Richard Gooch   30-APR-2003: Removed code to draw frame
  marker as this is now handled in the <dispdata> package.

    Updated by      Richard Gooch   7-JUN-2003: Made use of new experimental
  function <viewimg_register_destroy_func> for scatter plot images so that
  events processed after either of the images have been unloaded is safe.

    Updated by      Richard Gooch   14-JUN-2003: Created <renumber_func>.

    Updated by      Richard Gooch   6-AUG-2003: Made use of new experimental
  function <c_unregister_callback2>.

    Updated by      Richard Gooch   27-AUG-2003: Support swapped axis names for
  cursor tracking and improved axis name discrimination.

    Updated by      Richard Gooch   15-SEP-2003: Fixed bug in function
  <draw_marker_in_scatter_plot>: crash when other set has no KwcsAstro object.

    Updated by      Richard Gooch   17-SEP-2003: Avoid setting extra label in
  <_ViewDatasets_viewimg_position_func> if values hasn't changed so that
  that label isn't refreshed unless needed (label widget is too dumb).

    Updated by      Richard Gooch   31-DEC-2003: Added azimuthally averaged
  profile support.

    Updated by      Richard Gooch   18-MAR-2004: Enabled further processing of
  keysym events in <_ViewDatasets_viewimg_position_func>.

    Updated by      Richard Gooch   6-MAY-2004: Determined Karma colourmap
  for trace popup widget. Needed for pretty Renzogram display.

    Updated by      Richard Gooch   18-JUN-2004: Added ClassInitialise method
  and KCallbackList for listen event co-ordinate conversion.

    Updated by      Richard Gooch   13-JUL-2004: Allow user to change initial
  slice direction.

    Last updated by Richard Gooch   15-AUG-2004: Blank track labels when
  pointer leaves window.


*/
#include <unistd.h>
#include <stdio.h>
#include <math.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Dialog.h>
#include <karma.h>
#define NEW_WIN_SCALE
#include <k_event_codes.h>
#include <karma_dispdata.h>
#include <karma_scatplt.h>
#include <karma_viewimg.h>
#include <karma_contour.h>
#include <karma_xtmisc.h>
#include <karma_event.h>
#include <karma_kcmap.h>
#include <karma_wcs.h>
#include <karma_ext.h>
#include <karma_ex.h>
#include <karma_ds.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_c.h>
/*#define USE_TRANSIENT*/
#include <Xkw/ViewDatasetsP.h>
#include <Xkw/ImageDisplay.h>
#include <Xkw/ExclusiveMenu.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/TracePopup.h>
#include <Xkw/SimpleSlider.h>
#include <Xkw/ChoiceMenu.h>
#include <Xkw/SlaveImageDisplayPopup.h>
#include <Xkw/IntensityPolicy.h>
#include <Xkw/SimpleTextEntry.h>
#include <Xkw/ExportMenu.h>
#include <Xkw/Canvas.h>
#include <Xkw/OverlayMenu.h>
#include <Xkw/Postscript.h>
#include <Xkw/DataBrowser.h>


#define ELLIPSE_RADIUS     5


struct cursor_data
{
    double wx;
    double wy;
    double wz;
    ViewDatasetsWidget source;
    CONST char *xname;
    CONST char *yname;
    CONST char *zname;
    KwcsAstro ap;
};


/*  Private functions  */
STATIC_FUNCTION (void _ViewDatasets_ClassInitialise, () );
STATIC_FUNCTION (void _ViewDatasets_Initialise,
		 (Widget request, Widget new) );
STATIC_FUNCTION (void _ViewDatasets_Destroy, (Widget W) );
STATIC_FUNCTION (void _ViewDatasets_create_scatter_plot_shell,
		 (ViewDatasetsWidget w) );
STATIC_FUNCTION (void slice_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void recompute_profile_dir, (ViewDatasetsWidget w) );
STATIC_FUNCTION (void _ViewDatasets_show_browsers_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void profile_mode_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void profile_dir_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void new_frame_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _ViewDatasets_link_iscale_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _ViewDatasets_show_marker_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _ViewDatasets_title_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _ViewDatasets_text_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _ViewDatasets_all_realised_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _ViewDatasets_renumber_func, (ViewDatasetsWidget w) );
STATIC_FUNCTION (void _ViewDatasets_new_editor_context_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _ViewDatasets_refresh_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _ViewDatasets_canvas_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize, void **info,
		  PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (flag _ViewDatasets_viewimg_position_func,
		 (ViewableImage vimage, double x, double y,
		  void *value, unsigned int event_code,
		  void *e_info, void **f_info,
		  double x_lin, double y_lin,
		  unsigned int value_type,
		  double x_im, double y_im, int x_pix, int y_pix) );
STATIC_FUNCTION (void _ViewDatasets_draw_beam,
		 (KWorldCanvas canvas, ViewDatasetsWidget w,
		  KDisplayDataObject dataobject, flag name) );
STATIC_FUNCTION (flag _ViewDatasets_scatplot,
		 (ViewDatasetsWidget top, double lx[2], double ly[2]) );
STATIC_FUNCTION (flag _ViewDatasets_left_geom_func,
		 (KPixCanvasEditorContext context, KPixCanvas canvas,
		  unsigned int object_type, unsigned int num_coords,
		  CONST int *x, CONST int *y,
		  unsigned int raw_event_code, void *raw_e_info,
		  void **f_info) );
STATIC_FUNCTION (void create_tracewinpopup, (ViewDatasetsWidget top) );
STATIC_FUNCTION (flag _ViewDatasets_trace_position_func,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code, void *e_info,
		  void **f_info, double x_lin, double y_lin,
		  int x_pix, int y_pix) );
STATIC_FUNCTION (void _ViewDatasets_draw_marker,
		 (KWorldCanvas canvas, KPixCanvasRefreshArea *area,
		  double x, double y) );
STATIC_FUNCTION (void _ViewDatasets_save_ppm_movie_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (flag _ViewDatasets_scatter_position_func,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code, void *e_info,
		  void **f_info, double x_lin, double y_lin,
		  int x_pix, int y_pix) );
STATIC_FUNCTION (flag _ViewDatasets_scatter_cursor_func,
		 (KScatterPlot plot, double x, double y,
		  uaddr index, void **info) );
STATIC_FUNCTION (flag _ViewDatasets_draw_marker_in_scatter_plot,
		 (ViewDatasetsWidget top, ViewableImage vimage_a,
		  double x, double y,
		  double value_a) );
STATIC_FUNCTION (void _ViewDatasets_square_scatter_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _ViewDatasets_listen_func,
		 (ViewDatasetsWidget w, void *client1_data,
		  struct cursor_data *data, void *client2_data) );


#define offset(field) XtOffsetOf(ViewDatasetsRec, viewDatasets.field)

static XtResource ViewDatasetsResources[] = 
{
    {XkwNcursorCallback, XtCCallback, XtRCallback, sizeof (XtPointer),
     offset (cursorCallback), XtRCallback, (XtPointer) NULL},
    {XkwNverbose, XkwCVerbose, XtRBool, sizeof (Bool),
     offset (verbose), XtRImmediate, False},
    {XkwNwmtitle, XkwCWMtitle, XtRString, sizeof (String),
     offset (wmtitle), XtRImmediate, NULL},
};

#undef offset

ViewDatasetsClassRec viewDatasetsClassRec = 
{
  {
      /* CoreClassPart */
#ifdef USE_TRANSIENT
    (WidgetClass) &transientShellClassRec,       /* superclass */
#else
    (WidgetClass) &topLevelShellClassRec,       /* superclass */
#endif
    "ViewDatasets",                            /* class_name */
    sizeof (ViewDatasetsRec),                  /* widget_size */
    _ViewDatasets_ClassInitialise,             /* class_initialise */
    NULL,                                       /* class_part_initialise */
    FALSE,                                      /* class_init */
    (XtInitProc) _ViewDatasets_Initialise,     /* initialise */
    NULL,                                       /* initialise_hook */
    XtInheritRealize,                           /* realise */
    NULL,                                       /* actions */
    0,                                          /* num_actions */
    ViewDatasetsResources,                     /* resources */
    XtNumber (ViewDatasetsResources),          /* num_resources */
    NULLQUARK,                                  /* xrm_class */
    TRUE,                                       /* compress_motion */
    TRUE,                                       /* compress_exposure */
    TRUE,                                       /* compress_enterleave */
    TRUE,                                       /* visible_interest */
    _ViewDatasets_Destroy,                      /* destroy */
    XtInheritResize,                            /* resize */
    NULL,                                       /* expose */
    NULL,                                       /* set_values */
    NULL,                                       /* set_values_hook */
    XtInheritSetValuesAlmost,                   /* set_values_almost */
    NULL,                                       /* get_values_hook */
    NULL,                                       /* accept_focus */
    XtVersion,                                  /* version */
    NULL,                                       /* callback_private */
    NULL,                                       /* tm_translations */
    NULL,
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
    /* geometry_manager	  */	XtInheritGeometryManager,
    /* change_managed	  */	XtInheritChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension	  */	NULL
  },
  { /* Shell */
    /* extension	  */	NULL
  },
  { /* WMShell */
    /* extension	  */	NULL
  },
  { /* VendorShell */
    /* extension	  */	NULL
  },
#ifdef USE_TRANSIENT
  { /* TransientShell */
    /* extension	  */	NULL
  },
#else
  { /* TopLevelShell */
    /* extension	  */	NULL
  },
#endif
  {  /* ViewDatasetsClassPart */
    0 /* empty */
  }
};

WidgetClass viewDatasetsWidgetClass = (WidgetClass) &viewDatasetsClassRec;

static KPixCanvasBinding main_bindings[] =
{
  {"mouse, middle", "zoom in"},
  {"mouse, right",  "zoom out"},
  {"b",             "switch to next blinkstate"},
  {"<",             "switch to next blinkstate"},
  {">",             "switch to previous blinkstate"},
  {"B",             "popup active blinkstate browser"},
  {"b, control",    "popup all blinkstate browsers for this window"},
  {"c",             "compute and display scatterplot"},
  {"PgUp",          "jump to next frame (control: 10 frames)"},
  {"PgDn",          "jump to previous frame (control: 10 frames)"},
  {"Home",          "jump to first (0th) frame"},
  {"End",           "jump to last frame"},
  {NULL,           NULL}
};

static KCallbackList broadcast_list = NULL;
static KCallbackList coord_convert_list = NULL;  /*  For <listen_func>  */


static void _ViewDatasets_ClassInitialise ()
{
    ext_call (viewDatasetsClassRec.core_class.class_name, "ClassInitialise",
	      0, NULL, FALSE,
	      &coord_convert_list);
}   /*  End Function ClassInitialise  */

static void _ViewDatasets_Initialise (Widget Request, Widget New)
{
    /*ViewDatasetsWidget request = (ViewDatasetsWidget) Request;*/
    ViewDatasetsWidget new = (ViewDatasetsWidget) New;
    Widget form, parent, w, upper_widget, movie_btn, zoom_policy;
    static char function_name[] = "ViewDatasetsWidget::Initialise";

    dispdata_init ();
    new->viewDatasets.profile_dir = KDISPLAYDATA_PROFILE_DIR_Z;
    new->viewDatasets.profile_mode = KDISPLAYDATA_PROFILE_MODE_NONE;
    new->viewDatasets.trace_winpopup = NULL;
    new->viewDatasets.trace_canvas = NULL;
    new->viewDatasets.main_marker_area.startx = -100;
    new->viewDatasets.scatter_plot_shell = NULL;
    new->viewDatasets.scatter_plot_canvas = NULL;
    new->viewDatasets.scatter_plot = NULL;
    new->viewDatasets.scatter_plot_limits.hard.left_x = TOOBIG;
    new->viewDatasets.scatter_plot_limits.hard.right_x = TOOBIG;
    new->viewDatasets.scatter_plot_limits.hard.bottom_y = TOOBIG;
    new->viewDatasets.scatter_plot_limits.hard.top_y = TOOBIG;
    m_clear (&new->viewDatasets.scatter_plot_x_handle,
	     sizeof new->viewDatasets.scatter_plot_x_handle);
    m_clear (&new->viewDatasets.scatter_plot_y_handle,
	     sizeof new->viewDatasets.scatter_plot_y_handle);
    new->viewDatasets.scatter_plot_marker_area.startx = -100;
    new->viewDatasets.scale = 1.0;
    new->viewDatasets.offset = 0.0;
    m_clear (&new->viewDatasets.cursor_tracking,
	     sizeof new->viewDatasets.cursor_tracking);
    new->viewDatasets.cursor_tracking.listen_callback =
	c_register_callback (&broadcast_list,
			     ( flag (*) () ) _ViewDatasets_listen_func, new,
			     NULL, FALSE, NULL, FALSE, FALSE);
    /*  Search for ImageDisplay widget somewhere in my ancestry  */
    new->viewDatasets.image_display = NULL;
    for ( parent = XtParent (New); parent; parent = XtParent (parent) )
    {
	if ( XtIsImageDisplay (parent) )
	{
	    new->viewDatasets.image_display = parent;
	    break;
	}
    }
    if (new->viewDatasets.image_display == NULL)
    {
	fprintf (stderr, "Could not find an ImageDisplay ancestor\n");
	a_prog_bug (function_name);
    }
    /*  Get the ZoomPolicy widget  */
    if ( ( zoom_policy = XtNameToWidget (new->viewDatasets.image_display,
					 "zoomPolicyPopup") ) == NULL )
    {
	fprintf (stderr, "Could not find ZoomPolicy popup\n");
	a_prog_bug (function_name);
    }
    if ( ( w = XtNameToWidget (zoom_policy,"form.middleZoomToggle") ) == NULL )
    {
	fprintf (stderr, "Could not find middleZoomToggle\n");
	a_prog_bug (function_name);
    }
    new->viewDatasets.middle_zoom_tgl = w;
    if ( ( w = XtNameToWidget (zoom_policy, "form.rightZoomToggle") ) == NULL )
    {
	fprintf (stderr, "Could not find rightZoomToggle\n");
	a_prog_bug (function_name);
    }
    new->viewDatasets.right_zoom_tgl = w;
    XtAddCallback (new->viewDatasets.image_display,
		   XkwNafterMainRealisedCallback,
		   _ViewDatasets_all_realised_cbk, new);
    XtAddCallback (new->viewDatasets.image_display,
		   XkwNnewEditorContextCallback,
		   _ViewDatasets_new_editor_context_cbk, new);
    form = XtVaCreateManagedWidget ("form", formWidgetClass, New,
				    XtNborderWidth, 0,
				    NULL);
    w = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				 XtNlabel, "Close",
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popdown_cbk, New);
    upper_widget = w;
    w = XtVaCreateManagedWidget ("browsersButton", commandWidgetClass, form,
				 XtNlabel, "Browsers",
				 XtNfromHoriz, w,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, _ViewDatasets_show_browsers_cbk,
		   new->viewDatasets.image_display);
    w = XtVaCreateManagedWidget ("movieButton", commandWidgetClass, form,
				 XtNlabel, "Movie",
				 XtNfromHoriz, w,
				 XtNheight, 20,
				 NULL);
    movie_btn = w;
    w = XtVaCreateManagedWidget ("profileModeMenu", exclusiveMenuWidgetClass,
				 form,
				 XtNmenuName, "profileModeMenu",
				 XkwNchoiceName, "Profile Mode",
				 XtNfromHoriz, w,
				 XkwNnumItems,
				     KDISPLAYDATA_NUM_PROFILE_MODE_CHOICES,
				 XkwNitemStrings,dispdata_profile_mode_choices,
				 NULL);
    XtAddCallback (w, XkwNselectCallback, profile_mode_cbk, New);
    w = XtVaCreateManagedWidget ("sliceMenu", exclusiveMenuWidgetClass, form,
				 XtNmenuName, "sliceMenu",
				 XkwNchoiceName, "Slice Direction",
				 XtNfromVert, upper_widget,
				 XkwNnumItems,
				   KDISPLAYDATA_NUM_SLICE_DIRECTIONS,
				 XkwNitemStrings, dispdata_slice_choices,
				 XkwNvaluePtr, &new->viewDatasets.slice,
				 NULL);
    XtVaSetValues (w, XkwNvaluePtr, NULL, NULL);
    XtAddCallback (w, XkwNselectCallback, slice_cbk, New);
    w = XtVaCreateManagedWidget ("profileDirMenu", exclusiveMenuWidgetClass,
				 form,
				 XtNmenuName, "profileDirMenu",
				 XkwNchoiceName, "Profile Axis",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XkwNnumItems,
				     KDISPLAYDATA_NUM_PROFILE_DIR_CHOICES,
				 XkwNitemStrings, dispdata_profile_dir_choices,
				 NULL);
    upper_widget = w;
    new->viewDatasets.profile_dir_menu = w;
    XtAddCallback (w, XkwNselectCallback, profile_dir_cbk, New);
    w = XtVaCreatePopupShell ("animatepopup", animateControlWidgetClass, New,
			      NULL);
    new->viewDatasets.animatepopup = w;
    XtAddCallback (w, XkwNnewFrameCallback, new_frame_cbk, New);
    XtAddCallback (movie_btn, XtNcallback, xtmisc_popup_cbk, w);
    w = XtNameToWidget (w, "form.antiFlickerToggle");
    XtAddCallback (w, XtNcallback, XkwImageDisplaySetSmoothCallback,
		   new->viewDatasets.image_display);
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, form,
				 XtNlabel, "Freeze Displayed Intensity Range",
				 XkwNcrosses, False,
				 XtNfromVert, upper_widget,
				 NULL);
    XtAddCallback (w, XtNcallback, _ViewDatasets_link_iscale_cbk, New);
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, form,
				 XtNlabel, "Track Cursor",
				 XkwNcrosses, False,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_toggle_cbk,
		   &new->viewDatasets.cursor_tracking.listen);
    upper_widget = w;
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, form,
				 XtNlabel, "Show Marker in Line Profile",
				 XkwNcrosses, False,
				 XtNfromVert, upper_widget,
				 NULL);
    XtAddCallback (w, XtNcallback, _ViewDatasets_show_marker_cbk, New);
    new->viewDatasets.show_profile_marker_tgl = w;
    upper_widget = w;
    w = XtVaCreateManagedWidget ("titleName", labelWidgetClass, form,
				 XtNlabel, "Title String:",
				 XtNborderWidth, 0,
				 XtNfromVert, upper_widget,
				 NULL);
    w = XtVaCreateManagedWidget ("titleToggle", ktoggleWidgetClass, form,
				 XtNlabel, "Auto Title",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNhorizDistance, 200,
				 XtNright, XtChainRight,
				 XkwNcrosses, False,
				 XtNstate, TRUE,
				 NULL);
    XtAddCallback (w, XtNcallback, _ViewDatasets_title_cbk, New);
    upper_widget = w;
    new->viewDatasets.auto_title_tgl = w;
    w = XtVaCreateManagedWidget ("title", simpleTextEntryWidgetClass, form,
				 XtNdefaultDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNwidth, 440,
				 XtNeditType, XawtextEdit,
				 NULL);
    new->viewDatasets.title_txt = w;
    upper_widget = w;
    XtAddCallback (w, XkwNnewCharacterCallback, _ViewDatasets_text_cbk, New);
    w = XtVaCreateManagedWidget ("beamToggle", ktoggleWidgetClass, form,
				 XtNlabel, "Show Beam",
				 XtNfromVert, upper_widget,
				 XkwNcrosses, False,
				 NULL);
    XtAddCallback (w, XtNcallback, _ViewDatasets_refresh_cbk, New);
    new->viewDatasets.show_beam_tgl = w;
    w = XtVaCreateManagedWidget ("beamNameToggle", ktoggleWidgetClass, form,
				 XtNlabel, "Show Beam Name",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XkwNcrosses, False,
				 XtNstate, TRUE,
				 NULL);
    XtAddCallback (w, XtNcallback, _ViewDatasets_refresh_cbk, New);
    upper_widget = w;
    new->viewDatasets.show_beam_name_tgl = w;
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				 XtNfromVert, upper_widget,
				 XtNlabel, "Beam Xpos",
				 XkwNminimum, 0,
				 XkwNmaximum, 100,
				 XkwNmodifier, 1,
				 XtNvalue, 50,
				 XkwNscaledValuePtr,
				 &new->viewDatasets.beam_xpos,
				 XkwNshowValue, FALSE,
				 NULL);
    XkwSimpleSliderSetScale (w, 1e-2, 0.0, FALSE, FALSE);
    XtAddCallback (w, XkwNvalueChangeCallback, _ViewDatasets_refresh_cbk, New);
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNlabel, "Beam Ypos",
				 XkwNminimum, 0,
				 XkwNmaximum, 100,
				 XkwNmodifier, 1,
				 XtNvalue, 50,
				 XkwNscaledValuePtr,
				 &new->viewDatasets.beam_ypos,
				 XkwNshowValue, FALSE,
				 NULL);
    XkwSimpleSliderSetScale (w, 1e-2, 0.0, FALSE, FALSE);
    XtAddCallback (w, XkwNvalueChangeCallback, _ViewDatasets_refresh_cbk, New);
    upper_widget = w;
    /*  Create other popups  */
    if ( ( w = XtNameToWidget (new->viewDatasets.image_display,
			       "exportMenu") ) == NULL )
    {
	fprintf (stderr, "No \"exportMenu\" widget\n");
	a_prog_bug (function_name);
    }
    XtAddCallback (w, XkwNPPMWindowMovieCallback,
		   _ViewDatasets_save_ppm_movie_cbk, New);
}   /*  End Function Initialise  */

static void _ViewDatasets_Destroy (Widget W)
{
    ViewDatasetsWidget w = (ViewDatasetsWidget) W;

    c_unregister_callback (w->viewDatasets.cursor_tracking.listen_callback);
}   /*  End Function Destroy  */

static void _ViewDatasets_create_scatter_plot_shell (ViewDatasetsWidget top)
{
    KPixCanvas pixcanvas;
    KWorldCanvas worldcanvas;
    Widget shell, pswin, form, w, om, track_lbl1, square_btn;
    /*static char function_name[]="ViewDatasets::create_scatter_plot_shell";*/

    if (top->viewDatasets.scatter_plot)
	scatplt_destroy (top->viewDatasets.scatter_plot);
    top->viewDatasets.scatter_plot = NULL;
    if (top->viewDatasets.scatter_plot_shell) return;
    shell = XtVaCreatePopupShell ("scatterPlotPopup", topLevelShellWidgetClass,
				  (Widget) top,
				  NULL);
    top->viewDatasets.scatter_plot_shell = shell;
    _ViewDatasets_renumber_func (top);
    pswin = XtVaCreatePopupShell ("postscriptwinpopup", postscriptWidgetClass,
				  shell,
				  XtNtitle, "Postscript Window",
				  XkwNautoIncrement, TRUE,
				  NULL);
    form = XtVaCreateManagedWidget ("form", formWidgetClass, shell,
				    XtNborderWidth, 0,
				    NULL);
    w = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				 XtNlabel, "Close",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popdown_cbk,
		   top->viewDatasets.scatter_plot_shell);
    om = XtVaCreateManagedWidget ("overlayMenu", overlayMenuWidgetClass, form,
				  XtNtop, XtChainTop,
				  XtNbottom, XtChainTop,
				  XtNfromHoriz, w,
				  XtNheight, 20,
				  NULL);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Print",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromHoriz, om,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popup_cbk, pswin);
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, form,
				 XtNlabel, "Auto Range",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromHoriz, w,
				 XtNheight, 20,
				 XkwNcrosses, False,
				 XtNstate, True,
				 NULL);
    top->viewDatasets.scatter_plot_auto_tgl = w;
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Square",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromHoriz, w,
				 XtNheight, 20,
				 NULL);
    square_btn = w;
    w = XtVaCreateManagedWidget ("trackLabel", labelWidgetClass, form,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, w,
				 XtNwidth, 400,
				 XtNheight, 20,
				 NULL);
    track_lbl1 = w;
    w = XtVaCreateManagedWidget ("trackLabel", labelWidgetClass, form,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, w,
				 XtNwidth, 400,
				 XtNheight, 20,
				 XtVaTypedArg,XtNbackground,XtRString,"red", 4,
				 NULL);
    top->viewDatasets.scatter_plot_track_lbl2 = w;
    w = XtVaCreateManagedWidget ("scatterPlotCanvas", canvasWidgetClass, form,
				 XtNfromVert, w,
				 XtNwidth, 400,
				 XtNheight, 400,
				 XkwNsilenceUnconsumed, True,
				 NULL);
    XtRealizeWidget (shell);
    XtVaGetValues (w, XkwNmonoPixCanvas, &pixcanvas, NULL);
    worldcanvas = canvas_create (pixcanvas, NULL, NULL);
    scatplt_init (worldcanvas, SCATPLT_CANVAS_ATT_END);
    XkwOverlayMenuAddCanvas (om, worldcanvas);
    top->viewDatasets.scatter_plot_canvas = worldcanvas;
    canvas_set_dressing (worldcanvas,
			 KCD_ATT_SHOW_LABELS, TRUE,
			 KCD_ATT_TOP_TICKS, TRUE,
			 KCD_ATT_BOTTOM_TICKS, TRUE,
			 KCD_ATT_LEFT_TICKS, TRUE,
			 KCD_ATT_RIGHT_TICKS, TRUE,
			 KCD_ATT_END);
    canvas_register_position_event_func (worldcanvas,
					 _ViewDatasets_scatter_position_func,
					 track_lbl1);
    scatplt_create_smart_cursor (worldcanvas, 10.0,
				 _ViewDatasets_scatter_cursor_func, top);
    canvas_zoom_handle_mouse (worldcanvas,
			      &top->viewDatasets.scatter_plot_limits);
    canvas_zoom_handle_keys (worldcanvas,
			     &top->viewDatasets.scatter_plot_limits);
    XkwPostscriptRegisterImageAndName (pswin, pixcanvas, NULL);
    XtAddCallback (square_btn, XtNcallback, _ViewDatasets_square_scatter_cbk,
		   worldcanvas);
}   /*  End Function create_scatter_plot_shell  */

static void slice_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the slice callback. This is called when the slice direction is
    changed.
*/
{
    KDisplayDataWindow window;
    KDisplayDataObject dataobject;
    ViewDatasetsWidget top = (ViewDatasetsWidget) client_data;
    int new_slice = *(int *) call_data;
    unsigned int frame_number, num_frames;

    if (new_slice == top->viewDatasets.slice) return;
    top->viewDatasets.slice = new_slice;
    window = XkwImageDisplayGetWindow (top->viewDatasets.image_display);
    dispdata_set_window_attributes
	(window,
	 KDISPLAYDATA_WINDOW_ATT_SLICE_DIRECTION, top->viewDatasets.slice,
	 KDISPLAYDATA_WINDOW_ATT_END);
    dispdata_get_window_attributes
	(window,
	 KDISPLAYDATA_WINDOW_ATT_MOVIE_MASTER, &dataobject,
	 KDISPLAYDATA_WINDOW_ATT_END);
    if (!dataobject) return;
    dispdata_get_data_attributes
	(dataobject,
	 KDISPLAYDATA_DATA_ATT_FRAME_NUMBER, &frame_number,
	 KDISPLAYDATA_DATA_ATT_NUM_FRAMES, &num_frames,
	 KDISPLAYDATA_DATA_ATT_DIRECTION_FOR_FRAME, top->viewDatasets.slice,
	 KDISPLAYDATA_DATA_ATT_END);
    XtVaSetValues (top->viewDatasets.animatepopup,
		   XkwNnumFrames, num_frames,
		   XkwNcurrentFrame, frame_number,
		   NULL);
}   /*  End Function slice_cbk  */

static void recompute_profile_dir (ViewDatasetsWidget w)
/*  [SUMMARY] Set the profile direction based on the display mode.
    <w> The ViewDatasets widget.
    [RETURNS] Nothing.
*/
{
    int new_profile = 0;

    switch (w->viewDatasets.slice)
    {
      case KDISPLAYDATA_SLICE_DIRECTION_XY:
	new_profile = KDISPLAYDATA_PROFILE_DIR_Z;
	break;
      case KDISPLAYDATA_SLICE_DIRECTION_XZ:
	new_profile = KDISPLAYDATA_PROFILE_DIR_Y;
	break;
      case KDISPLAYDATA_SLICE_DIRECTION_ZY:
	new_profile = KDISPLAYDATA_PROFILE_DIR_X;
	break;
    }
    XtVaSetValues (w->viewDatasets.profile_dir_menu,
		   XkwNsetChoice, dispdata_profile_dir_choices[new_profile],
		   NULL);
    w->viewDatasets.profile_dir = new_profile;
}   /*  End Function recompute_profile_dir  */

static void _ViewDatasets_show_browsers_cbk (Widget w, XtPointer client_data,
					     XtPointer call_data)
{
    KDisplayDataWindow window;
    KDisplayDataBlinkState blinkstate;
    Widget image_display = client_data;

    window = XkwImageDisplayGetWindow (image_display);
    for ( blinkstate = dispdata_get_first_blinkstate (window, NULL);
	  blinkstate;
	  blinkstate = dispdata_get_next_blinkstate (blinkstate, FALSE, NULL) )
	XkwDataBrowserPopup (blinkstate);
}   /*  End Function show_browsers_cbk  */

static void profile_mode_cbk (Widget w, XtPointer client_data,
			      XtPointer call_data)
/*  This is the profile mode (shape) callback.
*/
{
    KDisplayDataWindow window;
    int new_mode = *(int *) call_data;
    Boolean sensitive = True;
    ViewDatasetsWidget top = (ViewDatasetsWidget) client_data;

    create_tracewinpopup (top);
    switch (top->viewDatasets.profile_mode = new_mode)
    {
      case KDISPLAYDATA_PROFILE_MODE_NONE:
	XtPopdown (top->viewDatasets.trace_winpopup);
	break;
      case KDISPLAYDATA_PROFILE_MODE_LINE:
	XtPopup (top->viewDatasets.trace_winpopup, XtGrabNone);
	m_clear (&top->viewDatasets.main_marker_area,
		 sizeof top->viewDatasets.main_marker_area);
	top->viewDatasets.main_marker_area.startx = -100;
	break;
      case KDISPLAYDATA_PROFILE_MODE_BOX_SUM:
      case KDISPLAYDATA_PROFILE_MODE_BOX_AVERAGE:
	XtPopup (top->viewDatasets.trace_winpopup, XtGrabNone);
	sensitive = False;
	break;
      case KDISPLAYDATA_PROFILE_MODE_BOXED_HORIZONTAL:
      case KDISPLAYDATA_PROFILE_MODE_BOXED_VERTICAL:
      case KDISPLAYDATA_PROFILE_MODE_RADIAL:
      case KDISPLAYDATA_PROFILE_MODE_RADIAL_GAUSS:
      case KDISPLAYDATA_PROFILE_MODE_AZIMUTHAL_AVERAGE:
	XtPopup (top->viewDatasets.trace_winpopup, XtGrabNone);
	break;
    }
    if (!sensitive) recompute_profile_dir (top);
    XtSetSensitive (top->viewDatasets.profile_dir_menu, sensitive);
    window = XkwImageDisplayGetWindow (top->viewDatasets.image_display);
    dispdata_set_window_attributes
	(window,
	 KDISPLAYDATA_WINDOW_ATT_PROFILE_MODE, new_mode,
	 KDISPLAYDATA_WINDOW_ATT_END);
}   /*  End Function profile_mode_cbk  */

static void profile_dir_cbk (Widget w, XtPointer client_data,
			     XtPointer call_data)
/*  This is the profile direction callback.
*/
{
    KDisplayDataWindow window;
    ViewDatasetsWidget top = (ViewDatasetsWidget) client_data;
    int new_direction = *(int *) call_data;

    top->viewDatasets.profile_dir = new_direction;
    window = XkwImageDisplayGetWindow (top->viewDatasets.image_display);
    dispdata_set_window_attributes
	(window,
	 KDISPLAYDATA_WINDOW_ATT_PROFILE_DIRECTION, new_direction,
	 KDISPLAYDATA_WINDOW_ATT_END);
}   /*  End Function profile_dir_cbk  */

static void new_frame_cbk (Widget w, XtPointer client_data,XtPointer call_data)
/*  This is the new frame callback.
*/
{
    KDisplayDataWindow window;
    KDisplayDataObject dataobject;
    iarray pseudo;
    int frame_number = *(int *) call_data;
    ViewDatasetsWidget top = client_data;
    Bool bool;

    window = XkwImageDisplayGetWindow (top->viewDatasets.image_display);
    dispdata_get_window_attributes
	(window,
	 KDISPLAYDATA_WINDOW_ATT_MOVIE_MASTER, &dataobject,
	 KDISPLAYDATA_WINDOW_ATT_END);
    if (!dataobject) return;
    dispdata_set_window_attributes
	(window,
	 KDISPLAYDATA_WINDOW_ATT_FRAME_NUMBER, frame_number,
	 KDISPLAYDATA_WINDOW_ATT_END);
    if (!w) XtVaSetValues (top->viewDatasets.animatepopup,
			   XkwNcurrentFrame, frame_number,
			   NULL);
    if (top->viewDatasets.trace_winpopup == NULL) return;
    dispdata_get_window_attributes
	(window,
	 KDISPLAYDATA_WINDOW_ATT_PROFILE_MASTER, &dataobject,
	 KDISPLAYDATA_WINDOW_ATT_END);
    if (!dataobject) return;
    dispdata_get_data_attributes (dataobject,
				  KDISPLAYDATA_DATA_ATT_PSEUDO_ARRAY, &pseudo,
				  KDISPLAYDATA_DATA_ATT_END);
    if (!pseudo) return;
    switch (top->viewDatasets.profile_mode)
    {
      case KDISPLAYDATA_PROFILE_MODE_LINE:
      case KDISPLAYDATA_PROFILE_MODE_BOX_SUM:
      case KDISPLAYDATA_PROFILE_MODE_BOX_AVERAGE:
	XtVaGetValues (top->viewDatasets.show_profile_marker_tgl,
		       XtNstate, &bool,
		       NULL);
	if (!bool) return;
	XkwTracePopupRefresh (top->viewDatasets.trace_winpopup, TRUE);
	break;
    }
}   /*  End Function new_frame_cbk  */

static void _ViewDatasets_link_iscale_cbk (Widget w, XtPointer client_data,
					   XtPointer call_data)
/*  This is the link iscale callback.
*/
{
    KDisplayDataWindow window;
    flag val;
    ViewDatasetsWidget top = (ViewDatasetsWidget) client_data;
    uaddr bool = (uaddr) call_data;

    window = XkwImageDisplayGetWindow (top->viewDatasets.image_display);
    val = bool ? FALSE : TRUE;
    dispdata_set_window_attributes (window,
				    KDISPLAYDATA_WINDOW_ATT_SET_ISCALE, val,
				    KDISPLAYDATA_WINDOW_ATT_END);
}   /*  End Function link_iscale_cbk  */

static void _ViewDatasets_show_marker_cbk (Widget w, XtPointer client_data,
					   XtPointer call_data)
{
    KDisplayDataWindow window;
    ViewDatasetsWidget top = (ViewDatasetsWidget) client_data;
    flag show_marker = (uaddr) call_data;

    window = XkwImageDisplayGetWindow (top->viewDatasets.image_display);
    dispdata_set_window_attributes
	(window,
	 KDISPLAYDATA_WINDOW_ATT_SHOW_MARKER_IN_PROFILE, show_marker,
	 KDISPLAYDATA_WINDOW_ATT_END);
}   /*  End Function show_marker_cbk  */

static void _ViewDatasets_title_cbk (Widget w, XtPointer client_data,
				     XtPointer call_data)
{
    KDisplayDataWindow window;
    ViewDatasetsWidget top = (ViewDatasetsWidget) client_data;
    flag auto_title = (uaddr) call_data;
    char *title;

    window = XkwImageDisplayGetWindow (top->viewDatasets.image_display);
    dispdata_set_window_attributes
	(window,
	 KDISPLAYDATA_WINDOW_ATT_AUTO_TITLE, auto_title,
	 KDISPLAYDATA_WINDOW_ATT_END);
    if (auto_title) return;
    XtVaGetValues (top->viewDatasets.title_txt, XtNstring, &title, NULL);
    _ViewDatasets_text_cbk (NULL, top, title);
}   /*  End Function title_cbk  */

static void _ViewDatasets_text_cbk (Widget w, XtPointer client_data,
				    XtPointer call_data)
/*  This is the title text callback.
*/
{
    KWorldCanvas pseudo_canvas, direct_canvas, true_canvas;
    Bool auto_title;
    ViewDatasetsWidget top = (ViewDatasetsWidget) client_data;
    char *title = call_data;

    XtVaGetValues (top->viewDatasets.auto_title_tgl,
		   XtNstate, &auto_title,
		   NULL);
    if (auto_title) return;
    XtVaGetValues (top->viewDatasets.image_display,
		   XkwNpseudoColourCanvas, &pseudo_canvas,
		   XkwNdirectColourCanvas, &direct_canvas,
		   XkwNtrueColourCanvas, &true_canvas,
		   NULL);
    if (pseudo_canvas) canvas_set_dressing (pseudo_canvas,
					    KCD_ATT_TITLE, title,
					    KCD_ATT_END);
    if (direct_canvas) canvas_set_dressing (direct_canvas,
					    KCD_ATT_TITLE, title,
					    KCD_ATT_END);
    if (true_canvas) canvas_set_dressing (true_canvas,
					  KCD_ATT_TITLE, title,
					  KCD_ATT_END);
    XkwImageDisplayRefresh (top->viewDatasets.image_display, FALSE);
}   /*  End Function text_cbk  */

static void _ViewDatasets_all_realised_cbk (Widget w, XtPointer client_data,
					    XtPointer call_data)
/*  This is the all realised callback.
*/
{
    KDisplayDataWindow window;
    KWorldCanvas pseudo_canvas, direct_canvas, true_canvas;
    ViewDatasetsWidget top = (ViewDatasetsWidget) client_data;
    KWorldCanvas *canvases = (KWorldCanvas *) call_data;
    /*static char function_name[] = "ViewDatasetsWidget::all_realised_cbk";*/

    for (; *canvases; ++canvases)
	canvas_register_refresh_func (*canvases,
				      _ViewDatasets_canvas_refresh_func, top);
    XtVaGetValues (top->viewDatasets.image_display,
		   XkwNpseudoColourCanvas, &pseudo_canvas,
		   XkwNdirectColourCanvas, &direct_canvas,
		   XkwNtrueColourCanvas, &true_canvas,
		   NULL);
    if (pseudo_canvas)
    {
	viewimg_register_position_event_func
	    (pseudo_canvas, _ViewDatasets_viewimg_position_func, top);
	kwin_register_bindings (canvas_get_pixcanvas (pseudo_canvas),
				main_bindings, FALSE);
    }
    if (direct_canvas)
    {
	viewimg_register_position_event_func
	    (direct_canvas, _ViewDatasets_viewimg_position_func, top);
	kwin_register_bindings (canvas_get_pixcanvas (direct_canvas),
				main_bindings, FALSE);
    }
    if (true_canvas)
    {
	viewimg_register_position_event_func
	    (true_canvas, _ViewDatasets_viewimg_position_func, top);
	kwin_register_bindings (canvas_get_pixcanvas (true_canvas),
				main_bindings, FALSE);
    }
    window = XkwImageDisplayGetWindow (top->viewDatasets.image_display);
    dispdata_set_window_attributes
	(window,
	 KDISPLAYDATA_WINDOW_ATT_SLICE_DIRECTION, top->viewDatasets.slice,
	 KDISPLAYDATA_WINDOW_ATT_END);
    if (dispdata_get_window_index (window) > 0)
	dispdata_register_window_renumber_func (window,
						_ViewDatasets_renumber_func,
						top);
    _ViewDatasets_renumber_func (top);
}   /*  End Function all_realised_cbk  */

static void _ViewDatasets_renumber_func (ViewDatasetsWidget w)
{
    KDisplayDataWindow window;
    unsigned int index;
    char title[STRING_LENGTH];

    window = XkwImageDisplayGetWindow (w->viewDatasets.image_display);
    index = dispdata_get_window_index (window) + 1;
    sprintf (title, "View Control for display window %u", index);
    XtVaSetValues ( (Widget) w, XtNtitle, title, NULL );
    sprintf (title, "Animate Control for display window %u", index);
    XtVaSetValues (w->viewDatasets.animatepopup, XtNtitle, title, NULL);
    if (w->viewDatasets.trace_winpopup)
    {
	sprintf (title, "Profile Window for display window %u", index);
	XtVaSetValues (w->viewDatasets.trace_winpopup, XtNtitle, title, NULL);
    }
    if (w->viewDatasets.scatter_plot_shell)
    {
	sprintf (title, "Scatter Plot for display window %u", index);
	XtVaSetValues (w->viewDatasets.scatter_plot_shell,XtNtitle,title,NULL);
    }
}   /*  End Function renumber_func  */

static void _ViewDatasets_new_editor_context_cbk (Widget w,
						  XtPointer client_data,
						  XtPointer call_data)
/*  This is the callback for when a new editor context is available.
*/
{
    KPixCanvasEditorContext ctx = (KPixCanvasEditorContext) call_data;

    kwin_editor_register_event_func (ctx, _ViewDatasets_left_geom_func,
				     client_data);
}   /*  End Function new_editor_context_cbk  */

static void _ViewDatasets_refresh_cbk (Widget w, XtPointer client_data,
				       XtPointer call_data)
/*  This is the refresh callback.
*/
{
    ViewDatasetsWidget top = (ViewDatasetsWidget) client_data;

    XkwImageDisplayRefresh (top->viewDatasets.image_display, FALSE);
}   /*  End Function refresh_cbk  */

static void _ViewDatasets_canvas_refresh_func(KWorldCanvas canvas,
					      int width, int height,
					      struct win_scale_type *win_scale,
					      Kcolourmap cmap,
					      flag cmap_resize,void **info,
					      PostScriptPage pspage,
					      unsigned int num_areas,
					      KPixCanvasRefreshArea *areas,
					      flag *honoured_areas)
/*  [SUMMARY] Refresh event callback.
    [PURPOSE] This routine is a refresh event consumer for a world canvas.
    <canvas> The world canvas being refreshed.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The colourmap associated with the canvas.
    <cmap_resize> TRUE if the refresh function was called as a result of a
    colourmap resize, else FALSE.
    <info> A pointer to the arbitrary canvas information pointer.
    <pspage> If not NULL, the PostScriptPage object the refresh is
    redirected to.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    KDisplayDataWindow window;
    KDisplayDataObject dataobject;
    KwcsAstro ap;
    iarray array;
    unsigned int num_restr;
    ViewDatasetsWidget top = *info;
    Bool show_beam, show_beam_name;
    char *xlabel, *ylabel;
    char **restr_names;
    double *restr_values;
    /*static char function_name[]="ViewDatasetsWidget::canvas_refresh_func";*/

    *honoured_areas = TRUE;  /*  Bullshit. But this should be drawn last, and I
				 want to avoid redrawing axis labels if not
				 required  */
    window = XkwImageDisplayGetWindow (top->viewDatasets.image_display);
    dispdata_get_window_attributes (window,
				    KDISPLAYDATA_WINDOW_ATT_IMAGE, &dataobject,
				    KDISPLAYDATA_WINDOW_ATT_END);
    if (!dataobject) return;
    dispdata_get_data_attributes (dataobject,
				  KDISPLAYDATA_DATA_ATT_PSEUDO_ARRAY, &array,
				  KDISPLAYDATA_DATA_ATT_END);
    if (!array) return;
    XtVaGetValues (top->viewDatasets.show_beam_tgl,
		   XtNstate, &show_beam,
		   NULL);
    XtVaGetValues (top->viewDatasets.show_beam_name_tgl,
		   XtNstate, &show_beam_name,
		   NULL);
    if (!show_beam) return;
    ap = iarray_get_associated_wcs_astro (array);
    if (!ap) return;
    canvas_get_specification (canvas, &xlabel, &ylabel,
			      &num_restr, &restr_names, &restr_values);
    if ( !wcs_astro_test_if_lon (ap, xlabel) ) return;
    if ( !wcs_astro_test_if_lat (ap, ylabel) ) return;
    _ViewDatasets_draw_beam (canvas, top, dataobject, show_beam_name);
}   /*  End Function canvas_refresh_func  */

static flag _ViewDatasets_viewimg_position_func (ViewableImage vimage,
						 double x, double y,
						 void *value,
						 unsigned int event_code,
						 void *e_info, void **f_info,
						 double x_lin, double y_lin,
						 unsigned int value_type,
						 double x_im, double y_im,
						 int x_pix, int y_pix)
/*  [SUMMARY] Position event callback for the main display canvas.
    [PURPOSE] This routine is a position event consumer for a world canvas
    which has a number of ViewableImage objects associated with it.
    <viewimg> The active viewable image.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    These values will have been transformed by the registered transform
    function (see [<canvas_register_transform_func>]) for the associated
    world canvas.
    <value> A pointer to the data value in the viewable image corresponding
    to the event co-ordinates.
    <event_code> The arbitrary event code.
    <e_info> The arbitrary event information.
    <f_info> The arbitrary function information pointer.
    <x_lin> The linear horizontal world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <y_lin> The linear vertical world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <value_type> The type of the data value. This may be K_DCOMPLEX or
    K_UB_RGB.
    <x_im> The quantised horizontal image pixel co-ordinate. 0.0 is the pixel
    centre.
    <y_im> The quantised vertical image pixel co-ordinate. 0.0 is the pixel
    centre.
    <x_pix> The horizontal screen pixel co-ordinate.
    <y_pix> The vertical screen pixel co-ordinate.
    [NOTE] All co-ordinates are clipped to pixel boundaries.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KDisplayDataWindow window;
    KDisplayDataBlinkState bs;
    KDisplayDataObject image, movie;
    KEvent event;
    Bool bool;
    int frame_jump, frame, num_frames;
    unsigned int old_event;
    unsigned long mask;
    ViewDatasetsWidget top = *f_info;
    Widget first_track_label, second_track_label, third_track_label;
    Widget image_display;
    KEventCoordinate *coordinate = NULL;
    char *key;
    char *ptr_x, *ptr_y, *ptr_z, *ptr_string;
    struct cursor_data broadcast_cursor;
    char pix_string[STRING_LENGTH];
    char world_string[STRING_LENGTH], extra_string[STRING_LENGTH];
    double lx[2], ly[2];
    static flag in_broadcast = FALSE;
    static char function_name[] = "ViewDatasets_viewimg::position_func";

    /*  First process internal events, consuming some and tapping others  */
    window = XkwImageDisplayGetWindow (top->viewDatasets.image_display);
    bs = dispdata_get_active_blinkstate (window, NULL);
    old_event = top->viewDatasets.last_event_code;
    top->viewDatasets.last_event_code = event_code;
    if (event_code == K_CANVAS_EVENT_POINTER_LEAVE)
    {
	pix_string[0] = '\0';
	world_string[0] = '\0';
	extra_string[0] = '\0';
    }
    else viewimg_track_compute (vimage, value, value_type, x, y, x_im, y_im,
				pix_string, world_string, extra_string);
    dispdata_get_window_attributes
	(window,
	 KDISPLAYDATA_WINDOW_ATT_IMAGE, &image,
	 KDISPLAYDATA_WINDOW_ATT_MOVIE_MASTER, &movie,
	 KDISPLAYDATA_WINDOW_ATT_END);
    switch (event_code)
    {
      case K_CANVAS_EVENT_MIDDLE_MOUSE_RELEASE:
	if (top->viewDatasets.profile_mode ==
	    KDISPLAYDATA_PROFILE_MODE_BOX_SUM) break;
	if (top->viewDatasets.profile_mode ==
	    KDISPLAYDATA_PROFILE_MODE_BOX_AVERAGE) break;
	if (top->viewDatasets.profile_mode ==
	    KDISPLAYDATA_PROFILE_MODE_RADIAL) break;
	if (top->viewDatasets.profile_mode ==
	    KDISPLAYDATA_PROFILE_MODE_RADIAL_GAUSS) break;
	if (top->viewDatasets.profile_mode ==
	    KDISPLAYDATA_PROFILE_MODE_BOXED_HORIZONTAL) break;
	if (top->viewDatasets.profile_mode ==
	    KDISPLAYDATA_PROFILE_MODE_BOXED_VERTICAL) break;
	XtVaGetValues (top->viewDatasets.middle_zoom_tgl,
		       XtNstate, &bool,
		       NULL);
	if (!bool) break;
	if (old_event != K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK) break;
	viewimg_zoom (vimage, x_lin, y_lin, FALSE, 2.0);
	canvas_warp_pointer (viewimg_get_worldcanvas (vimage),
			     x_lin, y_lin, TRUE, TRUE);
	return (TRUE);
	/*break;*/
      case K_CANVAS_EVENT_MOUSE_BUTTON5_RELEASE:
	if (old_event != K_CANVAS_EVENT_MOUSE_BUTTON5_CLICK) break;
	viewimg_zoom (vimage, x_lin, y_lin, FALSE, 2.0);
	canvas_warp_pointer (viewimg_get_worldcanvas (vimage),
			     x_lin, y_lin, TRUE, TRUE);
	return (TRUE);
	/*break;*/
      case K_CANVAS_EVENT_RIGHT_MOUSE_RELEASE:
	if (old_event != K_CANVAS_EVENT_RIGHT_MOUSE_CLICK) break;
	XtVaGetValues (top->viewDatasets.right_zoom_tgl,
		       XtNstate, &bool,
		       NULL);
	if (!bool) break;
	viewimg_zoom (vimage, 0.5, 0.5, TRUE, 0.5);
	canvas_warp_pointer (viewimg_get_worldcanvas (vimage),
			     x_lin, y_lin, TRUE, TRUE);
	return (TRUE);
	/*break;*/
      case K_CANVAS_EVENT_MOUSE_BUTTON4_RELEASE:
	if (old_event != K_CANVAS_EVENT_MOUSE_BUTTON4_CLICK) break;
	viewimg_zoom (vimage, 0.5, 0.5, TRUE, 0.5);
	canvas_warp_pointer (viewimg_get_worldcanvas (vimage),
			     x_lin, y_lin, TRUE, TRUE);
	return (TRUE);
	/*break;*/
      case K_CANVAS_EVENT_PLAIN_KEY_PRESS:
	key = e_info;
	switch (*key)
	{
	  case 'b':
	  case '>':
	    bs = dispdata_get_next_blinkstate (bs, TRUE, NULL);
	    dispdata_activate_blinkstate (bs);
	    return (TRUE);
	    /*break;*/
	  case '<':
	    bs = dispdata_get_prev_blinkstate (bs, TRUE, NULL);
	    dispdata_activate_blinkstate (bs);
	    return (TRUE);
	    /*break;*/
	  case 'B':
	    XkwDataBrowserPopup (bs);
	    return (TRUE);
	    /*break;*/
	  case 'c':
	    canvas_get_attributes (viewimg_get_worldcanvas (vimage),
				   CANVAS_ATT_LEFT_X, lx,
				   CANVAS_ATT_RIGHT_X, lx + 1,
				   CANVAS_ATT_BOTTOM_Y, ly,
				   CANVAS_ATT_TOP_Y, ly + 1,
				   CANVAS_ATT_END);
	    return _ViewDatasets_scatplot (top, lx, ly);
	    /*break;*/
	  case 'l':
	    puts (world_string);
	    return (TRUE);
	    /*break;*/
	}
	break;
      case K_CANVAS_EVENT_MODIFIED_KEY_PRESS:
	key = e_info;
	m_copy (&mask, key + strlen (key) + 1, sizeof mask);
	switch (*key)
	{
	  case 'b':
	    if (mask != K_CANVAS_EVENT_CONTROL_KEY_MASK) break;
	    for ( bs = dispdata_get_first_blinkstate (window, NULL);
		  bs;
		  bs = dispdata_get_next_blinkstate (bs, FALSE, NULL) )
		XkwDataBrowserPopup (bs);
	    return (TRUE);
	    /*break;*/
	}
	break;
      case K_CANVAS_EVENT_KEYSYM_PRESS:
	key = e_info;
	if (strncmp (key, "KP_", 3) == 0) key += 3;
	m_copy (&mask, key + strlen (key) + 1, sizeof mask);
	if (!movie) break;
	dispdata_get_data_attributes
	    (movie,
	     KDISPLAYDATA_DATA_ATT_FRAME_NUMBER, &frame,
	     KDISPLAYDATA_DATA_ATT_NUM_FRAMES, &num_frames,
	     KDISPLAYDATA_DATA_ATT_DIRECTION_FOR_FRAME,top->viewDatasets.slice,
	     KDISPLAYDATA_DATA_ATT_END);
	frame_jump = (mask & K_CANVAS_EVENT_CONTROL_KEY_MASK) ? 10 : 1;
	if (strcmp (key, "Next") == 0) frame -= frame_jump;        /*  PgDn  */
	else if (strcmp (key, "Prior") == 0) frame += frame_jump;  /*  PgUp  */
	else if (strcmp (key, "Home") == 0) frame = 0;
	else if (strcmp (key, "End") == 0) frame = num_frames - 1;
	else break;
	if (frame < 0) frame = num_frames - 1;
	else if (frame >= num_frames) frame = 0;
	new_frame_cbk (NULL, top, &frame);
	return (TRUE);
	/*break;*/
      case K_CANVAS_EVENT_POINTER_MOVE:
      case K_CANVAS_EVENT_POINTER_WARP:
	if (value_type == K_DCOMPLEX)
	{
	    flag ok = TRUE;
	    /*  Check if this ViewableImage is either of the ones used to
		generate the scatterplot  */
	    if (!top->viewDatasets.scatter_plot_x_handle.vimage ||
		!top->viewDatasets.scatter_plot_y_handle.vimage) ok = FALSE;
	    else if ( (top->viewDatasets.scatter_plot_x_handle.vimage
		       != vimage) &&
		      (top->viewDatasets.scatter_plot_y_handle.vimage
		       != vimage) )
		ok = FALSE;
	    /*  If vimage is original scatterplot X vimage, check limits now */
	    else if ( (top->viewDatasets.scatter_plot_x_handle.vimage
		       == vimage) &&
		      ( (x_im < top->viewDatasets.scatter_plot_xstart) ||
			(x_im >= top->viewDatasets.scatter_plot_xstart +
			 top->viewDatasets.scatter_plot_xlen) ||
			(y_im < top->viewDatasets.scatter_plot_ystart) ||
			(y_im >= top->viewDatasets.scatter_plot_ystart +
			 top->viewDatasets.scatter_plot_ylen) ) ) ok = FALSE;
	    else
	        ok = _ViewDatasets_draw_marker_in_scatter_plot
		    (top, vimage, x, y, *(double *) value);
	    if (!ok && top->viewDatasets.scatter_plot_shell)
	    {
		_ViewDatasets_draw_marker
		    (top->viewDatasets.scatter_plot_canvas,
		     &top->viewDatasets.scatter_plot_marker_area,
		     TOOBIG, TOOBIG);
		if (top->viewDatasets.scatter_plot_track_lbl2)
		    XtVaSetValues (top->viewDatasets.scatter_plot_track_lbl2,
				   XtNlabel, "",
				   NULL);
	    }
	}
	if (!in_broadcast)
	{
	    KWorldCanvas wc = viewimg_get_worldcanvas (vimage);
	    unsigned int num_restr;
	    char **restr_names;
	    double *restr_values;

	    in_broadcast = TRUE;
	    canvas_get_specification (wc, (char **) &broadcast_cursor.xname,
				      (char **) &broadcast_cursor.yname,
				      &num_restr, &restr_names, &restr_values);
	    canvas_get_attributes (wc,
				   CANVAS_ATT_WCS_ASTRO, &broadcast_cursor.ap,
				   CANVAS_ATT_END);
	    if (num_restr > 0)
	    {
		broadcast_cursor.wx = x_lin;
		broadcast_cursor.wy = y_lin;
		broadcast_cursor.wz = restr_values[0];
		broadcast_cursor.zname = restr_names[0];
		if (broadcast_cursor.ap)
		    wcs_astro_transform3 (broadcast_cursor.ap, FALSE, 1,
					  broadcast_cursor.xname,
					  &broadcast_cursor.wx, FALSE,
					  broadcast_cursor.yname,
					  &broadcast_cursor.wy, FALSE,
					  restr_names[0],
					  &broadcast_cursor.wz, FALSE,
					  num_restr,
					  (CONST char **) restr_names,
					  restr_values);
	    }
	    else broadcast_cursor.zname = NULL;
	    broadcast_cursor.wx = x;
	    broadcast_cursor.wy = y;
	    broadcast_cursor.source = top;
	    c_call_callbacks (broadcast_list, &broadcast_cursor);
	    in_broadcast = FALSE;
	}
	break;
      case K_CANVAS_EVENT_POINTER_LEAVE:
	_ViewDatasets_draw_marker
	    (top->viewDatasets.scatter_plot_canvas,
	     &top->viewDatasets.scatter_plot_marker_area,
	     TOOBIG, TOOBIG);
	if (!in_broadcast)
	{
	    in_broadcast = TRUE;
	    broadcast_cursor.wx = TOOBIG;
	    broadcast_cursor.wy = TOOBIG;
	    broadcast_cursor.wz = TOOBIG;
	    broadcast_cursor.source = top;
	    c_call_callbacks (broadcast_list, &broadcast_cursor);
	    in_broadcast = FALSE;
	}
	break;
      default:
	break;
    }
    /*  Now process events for dispatch  */
    m_clear (&event, sizeof event);
    event.type = event_code;
    switch (event_code)
    {
      case K_CANVAS_EVENT_LEFT_MOUSE_CLICK:
      case K_CANVAS_EVENT_LEFT_MOUSE_DRAG:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_DRAG:
      case K_CANVAS_EVENT_RIGHT_MOUSE_CLICK:
      case K_CANVAS_EVENT_RIGHT_MOUSE_DRAG:
      case K_CANVAS_EVENT_POINTER_MOVE:
      case K_CANVAS_EVENT_LEFT_MOUSE_RELEASE:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_RELEASE:
      case K_CANVAS_EVENT_RIGHT_MOUSE_RELEASE:
      case K_CANVAS_EVENT_POINTER_WARP:
      case K_CANVAS_EVENT_MOUSE_BUTTON4_CLICK:
      case K_CANVAS_EVENT_MOUSE_BUTTON4_DRAG:
      case K_CANVAS_EVENT_MOUSE_BUTTON4_RELEASE:
      case K_CANVAS_EVENT_MOUSE_BUTTON5_CLICK:
      case K_CANVAS_EVENT_MOUSE_BUTTON5_DRAG:
      case K_CANVAS_EVENT_MOUSE_BUTTON5_RELEASE:
      case K_CANVAS_EVENT_POINTER_LEAVE:
	coordinate = &event.data.point;
	break;
      case K_CANVAS_EVENT_PLAIN_KEY_PRESS:
      case K_CANVAS_EVENT_PLAIN_KEY_RELEASE:
	coordinate = &event.data.keypress.position;
	event.data.keypress.modifier_mask = 0;
	event.data.keypress.string = e_info;
	break;
      case K_CANVAS_EVENT_MODIFIED_KEY_PRESS:
      case K_CANVAS_EVENT_MODIFIED_KEY_RELEASE:
      case K_CANVAS_EVENT_KEYSYM_PRESS:
      case K_CANVAS_EVENT_KEYSYM_RELEASE:
	coordinate = &event.data.keypress.position;
	ptr_string = e_info;
	m_copy (&event.data.keypress.modifier_mask,
		ptr_string + strlen (ptr_string) + 1,
		sizeof event.data.keypress.modifier_mask);
	event.data.keypress.string = ptr_string;
	break;
      case K_CANVAS_EVENT_POINTER_ENTER:
	return (FALSE);
	/*break;*/
      default:
	fprintf (stderr, "%s: unsupported event type: %u\n",
		 function_name, event_code);
	return (TRUE);
	/*break;*/
    }
    coordinate->screen_pixel.defined = TRUE;
    coordinate->screen_pixel.x = x_pix;
    coordinate->screen_pixel.y = y_pix;
    coordinate->linear.defined = TRUE;
    coordinate->linear.x = x_lin;
    coordinate->linear.y = y_lin;
    coordinate->world.defined = TRUE;
    coordinate->world.x = x;
    coordinate->world.y = y;
    coordinate->image_pixel.defined = TRUE;
    coordinate->image_pixel.x = x_im;
    coordinate->image_pixel.y = y_im;
    event_dispatch (event);
    dispdata_get_data_attributes
	(image,
	 KDISPLAYDATA_DATA_ATT_NUM_FRAMES, &num_frames,
	 KDISPLAYDATA_DATA_ATT_DIRECTION_FOR_FRAME, top->viewDatasets.slice,
	 KDISPLAYDATA_DATA_ATT_END);
    if ( pix_string[0] && (num_frames > 0) )
    {   /*  Modify pix_string  */
	if ( ( ptr_x = strchr (pix_string, 'x') ) == NULL )
	{
	    fprintf (stderr, "%s: error searching for 'x' in: \"%s\"\n",
		     function_name, pix_string);
	    return (FALSE);
	}
	if ( ( ptr_y = strchr (pix_string, 'y') ) == NULL )
	{
	    fprintf (stderr, "%s: error searching for 'y' in: \"%s\"\n",
		     function_name, pix_string);
	    return (FALSE);
	}
	if ( ( ptr_z = strchr (pix_string, 'z') ) == NULL )
	{
	    fprintf (stderr, "%s: error searching for 'z' in: \"%s\"\n",
		     function_name, pix_string);
	    return (FALSE);
	}
	switch (top->viewDatasets.slice)
	{
	  case KDISPLAYDATA_SLICE_DIRECTION_XY:
	    ptr_z[0] = ' ';
	    ptr_z[1] = 'z';
	    break;
	  case KDISPLAYDATA_SLICE_DIRECTION_XZ:
	    *ptr_y = 'z';
	    ptr_z[0] = ' ';
	    ptr_z[1] = 'y';
	    break;
	  case KDISPLAYDATA_SLICE_DIRECTION_ZY:
	    *ptr_x = 'z';
	    ptr_z[0] = ' ';
	    ptr_z[1] = 'x';
	    break;
	}
    }
    image_display = top->viewDatasets.image_display;
    first_track_label = XtNameToWidget (image_display, "trackLabel0");
    second_track_label = XtNameToWidget (image_display, "trackLabel1");
    third_track_label = XtNameToWidget (image_display, "trackLabel2");
    XtVaSetValues (first_track_label, XtNlabel, pix_string, NULL);
    XtVaSetValues (second_track_label, XtNlabel, world_string, NULL);
    XtVaGetValues (third_track_label, XtNlabel, &key, NULL);
    if (strcmp (key, extra_string) != 0)
	XtVaSetValues (third_track_label, XtNlabel, extra_string, NULL);
    return (FALSE);
}   /*  End Function _ViewDatasets_viewimg_position_func  */

static void _ViewDatasets_draw_beam (KWorldCanvas canvas, ViewDatasetsWidget w,
				     KDisplayDataObject dataobject, flag name)
/*  [SUMMARY] Draw a beam on a canvas.
    <canvas> The world canvas.
    <w> The ViewDatasetWidget.
    <dataobject> The beam data.
    <name> If TRUE, the beam name is displayed.
    [RETURNS] Nothing.
*/
{
    KPixCanvas pixcanvas;
    KPixCanvasHersheyString string;
    iarray array;
    unsigned int array_num;
    unsigned long pixel_value = 0;
    double left_x, right_x, bottom_y, top_y;
    double cx, cy, x, y, width, height;
    multi_array *multi_desc;
    char *colourname, *object;
    double bmaj, bmin, bpa, value[2];

    dispdata_get_data_attributes (dataobject,
				  KDISPLAYDATA_DATA_ATT_NAME, &colourname,
				  KDISPLAYDATA_DATA_ATT_OBJECT, &object,
				  KDISPLAYDATA_DATA_ATT_PSEUDO_ARRAY, &array,
				  KDISPLAYDATA_DATA_ATT_END);
    if (!object) object = colourname;
    pixcanvas = canvas_get_pixcanvas (canvas);
    canvas_get_dressing (canvas,
			 KCD_ATT_LABEL_COLOUR, &colourname,
			 KCD_ATT_END);
    if (colourname == NULL) colourname = "green";
    kwin_get_colour_harder (pixcanvas, colourname, &pixel_value);
    canvas_get_attributes (canvas,
			   CANVAS_ATT_LEFT_X, &left_x,
			   CANVAS_ATT_RIGHT_X, &right_x,
			   CANVAS_ATT_BOTTOM_Y, &bottom_y,
			   CANVAS_ATT_TOP_Y, &top_y,
			   CANVAS_ATT_END);
    multi_desc = iarray_multi_array (array);
    array_num = iarray_multi_index (array);
    if ( !ds_get_unique_named_value (multi_desc->headers[array_num],
				     multi_desc->data[array_num],
				     "BMAJ", NULL, value) ) return;
    bmaj = value[0] * 0.5;
    if ( !ds_get_unique_named_value (multi_desc->headers[array_num],
				     multi_desc->data[array_num],
				     "BMIN", NULL, value) ) return;
    bmin = value[0] * 0.5;
    if ( !ds_get_unique_named_value (multi_desc->headers[array_num],
				     multi_desc->data[array_num],
				     "BPA", NULL, value) ) return;
    bpa = 90.0 - value[0];
    cx = left_x + w->viewDatasets.beam_xpos * (right_x - left_x);
    cy = bottom_y + w->viewDatasets.beam_ypos * (top_y - bottom_y);
    canvas_coords_transform (canvas, 1, &cx, FALSE, &cy, FALSE);
    canvas_convert_ellipse_coords (canvas, FALSE, FALSE,
				   &cx, &cy, &bmaj, &bmin, &bpa);
    if (bpa <= -90.0) bpa += 180.0;
    else if (bpa >= 90.0) bpa -= 180.0;
    if (fabs (bmaj - bmin) < 1.0) bpa = 0.0;
    /*  Draw the ellipse now that parameters are in pixel canvas  */
    kwin_draw_arc (pixcanvas, cx, cy, bmaj, bmin, 0, 360, bpa,
		   pixel_value, FALSE);
    if (!name) return;
    if ( ( string = kwin_hershey_get_string (object, bpa, 12.0,
					     &width, &height, NULL, NULL) )
	 == NULL ) return;
    x = cos (bpa * PION180 + PI_ON_2) * (bmin + height);
    y = sin (bpa * PION180 + PI_ON_2) * (bmin + height);
    x -= cos (bpa * PION180) * width * 0.5;
    y -= sin (bpa * PION180) * width * 0.5;
    kwin_hershey_draw_string (pixcanvas, string, cx + x, cy + y, 12.0,
			      pixel_value);
}   /*  End Function draw_beam  */

static flag _ViewDatasets_scatplot (ViewDatasetsWidget top,
				    double lx[2], double ly[2])
/*  [SUMMARY] Compute a scatter plot.
    <top> The ViewDatasets widget.
    <lx> The X corners, in linear co-ordinates.
    <ly> The Y corners, in linear co-ordinates.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KDisplayDataWindow window;
    KDisplayDataObject set_a, set_b;
    KDisplayDataBlinkEntry be_b;
    ViewableImage vimage_a, vimage_b;
    KScatterPlot plot;
    KwcsAstro ap_a, ap_b;
    KWorldCanvas wc;
    iarray arr_a, arr_b;
    unsigned int hdim_a, vdim_a, hdim_b, vdim_b, elem_index_a, elem_index_b;
    unsigned int xstart, xlen, ystart, ylen, num_restr_a, num_restr_b;
    double min_a = TOOBIG;
    double max_a = -TOOBIG;
    double min_b = TOOBIG;
    double max_b = -TOOBIG;
    double scale_a, offset_a, scale_b, offset_b;
    Bool auto_range;
    dim_desc *hd_a, *vd_a;
    array_desc *arr_desc_a, *arr_desc_b;
    char *name_a, *name_b;
    char **restr_names_a, **restr_names_b;
    double *restr_values_a, *restr_values_b;
    CONST char *array_a, *array_b;
    char txt[STRING_LENGTH];
    /*static char function_name[] = "ViewDatasetsWidget::scatplot";*/

    window = XkwImageDisplayGetWindow (top->viewDatasets.image_display);
    XtVaGetValues (top->viewDatasets.image_display,
		   XkwNvisibleCanvas, &wc,
		   NULL);
    dispdata_get_window_attributes
	(window,
	 KDISPLAYDATA_WINDOW_ATT_IMAGE, &set_a,
	 KDISPLAYDATA_WINDOW_ATT_END);
    if (!set_a) return (FALSE);
    dispdata_get_data_attributes (set_a,
				  KDISPLAYDATA_DATA_ATT_PSEUDO_ARRAY, &arr_a,
				  KDISPLAYDATA_DATA_ATT_NAME, &name_a,
				  KDISPLAYDATA_DATA_ATT_END);
    if (!arr_a) return (FALSE);
    be_b = XkwDataBrowserGetAlternate
	( dispdata_get_active_blinkstate (window, NULL) );
    if (!be_b) return (FALSE);
    set_b = dispdata_get_dataobject_from_blinkentry (be_b);
    dispdata_get_data_attributes (set_b,
				  KDISPLAYDATA_DATA_ATT_PSEUDO_ARRAY, &arr_b,
				  KDISPLAYDATA_DATA_ATT_NAME, &name_b,
				  KDISPLAYDATA_DATA_ATT_END);
    vimage_a = viewimg_get_active (wc);
    viewimg_get_attributes (vimage_a,
			    VIEWIMG_VATT_ARRAY_DESC, &arr_desc_a,
			    VIEWIMG_VATT_SLICE, &array_a,
			    VIEWIMG_VATT_HDIM, &hdim_a,
			    VIEWIMG_VATT_VDIM, &vdim_a,
			    VIEWIMG_VATT_PSEUDO_INDEX, &elem_index_a,
			    VIEWIMG_VATT_NUM_RESTRICTIONS, &num_restr_a,
			    VIEWIMG_VATT_RESTRICTION_NAMES, &restr_names_a,
			    VIEWIMG_VATT_RESTRICTION_VALUES, &restr_values_a,
			    VIEWIMG_VATT_END);
    hd_a = arr_desc_a->dimensions[hdim_a];
    vd_a = arr_desc_a->dimensions[vdim_a];
    /*  Convert to image pixel co-ordinates in image A  */
    ds_convert_coordinates (hd_a, 2, lx, TRUE, TRUE, TRUE);
    ds_convert_coordinates (vd_a, 2, ly, TRUE, TRUE, TRUE);
    array_a += ds_get_element_offset (arr_desc_a->packet, elem_index_a);
    if (lx[0] < lx[1])
    {
	xstart = lx[0];
	xlen = lx[1] - lx[0] + 1.0;
    }
    else
    {
	xstart = lx[1];
	xlen = lx[0] - lx[1] + 1.0;
    }
    if (ly[0] < ly[1])
    {
	ystart = ly[0];
	ylen = ly[1] - ly[0] + 1.0;
    }
    else
    {
	ystart = ly[1];
	ylen = ly[0] - ly[1] + 1.0;
    }
    dispdata_get_blinkentry_attributes
	(be_b,
	 KDISPLAYDATA_BLINKENTRY_ATT_VIMAGE, &vimage_b,
	 KDISPLAYDATA_BLINKENTRY_ATT_END);
    viewimg_get_attributes (vimage_b,
			    VIEWIMG_VATT_ARRAY_DESC, &arr_desc_b,
			    VIEWIMG_VATT_SLICE, &array_b,
			    VIEWIMG_VATT_HDIM, &hdim_b,
			    VIEWIMG_VATT_VDIM, &vdim_b,
			    VIEWIMG_VATT_PSEUDO_INDEX, &elem_index_b,
			    VIEWIMG_VATT_NUM_RESTRICTIONS, &num_restr_b,
			    VIEWIMG_VATT_RESTRICTION_NAMES, &restr_names_b,
			    VIEWIMG_VATT_RESTRICTION_VALUES, &restr_values_b,
			    VIEWIMG_VATT_END);
    array_b += ds_get_element_offset (arr_desc_b->packet, elem_index_b);
    ap_a = iarray_get_associated_wcs_astro (arr_a);
    ap_b = iarray_get_associated_wcs_astro (arr_b);
    iarray_get_data_scaling (arr_a, &scale_a, &offset_a);
    iarray_get_data_scaling (arr_b, &scale_b, &offset_b);
    _ViewDatasets_create_scatter_plot_shell (top);
    plot = scatplt_create_vv (top->viewDatasets.scatter_plot_canvas,
			      array_a, arr_desc_a, hdim_a, vdim_a,
			      arr_desc_a->packet->element_types[elem_index_a],
			      scale_a, offset_a, ap_a, num_restr_a,
			      (CONST char **) restr_names_a, restr_values_a,
			      &min_a, &max_a,
			      xstart, xlen, ystart, ylen,
			      array_b, arr_desc_b, hdim_b, vdim_b,
			      arr_desc_b->packet->element_types[elem_index_b],
			      scale_b, offset_b, ap_b, num_restr_b,
			      (CONST char **) restr_names_b, restr_values_b,
			      &min_b, &max_b);
    if (plot == NULL)
    {
	fprintf (stderr, "Unable to create scatter plot\n");
	return (TRUE);
    }
    XtPopup ( (Widget) top->viewDatasets.scatter_plot_shell, XtGrabNone );
    c_unregister_callback2 (top->viewDatasets.scatter_plot_x_handle.cbk);
    c_unregister_callback2 (top->viewDatasets.scatter_plot_y_handle.cbk);
    top->viewDatasets.scatter_plot_xstart = xstart;
    top->viewDatasets.scatter_plot_xlen = xlen;
    top->viewDatasets.scatter_plot_ystart = ystart;
    top->viewDatasets.scatter_plot_ylen = ylen;
    top->viewDatasets.scatter_plot_limits.soft.left_x = min_a;
    top->viewDatasets.scatter_plot_limits.soft.right_x = max_a;
    top->viewDatasets.scatter_plot_limits.soft.bottom_y = min_b;
    top->viewDatasets.scatter_plot_limits.soft.top_y = max_b;
    top->viewDatasets.scatter_plot_x_handle.vimage = vimage_a;
    top->viewDatasets.scatter_plot_y_handle.vimage = vimage_b;
    top->viewDatasets.scatter_plot_x_handle.cbk =
	viewimg_register_destroy_func(vimage_a, c_write_2nulls_func,
				      &top->viewDatasets.scatter_plot_x_handle);
    top->viewDatasets.scatter_plot_y_handle.cbk =
	viewimg_register_destroy_func(vimage_b, c_write_2nulls_func,
				      &top->viewDatasets.scatter_plot_y_handle);
    XtVaGetValues (top->viewDatasets.scatter_plot_auto_tgl,
		   XtNstate, &auto_range,
		   NULL);
    if (auto_range)
    {
	canvas_set_attributes (top->viewDatasets.scatter_plot_canvas,
			       CANVAS_ATT_LEFT_X, min_a,
			       CANVAS_ATT_RIGHT_X, max_a,
			       CANVAS_ATT_BOTTOM_Y, min_b,
			       CANVAS_ATT_TOP_Y, max_b,
			       CANVAS_ATT_END);
    }
    canvas_specify (top->viewDatasets.scatter_plot_canvas,
		    iarray_value_name (arr_a), iarray_value_name (arr_b),
		    0, NULL, NULL);
    sprintf (txt, "X: %s, Y: %s", name_a, name_b);
    canvas_set_dressing (top->viewDatasets.scatter_plot_canvas,
			 KCD_ATT_TITLE, txt,
			 KCD_ATT_END);
    scatplt_set_active (plot, TRUE, FALSE, TRUE, FALSE, 0);
    top->viewDatasets.scatter_plot = plot;
    return (TRUE);
}   /*  End Function scatplot  */

static flag _ViewDatasets_left_geom_func (KPixCanvasEditorContext context,
					  KPixCanvas canvas,
					  unsigned int object_type,
					  unsigned int num_coords,
					  CONST int *x, CONST int *y,
					  unsigned int raw_event_code,
					  void *raw_e_info, void **f_info)
/*  [SUMMARY] Process a geometry editor event on a pixel canvas.
    <context> The KPixCanvasEditorContext.
    <canvas> The pixel canvas on which the event occurred.
    <object_type> The type of the geometric object.
    <num_coords> The number of co-ordinates in the geometric object.
    <x> The array of horizontal co-ordinates for the geometric object.
    <y> The array of vertical co-ordinates for the geometric object.
    <raw_event_code> The raw arbitrary event code.
    <raw_e_info> A pointer to the raw arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KWorldCanvas worldcanvas;
    ViewDatasetsWidget top = (ViewDatasetsWidget) *f_info;
    unsigned long mask;
    CONST char *string = raw_e_info;
    double lx[2], ly[2];
    /*static char function_name[] = "ViewDatasetsWidget::left_geom_func";*/

    if (raw_event_code != K_CANVAS_EVENT_MODIFIED_KEY_PRESS) return (FALSE);
    m_copy (&mask, string + strlen (string) + 1, sizeof mask);
    if (mask != K_CANVAS_EVENT_LEFT_MOUSE_MASK) return (FALSE);
    if (strcmp (string, "c") != 0) return (FALSE);
    XtVaGetValues (top->viewDatasets.image_display,
		   XkwNvisibleCanvas, &worldcanvas,
		   NULL);
    lx[0] = x[0];
    ly[0] = y[0];
    lx[1] = x[1];
    ly[1] = y[1];
    canvas_convert_to_canvas_coords (worldcanvas,
				     TRUE, 2, lx, ly, lx, ly, NULL, NULL);
    return _ViewDatasets_scatplot (top, lx, ly);
}   /*  End Function left_geom_func  */

static void create_tracewinpopup (ViewDatasetsWidget top)
/*  [SUMMARY] Create the tracewinpopup widget.
    <top> The ViewDatasets widget.
    [RETURNS] Nothing.
*/
{
    KDisplayDataWindow window;
    KWorldCanvas canvas;
    Kcolourmap kcmap;
    Widget w, pc_cnv, dc_cnv, tc_cnv;
    Widget TOP = (Widget) top;
    Visual *my_visual, *pc_visual, *dc_visual, *tc_visual;

    if (top->viewDatasets.trace_winpopup) return;
    XtVaGetValues (TOP , XtNvisual, &my_visual, NULL);
    if (!my_visual) my_visual = DefaultVisualOfScreen ( XtScreen (TOP) );
    pc_cnv = XtNameToWidget (top->viewDatasets.image_display,
			     "multiCanvas.pseudoColourCanvas");
    dc_cnv = XtNameToWidget (top->viewDatasets.image_display,
			     "multiCanvas.directColourCanvas");
    tc_cnv = XtNameToWidget (top->viewDatasets.image_display,
			     "multiCanvas.trueColourCanvas");
    if (pc_cnv) XtVaGetValues (pc_cnv, XtNvisual, &pc_visual, NULL);
    else pc_visual = NULL;
    if (dc_cnv) XtVaGetValues (dc_cnv, XtNvisual, &dc_visual, NULL);
    else dc_visual = NULL;
    if (tc_cnv) XtVaGetValues (tc_cnv, XtNvisual, &tc_visual, NULL);
    else tc_visual = NULL;
    if (my_visual == pc_visual)
	XtVaGetValues (top->viewDatasets.image_display,
		       XkwNrenzoPseudoCmap, &kcmap,
		       NULL);
    else if (my_visual == dc_visual) kcmap = NULL;
    else XtVaGetValues (top->viewDatasets.image_display,
			XkwNrenzoTrueCmap, &kcmap,
			NULL);
    w = XtVaCreatePopupShell ("tracewinpopup", tracePopupWidgetClass,
			      (Widget) top,
			      XkwNkarmaColourmap, kcmap,
			      NULL);
    top->viewDatasets.trace_winpopup = w;
    _ViewDatasets_renumber_func (top);
    XtRealizeWidget (w);
    XtVaGetValues (w, XkwNworldCanvas, &canvas, NULL);
    top->viewDatasets.trace_canvas = canvas;
    canvas_register_position_event_func (canvas,
					 _ViewDatasets_trace_position_func,
					 top);
    window = XkwImageDisplayGetWindow (top->viewDatasets.image_display);
    dispdata_set_window_attributes (window,
				    KDISPLAYDATA_WINDOW_ATT_PROFILE_CANVAS,
				        canvas,
				    KDISPLAYDATA_WINDOW_ATT_END);
}   /*  End Function create_tracewinpopup  */

static flag _ViewDatasets_trace_position_func (KWorldCanvas canvas,
					       double x, double y,
					       unsigned int event_code,
					       void *e_info,
					       void **f_info,
					       double x_lin, double y_lin,
					       int x_pix, int y_pix)
/*  [SUMMARY] Position event callback for the trace (profile) canvas.
    [PURPOSE] This routine is a position event consumer for a world canvas.
    <canvas> The canvas on which the event occurred.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    <event_code> The arbitrary event code.
    <e_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    <x_lin> The horizontal linear world co-ordinate prior to the transform
    function being called.
    <y_lin> The vertical linear world co-ordinate prior to the transform
    function being called.
    <x_pix> The horizontal pixel co-ordinate.
    <y_pix> The vertical pixel co-ordinate.
    [NOTE] The world co-ordinate values will have been transformed by the
    registered transform function (see [<canvas_register_transform_func>]).
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KWorldCanvas image_worldcanvas;
    unsigned int trace_num_restr, count;
    double image_x = TOOBIG;
    double image_y = TOOBIG;
    ViewDatasetsWidget top = (ViewDatasetsWidget) *f_info;
    char *trace_xlabel, *image_xlabel, *image_ylabel;
    char **trace_restr_names;
    double *trace_restr_values;

    if (top->viewDatasets.profile_mode != KDISPLAYDATA_PROFILE_MODE_LINE)
	return (FALSE);
    XtVaGetValues (top->viewDatasets.image_display,
		   XkwNvisibleCanvas, &image_worldcanvas,
		   NULL);
    if (event_code == K_CANVAS_EVENT_POINTER_LEAVE)
    {
	_ViewDatasets_draw_marker (image_worldcanvas,
				   &top->viewDatasets.main_marker_area,
				   TOOBIG, TOOBIG);
	return (FALSE);
    }
    canvas_get_specification (canvas, &trace_xlabel, NULL, &trace_num_restr,
			      &trace_restr_names, &trace_restr_values);
    canvas_get_specification (image_worldcanvas, &image_xlabel, &image_ylabel,
			      NULL, NULL, NULL);
    if (!trace_xlabel || !image_xlabel) return (FALSE);
    if (strcmp (trace_xlabel, image_xlabel) == 0) image_x = x_lin;
    else if (strcmp (trace_xlabel, image_ylabel) == 0) image_y = x_lin;
    for (count = 0; count < trace_num_restr; ++count)
    {
	if (strcmp (trace_restr_names[count], image_xlabel) == 0)
	    image_x = trace_restr_values[count];
	else if (strcmp (trace_restr_names[count], image_ylabel) == 0)
	    image_y = trace_restr_values[count];
    }
    _ViewDatasets_draw_marker (image_worldcanvas,
			       &top->viewDatasets.main_marker_area,
			       image_x, image_y);
    return (FALSE);
}   /*  End Function trace_position_func  */

static void _ViewDatasets_draw_marker (KWorldCanvas canvas,
				       KPixCanvasRefreshArea *area,
				       double x, double y)
/*  [SUMMARY] Draw the marker.
    <canvas> The world canvas to draw to.
    <area> State information for undrawing the old marker. This is updated.
    <x> The horizontal linear co-ordinate of the marker.
    <y> The vertical linear co-ordinate of the marker.
    [RETURNS] Nothing.
*/
{
    KPixCanvas pixcanvas;
    unsigned long pixel_value = 0;

    if (canvas == NULL) return;
    pixcanvas = canvas_get_pixcanvas (canvas);
    if (area->startx > -100)
    {
	/*  Undraw the old circle  */
	kwin_partial_refresh (pixcanvas, 1, area, FALSE);
	area->startx = -100;
    }
    if ( DOUBLE_IS_BLANK (x) || DOUBLE_IS_BLANK (y) ) return;
    canvas_convert_from_canvas_coords (canvas, FALSE, TRUE, 1, &x, &y, &x, &y);
    /*  Setup undraw parameters for next time  */
    area->startx = x - ELLIPSE_RADIUS - 1;
    area->endx = x + ELLIPSE_RADIUS + 1;
    area->starty = y - ELLIPSE_RADIUS - 1;
    area->endy = y + ELLIPSE_RADIUS + 1;
    area->clear = TRUE;
    area->geom_clear = FALSE;
    /*  And draw this one  */
    kwin_get_colour_harder (pixcanvas, "red", &pixel_value);
    kwin_draw_ellipse (pixcanvas, x, y, ELLIPSE_RADIUS, ELLIPSE_RADIUS,
		       pixel_value);
    kwin_get_colour_harder (pixcanvas, "yellow", &pixel_value);
    kwin_draw_ellipse (pixcanvas, x, y, ELLIPSE_RADIUS - 1, ELLIPSE_RADIUS - 1,
		       pixel_value);
}   /*  End Function draw_marker  */

static void _ViewDatasets_save_ppm_movie_cbk (Widget w, XtPointer client_data,
					      XtPointer call_data)
/*  [SUMMARY] Save PPM window movie callback.
    <w> The widget the event occurred on.
    <client_data> Client data.
    <call_data> Call data.
    [RETURNS] Nothing.
*/
{
    KDisplayDataWindow window;
    KDisplayDataObject dataobject;
    int precision;
    unsigned int frame_number, num_frames, count;
    ViewDatasetsWidget top = client_data;
    CONST char *basename = call_data;
    char filename[STRING_LENGTH];

    window = XkwImageDisplayGetWindow (top->viewDatasets.image_display);
    dispdata_get_window_attributes
	(window,
	 KDISPLAYDATA_WINDOW_ATT_MOVIE_MASTER, &dataobject,
	 KDISPLAYDATA_WINDOW_ATT_END);
    if (!dataobject)
    {
	fprintf (stderr, "No frames to display\n");
	return;
    }
    dispdata_get_data_attributes
	(dataobject,
	 KDISPLAYDATA_DATA_ATT_FRAME_NUMBER, &frame_number,  /*  Save it  */
	 KDISPLAYDATA_DATA_ATT_NUM_FRAMES, &num_frames,
	 KDISPLAYDATA_DATA_ATT_DIRECTION_FOR_FRAME, top->viewDatasets.slice,
	 KDISPLAYDATA_DATA_ATT_END);
    precision = ceil ( log10 (num_frames) );
    for (count = 0; count < num_frames; ++count)
    {
	sprintf (filename, "%s.%.*u.ppm", basename, precision, count);
	dispdata_set_window_attributes
	    (window,
	     KDISPLAYDATA_WINDOW_ATT_FRAME_NUMBER, count,
	     KDISPLAYDATA_DATA_ATT_END);
	ExportMenuWritePPMWindow (w, filename);
    }
    /*  Put display back where it was  */
    dispdata_set_window_attributes
	(window,
	 KDISPLAYDATA_WINDOW_ATT_FRAME_NUMBER, frame_number,
	 KDISPLAYDATA_DATA_ATT_END);
}   /*  End Function save_ppm_movie_cbk  */

static flag _ViewDatasets_scatter_position_func (KWorldCanvas canvas,
						 double x, double y,
						 unsigned int event_code,
						 void *e_info, void **f_info,
						 double x_lin, double y_lin,
						 int x_pix, int y_pix)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas.
    <canvas> The canvas on which the event occurred.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    <event_code> The arbitrary event code.
    <e_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    <x_lin> The horizontal linear world co-ordinate prior to the transform
    function being called.
    <y_lin> The vertical linear world co-ordinate prior to the transform
    function being called.
    <x_pix> The horizontal pixel co-ordinate.
    <y_pix> The vertical pixel co-ordinate.
    [NOTE] The world co-ordinate values will have been transformed by the
    registered transform function (see [<canvas_register_transforms_func>]).
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    Widget lbl = *f_info;
    char txt[STRING_LENGTH];

    canvas_track_compute (canvas, x, y, x_lin, y_lin, txt, NULL);
    XtVaSetValues (lbl, XtNlabel, txt, NULL);
    return (FALSE);
}   /*  End Function scatter_position_func  */

static flag _ViewDatasets_scatter_cursor_func (KScatterPlot plot,
					       double x, double y,
					       uaddr index, void **info)
/*  [SUMMARY] Process a smart cursor event on a scatter plot.
    <plot> The KScatterPlot object the cursor event matched.
    <x> The X co-ordinate of the point matched.
    <y> The Y co-ordinate of the point matched.
    <index> The index of the point matched.
    <info> A pointer to the arbitrary information pointer.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas main_wc;
    KwcsAstro ap = NULL;
    unsigned int xpos, ypos, hdim, vdim, num_restr;
    ViewDatasetsWidget top = *info;
    array_desc *arr_desc;
    dim_desc *hd, *vd;
    char **restr_names;
    double *restr_values;
    char txt[STRING_LENGTH];

    if (!top->viewDatasets.scatter_plot_x_handle.vimage) return (FALSE);
    XtVaGetValues (top->viewDatasets.image_display,
		   XkwNvisibleCanvas, &main_wc,
		   NULL);
    if ( DOUBLE_IS_BLANK (x) || DOUBLE_IS_BLANK (y) )
    {
	_ViewDatasets_draw_marker (main_wc,
				   &top->viewDatasets.main_marker_area,
				   TOOBIG, TOOBIG);
	XtVaSetValues (top->viewDatasets.scatter_plot_track_lbl2,
		       XtNlabel, "",
		       NULL);
	return (FALSE);
    }
    canvas_track_compute (scatplt_get_worldcanvas (plot), x, y, x, y, txt,
			  NULL);
    XtVaSetValues (top->viewDatasets.scatter_plot_track_lbl2,
		   XtNlabel, txt,
		   NULL);
    ypos = index / top->viewDatasets.scatter_plot_xlen;
    ypos += top->viewDatasets.scatter_plot_ystart;
    xpos = index % top->viewDatasets.scatter_plot_xlen;
    xpos += top->viewDatasets.scatter_plot_xstart;
    viewimg_get_attributes (top->viewDatasets.scatter_plot_x_handle.vimage,
			    VIEWIMG_VATT_ARRAY_DESC, &arr_desc,
			    VIEWIMG_VATT_WCS_ASTRO, &ap,
			    VIEWIMG_VATT_HDIM, &hdim,
			    VIEWIMG_VATT_VDIM, &vdim,
			    VIEWIMG_VATT_NUM_RESTRICTIONS, &num_restr,
			    VIEWIMG_VATT_RESTRICTION_NAMES, &restr_names,
			    VIEWIMG_VATT_RESTRICTION_VALUES, &restr_values,
			    VIEWIMG_VATT_END);
    hd = arr_desc->dimensions[hdim];
    vd = arr_desc->dimensions[vdim];
    x = xpos;
    y = ypos;
    ds_convert_coordinates (hd, 1, &x, FALSE, FALSE, FALSE);
    ds_convert_coordinates (vd, 1, &y, FALSE, FALSE, FALSE);
    if (viewimg_get_active (main_wc)
	!= top->viewDatasets.scatter_plot_x_handle.vimage)
    {   /*  Convert linear co-ordinates to linear co-ordinates for main canvas
	 */
	if (ap) wcs_astro_transform3 (ap, FALSE, 1,
				      hd->name, &x, FALSE,
				      vd->name, &y, FALSE,
				      NULL, NULL, FALSE,
				      num_restr, (CONST char **) restr_names,
				      restr_values);
	canvas_coords_transform (main_wc, 1, &x, TRUE, &y, TRUE);
    }
    _ViewDatasets_draw_marker (main_wc, &top->viewDatasets.main_marker_area,
			       x, y);
    return (FALSE);
}   /*  End Function scatter_cursor_func  */

static flag _ViewDatasets_draw_marker_in_scatter_plot (ViewDatasetsWidget top,
						       ViewableImage vimage_a,
						       double x, double y,
						       double value_a)
/*  [SUMMARY] Draw marker in scatter plot.
    <top> The ViewDatasets widget.
    <vimage_a> The ViewableImage for the event.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    <value_a> The value from the image.
    [RETURNS] TRUE on success, else FALSE, indicating that the old marker
    should be cleared.
*/
{
    ViewableImage vimage_b;
    KwcsAstro ap_b;
    unsigned int hdim_b, vdim_b, elem_index_b, num_restr_b;
    double scale, offset;
    array_desc *arr_desc_b;
    dim_desc *hd_b, *vd_b;
    CONST char *array_b;
    char **restr_names_b;
    double *restr_values_b;
    char txt[STRING_LENGTH];
    double values[2];

    if ( DOUBLE_IS_BLANK (value_a) ) return (FALSE);
    if (top->viewDatasets.scatter_plot == NULL) return (FALSE);
    viewimg_get_attributes (vimage_a,
			    VIEWIMG_VATT_DATA_SCALE, &scale,
			    VIEWIMG_VATT_DATA_OFFSET, &offset,
			    VIEWIMG_VATT_END);
    value_a = value_a * scale + offset;
    if (top->viewDatasets.scatter_plot_x_handle.vimage == vimage_a)
	vimage_b = top->viewDatasets.scatter_plot_y_handle.vimage;
    else vimage_b = top->viewDatasets.scatter_plot_x_handle.vimage;
    viewimg_get_attributes (vimage_b,
			    VIEWIMG_VATT_ARRAY_DESC, &arr_desc_b,
			    VIEWIMG_VATT_SLICE, &array_b,
			    VIEWIMG_VATT_HDIM, &hdim_b,
			    VIEWIMG_VATT_VDIM, &vdim_b,
			    VIEWIMG_VATT_PSEUDO_INDEX, &elem_index_b,
			    VIEWIMG_VATT_DATA_SCALE, &scale,
			    VIEWIMG_VATT_DATA_OFFSET, &offset,
			    VIEWIMG_VATT_WCS_ASTRO, &ap_b,
			    VIEWIMG_VATT_NUM_RESTRICTIONS, &num_restr_b,
			    VIEWIMG_VATT_RESTRICTION_NAMES, &restr_names_b,
			    VIEWIMG_VATT_RESTRICTION_VALUES, &restr_values_b,
			    VIEWIMG_VATT_END);
    hd_b = arr_desc_b->dimensions[hdim_b];
    vd_b = arr_desc_b->dimensions[vdim_b];
    /*  Convert world co-ordinates to image pixel co-ordinates in other set  */
    if (ap_b) wcs_astro_transform3 (ap_b, FALSE, 1,
				    hd_b->name, &x, TRUE,
				    vd_b->name, &y, TRUE,
				    NULL, NULL, FALSE,
				    num_restr_b, (CONST char **) restr_names_b,
				    restr_values_b);
    ds_convert_coordinates (hd_b, 1, &x, TRUE, TRUE, TRUE);
    ds_convert_coordinates (vd_b, 1, &y, TRUE, TRUE, TRUE);
    if ( DOUBLE_IS_BLANK (x) || DOUBLE_IS_BLANK (y) ) return (FALSE);
    /* If vimage is original scatterplot Y vimage, now ready to check limits */
    if ( (top->viewDatasets.scatter_plot_y_handle.vimage == vimage_a) &&
	 ( (x < top->viewDatasets.scatter_plot_xstart) ||
	   (x >= top->viewDatasets.scatter_plot_xstart +
	    top->viewDatasets.scatter_plot_xlen) ||
	   (y < top->viewDatasets.scatter_plot_ystart) ||
	   (y >= top->viewDatasets.scatter_plot_ystart +
	    top->viewDatasets.scatter_plot_ylen) ) )
	return (FALSE);
    ds_get_element (array_b +
		    ds_get_element_offset (arr_desc_b->packet, elem_index_b) +
		    arr_desc_b->offsets[vdim_b][(long) y] +
		    arr_desc_b->offsets[hdim_b][(long) x],
		    arr_desc_b->packet->element_types[elem_index_b],
		    values, NULL);
    if ( DOUBLE_IS_BLANK (values[0]) ) return (FALSE);
    values[0] = values[0] * scale + offset;
    if (top->viewDatasets.scatter_plot_x_handle.vimage == vimage_a)
    {
	x = value_a;
	y = values[0];
    }
    else
    {
	x = values[0];
	y = value_a;
    }
    _ViewDatasets_draw_marker (top->viewDatasets.scatter_plot_canvas,
			       &top->viewDatasets.scatter_plot_marker_area,
			       x, y);
    canvas_track_compute (top->viewDatasets.scatter_plot_canvas,
			  x, y, x, y, txt, NULL);
    XtVaSetValues (top->viewDatasets.scatter_plot_track_lbl2,
		   XtNlabel, txt,
		   NULL);
    return (TRUE);
}   /*  End Function draw_marker_in_scatter_plot  */

static void _ViewDatasets_square_scatter_cbk (Widget w, XtPointer client_data,
					      XtPointer call_data)
/*  [SUMMARY] Make the scatter plot square.
    <w> The widget the event occurred on.
    <client_data> Client data.
    <call_data> Call data.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas canvas = client_data;
    double left_x, right_x, bottom_y, top_y;

    canvas_get_attributes (canvas,
			   CANVAS_ATT_LEFT_X, &left_x,
			   CANVAS_ATT_RIGHT_X, &right_x,
			   CANVAS_ATT_BOTTOM_Y, &bottom_y,
			   CANVAS_ATT_TOP_Y, &top_y,
			   CANVAS_ATT_END);
    if (left_x < bottom_y) bottom_y = left_x;
    else if (bottom_y < left_x) left_x = bottom_y;
    if (right_x > top_y) top_y = right_x;
    else if (top_y > right_x) right_x = top_y;
    canvas_set_attributes (canvas,
			   CANVAS_ATT_LEFT_X, left_x,
			   CANVAS_ATT_RIGHT_X, right_x,
			   CANVAS_ATT_BOTTOM_Y, bottom_y,
			   CANVAS_ATT_TOP_Y, top_y,
			   CANVAS_ATT_END);
    canvas_refresh_if_visible (canvas, FALSE);
}   /*  End Function square_scatter_cbk  */

static void _ViewDatasets_listen_func (ViewDatasetsWidget w,void *client1_data,
				       struct cursor_data *data,
				       void *client2_data)
/*  [SUMMARY] Broadcast cursor listen callback.
    <w> The ViewDatasets widget receiving the event.
    <data> The event data.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas wc;
    unsigned int event_code;
    double x, y;

    if (data->source == w) return;
    if (!w->viewDatasets.cursor_tracking.listen) return;
    XtVaGetValues (w->viewDatasets.image_display, XkwNvisibleCanvas, &wc,NULL);
    x = data->wx;
    y = data->wy;
    if ( DOUBLE_IS_VALID (x) && DOUBLE_IS_VALID (y) )
    {
	KwcsAstro ap;
	flag ok = FALSE;
	XkwViewDatasetsCoordConvertData ext_data;
	unsigned int num_restr;
	double tmp;
	char *xname, *yname;
	char **restr_names;
	double *restr_values;

	canvas_get_specification (wc, &xname, &yname, &num_restr, &restr_names,
				  &restr_values);
	canvas_get_attributes (wc, CANVAS_ATT_WCS_ASTRO, &ap, CANVAS_ATT_END);
	ext_data.image_display = w->viewDatasets.image_display;
	ext_data.inp_xname = data->xname;
	ext_data.inp_yname = data->yname;
	ext_data.inp_zname = data->zname;
	ext_data.out_xname = xname;
	ext_data.out_yname = yname;
	ext_data.out_zname = (num_restr > 0) ? restr_names[0] : NULL;
	ext_data.x = x;
	ext_data.y = y;
	ext_data.z = data->wz;
	if ( c_call_callbacks (coord_convert_list, &ext_data) )
	{
	    ok = TRUE;
	    x = ext_data.x;
	    y = ext_data.y;
	}
	else if ( wcs_astro_compare_axes (data->ap, data->xname, ap, xname) &&
		  wcs_astro_compare_axes (data->ap, data->yname, ap, yname) )
	    ok = TRUE;
	else if ( wcs_astro_compare_axes (data->ap, data->xname, ap, yname) &&
		  wcs_astro_compare_axes (data->ap, data->yname, ap, xname) )
	{
	    swap (x, y, tmp);
	    ok = TRUE;
	}
	else x = y = TOOBIG;
	if (ok) canvas_coords_transform (wc, 1, &x, TRUE, &y, TRUE);
    }
    if ( DOUBLE_IS_VALID (x) && DOUBLE_IS_VALID (y) )
    {
	event_code = K_CANVAS_EVENT_POINTER_MOVE;
	w->viewDatasets.cursor_tracking.last_was_inside = TRUE;
    }
    else if (!w->viewDatasets.cursor_tracking.last_was_inside) return;
    else
    {
	event_code = K_CANVAS_EVENT_POINTER_LEAVE;
	w->viewDatasets.cursor_tracking.last_was_inside = FALSE;
    }
    _ViewDatasets_draw_marker (wc, &w->viewDatasets.main_marker_area, x, y);
    canvas_convert_from_canvas_coords (wc, FALSE, TRUE, 1, &x, &y, &x, &y);
    kwin_process_position_event (canvas_get_pixcanvas (wc), x, y, FALSE,
				 event_code, NULL);
}   /*  End Function listen_func  */


/*  Public functions follow  */
