/*LINTLIBRARY*/
/*  array.c

    This code provides array-type DisplayData objects.

    Copyright (C) 2002-2004  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for manipulating array-type display
    data.


    Written by      Richard Gooch   12-NOV-2002

    Updated by      Richard Gooch   23-NOV-2002: First cut of code. Compiles.

    Updated by      Richard Gooch   11-DEC-2002: Second cut of code. Tested
  with new <kvis> application.

    Updated by      Richard Gooch   14-DEC-2002: Change raw min&max of canvas
  when <set_iscale> is FALSE, so that scaled min&max do not change.

    Updated by      Richard Gooch   20-DEC-2002: Added selective refresh in
  <_dispdata_array_set_blinkentry_attributes>. Trap for setting Renzogram mode
  when no colourmap(s) available.

    Updated by      Richard Gooch   24-DEC-2002: Supported multiple general
  data structures, each with an array.

    Updated by      Richard Gooch   29-DEC-2002: Added basic profile support.

    Updated by      Richard Gooch   2-JAN-2002: Added boxed sum/average profile
  support.

    Updated by      Richard Gooch   18-JAN-2002: Schedule profile unzoom when
  mode or direction are changed.

    Updated by      Richard Gooch   20-JAN-2002: Added boxed horizontal and
  vertical profiles and radial profiles, including Gaussian fitting.

    Updated by      Richard Gooch   24-JAN-2003: Took account of API change to
  KDisplayDataFuncActivateBlinkEntry. Added KDisplayDataFuncCopyDataAttributes
  hook.

    Updated by      Richard Gooch   27-JAN-2003: Added return flag for all
  get/set hooks.

    Updated by      Richard Gooch   8-FEB-2003: Added sanity checks for
  intensity range in <set_data_attributes>. Fixed off-by-one error in
  <set_matching_frame> caused by unnecessary call to <ds_get_coord_num>. Fixed
  tiny memory leak in <data_event_func> and don't terminate early if a general
  data structure has no array.

    Updated by      Richard Gooch   17-FEB-2003: Added auto title feature.

    Updated by      Richard Gooch   20-FEB-2003: First cut of working Renzogram
  support code. Use of colourmap for profiles not supported. Might never be.

    Updated by      Richard Gooch   22-FEB-2003: Added channels parsing code
  for Renzograms.

    Updated by      Richard Gooch   2-MAR-2003: Improved radial profile output
  to show sum (raw and background-subtracted).

    Updated by      Richard Gooch   16-MAR-2003: Fixed bug in
  <set_matching_frame> where frame0 was not reachable under some conditions.

    Updated by      Richard Gooch   21-MAR-2003: Fixed bug in
  <copy_data_attributes>: frame number needs to be checked against num_frames.

    Updated by      Richard Gooch   23-APR-2003: Fixed bug with auto title
  feature: title was not being updated when viewable destroyed.

    Updated by      Richard Gooch   2-MAY-2003: Finished adding profile marker
  support. Each dataset has its own marker. Mouse position can change marker
  position if profile axis is aligned along a visible axis.

    Updated by      Richard Gooch   11-JUN-2003: Added display of flux in
  radial profile output, if possible.

    Updated by      Richard Gooch   16-JUN-2003: Fixed bug with profile display
  when unloading master: canvas could be left with dangling KwcsAstro object.
  Added call to <canvas_use_astro_transform> in <destroy_blinkentry> to fix it.

    Updated by      Richard Gooch   9-JUL-2003: Fixed bug introduced by
  previous fix: did not check if profile canvas was NULL.

    Updated by      Richard Gooch   3-AUG-2003: Split image and profile ranges
  so that they may be changed independently.

    Updated by      Richard Gooch   5-AUG-2003: Created experimental function
  <dispdata_array_register_irange_change_func>.

    Updated by      Richard Gooch   6-AUG-2003: Made use of new experimental
  function <c_unregister_callback2>.

    Updated by      Richard Gooch   13-AUG-2003: Made use of new experimental
  functions <viewimg_create_rgb2> and <viewimg_create_rgb_sequence2>.

    Updated by      Richard Gooch   27-AUG-2003: Made use of new experimental
  function <wcs_astro_compare_axes>.

    Updated by      Richard Gooch   10-SEP-2003: Added return value to
  <destroy_blinkentry>.

    Updated by      Richard Gooch   26-SEP-2003: Took account of negative scale
  in <set_canvas_iscale>.

    Updated by      Richard Gooch   2-OCT-2003: Took account of negative scale
  when setting and getting raw clip limits.

    Updated by      Richard Gooch   6-OCT-2003: Fixed bug in function
  <_dispdata_array_collect_radial_points>: was prematurely refreshing profile
  canvas in call to <graph1d_set_active>. Not required as canvas is refreshed
  later, after canvas is completely configured.

    Updated by      Richard Gooch   31-DEC-2003: Added azimuthally averaged
  profile support.

    Updated by      Richard Gooch   2-JAN-2004: Added missing newline to
  radial profile output.

    Updated by      Richard Gooch   17-MAR-2004: Added temporary hack to split
  4-dimensional arrays into separate 3-dimensional arrays.

    Updated by      Richard Gooch   1-MAY-2004: Took account of API change in
  <contour> package.

    Updated by      Richard Gooch   2-MAY-2004: Made use of new grouping
  functions in <contour> package.

    Updated by      Richard Gooch   6-MAY-2004: Enable/disable colourmap for
  line profile of array shown as renzogram.

    Updated by      Richard Gooch   16-MAY-2004: Passed reference position to
  <graph1d_orthogonal_show_trace> and <graph1d_orthogonal_show_boxed_trace> so
  that chromatic aberration can be corrected.

    Updated by      Richard Gooch   18-MAY-2004: Added support for slaving
  line profile H-range to main canvas.

    Updated by      Richard Gooch   21-MAY-2004: Only enable correction for
  chromatic aberration when line profile is perpendicular to screen.

    Updated by      Richard Gooch   31-MAY-2004: Select a new image viewable in
  <_dispdata_array_destroy_blinkentry>. Fixed bug in function
  <dispdata_array_clone_data>: was not creating ContourImage object(s), which
  is now required due to <contour> package API change.

    Updated by      Richard Gooch   3-JUN-2004: Make viewable the image
  viewable in <_dispdata_array_create_blinkentry> if one not yet set.

    Updated by      Richard Gooch   20-JUN-2004: Changed from
  KDISPLAYDATA_DATA_ATT_END to KDISPLAYDATA_WINDOW_ATT_END in
  <_dispdata_array_{sg}et_window_attributes>. Since both are currently the same
  this did not cause any problems.

    Updated by      Richard Gooch   27-JUN-2004: Set contour levels in
  <_dispdata_array_copy_data_attributes>.

    Updated by      Richard Gooch   18-JUL-2004: Created experimental function
  <dispdata_array_convert_3coord>.

    Updated by      Richard Gooch   19-JUL-2004: Finished initial
  implementation of new output format in <_dispdata_array_radial_profile>.

    Updated by      Richard Gooch   25-JUL-2004: Handle negative declinations
  for new output format in <_dispdata_array_radial_profile> and
  non-astronomical co-ordinates.

    Last updated by Richard Gooch   13-NOV-2004: Bug fix in function
  <dispdata_array_clone_data>: make sure cimages are created before data
  attributes are copied.


*/
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <karma.h>
#include <k_list.h>
#define NEW_WIN_SCALE
#include <k_event_codes.h>
#include <karma_dispdata.h>
#include <karma_viewimg.h>
#include <karma_contour.h>
#include <karma_graph1d.h>
#include <karma_canvas.h>
#include <karma_iarray.h>
#include <karma_kcmap.h>
#include <karma_kwin.h>
#include <karma_ds.h>
#include <karma_wcs.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_c.h>

#include <karma_dispdata_hooks.h>


#define MAX_CONTOUR_LEVELS  100
#define MAX_HISTOGRAM_BINS  256
#define GAUSSCURVE_LENGTH   512

struct minmax
{
    double min, max;
};

struct frame_pos
{
    unsigned int index;
    double coord;                   /*  Other axis positions are length / 2  */
};

struct data_attributes
{
    double input_lower_clip;                    /*  Scaled          */
    double input_upper_clip;                    /*  Scaled          */
    double image_lower_clip;                    /*  Scaled          */
    double image_upper_clip;                    /*  Scaled          */
    double profile_lower_clip;                  /*  Scaled          */
    double profile_upper_clip;                  /*  Scaled          */
    char *contour_neg_colour;
    char *contour_pos_colour;
    double contour_neg_linewidth;
    double contour_pos_linewidth;
    char *contour_levels_string;
    char *renzogram_levels_string;
    char *renzogram_channels_string;
    char *profile_dark_colour;
    char *profile_light_colour;
    struct frame_pos xy_frame;
    struct frame_pos xz_frame;
    struct frame_pos zy_frame;
};

struct data_container
{
    unsigned int refcount;                      /*  Number of datasets now   */
    unsigned int counter;                       /*  Number of datasets ever  */
    char *name;
    char *object;
    iarray pseudo;
    iarray red;
    iarray green;
    iarray blue;
    double scale;
    double offset;
    struct minmax raw;                          /*  Raw (unscaled)  */
    struct minmax scaled;                       /*  Scaled          */
    KwcsAstro ap;
    unsigned int num_xy_frames;
    unsigned int num_xz_frames;
    unsigned int num_zy_frames;
};

struct dataset_direction
{
    KContourImage *map_cimages;
    KContourImage *renzo_cimages;
};

struct dataset
{
    KDisplayDataObject superclass;
    struct data_container *data;
    /*  State information  */
    struct data_attributes attributes;
    KCallbackList irange_change_list;    /*  When image range changes        */
    KContourImageGroup map_cimage_group; /*  Group of normal contours        */
    KContourImageGroup renzo_cimage_group;/* Group of Renzogram contours     */
    KContourImage map_cimage;            /*  Normal contours                 */
    /*  Image caches for 3-D dataset  */
    struct dataset_direction xy;
    struct dataset_direction xz;
    struct dataset_direction zy;
    struct viewable *first;
    struct viewable *last;
};

struct canvas
{
    KWorldCanvas canvas;
    struct window *window;
};

struct window
{
    KDisplayDataWindow superclass;
    KDisplayDataCanvas pc;         /*  This might be really be RGB           */
    KDisplayDataCanvas rgb;        /*  May be NULLs if <<pc>> is really RGB  */
    KDisplayDataCanvas *best_rgb;
    KPixCanvasEditorContext ctx_pc;
    KPixCanvasEditorContext ctx_rgb;
    struct canvas pc_canvas;
    struct canvas rgb_canvas;
    KWorldCanvas profile_canvas;
    KPixCanvas profile_pixcanvas;
    struct blinkstate *active_blinkstate;
    unsigned int slice_direction;
    unsigned int profile_mode;
    unsigned int profile_direction;
    flag profile_unzoom;
    double min;                                 /*  Scaled                   */
    double max;                                 /*  Scaled                   */
    flag set_iscale;     /*  If TRUE, set canvas iscale when vimage changes  */
    flag auto_title;
    flag show_marker_in_profile;
    KCallbackFunc pc_cmap_callback;
    KCallbackFunc rgb_cmap_callback;
    flag movie_refresh;                         /*  For <<change_frame>>     */
    struct viewable *first;                     /*  List of viewables        */
    struct viewable *last;                      /*  List of viewables        */
    struct window *prev;
    struct window *next;
};

struct blinkstate
{
    KDisplayDataClassBlinkState superclass;
    struct window *window;
    struct viewable *image_viewable;
    struct viewable *movie_master;
    struct viewable *profile_master;
};

struct blinkentry
{
    struct blinkstate *blinkstate;
    struct viewable *viewable;
    unsigned char contour_mode;
    unsigned char movie_slave;
    unsigned char profile_slave;
};

struct viewable_direction
{
    ViewableImage *vimages;
    ViewableImage *mag_vimages;
    ViewableImage *pan_vimages;
    KViewableContourImage *map_pc_vcimages;
    KViewableContourImage *map_rgb_vcimages;
    KViewableContourImage *renzo_pc_vcimages;
    KViewableContourImage *renzo_rgb_vcimages;
};

struct chain_entry
{
    struct viewable *prev;
    struct viewable *next;
};

struct viewable
{
    struct dataset *dataset;
    struct window *window;
    struct blinkentry *active_blinkentry;
    /*  Image caches for 2-D dataset  */
    ViewableImage vimage;               /*  On preferred canvas for image    */
    ViewableImage mag_vimage;           /*  Magnified image on same canvas   */
    ViewableImage pan_vimage;           /*  Panner image on same canvas      */
    KViewableContourImageGroup map_pc_vcimage_group;     /*  Normal contours */
    KViewableContourImageGroup map_rgb_vcimage_group;    /*  Normal contours */
    KViewableContourImageGroup renzo_pc_vcimage_group;   /*  Renzo contours  */
    KViewableContourImageGroup renzo_rgb_vcimage_group;  /*  Renzo contours  */
    KViewableContourImage map_pc_vcimage;      /* Contours on best-PC canvas */
    KViewableContourImage map_rgb_vcimage;
    /*  Image caches for 3-D dataset  */
    struct viewable_direction xy;
    struct viewable_direction xz;
    struct viewable_direction zy;
    KOrthogonalProfile profile;
    KViewableGraph vgraph;
    KViewableGraph gausscurve;
    struct chain_entry dataset_chain;
    struct chain_entry window_chain;
};


/*  Private data  */
static KDisplayDataClass dataclass = NULL;
static unsigned int clip_mode =KDISPLAYDATA_CLIP_MODE_CHANGE_IMAGE_AND_PROFILE;
static flag slave_profile_h_zoom = FALSE;
static struct window *first_window = NULL;
static struct window *last_window = NULL;


/*  Hook functions  */
STATIC_FUNCTION (void _dispdata_array_set_class_attributes, (va_list argp) );
STATIC_FUNCTION (void _dispdata_array_destroy_dataobject, (void *dataobject) );
STATIC_FUNCTION (flag _dispdata_array_copy_data_attributes,
		 (void *dest_p, void *source_p) );
STATIC_FUNCTION (flag _dispdata_array_set_data_attributes,
		 (void *data, va_list argp) );
STATIC_FUNCTION (flag _dispdata_array_get_data_attributes,
		 (void *data, va_list argp) );
STATIC_FUNCTION (void *_dispdata_array_create_window,
		 (KDisplayDataWindow superclass,
		  KDisplayDataCanvas *pc, KDisplayDataCanvas *rgb) );
STATIC_FUNCTION (void _dispdata_array_destroy_window, (void *window_p) );
STATIC_FUNCTION (flag _dispdata_array_set_window_attributes,
		 (void *win, va_list argp) );
STATIC_FUNCTION (flag _dispdata_array_get_window_attributes,
		 (void *win, va_list argp) );
STATIC_FUNCTION (void *_dispdata_array_create_viewable,
		 (void *data, void *window_p) );
STATIC_FUNCTION (void _dispdata_array_destroy_viewable,
		 (void *viewable_p, flag canvas_destroy) );
STATIC_FUNCTION (void *_dispdata_array_create_blinkstate,
		 (KDisplayDataClassBlinkState superclass, void *window) );
STATIC_FUNCTION (void _dispdata_array_destroy_blinkstate,
		 (void *blinkstate_p) );
STATIC_FUNCTION (void _dispdata_array_activate_blinkstate,
		 (void *blinkstate_p, flag before_blinkentries) );
STATIC_FUNCTION (void *_dispdata_array_create_blinkentry,
		 (void *blinkstate, void *viewable) );
STATIC_FUNCTION (flag _dispdata_array_destroy_blinkentry,
		 (void *blinkentry_p) );
STATIC_FUNCTION (void _dispdata_array_copy_blinkentry,
		 (void *dest_p, void *source_p) );
STATIC_FUNCTION (KWorldCanvas _dispdata_array_activate_blinkentry,
		 (void *blinkentry_p, flag *visible) );
STATIC_FUNCTION (flag _dispdata_array_set_blinkentry_attributes,
		 (void *blinkentry_p, va_list argp) );
STATIC_FUNCTION (flag _dispdata_array_get_blinkentry_attributes,
		 (void *blinkentry_p, va_list argp) );


/*  Public data  */
char *dispdata_slice_choices[KDISPLAYDATA_NUM_SLICE_DIRECTIONS] =
{
    "XY", "XZ", "ZY",
};

char *dispdata_profile_mode_choices[KDISPLAYDATA_NUM_PROFILE_MODE_CHOICES] =
{
    "None", "Line", "Box Sum", "Box Average", "Radial", "Radial+Fit",
    "Azimuth Avg", "Boxed Horiz", "Boxed Vert",
};

char *dispdata_profile_dir_choices[KDISPLAYDATA_NUM_PROFILE_DIR_CHOICES] =
{
    "Z", "Y", "X",
};

char *dispdata_clip_mode_choices[KDISPLAYDATA_NUM_CLIP_MODE_CHOICES] =
{
    "Images and Profiles",
    "Images only",
    "Profiles only",
    "Both later",
};


/*  Private functions  */
STATIC_FUNCTION (flag _dispdata_array_data_event_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );
STATIC_FUNCTION (flag _dispdata_array_create_dataobject,
		 (multi_array *multi_desc, unsigned int index,
		  CONST char *domain, CONST char *name) );
STATIC_FUNCTION (flag _dispdata_array_create_dataobject2,
		 (iarray pseudo, iarray red, iarray green, iarray blue,
		  CONST char *domain, CONST char *name, flag quiet,
		  flag set_range_ok) );
STATIC_FUNCTION (flag _dispdata_array_split_hypercube,
		 (iarray pseudo, iarray red, iarray green, iarray blue,
		  CONST char *domain, CONST char *name) );
STATIC_FUNCTION (flag _dispdata_array_create_cimages,
		 (struct dataset *set, iarray pseudo) );
STATIC_FUNCTION (flag _dispdata_array_setup_2D_pseudo,
		 (struct viewable *viewable, iarray array) );
STATIC_FUNCTION (flag _dispdata_array_setup_3D_pseudo,
		 (struct viewable *viewable, iarray array) );
STATIC_FUNCTION (flag _dispdata_array_setup_2D_rgb,
		 (struct viewable *viewable, iarray red, iarray green,
		  iarray blue) );
STATIC_FUNCTION (flag _dispdata_array_setup_3D_rgb,
		 (struct viewable *viewable, iarray red, iarray green,
		  iarray blue) );
STATIC_FUNCTION (void _dispdata_array_destroy_data_attributes,
		 (struct data_attributes *att) );
STATIC_FUNCTION (void _dispdata_array_dealloc_direction_viewable,
		 (struct viewable_direction *dir, unsigned int num_frames) );
STATIC_FUNCTION (void _dispdata_array_set_contour_levels,
		 (struct dataset *dataset,
		  flag set_map_levels, flag set_renzogram_levels) );
STATIC_FUNCTION (void _dispdata_array_set_contour_styles,
		 (struct viewable *viewable) );
STATIC_FUNCTION (void _dispdata_array_set_renzogram_styles,
		 (struct viewable *viewable) );
STATIC_FUNCTION (void _dispdata_array_cmap_modify_func,
		 (Kcolourmap cmap, void **info) );
STATIC_FUNCTION (void _dispdata_array_activate_frames,
		 (struct viewable *viewable) );
STATIC_FUNCTION (flag _dispdata_array_canvas_position_func,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code, void *e_info, void **f_info,
		  double x_lin, double y_lin, int x_pix, int y_pix) );
STATIC_FUNCTION (flag _dispdata_array_middle_geom_func,
		 (KPixCanvasEditorContext context, KPixCanvas canvas,
		  unsigned int object_type, unsigned int num_coords,
		  CONST int *x, CONST int *y, unsigned int raw_event_code,
		  void *raw_e_info, void **f_info) );
STATIC_FUNCTION (flag _dispdata_array_change_frame,
		 (struct window *sourcewin, unsigned int framenum) );
STATIC_FUNCTION (flag _dispdata_array_set_matching_frame,
		 (struct viewable *dest, struct viewable *source,
		  CONST struct frame_pos *frame) );
STATIC_FUNCTION (void _dispdata_array_compute_frame_coord,
		 (struct dataset *dataset, unsigned int dim_num) );
STATIC_FUNCTION (void _dispdata_array_set_canvas_iscale,
		 (struct viewable *viewable) );
STATIC_FUNCTION (flag _dispdata_array_set_profile_vscale,
		 (struct viewable *viewable) );
STATIC_FUNCTION (ViewableImage _dispdata_array_get_vimage,
		 (struct viewable *viewable) );
STATIC_FUNCTION (flag _dispdata_array_convert_3coord,
		 (struct dataset *dataset, unsigned int slice_direction,
		  double *x, double *y, double *z, double h, double v) );
STATIC_FUNCTION (flag _dispdata_array_convert_2coord,
		 (struct viewable *viewable, double *h, double *v) );
STATIC_FUNCTION (void _dispdata_array_line_profile,
		 (struct viewable *viewable, double x, double y) );
STATIC_FUNCTION (void _dispdata_array_box_event,
		 (struct viewable *viewable,
		  CONST double *h, CONST double *v, flag configure) );
STATIC_FUNCTION (void _dispdata_array_box_profile,
		 (struct viewable *viewable, flag sum,
		  CONST double *h, CONST double *v, flag configure) );
STATIC_FUNCTION (void _dispdata_array_boxed_profile,
		 (struct viewable *viewable, flag horizontal,
		  CONST double *h, CONST double *v, flag configure) );
STATIC_FUNCTION (flag _dispdata_array_collapse_dim,
		 (double *out, CONST char *in, unsigned int in_type,
		  uaddr start_x, uaddr end_x, CONST uaddr *xoffsets,
		  uaddr start_y, uaddr end_y, CONST uaddr *yoffsets,
		  double scale, double offset) );
STATIC_FUNCTION (void _dispdata_array_circle_event,
		 (struct viewable *viewable,
		  CONST double *h, CONST double *v, flag configure) );
STATIC_FUNCTION (void _dispdata_array_radial_profile,
		 (struct viewable *viewable, flag do_fit,
		  CONST double *h, CONST double *v, flag configure) );
STATIC_FUNCTION (flag _dispdata_array_collect_radial_points,
		 (KViewableGraph vgraph, CONST char *in, unsigned int in_type,
		  uaddr xlen, CONST uaddr *xoffsets, double xcentre,
		  signed int xradius, uaddr ylen, CONST uaddr *yoffsets,
		  double ycentre, signed int yradius,
		  float scale, float offset) );
STATIC_FUNCTION (void _dispdata_array_show_fwhm,
		 (ViewableImage vimage, double x, double y, double pixels) );
STATIC_FUNCTION (void _dispdata_array_show_fwhm2,
		 (ViewableImage vimage, double x, double y, double pixels,
		  char *unit_str, char *val_str) );
STATIC_FUNCTION (void _dispdata_array_azimuthal_avg_profile,
		 (struct viewable *viewable, CONST double *h, CONST double *v,
		  flag configure) );
STATIC_FUNCTION (flag _dispdata_array_azimuthally_average,
		 (KViewableGraph vgraph, CONST char *in, unsigned int in_type,
		  uaddr xlen, CONST uaddr *xoffsets,
		  double xcentre, double xradius,
		  uaddr ylen, CONST uaddr *yoffsets,
		  double ycentre, double yradius,
		  double rotation, float scale, float offset) );
STATIC_FUNCTION (void _dispdata_array_set_title, (struct window *window) );
STATIC_FUNCTION (flag _dispdata_array_irange_change_func,
		 (void *object, void *client1_data, void *call_data,
		  void *client2_data) );
STATIC_FUNCTION (flag _dispdata_array_test_profile_perpendicular,
		 (const struct window *window) );


/* Public functions follow */

/*EXPERIMENTAL_FUNCTION*/
KDisplayDataClass dispdata_array_create_dataclass (void)
/*  [SUMMARY] Initialise array data support for the package.
    [RETURNS] A KDisplayDataClass object on success, else NULL.
*/
{
    KDisplayDataClass dc;

    if (dataclass) return (dataclass);
    dc = dispdata_create_dataclass ("Arrays",
				    _dispdata_array_set_class_attributes,
				    _dispdata_array_destroy_dataobject,
				    _dispdata_array_copy_data_attributes,
				    _dispdata_array_set_data_attributes,
				    _dispdata_array_get_data_attributes,
				    _dispdata_array_create_window,
				    _dispdata_array_destroy_window,
				    _dispdata_array_set_window_attributes,
				    _dispdata_array_get_window_attributes,
				    _dispdata_array_create_viewable,
				    _dispdata_array_destroy_viewable,
				    _dispdata_array_create_blinkstate,
				    _dispdata_array_destroy_blinkstate,
				    _dispdata_array_activate_blinkstate,
				    _dispdata_array_create_blinkentry,
				    _dispdata_array_destroy_blinkentry,
				    _dispdata_array_copy_blinkentry,
				    _dispdata_array_activate_blinkentry,
				    _dispdata_array_set_blinkentry_attributes,
				    _dispdata_array_get_blinkentry_attributes);
    dataclass = dc;
    ds_event_register_func (_dispdata_array_data_event_func, NULL);
    return (dc);
}   /*  End Function dispdata_array_create_dataclass  */

/*EXPERIMENTAL_FUNCTION*/
void dispdata_array_clone_data (KDisplayDataObject dataobject)
/*  [SUMMARY] Clone an array dataset.
    <dataobject> The KDisplayDataObject to clone.
    [RETURNS] Nothing.
*/
{
    unsigned int namelen;
    struct dataset *old, *new;
    char *domain;
    char name[STRING_LENGTH], number[STRING_LENGTH];
    static char function_name[] = "dispdata_array_clone_data";

    old = dispdata_get_dataobject_subclass (dataobject, dataclass);
    dispdata_get_data_attributes (dataobject,
				  KDISPLAYDATA_DATA_ATT_DOMAIN, &domain,
				  KDISPLAYDATA_DATA_ATT_END);
    if ( ( new = m_calloc (sizeof *new) ) == NULL )
    {
	m_error_notify (function_name, "data object");
	return;
    }
    new->data = old->data;
    ++new->data->refcount;
    if (new->data->pseudo)
    {
	if ( !_dispdata_array_create_cimages (new, new->data->pseudo) )
	{
	    _dispdata_array_destroy_dataobject (new);
	    return;
	}
    }
    if ( !_dispdata_array_copy_data_attributes (new, old) )
    {
	_dispdata_array_destroy_dataobject (new);
	return;
    }
    /*  Construct name[%u] safely (prevent buffer overruns)  */
    sprintf (number, "[%u]", new->data->counter);
    namelen = strlen (new->data->name);
    if (namelen + strlen (number) + 1 >= STRING_LENGTH)
	namelen = STRING_LENGTH - strlen (number) - 1;
    strncpy (name, new->data->name, namelen);
    strcpy (name + namelen, number);
    if ( ( new->superclass = dispdata_create_dataobject (dataclass, domain,
							 name, new) )
	 == NULL )
    {
	_dispdata_array_destroy_dataobject (new);
	return;
    }
    ++new->data->counter;
}   /*  End Function dispdata_array_clone_data  */

/*EXPERIMENTAL_FUNCTION*/
KCallbackFunc dispdata_array_register_irange_change_func
    (KDisplayDataObject dataobject,
     void (*func) (KDisplayDataObject dataobject, void *info),
     void *info)
/*  [SUMMARY] Register a function to be called the image itensity range changes
    <dataobject> The KDisplayDataObject to clone.
    <func> The function that is called when the image intensity range changes.
    <info> An arbitrary information pointer passed to <<func>>.
    [RETURNS] A KCallbackFunc object. On failure, the process aborts.
*/
{
    struct dataset *dataset;

    dataset = dispdata_get_dataobject_subclass (dataobject, dataclass);
    return c_register_callback (&dataset->irange_change_list,
				_dispdata_array_irange_change_func, dataobject,
				(void *) func, FALSE, info, FALSE, FALSE);
}   /*  End Function dispdata_array_register_irange_change_func  */

/*EXPERIMENTAL_FUNCTION*/
flag dispdata_array_convert_3coord (KDisplayDataObject dataobject,
				    KDisplayDataWindow window,
				    double *x, double *y, double *z,
				    double h, double v)
/*  [SUMMARY] Convert canvas co-ordinate to 3D dataset co-ordinate.
    <dataobject> The KDisplayDataObject to clone.
    <window> The KDisplayDataWindow.
    <x> The dataset X image pixel co-ordinate is written here.
    <y> The dataset Y image pixel co-ordinate is written here.
    <z> The dataset Z image pixel co-ordinate is written here.
    <h> The horizontal co-ordinate on the world canvas.
    <v> The vertical co-ordinate on the world canvas.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int slice_direction;
    struct dataset *dataset;

    dataset = dispdata_get_dataobject_subclass (dataobject, dataclass);
    dispdata_get_window_attributes (window,
				    KDISPLAYDATA_WINDOW_ATT_SLICE_DIRECTION,
				        &slice_direction,
				    KDISPLAYDATA_WINDOW_ATT_END);
    return _dispdata_array_convert_3coord (dataset, slice_direction, x, y, z,
					   h, v);
}   /*  End Function dispdata_array_convert_3coord  */


/*  Hook functions follow  */

static void _dispdata_array_set_class_attributes (va_list argp)
/*  [SUMMARY] Set global attributes for the array dataclass.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    KDISPLAYDATA_CLASS_ATT_END.
    [NOTE] Some windows may be refreshed.
    [RETURNS] Nothing.
*/
{
    KDisplayDataObject dataobject;
    flag sync_clips = FALSE;
    unsigned int att_key, uival;
    static char function_name[] = "_dispdata_array_set_class_attributes";

    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    KDISPLAYDATA_CLASS_ATT_END )
    {
	switch (att_key)
	{
	  case KDISPLAYDATA_CLASS_ATT_CLIP_MODE:
	    uival = va_arg (argp, unsigned int);
	    if (clip_mode == uival) break;
	    switch (uival)
	    {
	      case KDISPLAYDATA_CLIP_MODE_CHANGE_IMAGE_AND_PROFILE:
		sync_clips = TRUE;
		/*  Fall through  */
	      case KDISPLAYDATA_CLIP_MODE_CHANGE_IMAGE_ONLY:
	      case KDISPLAYDATA_CLIP_MODE_CHANGE_PROFILE_ONLY:
	      case KDISPLAYDATA_CLIP_MODE_CHANGE_BOTH_LATER:
		break;
	      default:
		fprintf (stderr, "Invalid clip mode: %u\n", uival);
		a_prog_bug (function_name);
		break;
	    }
	    clip_mode = uival;
	    break;
	  case KDISPLAYDATA_CLASS_ATT_SLAVE_PROFILE_H_ZOOM:
	    slave_profile_h_zoom = va_arg (argp, flag);
	    FLAG_VERIFY (slave_profile_h_zoom);
	    break;
	}
    }
    if (!sync_clips) return;
    for ( dataobject = dispdata_get_first_dataobject (dataclass);
	  dataobject;
	  dataobject = dispdata_get_next_dataobject (dataobject) )
    {
	double lower_clip, upper_clip;

	dispdata_get_data_attributes (dataobject,
				      KDISPLAYDATA_DATA_ATT_SCALED_LOWER_CLIP,
				          &lower_clip,
				      KDISPLAYDATA_DATA_ATT_SCALED_UPPER_CLIP,
				          &upper_clip,
				      KDISPLAYDATA_DATA_ATT_END);
	dispdata_set_data_attributes (dataobject,
				      KDISPLAYDATA_DATA_ATT_SCALED_LOWER_CLIP,
				          lower_clip,
				      KDISPLAYDATA_DATA_ATT_SCALED_UPPER_CLIP,
				          upper_clip,
				      KDISPLAYDATA_DATA_ATT_END);
    }
}   /*  End Function _dispdata_array_set_class_attributes  */

static void _dispdata_array_destroy_dataobject (void *dataobject)
/*  [SUMMARY] Destroy a data object.
    <dataobject> The dataobject to destroy.
    [RETURNS] Nothing.
*/
{
    struct dataset *dataset = dataobject;

    if (dataset->data->refcount <= 1)
    {   /*  No other references to these data  */
	m_free2 (dataset->data->object);
	m_free2 (dataset->data->name);
	iarray_dealloc2 (dataset->data->pseudo);
	iarray_dealloc2 (dataset->data->red);
	iarray_dealloc2 (dataset->data->green);
	iarray_dealloc2 (dataset->data->blue);
	m_free (dataset->data);
    }
    else --dataset->data->refcount;
    _dispdata_array_destroy_data_attributes (&dataset->attributes);
    if (dataset->map_cimage_group)
	contour_destroy_group (dataset->map_cimage_group);
    if (dataset->renzo_cimage_group)
	contour_destroy_group (dataset->renzo_cimage_group);
    m_free2 (dataset->xy.map_cimages);
    m_free2 (dataset->xy.renzo_cimages);
    m_free2 (dataset->xz.map_cimages);
    m_free2 (dataset->xz.renzo_cimages);
    m_free2 (dataset->zy.map_cimages);
    m_free2 (dataset->zy.renzo_cimages);
    c_destroy_list (dataset->irange_change_list);
    m_free (dataset);
}   /*  End Function _dispdata_array_destroy_dataobject  */

static flag _dispdata_array_copy_data_attributes (void *dest_p, void *source_p)
/*  [SUMMARY] Copy data attributes.
    <dest_p> The destination data object.
    <source_p> The source data object.
    [RETURNS] TRUE on success, else FALSE..
*/
{
    struct dataset *dest = dest_p;
    struct dataset *source = source_p;
    static char function_name[] = "_dispdata_array_copy_data_attributes";

    dest->attributes.xy_frame.index = source->attributes.xy_frame.index;
    if (dest->attributes.xy_frame.index >= dest->data->num_xy_frames)
	dest->attributes.xy_frame.index = dest->data->num_xy_frames - 1;
    _dispdata_array_compute_frame_coord (dest, 0);
    dest->attributes.xz_frame.index = source->attributes.xz_frame.index;
    if (dest->attributes.xz_frame.index >= dest->data->num_xz_frames)
	dest->attributes.xz_frame.index = dest->data->num_xz_frames - 1;
    _dispdata_array_compute_frame_coord (dest, 1);
    dest->attributes.zy_frame.index = source->attributes.zy_frame.index;
    if (dest->attributes.zy_frame.index >= dest->data->num_zy_frames)
	dest->attributes.zy_frame.index = dest->data->num_zy_frames - 1;
    _dispdata_array_compute_frame_coord (dest, 2);
    if (!dest->data->pseudo || !source->data->pseudo) return (TRUE);
    dest->attributes.input_lower_clip = source->attributes.input_lower_clip;
    dest->attributes.input_upper_clip = source->attributes.input_upper_clip;
    dest->attributes.image_lower_clip = source->attributes.image_lower_clip;
    dest->attributes.image_upper_clip = source->attributes.image_upper_clip;
    dest->attributes.profile_lower_clip =source->attributes.profile_lower_clip;
    dest->attributes.profile_upper_clip =source->attributes.profile_upper_clip;
    dest->attributes.contour_neg_linewidth =
	source->attributes.contour_neg_linewidth;
    dest->attributes.contour_pos_linewidth =
	source->attributes.contour_pos_linewidth;
    if (dest->attributes.contour_neg_colour)
	m_free (dest->attributes.contour_neg_colour);
    if ( ( dest->attributes.contour_neg_colour =
	   st_dup (source->attributes.contour_neg_colour) ) == NULL )
    {
	m_error_notify (function_name, "negative contour colour");
	return (FALSE);
    }
    if (dest->attributes.contour_pos_colour)
	m_free (dest->attributes.contour_pos_colour);
    if ( ( dest->attributes.contour_pos_colour =
	   st_dup (source->attributes.contour_pos_colour) ) == NULL )
    {
	m_error_notify (function_name, "positive contour colour");
	return (FALSE);
    }
    if (source->attributes.contour_levels_string)
    {
	if ( ( dest->attributes.contour_levels_string =
	       st_dup (source->attributes.contour_levels_string) ) == NULL )
	{
	    m_error_notify (function_name, "contour levels string");
	    return (FALSE);
	}
    }
    if (source->attributes.renzogram_levels_string)
    {
	if ( ( dest->attributes.renzogram_levels_string =
	       st_dup (source->attributes.renzogram_levels_string) ) == NULL )
	{
	    m_error_notify (function_name, "renzogram levels string");
	    return (FALSE);
	}
    }
    if (source->attributes.renzogram_channels_string)
    {
	if (dest->attributes.renzogram_channels_string)
	    m_free (dest->attributes.renzogram_channels_string);
	if ( ( dest->attributes.renzogram_channels_string =
	       st_dup (source->attributes.renzogram_channels_string) ) ==NULL )
	{
	    m_error_notify (function_name, "renzogram channels string");
	    return (FALSE);
	}
    }
    _dispdata_array_set_contour_levels (dest, TRUE, TRUE);
    if (dest->attributes.profile_dark_colour)
	m_free (dest->attributes.profile_dark_colour);
    if ( ( dest->attributes.profile_dark_colour =
	   st_dup (source->attributes.profile_dark_colour) ) == NULL )
    {
	m_error_notify (function_name, "dark profile colour");
	return (FALSE);
    }
    if (dest->attributes.profile_light_colour)
	m_free (dest->attributes.profile_light_colour);
    if ( ( dest->attributes.profile_light_colour =
	   st_dup (source->attributes.profile_light_colour) ) == NULL )
    {
	m_error_notify (function_name, "light profile colour");
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function _dispdata_array_copy_data_attributes  */

static flag _dispdata_array_set_data_attributes (void *data, va_list argp)
{
    flag set_range = FALSE;
    flag changed_irange = FALSE;
    flag changed_prange = FALSE;
    flag changed_contour_levels = FALSE;
    flag changed_contour_styles = FALSE;
    flag changed_renzogram_levels = FALSE;
    flag changed_renzogram_channels = FALSE;
    flag changed_profile_colours = FALSE;
    flag all_known = TRUE;
    char ch;
    unsigned int att_key, dummy;
    double dval;
    struct dataset *dataset = data;
    struct data_attributes *att = &dataset->attributes;
    struct viewable *viewable;
    char *string;
    static char function_name[] = "_dispdata_array_set_data_attributes";

    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    KDISPLAYDATA_DATA_ATT_END )
    {
	switch (att_key)
	{
	  case KDISPLAYDATA_DATA_ATT_RAW_LOWER_CLIP:
	    dval = va_arg (argp, double);
	    if (!dataset->data->pseudo) break;
	    dval = dval * dataset->data->scale + dataset->data->offset;
	    if (dataset->data->scale < 0.0) att->input_upper_clip = dval;
	    else att->input_lower_clip = dval;
	    set_range = TRUE;
	    break;
	  case KDISPLAYDATA_DATA_ATT_RAW_UPPER_CLIP:
	    dval = va_arg (argp, double);
	    if (!dataset->data->pseudo) break;
	    dval = dval * dataset->data->scale + dataset->data->offset;
	    if (dataset->data->scale < 0.0) att->input_lower_clip = dval;
	    else att->input_upper_clip = dval;
	    set_range = TRUE;
	    break;
	  case KDISPLAYDATA_DATA_ATT_SCALED_LOWER_CLIP:
	    dval = va_arg (argp, double);
	    if (!dataset->data->pseudo) break;
	    att->input_lower_clip = dval;
	    set_range = TRUE;
	    break;
	  case KDISPLAYDATA_DATA_ATT_SCALED_UPPER_CLIP:
	    dval = va_arg (argp, double);
	    if (!dataset->data->pseudo) break;
	    att->input_upper_clip = dval;
	    set_range = TRUE;
	    break;
	  case KDISPLAYDATA_DATA_ATT_CONTOUR_NEG_COLOUR:
	    string = va_arg (argp, char *);
	    if (!dataset->data->pseudo) break;
	    m_free (att->contour_neg_colour);
	    if ( ( att->contour_neg_colour = st_dup (string) ) == NULL )
		m_abort (function_name, "negative contour colour");
	    changed_contour_styles = TRUE;
	    break;
	  case KDISPLAYDATA_DATA_ATT_CONTOUR_POS_COLOUR:
	    string = va_arg (argp, char *);
	    if (!dataset->data->pseudo) break;
	    m_free (att->contour_pos_colour);
	    if ( ( att->contour_pos_colour = st_dup (string) ) == NULL )
		m_abort (function_name, "positive contour colour");
	    changed_contour_styles = TRUE;
	    break;
	  case KDISPLAYDATA_DATA_ATT_CONTOUR_NEG_LINEWIDTH:
	    att->contour_neg_linewidth = va_arg (argp, double);
	    if (dataset->data->pseudo) changed_contour_styles = TRUE;
	    break;
	  case KDISPLAYDATA_DATA_ATT_CONTOUR_POS_LINEWIDTH:
	    att->contour_pos_linewidth = va_arg (argp, double);
	    if (dataset->data->pseudo) changed_contour_styles = TRUE;
	    break;
	  case KDISPLAYDATA_DATA_ATT_CONTOUR_LEVELS:
	    string = va_arg (argp, char *);
	    if (!dataset->data->pseudo) break;
	    if (att->contour_levels_string)
	    {
		if (strcmp (string, att->contour_levels_string) == 0) break;
		m_free (att->contour_levels_string);
	    }
	    if ( ( att->contour_levels_string = st_dup (string) ) == NULL )
		m_abort (function_name, "contour string");
	    changed_contour_levels = TRUE;
	    changed_contour_styles = TRUE;
	    break;
	  case KDISPLAYDATA_DATA_ATT_RENZOGRAM_LEVELS:
	    string = va_arg (argp, char *);
	    if (!dataset->data->pseudo) break;
	    if (dataset->data->num_xy_frames < 1) break;
	    if (att->renzogram_levels_string)
		m_free (att->renzogram_levels_string);
	    if ( ( att->renzogram_levels_string = st_dup (string) ) == NULL )
		m_abort (function_name, "renzogram levels string");
	    changed_renzogram_levels = TRUE;
	    break;
	  case KDISPLAYDATA_DATA_ATT_RENZOGRAM_CHANNELS:
	    string = va_arg (argp, char *);
	    if (!dataset->data->pseudo) break;
	    if (dataset->data->num_xy_frames < 1) break;
	    ch = toupper (string[0]);
	    if ( (ch != 'I') && (ch != 'W') ) break;
	    if (att->renzogram_channels_string)
		m_free (att->renzogram_channels_string);
	    if ( ( att->renzogram_channels_string = st_dup (string) ) == NULL )
		m_abort (function_name, "renzogram channels string");
	    att->renzogram_channels_string[0] = ch;
	    changed_renzogram_channels = TRUE;
	    break;
	  case KDISPLAYDATA_DATA_ATT_PROFILE_DARK_COLOUR:
	    string = va_arg (argp, char *);
	    if (!dataset->data->pseudo) break;
	    m_free (att->profile_dark_colour);
	    if ( ( att->profile_dark_colour = st_dup (string) ) == NULL )
		m_abort (function_name, "dark profile colour");
	    changed_profile_colours = TRUE;
	    break;
	  case KDISPLAYDATA_DATA_ATT_PROFILE_LIGHT_COLOUR:
	    string = va_arg (argp, char *);
	    if (!dataset->data->pseudo) break;
	    m_free (att->profile_light_colour);
	    if ( ( att->profile_light_colour = st_dup (string) ) == NULL )
		m_abort (function_name, "light profile colour");
	    changed_profile_colours = TRUE;
	    break;
	  default:
	    dummy = va_arg (argp, unsigned int);
	    all_known = FALSE;
	    break;
	}
    }
    if (set_range)
    {   /*  Sanity checks  */
	if (att->input_upper_clip <= att->input_lower_clip)
	    att->input_upper_clip = att->input_lower_clip + 1.0;
	switch ( iarray_type (dataset->data->pseudo) )
	{
	  case K_FLOAT:
	  case K_DOUBLE:
	  case K_COMPLEX:
	  case K_DCOMPLEX:
	    break;
	  default:
	    if ( fabs (att->input_upper_clip - att->input_lower_clip) <=
		 dataset->data->scale )
		att->input_upper_clip = att->input_lower_clip +
		    2.0 * dataset->data->scale;
	    break;
	}
	switch (clip_mode)
	{
	  case KDISPLAYDATA_CLIP_MODE_CHANGE_IMAGE_ONLY:
	    if ( (att->image_lower_clip != att->input_lower_clip) ||
		 (att->image_upper_clip != att->input_upper_clip) )
	    {
		att->image_lower_clip = att->input_lower_clip;
		att->image_upper_clip = att->input_upper_clip;
		changed_irange = TRUE;
	    }
	    break;
	  default:
	    if ( (att->image_lower_clip != att->input_lower_clip) ||
		 (att->image_upper_clip != att->input_upper_clip) )
	    {
		att->image_lower_clip = att->input_lower_clip;
		att->image_upper_clip = att->input_upper_clip;
		changed_irange = TRUE;
	    }
	    /*  Fall through  */
	  case KDISPLAYDATA_CLIP_MODE_CHANGE_PROFILE_ONLY:
	    if ( (att->profile_lower_clip != att->input_lower_clip) ||
		 (att->profile_upper_clip != att->input_upper_clip) )
	    {
		att->profile_lower_clip = att->input_lower_clip;
		att->profile_upper_clip = att->input_upper_clip;
		changed_prange = TRUE;
	    }
	    break;
	}
    }
    _dispdata_array_set_contour_levels (dataset, changed_contour_levels,
					changed_renzogram_levels);
    /*  Scan through list of viewables, applying changes  */
    for (viewable = dataset->first; viewable;
	 viewable = viewable->dataset_chain.next)
    {
	flag refresh_main = FALSE;
	flag refresh_profile = FALSE;
	struct blinkentry *blinkentry = viewable->active_blinkentry;
	struct window *window = viewable->window;

	if (changed_irange &&
	    (window->active_blinkstate->image_viewable == viewable) )
	{
	    viewable->window->min = att->image_lower_clip;
	    viewable->window->max = att->image_upper_clip;
	    _dispdata_array_set_canvas_iscale (viewable);
	    refresh_main = TRUE;
	}
	if (changed_contour_styles && blinkentry &&
	    (blinkentry->contour_mode == KDISPLAYDATA_CONTOUR_NORMAL) )
	{
	    _dispdata_array_set_contour_styles (viewable);
	    refresh_main = TRUE;
	}
	if (changed_renzogram_levels && blinkentry &&
	    (blinkentry->contour_mode == KDISPLAYDATA_CONTOUR_RENZOGRAM) )
	{
	    _dispdata_array_set_renzogram_styles (viewable);
	    refresh_main = TRUE;
	}
	if (changed_renzogram_channels && blinkentry &&
	    (blinkentry->contour_mode == KDISPLAYDATA_CONTOUR_RENZOGRAM) )
	{
	    _dispdata_array_activate_frames (viewable);
	    refresh_main = TRUE;
	}
	if (refresh_main) dispdata_refresh_window (window->superclass, NULL);
	if (changed_profile_colours && viewable->vgraph)
	{
	    graph1d_set_vgraph_attributes
		(viewable->vgraph,
		 GRAPH1D_VGRAPH_ATT_DARK_COLOUR, att->profile_dark_colour,
		 GRAPH1D_VGRAPH_ATT_LIGHT_COLOUR, att->profile_light_colour,
		 GRAPH1D_VGRAPH_ATT_END);
	    graph1d_set_vgraph_attributes
		(viewable->gausscurve,
		 GRAPH1D_VGRAPH_ATT_DARK_COLOUR, att->profile_dark_colour,
		 GRAPH1D_VGRAPH_ATT_LIGHT_COLOUR, att->profile_light_colour,
		 GRAPH1D_VGRAPH_ATT_END);
	    if ( (blinkentry->blinkstate->profile_master == viewable) ||
		 blinkentry->profile_slave ) refresh_profile = TRUE;
	}
	if ( changed_prange &&
	     window->profile_canvas &&
	     _dispdata_array_set_profile_vscale (viewable) )
	    refresh_profile = TRUE;
	if (refresh_profile)
	    kwin_refresh_if_visible (window->profile_pixcanvas, TRUE);
    }
    if (changed_irange) c_call_callbacks (dataset->irange_change_list, NULL);
    return (all_known);
}   /*  End Function _dispdata_array_set_data_attributes  */

static flag _dispdata_array_get_data_attributes (void *data, va_list argp)
{
    flag all_known = TRUE;
    unsigned int att_key;
    double lower_clip, upper_clip;
    unsigned int *frame_number = NULL;
    unsigned int *num_frames = NULL;
    signed int direction = -1;
    struct dataset *dataset = data;
    struct data_attributes *att = &dataset->attributes;
    void *dummy;
    static char function_name[] = "_dispdata_array_get_data_attributes";

    switch (clip_mode)
    {
      case KDISPLAYDATA_CLIP_MODE_CHANGE_IMAGE_ONLY:
	lower_clip = att->image_lower_clip;
	upper_clip = att->image_upper_clip;
	break;
      case KDISPLAYDATA_CLIP_MODE_CHANGE_PROFILE_ONLY:
	lower_clip = att->profile_lower_clip;
	upper_clip = att->profile_upper_clip;
	break;
      default:
	lower_clip = att->input_lower_clip;
	upper_clip = att->input_upper_clip;
	break;
    }
    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    KDISPLAYDATA_DATA_ATT_END )
    {
	switch (att_key)
	{
	  case KDISPLAYDATA_DATA_ATT_RAW_LOWER_CLIP:
	    if (dataset->data->scale < 0.0)
		*va_arg (argp, double *) =
		(upper_clip - dataset->data->offset) / dataset->data->scale;
	    else *va_arg (argp, double *) =
		(lower_clip - dataset->data->offset) / dataset->data->scale;
	    break;
	  case KDISPLAYDATA_DATA_ATT_RAW_UPPER_CLIP:
	    if (dataset->data->scale < 0.0)
		*va_arg (argp, double *) =
		(lower_clip - dataset->data->offset) / dataset->data->scale;
	    else *va_arg (argp, double *) =
		(upper_clip - dataset->data->offset) / dataset->data->scale;
	    break;
	  case KDISPLAYDATA_DATA_ATT_SCALED_LOWER_CLIP:
	    *va_arg (argp, double *) = lower_clip;
	    break;
	  case KDISPLAYDATA_DATA_ATT_SCALED_UPPER_CLIP:
	    *va_arg (argp, double *) = upper_clip;
	    break;
	  case KDISPLAYDATA_DATA_ATT_CONTOUR_NEG_COLOUR:
	    *va_arg (argp, char **) = att->contour_neg_colour;
	    break;
	  case KDISPLAYDATA_DATA_ATT_CONTOUR_POS_COLOUR:
	    *va_arg (argp, char **) = att->contour_pos_colour;
	    break;
	  case KDISPLAYDATA_DATA_ATT_CONTOUR_NEG_LINEWIDTH:
	    *va_arg (argp, double *) = att->contour_neg_linewidth;
	    break;
	  case KDISPLAYDATA_DATA_ATT_CONTOUR_POS_LINEWIDTH:
	    *va_arg (argp, double *) = att->contour_pos_linewidth;
	    break;
	  case KDISPLAYDATA_DATA_ATT_CONTOUR_LEVELS:
	    *va_arg (argp, char **) = att->contour_levels_string;
	    break;
	  case KDISPLAYDATA_DATA_ATT_RENZOGRAM_LEVELS:
	    *va_arg (argp, char **) = att->renzogram_levels_string;
	    break;
	  case KDISPLAYDATA_DATA_ATT_RENZOGRAM_CHANNELS:
	    *va_arg (argp, char **) = att->renzogram_channels_string;
	    break;
	  case KDISPLAYDATA_DATA_ATT_PROFILE_DARK_COLOUR:
	    *va_arg (argp, char **) = att->profile_dark_colour;
	    break;
	  case KDISPLAYDATA_DATA_ATT_PROFILE_LIGHT_COLOUR:
	    *va_arg (argp, char **) = att->profile_light_colour;
	    break;
	  case KDISPLAYDATA_DATA_ATT_FRAME_NUMBER:
	    frame_number = va_arg (argp, unsigned int *);
	    break;
	  case KDISPLAYDATA_DATA_ATT_NUM_FRAMES:
	    num_frames = va_arg (argp, unsigned int *);
	    break;
	  case KDISPLAYDATA_DATA_ATT_DIRECTION_FOR_FRAME:
	    direction = va_arg (argp, unsigned int);
	    break;
	  case KDISPLAYDATA_DATA_ATT_OBJECT:
	    *va_arg (argp, char **) = dataset->data->object;
	    break;
	  case KDISPLAYDATA_DATA_ATT_PSEUDO_ARRAY:
	    *va_arg (argp, iarray *) = dataset->data->pseudo;
	    break;
	  case KDISPLAYDATA_DATA_ATT_RED_ARRAY:
	    *va_arg (argp, iarray *) = dataset->data->red;
	    break;
	  case KDISPLAYDATA_DATA_ATT_GREEN_ARRAY:
	    *va_arg (argp, iarray *) = dataset->data->green;
	    break;
	  case KDISPLAYDATA_DATA_ATT_BLUE_ARRAY:
	    *va_arg (argp, iarray *) = dataset->data->blue;
	    break;
	  case KDISPLAYDATA_DATA_ATT_RAW_MINIMUM:
	    *va_arg (argp, double *) = dataset->data->raw.min;
	    break;
	  case KDISPLAYDATA_DATA_ATT_RAW_MAXIMUM:
	    *va_arg (argp, double *) = dataset->data->raw.max;
	    break;
	  case KDISPLAYDATA_DATA_ATT_SCALED_MINIMUM:
	    *va_arg (argp, double *) = dataset->data->scaled.min;
	    break;
	  case KDISPLAYDATA_DATA_ATT_SCALED_MAXIMUM:
	    *va_arg (argp, double *) = dataset->data->scaled.max;
	    break;
	  default:
	    dummy = va_arg (argp, void *);
	    all_known = FALSE;
	    break;
	}
    }
    if (frame_number)
    {
	if (direction < 0)
	{
	    fprintf (stderr, "Direction not specified\n");
	    a_prog_bug (function_name);
	}
	switch (direction)
	{
	  case KDISPLAYDATA_SLICE_DIRECTION_XY:
	    *frame_number = att->xy_frame.index;
	    break;
	  case KDISPLAYDATA_SLICE_DIRECTION_XZ:
	    *frame_number = att->xz_frame.index;
	    break;
	  case KDISPLAYDATA_SLICE_DIRECTION_ZY:
	    *frame_number = att->zy_frame.index;
	    break;
	  default:
	    fprintf (stderr, "Invalid direction: %d\n", direction);
	    a_prog_bug (function_name);
	    break;
	}
    }
    if (num_frames)
    {
	if (direction < 0)
	{
	    fprintf (stderr, "Direction not specified\n");
	    a_prog_bug (function_name);
	}
	switch (direction)
	{
	  case KDISPLAYDATA_SLICE_DIRECTION_XY:
	    *num_frames = dataset->data->num_xy_frames;
	    break;
	  case KDISPLAYDATA_SLICE_DIRECTION_XZ:
	    *num_frames = dataset->data->num_xz_frames;
	    break;
	  case KDISPLAYDATA_SLICE_DIRECTION_ZY:
	    *num_frames = dataset->data->num_zy_frames;
	    break;
	  default:
	    fprintf (stderr, "Invalid direction: %d\n", direction);
	    a_prog_bug (function_name);
	    break;
	}
    }
    return (all_known);
}   /*  End Function _dispdata_array_get_data_attributes  */

static void *_dispdata_array_create_window (KDisplayDataWindow superclass,
					    KDisplayDataCanvas *pc,
					    KDisplayDataCanvas *rgb)
/*  [SUMMARY] Create a window.
    <superclass> The superclass KDisplayDataWindow object.
    <pc> The PseudoColour canvas. This may contain NULL KWorldCanvas objects.
    <rgb> The TrueColour/DirectColour canvas. This may contain NULL
    KWorldCanvas objects.
    [RETURNS] The window object.
*/
{
    KPixCanvas pixcanvas;
    unsigned long pixel_value = 0;
    struct window *window;
    static char function_name[] = "_dispdata_array_create_window";

    if ( ( window = m_calloc (sizeof *window) ) == NULL )
    {
	m_error_notify (function_name, "window information");
	return (NULL);
    }
    window->superclass = superclass;
    if (pc->main)
    {
	window->pc = *pc;
	window->rgb = *rgb;
	window->best_rgb = rgb->main ? &window->rgb : &window->pc;
    }
    else
    {
	window->pc = *rgb;
	window->best_rgb = &window->pc;
    }
    if (pc->main)
    {
	canvas_register_position_event_func
	    (pc->main,
	     _dispdata_array_canvas_position_func,
	     window);
	pixcanvas = canvas_get_pixcanvas (pc->main);
	kwin_get_colour_harder (pixcanvas, "green", &pixel_value);
	window->ctx_pc =
	    kwin_editor_create_context (pixcanvas,
					KWIN_EDITOR_CONTEXT_ATT_MOUSE,
					    K_CANVAS_EVENT_MIDDLE_MOUSE_MASK,
					KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT,
					    KWIN_EDIT_OBJECT_NONE,
					KWIN_EDITOR_CONTEXT_ATT_PIXEL_VALUE,
					    pixel_value,
					KWIN_EDITOR_CONTEXT_ATT_CLEAR_OBJECT,
					    FALSE,
					KWIN_EDITOR_CONTEXT_ATT_END);
	kwin_editor_register_event_func (window->ctx_pc,
					 _dispdata_array_middle_geom_func,
					 window);
	window->pc_canvas.canvas = pc->main;
	window->pc_canvas.window = window;
    }
    if (rgb->main)
    {
	canvas_register_position_event_func
	    (rgb->main, _dispdata_array_canvas_position_func,
	     window);
	pixcanvas = canvas_get_pixcanvas (rgb->main);
	kwin_get_colour_harder (pixcanvas, "green", &pixel_value);
	window->ctx_rgb =
	    kwin_editor_create_context (pixcanvas,
					KWIN_EDITOR_CONTEXT_ATT_MOUSE,
					    K_CANVAS_EVENT_MIDDLE_MOUSE_MASK,
					KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT,
					    KWIN_EDIT_OBJECT_NONE,
					KWIN_EDITOR_CONTEXT_ATT_PIXEL_VALUE,
					    pixel_value,
					KWIN_EDITOR_CONTEXT_ATT_CLEAR_OBJECT,
					    FALSE,
					KWIN_EDITOR_CONTEXT_ATT_END);
	kwin_editor_register_event_func (window->ctx_rgb,
					 _dispdata_array_middle_geom_func,
					 window);
	window->rgb_canvas.canvas = rgb->main;
	window->rgb_canvas.window = window;
    }
    if (pc->renzogram_cmap)
	window->pc_cmap_callback =
	    kcmap_register_resize_func (pc->renzogram_cmap,
					_dispdata_array_cmap_modify_func,
					&window->pc_canvas);
    if (rgb->renzogram_cmap)
	window->rgb_cmap_callback =
	    kcmap_register_resize_func (rgb->renzogram_cmap,
					_dispdata_array_cmap_modify_func,
					&window->rgb_canvas);
    window->auto_title = TRUE;
    KLIST_APPEND2 (first_window, last_window, window);
    return (window);
}   /*  End Function _dispdata_array_create_window  */

static void _dispdata_array_destroy_window (void *window_p)
/*  [SUMMARY] Deallocate a window.
    [RETURNS] Nothing.
*/
{
    struct window *window = window_p;

    KLIST_REMOVE2 (first_window, last_window, window);
    c_unregister_callback2 (window->pc_cmap_callback);
    c_unregister_callback2 (window->rgb_cmap_callback);
    m_free (window);
}   /*  End Function _dispdata_array_destroy_window  */

static flag _dispdata_array_set_window_attributes (void *win, va_list argp)
{
    KWorldCanvas profile_canvas = NULL;
    flag slice_changed = FALSE;
    flag refresh_main = FALSE;
    flag create_profiles = FALSE;
    flag do_loop = FALSE;
    flag invalidate_profiles = FALSE;
    flag refresh_profiles = FALSE;
    flag all_known = TRUE;
    flag set_title = FALSE;
    flag update_vlines = FALSE;
    flag update_can_use_kcmap = FALSE;
    flag can_show_vlines = FALSE;
    flag bool;
    unsigned int att_key, uival, dummy;
    unsigned int editor_object = 0;  /*  Initialised to keep compiler happy  */
    unsigned int draw_style = GRAPH1D_DRAW_STYLE_DEFAULT;
    unsigned int num_frames = 0;
    struct window *window = win;
    struct viewable *viewable;
    struct blinkstate *blinkstate = window->active_blinkstate;
    unsigned int *framenum = NULL;
    static char function_name[] = "_dispdata_array_set_window_attributes";

    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    KDISPLAYDATA_WINDOW_ATT_END )
    {
	switch (att_key)
	{
	  case KDISPLAYDATA_WINDOW_ATT_SLICE_DIRECTION:
	    uival = va_arg (argp, unsigned int);
	    switch (uival)
	    {
	      case KDISPLAYDATA_SLICE_DIRECTION_XY:
	      case KDISPLAYDATA_SLICE_DIRECTION_XZ:
	      case KDISPLAYDATA_SLICE_DIRECTION_ZY:
		break;
	      default:
		fprintf (stderr, "Invalid direction: %u\n", uival);
		a_prog_bug (function_name);
		break;
	    }
	    if (window->slice_direction == uival) break;
	    window->slice_direction = uival;
	    slice_changed = TRUE;
	    update_vlines = TRUE;
	    update_can_use_kcmap = TRUE;
	    refresh_profiles = TRUE;
	    do_loop = TRUE;
	    break;
	  case KDISPLAYDATA_WINDOW_ATT_SET_ISCALE:
	    bool = va_arg (argp, flag);
	    FLAG_VERIFY (bool);
	    window->set_iscale = bool;
	    break;
	  case KDISPLAYDATA_WINDOW_ATT_FRAME_NUMBER:
	    uival = va_arg (argp, unsigned int);
	    if (!blinkstate) break;
	    viewable = blinkstate->movie_master;
	    if (!viewable)
	    {
		fprintf (stderr, "%s: no movie master found\n", function_name);
		break;
	    }
	    switch (window->slice_direction)
	    {
	      case KDISPLAYDATA_SLICE_DIRECTION_XY:
		framenum = &viewable->dataset->attributes.xy_frame.index;
		num_frames = viewable->dataset->data->num_xy_frames;
		break;
	      case KDISPLAYDATA_SLICE_DIRECTION_XZ:
		framenum = &viewable->dataset->attributes.xz_frame.index;
		num_frames = viewable->dataset->data->num_xz_frames;
		break;
	      case KDISPLAYDATA_SLICE_DIRECTION_ZY:
		framenum = &viewable->dataset->attributes.zy_frame.index;
		num_frames = viewable->dataset->data->num_zy_frames;
		break;
	    }
	    if (uival >= num_frames) uival = num_frames - 1;
	    *framenum = uival;
	    refresh_profiles = TRUE;
	    break;
	  case KDISPLAYDATA_WINDOW_ATT_PROFILE_CANVAS:
	    profile_canvas = va_arg (argp, KWorldCanvas);
	    if (window->profile_canvas == profile_canvas) break;
	    if (window->profile_canvas)
	    {
		fprintf (stderr,
			 "Cannot change profile canvas once created\n");
		a_prog_bug (function_name);
	    }
	    window->profile_canvas = profile_canvas;
	    window->profile_pixcanvas = canvas_get_pixcanvas (profile_canvas);
	    graph1d_set_canvas_attributes (window->profile_canvas,
					   GRAPH1D_CANVAS_ATT_SHOW_VLINE,
					       window->show_marker_in_profile,
					   GRAPH1D_CANVAS_ATT_END);
	    create_profiles = TRUE;
	    update_vlines = TRUE;
	    update_can_use_kcmap = TRUE;
	    refresh_profiles = TRUE;
	    do_loop = TRUE;
	    break;
	  case KDISPLAYDATA_WINDOW_ATT_PROFILE_MODE:
	    uival = va_arg (argp, unsigned int);
	    switch (uival)
	    {
	      case KDISPLAYDATA_PROFILE_MODE_NONE:
	      case KDISPLAYDATA_PROFILE_MODE_LINE:
		editor_object = KWIN_EDIT_OBJECT_NONE;
		break;
	      case KDISPLAYDATA_PROFILE_MODE_BOX_SUM:
	      case KDISPLAYDATA_PROFILE_MODE_BOX_AVERAGE:
		editor_object = KWIN_EDIT_OBJECT_RECTANGLE;
		break;
	      case KDISPLAYDATA_PROFILE_MODE_RADIAL:
	      case KDISPLAYDATA_PROFILE_MODE_RADIAL_GAUSS:
		draw_style = GRAPH1D_DRAW_STYLE_CROSSES;
		/*  Fall through  */
	      case KDISPLAYDATA_PROFILE_MODE_AZIMUTHAL_AVERAGE:
		editor_object = KWIN_EDIT_OBJECT_CIRCLE;
		break;
	      case KDISPLAYDATA_PROFILE_MODE_BOXED_HORIZONTAL:
	      case KDISPLAYDATA_PROFILE_MODE_BOXED_VERTICAL:
		editor_object = KWIN_EDIT_OBJECT_RECTANGLE;
		break;
	      default:
		fprintf (stderr, "Invalid profile mode: %u\n", uival);
		a_prog_bug (function_name);
		break;
	    }
	    if (window->profile_mode == uival) break;
	    window->profile_mode = uival;
	    if (window->ctx_pc)
		kwin_editor_set_attributes(window->ctx_pc,
					   KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT,
					       editor_object,
					   KWIN_EDITOR_CONTEXT_ATT_END);
	    if (window->ctx_rgb)
		kwin_editor_set_attributes(window->ctx_rgb,
					   KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT,
					       editor_object,
					   KWIN_EDITOR_CONTEXT_ATT_END);
	    window->profile_unzoom = TRUE;
	    if (!window->profile_canvas) break;
	    invalidate_profiles = TRUE;
	    update_vlines = TRUE;
	    update_can_use_kcmap = TRUE;
	    refresh_profiles = TRUE;
	    do_loop = TRUE;
	    break;
	  case KDISPLAYDATA_WINDOW_ATT_PROFILE_DIRECTION:
	    uival = va_arg (argp, unsigned int);
	    switch (uival)
	    {
	      case KDISPLAYDATA_PROFILE_DIR_Z:
	      case KDISPLAYDATA_PROFILE_DIR_Y:
	      case KDISPLAYDATA_PROFILE_DIR_X:
		break;
	      default:
		fprintf (stderr, "Invalid profile direction: %u\n", uival);
		a_prog_bug (function_name);
		break;
	    }
	    if (window->profile_direction == uival) break;
	    window->profile_direction = uival;
	    window->profile_unzoom = TRUE;
	    if (!window->profile_canvas) break;
	    invalidate_profiles = TRUE;
	    update_vlines = TRUE;
	    update_can_use_kcmap = TRUE;
	    refresh_profiles = TRUE;
	    do_loop = TRUE;
	    break;
	  case KDISPLAYDATA_WINDOW_ATT_AUTO_TITLE:
	    bool = va_arg (argp, flag);
	    FLAG_VERIFY (bool);
	    if (window->auto_title == bool) break;
	    window->auto_title = bool;
	    if (bool) set_title = TRUE, refresh_main = TRUE;
	    break;
	  case KDISPLAYDATA_WINDOW_ATT_SHOW_MARKER_IN_PROFILE:
	    bool = va_arg (argp, flag);
	    FLAG_VERIFY (bool);
	    if (window->show_marker_in_profile == bool) break;
	    window->show_marker_in_profile = bool;
	    if (window->profile_canvas)
		graph1d_set_canvas_attributes (window->profile_canvas,
					       GRAPH1D_CANVAS_ATT_SHOW_VLINE,
					           bool,
					       GRAPH1D_CANVAS_ATT_END);
	    refresh_profiles = TRUE;
	    break;
	  default:
	    dummy = va_arg (argp, unsigned int);
	    all_known = FALSE;
	    break;
	}
    }
    if (!window->first) return (all_known);
    if (set_title) _dispdata_array_set_title (window);
    if ( framenum && _dispdata_array_change_frame (window, *framenum) )
	refresh_main = TRUE;
    if ( _dispdata_array_test_profile_perpendicular (window) )
	can_show_vlines = TRUE;
    switch (window->profile_mode)
    {
      case KDISPLAYDATA_PROFILE_MODE_LINE:
      case KDISPLAYDATA_PROFILE_MODE_BOX_SUM:
      case KDISPLAYDATA_PROFILE_MODE_BOX_AVERAGE:
	break;
      default:
	can_show_vlines = FALSE;
	break;
    }
    for (viewable = window->first; viewable && do_loop;
	 viewable = viewable->window_chain.next)
    {
	struct data_container *data = viewable->dataset->data;
	struct blinkentry *blinkentry = viewable->active_blinkentry;

	if (create_profiles && data->pseudo)
	{
	    struct data_attributes *att = &viewable->dataset->attributes;

	    viewable->profile =
		graph1d_orthogonal_create (profile_canvas, data->pseudo,
					   &viewable->vgraph);
	    graph1d_orthogonal_register_destroy_func
		(viewable->profile, c_write_null_func, &viewable->profile);
	    graph1d_set_vgraph_attributes
		(viewable->vgraph,
		 GRAPH1D_CANVAS_ATT_DRAW_STYLE, draw_style,
		 GRAPH1D_VGRAPH_ATT_DARK_COLOUR, att->profile_dark_colour,
		 GRAPH1D_VGRAPH_ATT_LIGHT_COLOUR, att->profile_light_colour,
		 GRAPH1D_VGRAPH_ATT_END);
	    viewable->gausscurve = graph1d_create
		(profile_canvas, 0, NULL, NULL, 0,
		 GRAPH1D_CANVAS_ATT_DRAW_STYLE, GRAPH1D_DRAW_STYLE_LINES,
		 GRAPH1D_VGRAPH_ATT_DARK_COLOUR, att->profile_dark_colour,
		 GRAPH1D_VGRAPH_ATT_LIGHT_COLOUR, att->profile_light_colour,
		 GRAPH1D_VGRAPH_ATT_END);
	    graph1d_register_destroy_func
		(viewable->gausscurve,c_write_null_func,&viewable->gausscurve);
	}
	if (invalidate_profiles && viewable->profile)
	{   /* No profile for RGB */
	    graph1d_set_vgraph_attributes
		(viewable->vgraph,
		 GRAPH1D_VGRAPH_ATT_DRAW_STYLE, draw_style,
		 GRAPH1D_VGRAPH_ATT_END);
	    graph1d_orthogonal_invalidate (viewable->profile);
	    graph1d_set_size (viewable->gausscurve, 0);
	}
	if (slice_changed && blinkentry && (data->num_xy_frames > 0) &&
	     ( (blinkentry->blinkstate->image_viewable == viewable) ||
	       (blinkentry->contour_mode != KDISPLAYDATA_CONTOUR_OFF) ) &&
	    !framenum)
	{
	    _dispdata_array_activate_frames (viewable);
	    refresh_main = TRUE;
	}
	if (update_vlines && !framenum && viewable->vgraph)
	{
	    double coord = TOOBIG;

	    if (can_show_vlines)
	    {
		switch (window->slice_direction)
		{
		  case KDISPLAYDATA_SLICE_DIRECTION_XY:
		    coord = viewable->dataset->attributes.xy_frame.coord;
		    break;
		  case KDISPLAYDATA_SLICE_DIRECTION_XZ:
		    coord = viewable->dataset->attributes.xz_frame.coord;
		    break;
		  case KDISPLAYDATA_SLICE_DIRECTION_ZY:
		    coord = viewable->dataset->attributes.zy_frame.coord;
		    break;
		}
	    }
	    graph1d_set_vgraph_attributes (viewable->vgraph,
					   GRAPH1D_VGRAPH_ATT_VLINE_X, coord,
					   GRAPH1D_VGRAPH_ATT_END);
	}
	if (update_can_use_kcmap && viewable->vgraph)
	{
	    flag can_use_kcmap = FALSE;

	    if ( (blinkentry->contour_mode == KDISPLAYDATA_CONTOUR_RENZOGRAM)
		 && can_show_vlines )
		can_use_kcmap = TRUE;
	    graph1d_set_vgraph_attributes (viewable->vgraph,
					   GRAPH1D_VGRAPH_ATT_USE_COLOURMAP,
					       can_use_kcmap,
					   GRAPH1D_VGRAPH_ATT_END);
	}
    }
    if (refresh_main) dispdata_refresh_window (window->superclass, NULL);
    if (window->active_blinkstate && (create_profiles || invalidate_profiles)
	&& window->active_blinkstate->profile_master)
    {
	viewable = window->active_blinkstate->profile_master;
	_dispdata_array_set_profile_vscale (viewable);
	canvas_use_astro_transform (window->profile_canvas,
				    viewable->dataset->data->ap);
    }
    if (create_profiles)
    {
	graph1d_change_draw_order (blinkstate->profile_master->vgraph, NULL,
				   FALSE);
	graph1d_change_draw_order (blinkstate->profile_master->gausscurve,
				   viewable->vgraph, TRUE);
    }
    if (refresh_profiles && window->profile_canvas)
	graph1d_refresh_if_needed (window->profile_canvas);
    return (all_known);
}   /*  End Function _dispdata_array_set_window_attributes  */

static flag _dispdata_array_get_window_attributes (void *win, va_list argp)
{
    flag all_known = TRUE;
    unsigned int att_key;
    struct window *window = win;
    struct blinkstate *blinkstate = window->active_blinkstate;
    void *data;
    /*static char function_name[] = "_dispdata_array_get_window_attributes";*/

    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    KDISPLAYDATA_WINDOW_ATT_END )
    {
	switch (att_key)
	{
	  case KDISPLAYDATA_WINDOW_ATT_SLICE_DIRECTION:
	    *va_arg (argp, unsigned int *) = window->slice_direction;
	    break;
	  case KDISPLAYDATA_WINDOW_ATT_SET_ISCALE:
	    *va_arg (argp, flag *) = window->set_iscale;
	    break;
	  case KDISPLAYDATA_WINDOW_ATT_IMAGE:
	    if (blinkstate && blinkstate->image_viewable)
		data = blinkstate->image_viewable->dataset->superclass;
	    else data = NULL;
	    *va_arg (argp, void **) = data;
	    break;
	  case KDISPLAYDATA_WINDOW_ATT_MOVIE_MASTER:
	    if (blinkstate && blinkstate->movie_master)
		data = blinkstate->movie_master->dataset->superclass;
	    else data = NULL;
	    *va_arg (argp, void **) = data;
	    break;
	  case KDISPLAYDATA_WINDOW_ATT_PROFILE_MASTER:
	    if (blinkstate && blinkstate->profile_master)
		data = blinkstate->profile_master->dataset->superclass;
	    else data = NULL;
	    *va_arg (argp, void **) = data;
	    break;
	  default:
	    data = va_arg (argp, void *);
	    all_known = FALSE;
	    break;
	}
    }
    return (all_known);
}   /*  End Function _dispdata_array_get_window_attributes  */

static void *_dispdata_array_create_viewable (void *data, void *window_p)
{
    struct viewable *viewable;
    struct window *window = window_p;
    struct dataset *set = data;
    struct data_attributes *att = &set->attributes;
    static char function_name[] = "_dispdata_array_create_viewable";

    if ( ( viewable = m_calloc (sizeof *viewable) ) == NULL )
    {
	m_error_notify (function_name, "window/data information");
	return (NULL);
    }
    viewable->dataset = set;
    viewable->window = window;
    KLIST_APPEND3 (set, dataset_chain, viewable);
    KLIST_APPEND3 (window, window_chain, viewable);
    if (set->data->pseudo)
    {
	if (window->pc.main)
	{
	    if ( ( viewable->map_pc_vcimage_group =
		   contour_create_viewable_group () ) == NULL )
	    {
		_dispdata_array_destroy_viewable (viewable, FALSE);
		return (NULL);
	    }
	    if ( ( viewable->renzo_pc_vcimage_group =
		   contour_create_viewable_group () ) == NULL )
	    {
		_dispdata_array_destroy_viewable (viewable, FALSE);
		return (NULL);
	    }
	}
	if (window->rgb.main)
	{
	    if ( ( viewable->map_rgb_vcimage_group =
		   contour_create_viewable_group () ) == NULL )
	    {
		_dispdata_array_destroy_viewable (viewable, FALSE);
		return (NULL);
	    }
	    if ( ( viewable->renzo_rgb_vcimage_group =
		   contour_create_viewable_group () ) == NULL )
	    {
		_dispdata_array_destroy_viewable (viewable, FALSE);
		return (NULL);
	    }
	}
    }
    if ( (set->data->pseudo) && (iarray_num_dim (set->data->pseudo) == 2) )
    {
	if ( !_dispdata_array_setup_2D_pseudo (viewable, set->data->pseudo) )
	{
	    _dispdata_array_destroy_viewable (viewable, FALSE);
	    return (NULL);
	}
    }
    if ( (set->data->pseudo) && (iarray_num_dim (set->data->pseudo) == 3) )
    {
	if ( !_dispdata_array_setup_3D_pseudo (viewable, set->data->pseudo) )
	{
	    _dispdata_array_destroy_viewable (viewable, FALSE);
	    return (NULL);
	}
    }
    if ( (set->data->red) && (iarray_num_dim (set->data->red) == 2) )
    {
	if ( !_dispdata_array_setup_2D_rgb (viewable, set->data->red,
					    set->data->green,set->data->blue) )
	{
	    _dispdata_array_destroy_viewable (viewable, FALSE);
	    return (NULL);
	}
    }
    if ( (set->data->red) && (iarray_num_dim (set->data->red) == 3) )
    {
	if ( !_dispdata_array_setup_3D_rgb (viewable, set->data->red,
					    set->data->green,set->data->blue) )
	{
	    _dispdata_array_destroy_viewable (viewable, FALSE);
	    return (NULL);
	}
    }
    if (window->profile_canvas && set->data->pseudo)
    {
	unsigned int draw_style = GRAPH1D_DRAW_STYLE_DEFAULT;
	double coord = TOOBIG;

	switch (window->slice_direction)
	{
	  case KDISPLAYDATA_SLICE_DIRECTION_XY:
	    if (window->profile_direction == KDISPLAYDATA_PROFILE_DIR_Z)
		coord = viewable->dataset->attributes.xy_frame.coord;
	    break;
	  case KDISPLAYDATA_SLICE_DIRECTION_XZ:
	    if (window->profile_direction == KDISPLAYDATA_PROFILE_DIR_Y)
		coord = viewable->dataset->attributes.xz_frame.coord;
	    break;
	  case KDISPLAYDATA_SLICE_DIRECTION_ZY:
	    if (window->profile_direction == KDISPLAYDATA_PROFILE_DIR_X)
		coord = viewable->dataset->attributes.zy_frame.coord;
	    break;
	}
	switch (window->profile_mode)
	{
	  case KDISPLAYDATA_PROFILE_MODE_LINE:
	  case KDISPLAYDATA_PROFILE_MODE_BOX_SUM:
	  case KDISPLAYDATA_PROFILE_MODE_BOX_AVERAGE:
	    break;
	  case KDISPLAYDATA_PROFILE_MODE_RADIAL:
	  case KDISPLAYDATA_PROFILE_MODE_RADIAL_GAUSS:
	    draw_style = GRAPH1D_DRAW_STYLE_CROSSES;
	    /*  Fall through  */
	  case KDISPLAYDATA_PROFILE_MODE_AZIMUTHAL_AVERAGE:
	    coord = TOOBIG;
	    break;
	  default:
	    coord = TOOBIG;
	    break;
	}
	viewable->profile =
	    graph1d_orthogonal_create (window->profile_canvas,
				       set->data->pseudo, &viewable->vgraph);
	graph1d_orthogonal_register_destroy_func
	    (viewable->profile, c_write_null_func, &viewable->profile);
	graph1d_set_vgraph_attributes
	    (viewable->vgraph,
	     GRAPH1D_CANVAS_ATT_DRAW_STYLE, draw_style,
	     GRAPH1D_VGRAPH_ATT_DARK_COLOUR, att->profile_dark_colour,
	     GRAPH1D_VGRAPH_ATT_LIGHT_COLOUR, att->profile_light_colour,
	     GRAPH1D_VGRAPH_ATT_VLINE_X, coord,
	     GRAPH1D_VGRAPH_ATT_END);
	viewable->gausscurve = graph1d_create
	    (window->profile_canvas, 0, NULL, NULL, 0,
	     GRAPH1D_CANVAS_ATT_DRAW_STYLE, GRAPH1D_DRAW_STYLE_LINES,
	     GRAPH1D_VGRAPH_ATT_DARK_COLOUR, att->profile_dark_colour,
	     GRAPH1D_VGRAPH_ATT_LIGHT_COLOUR, att->profile_light_colour,
	     GRAPH1D_VGRAPH_ATT_END);
	graph1d_register_destroy_func
	    (viewable->gausscurve, c_write_null_func, &viewable->gausscurve);
	graph1d_change_draw_order (viewable->gausscurve, NULL, TRUE);
	graph1d_change_draw_order (viewable->vgraph, NULL, TRUE);
    }
    return (viewable);
}   /*  End Function _dispdata_array_create_viewable  */

static void _dispdata_array_destroy_viewable (void *viewable_p,
					      flag canvas_destroy)
/*  [SUMMARY] Deallocate the cache for data on a window.
    [RETURNS] Nothing.
*/
{
    struct viewable *viewable = viewable_p;
    struct dataset *dataset = viewable->dataset;
    struct data_container *data = dataset->data;

    KLIST_REMOVE3 (dataset, dataset_chain, viewable);
    KLIST_REMOVE3 (viewable->window, window_chain, viewable);
    if (!canvas_destroy)
    {
	if (viewable->vimage) viewimg_destroy (viewable->vimage);
	if (viewable->mag_vimage) viewimg_destroy (viewable->mag_vimage);
	if (viewable->pan_vimage) viewimg_destroy (viewable->pan_vimage);
	_dispdata_array_dealloc_direction_viewable (&viewable->xy,
						    data->num_xy_frames);
	_dispdata_array_dealloc_direction_viewable (&viewable->xz,
						    data->num_xz_frames);
	_dispdata_array_dealloc_direction_viewable (&viewable->zy,
						    data->num_zy_frames);
    }
    if (viewable->map_pc_vcimage_group)
	contour_destroy_viewable_group (viewable->map_pc_vcimage_group);
    if (viewable->renzo_pc_vcimage_group)
	contour_destroy_viewable_group (viewable->renzo_pc_vcimage_group);
    if (viewable->map_rgb_vcimage_group)
	contour_destroy_viewable_group (viewable->map_rgb_vcimage_group);
    if (viewable->renzo_rgb_vcimage_group)
	contour_destroy_viewable_group (viewable->renzo_rgb_vcimage_group);
    if (viewable->profile) graph1d_orthogonal_destroy (viewable->profile);
    if (viewable->gausscurve) graph1d_destroy (viewable->gausscurve);
    m_free (viewable);
    if (viewable->window->auto_title)
	_dispdata_array_set_title (viewable->window);
}   /*  End Function _dispdata_array_destroy_viewable  */

static void *_dispdata_array_create_blinkstate
    (KDisplayDataClassBlinkState superclass, void *window)
{
    struct blinkstate *blinkstate;

    if ( ( blinkstate = m_calloc (sizeof *blinkstate) ) == NULL ) return NULL;
    blinkstate->superclass = superclass;
    blinkstate->window = window;
    return (blinkstate);
}   /*  End Function _dispdata_array_create_blinkstate  */

static void _dispdata_array_destroy_blinkstate (void *blinkstate_p)
{
    struct blinkstate *blinkstate = blinkstate_p;

    if (blinkstate->window->active_blinkstate == blinkstate)
	blinkstate->window->active_blinkstate = NULL;
    m_free (blinkstate);
}   /*  End Function _dispdata_array_destroy_blinkstate  */

static void _dispdata_array_activate_blinkstate (void *blinkstate_p,
						 flag before_blinkentries)
{
    struct blinkstate *blinkstate = blinkstate_p;
    /*static char function_name[] = "_dispdata_array_activate_blinkstate";*/

    blinkstate->window->active_blinkstate = blinkstate;
    if (!before_blinkentries && blinkstate->window->auto_title)
	_dispdata_array_set_title (blinkstate->window);
    if (!before_blinkentries && blinkstate->window->profile_canvas)
	graph1d_refresh_if_needed (blinkstate->window->profile_canvas);
}   /*  End Function _dispdata_array_activate_blinkstate  */

static void *_dispdata_array_create_blinkentry (void *blinkstate,
						void *viewable)
{
    struct blinkentry *blinkentry;

    if ( ( blinkentry = m_calloc (sizeof *blinkentry) ) == NULL )
	return (NULL);
    blinkentry->blinkstate = blinkstate;
    blinkentry->viewable = viewable;
    if (!blinkentry->blinkstate->image_viewable)
    {
	/*  Since no image viewable is yet set, make me viewable. Ideally, this
	    should be done by the user interface, but the current interface
	    doesn't (because blinkstates can be "hidden" and thus not fully
	    managed). Fix this by writing a lower-level user interface which
	    contains policy decisions, which the current GUI sits on top of  */
	blinkentry->blinkstate->image_viewable = viewable;
    }
    return (blinkentry);
}   /*  End Function _dispdata_array_create_blinkentry  */

static flag _dispdata_array_destroy_blinkentry (void *blinkentry_p)
/*  [SUMMARY] Destroy a blinkentry.
    <blinkentry_p> The blinkentry subclass data.
    [RETURNS] TRUE if the viewable was set to visible in this blinkentry.
*/
{
    flag visible = FALSE;
    struct blinkentry *blinkentry = blinkentry_p;
    struct blinkstate *blinkstate = blinkentry->blinkstate;
    struct viewable *viewable = blinkentry->viewable;

    if (blinkstate->image_viewable == viewable)
    {
	struct viewable *new_viewable;

	/*  Find a new viewable to make the image viewable. Ideally, this
	    should be done by the user interface, but the current interface
	    doesn't (because blinkstates can be "hidden" and thus not fully
	    managed). Fix this by writing a lower-level user interface which
	    contains policy decisions, which the current GUI sits on top of  */
	new_viewable = viewable->window->first;
	if (new_viewable == viewable) new_viewable = NULL;
	blinkstate->image_viewable = new_viewable;
	visible = TRUE;
    }
    if (blinkentry->contour_mode != KDISPLAYDATA_CONTOUR_OFF) visible = TRUE;
    if (blinkstate->movie_master == viewable) blinkstate->movie_master = NULL;
    if (blinkstate->profile_master == viewable)
    {
	blinkstate->profile_master = NULL;
	if ( (blinkstate->window->active_blinkstate == blinkstate) &&
	     blinkstate->window->profile_canvas )
	{
	    canvas_use_astro_transform (blinkstate->window->profile_canvas,
					NULL);
	    canvas_specify (blinkstate->window->profile_canvas, NULL, NULL,
			    0, NULL, NULL);
	}
    }
    if (viewable->active_blinkentry == blinkentry)
	viewable->active_blinkentry = NULL;
    m_free (blinkentry);
    return (visible);
}   /*  End Function _dispdata_array_destroy_blinkentry  */

static void _dispdata_array_copy_blinkentry (void *dest_p, void *source_p)
{
    struct blinkentry *dest = dest_p;
    struct blinkentry *source = source_p;

    if (source->blinkstate->image_viewable == source->viewable)
	dest->blinkstate->image_viewable = dest->viewable;
    if (source->blinkstate->movie_master == source->viewable)
	dest->blinkstate->movie_master = dest->viewable;
    if (source->blinkstate->profile_master == source->viewable)
	dest->blinkstate->profile_master = dest->viewable;
    dest->contour_mode = source->contour_mode;
    dest->movie_slave = source->movie_slave;
    dest->profile_slave = source->profile_slave;
}   /*  End Function _dispdata_array_copy_blinkentry  */

static KWorldCanvas _dispdata_array_activate_blinkentry (void *blinkentry_p,
							 flag *visible)
{
    KWorldCanvas wc = NULL;
    struct blinkentry *blinkentry = blinkentry_p;
    struct viewable *viewable = blinkentry->viewable;
    struct dataset *dataset = viewable->dataset;
    struct window *window = viewable->window;
    /*static char function_name[] = "_dispdata_array_activate_blinkentry";*/

    viewable->active_blinkentry = blinkentry;
    _dispdata_array_activate_frames (viewable);
    if (blinkentry->blinkstate->image_viewable == viewable)
    {
	if (visible) *visible = TRUE;
	if (dataset->data->pseudo)
	{
	    double min;
	    double max;

	    wc = window->pc.main;
	    canvas_get_attributes (wc,
				   CANVAS_ATT_VALUE_MIN, &min,
				   CANVAS_ATT_VALUE_MAX, &max,
				   CANVAS_ATT_END);
	    if ( window->set_iscale || (min >= max) )
	    {
		window->min = dataset->attributes.image_lower_clip;
		window->max = dataset->attributes.image_upper_clip;
	    }
	    _dispdata_array_set_canvas_iscale (viewable);
	}
	else
	{
	    wc = window->best_rgb->main;
	}
    }
    if (!viewable->dataset->data->pseudo) return (wc);
    switch (blinkentry->contour_mode)
    {
      case KDISPLAYDATA_CONTOUR_NORMAL:
	if (visible) *visible = TRUE;
	_dispdata_array_set_contour_styles (viewable);
	break;
      case KDISPLAYDATA_CONTOUR_RENZOGRAM:
	if (visible) *visible = TRUE;
	_dispdata_array_set_renzogram_styles (viewable);
	break;
    }
    if (window->profile_canvas)
    {
	flag master = (blinkentry->blinkstate->profile_master ==
		       viewable) ? TRUE : FALSE;
	flag show = (blinkentry->profile_slave || master) ? TRUE : FALSE;
	flag can_use_kcmap = FALSE;

	graph1d_set_active (viewable->vgraph, show, FALSE, FALSE, FALSE,0);
	graph1d_set_active (viewable->gausscurve, show, FALSE, FALSE, FALSE,0);
	_dispdata_array_set_profile_vscale (viewable);
	if (master)
	{
	    canvas_use_astro_transform (window->profile_canvas,
					dataset->data->ap);
	    if ( !graph1d_test_if_zoomed_horizontally(window->profile_canvas) )
		window->profile_unzoom = TRUE;
	    graph1d_change_draw_order (viewable->vgraph, NULL, FALSE);
	    graph1d_change_draw_order (viewable->gausscurve, viewable->vgraph,
				       TRUE);
	}
	if ( _dispdata_array_test_profile_perpendicular (window)
	     && (blinkentry->contour_mode == KDISPLAYDATA_CONTOUR_RENZOGRAM) )
	{
	    switch (window->profile_mode)
	    {
	      case KDISPLAYDATA_PROFILE_MODE_LINE:
	      case KDISPLAYDATA_PROFILE_MODE_BOX_SUM:
	      case KDISPLAYDATA_PROFILE_MODE_BOX_AVERAGE:
		can_use_kcmap = TRUE;
		break;
	    }
	}
	graph1d_set_vgraph_attributes (viewable->vgraph,
				       GRAPH1D_VGRAPH_ATT_USE_COLOURMAP,
				           can_use_kcmap,
				       GRAPH1D_VGRAPH_ATT_END);
    }
    return (wc);
}   /*  End Function _dispdata_array_activate_blinkentry  */

static flag _dispdata_array_set_blinkentry_attributes (void *blinkentry_p,
						       va_list argp)
{
    KWorldCanvas wc;
    flag bool;
    flag refresh_main = FALSE;
    flag all_known = TRUE;
    unsigned int att_key, contour_mode, dummy;
    struct blinkentry *blinkentry = blinkentry_p;
    struct blinkstate *blinkstate = blinkentry->blinkstate;
    struct viewable *viewable = blinkentry->viewable;
    struct viewable *old_profile = NULL;
    struct dataset *dataset = viewable->dataset;
    struct window *window = viewable->window;

    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    KDISPLAYDATA_BLINKENTRY_ATT_END )
    {
	switch (att_key)
	{
	  case KDISPLAYDATA_BLINKENTRY_ATT_IMAGE:
	    bool = va_arg (argp, flag);
	    if (bool) blinkstate->image_viewable = viewable;
	    else if (blinkstate->image_viewable == viewable)
		blinkstate->image_viewable = NULL;
	    refresh_main = TRUE;
	    break;
	  case KDISPLAYDATA_BLINKENTRY_ATT_CONTOUR_MODE:
	    contour_mode = va_arg (argp, unsigned int);
	    if ( (contour_mode == KDISPLAYDATA_CONTOUR_RENZOGRAM) &&
		 ( (window->pc.main && !window->pc.renzogram_cmap) ||
		   (window->rgb.main && !window->rgb.renzogram_cmap) ) )
	    {
		fprintf (stderr, "No colourmaps for Renzograms available\n");
		break;
	    }
	    if (blinkentry->contour_mode == contour_mode) break;
	    if (viewable->dataset->data->pseudo)
		blinkentry->contour_mode = contour_mode;
	    refresh_main = TRUE;
	    break;
	  case KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_MASTER:
	    bool = va_arg (argp, flag);
	    if (dataset->data->num_xy_frames > 0)
	    {
		if (bool) blinkstate->movie_master = viewable;
		else if (blinkstate->movie_master == viewable)
		    blinkstate->movie_master = NULL;
	    }
	    break;
	  case KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_SLAVE:
	    bool = va_arg (argp, flag);
	    if (dataset->data->num_xy_frames > 0)
		blinkentry->movie_slave = bool;
	    break;
	  case KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_MASTER:
	    bool = va_arg (argp, flag);
	    if (!dataset->data->pseudo) break;
	    if (bool)
	    {
		if (blinkstate->profile_master == viewable) break;
		old_profile = blinkstate->profile_master;
		blinkstate->profile_master = viewable;
	    }
	    else if (blinkstate->profile_master == viewable)
		blinkstate->profile_master = NULL;
	    break;
	  case KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_SLAVE:
	    bool = va_arg (argp, flag);
	    if (!dataset->data->pseudo) break;
	    if (blinkentry->profile_slave == bool) break;
	    blinkentry->profile_slave = bool;
	    break;
	  default:
	    dummy = va_arg (argp, unsigned int);
	    all_known = FALSE;
	    break;
	}
    }
    if (window->active_blinkstate != blinkstate) return (all_known);
    wc = _dispdata_array_activate_blinkentry (blinkentry, NULL);
    if (refresh_main)
    {
	if (window->auto_title) _dispdata_array_set_title (window);
	dispdata_refresh_window (window->superclass, wc);
    }
    if (!window->profile_canvas) return (all_known);
    if (old_profile)
    {
	if (old_profile->active_blinkentry &&
	    !old_profile->active_blinkentry->profile_slave)
	{
	    graph1d_set_active (old_profile->vgraph,FALSE,FALSE,FALSE,FALSE,0);
	    graph1d_set_active (old_profile->gausscurve, FALSE, FALSE, FALSE,
				FALSE, 0);
	}
    }
    graph1d_refresh_if_needed (window->profile_canvas);
    return (all_known);
}   /*  End Function _dispdata_array_set_blinkentry_attributes  */

static flag _dispdata_array_get_blinkentry_attributes (void *blinkentry_p,
						       va_list argp)
{
    flag all_known = TRUE;
    unsigned int att_key;
    struct blinkentry *blinkentry = blinkentry_p;
    struct viewable *viewable = blinkentry->viewable;
    void *dummy;

    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    KDISPLAYDATA_BLINKENTRY_ATT_END )
    {
	switch (att_key)
	{
	  case KDISPLAYDATA_BLINKENTRY_ATT_IMAGE:
	    *va_arg (argp, flag *) = (blinkentry->blinkstate->image_viewable ==
				      viewable) ? TRUE : FALSE;
	    break;
	  case KDISPLAYDATA_BLINKENTRY_ATT_CONTOUR_MODE:
	    *va_arg (argp, unsigned int *) = blinkentry->contour_mode;
	    break;
	  case KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_MASTER:
	    *va_arg (argp, flag *) = (blinkentry->blinkstate->movie_master ==
				      viewable) ? TRUE : FALSE;
	    break;
	  case KDISPLAYDATA_BLINKENTRY_ATT_MOVIE_SLAVE:
	    *va_arg (argp, flag *) = blinkentry->movie_slave;
	    break;
	  case KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_MASTER:
	    *va_arg (argp, flag *) = (blinkentry->blinkstate->profile_master ==
				      viewable) ? TRUE : FALSE;
	    break;
	  case KDISPLAYDATA_BLINKENTRY_ATT_PROFILE_SLAVE:
	    *va_arg (argp, flag *) = blinkentry->profile_slave;
	    break;
	  case KDISPLAYDATA_BLINKENTRY_ATT_VIMAGE:
	    *va_arg (argp, ViewableImage *) =
		_dispdata_array_get_vimage (viewable);
	    break;
	  default:
	    dummy = va_arg (argp, void *);
	    break;
	}
    }
    return (all_known);
}   /*  End Function _dispdata_array_get_blinkentry_attributes  */


/*  Private functions follow  */

static flag _dispdata_array_data_event_func (void *object,
					     multi_array *multi_desc,
					     CONST char *domain,
					     CONST char *name)
/*  [SUMMARY] Process an event.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    flag captured_all = TRUE;
    unsigned int count;
    static char function_name[] = "_dispdata_array_data_event_func";

    if (!multi_desc) return (FALSE);
    if (multi_desc->num_arrays == 1)
	return _dispdata_array_create_dataobject (multi_desc, 0, domain, name);
    for (count = 0; count < multi_desc->num_arrays; ++count)
    {
	unsigned int namelen = strlen (name);
	char subname[STRING_LENGTH];

	if (strlen (multi_desc->array_names[count]) + 2 + namelen >
	    STRING_LENGTH)
	{
	    fprintf (stderr, "%s: combined names too large\n", function_name);
	    return (TRUE);
	}
	strcpy (subname, name);
	subname[namelen] = ':';
	strcpy (subname + namelen + 1, multi_desc->array_names[count]);
	if ( !_dispdata_array_create_dataobject (multi_desc, count, domain,
						 subname) )
	    captured_all = FALSE;
    }
    return (captured_all);
}   /*  End Function _dispdata_array_data_event_func  */

static flag _dispdata_array_create_dataobject (multi_array *multi_desc,
					       unsigned int index,
					       CONST char *domain,
					       CONST char *name)
/*  [SUMMARY] Create a dataobject.
    <multi_desc> The multi_array descriptor.
    <index> The index of the general data structure within <<multi_desc>>.
    <domain> The domain from where the event was generated. The string contents
    are copied.
    <name> The name of the event in its domain. The string contents are copied.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag retval;
    iarray pseudo, red, green, blue;
    unsigned int cmap_index;
    char *array_name = NULL;
    /*static char function_name[] = "_dispdata_array_create_dataobject";*/

    if (multi_desc->num_arrays > 1)
    {
	unsigned int count;
	packet_desc *pack_desc = multi_desc->headers[index];

	/*  Quick check for array descriptor  */
	for (count = 0; count < pack_desc->num_elements; ++count)
	    if (pack_desc->element_types[count] == K_ARRAY) break;
	if (count >= pack_desc->num_elements) return (TRUE);
    }
    if (multi_desc->num_arrays > 1) array_name =multi_desc->array_names[index];
    if ( iarray_get_nD_from_multi (multi_desc, array_name, 4, &pseudo,
				   &red, &green, &blue, &cmap_index) )
	return _dispdata_array_split_hypercube (pseudo, red, green, blue,
						domain, name);
    /*  Try to get 2-D array (image)  */
    if ( !iarray_get_nD_from_multi (multi_desc, array_name, 2, &pseudo,
				    &red, &green, &blue, &cmap_index) )
    {   /*  That failed, so try 3-D array (movie)  */
	if ( !iarray_get_nD_from_multi (multi_desc, array_name, 3, &pseudo,
					&red, &green, &blue, &cmap_index) )
	{
	    if (array_name)
		fprintf (stderr, "Error getting image/movie in: \"%s\"\n",
			 array_name);
	    else fprintf (stderr, "Error getting image/movie\n");
	    return (FALSE);
	}
    }
    retval = _dispdata_array_create_dataobject2 (pseudo, red, green, blue,
						 domain, name, FALSE, TRUE);
    if (!retval)
    {
	if (pseudo) iarray_dealloc (pseudo);
	if (red) iarray_dealloc (red);
	if (green) iarray_dealloc (green);
	if (blue) iarray_dealloc (blue);
    }
    return (retval);
}   /*  End Function _dispdata_array_create_dataobject  */

static flag _dispdata_array_create_dataobject2 (iarray pseudo, iarray red,
						iarray green, iarray blue,
						CONST char *domain,
						CONST char *name, flag quiet,
						flag set_range_ok)
/*  [SUMMARY] Create a dataobject from Intelligent Arrays.
    <pseudo> The PseudoColour Intelligent Array.
    <red> The red component Intelligent Array for RGB data.
    <green> The green component Intelligent Array for RGB data.
    <blue> The blue component Intelligent Array for RGB data.
    <domain> The domain from where the event was generated. The string contents
    are copied.
    <name> The name of the event in its domain. The string contents are copied.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    iarray some_arr = pseudo ? pseudo : red;
    struct dataset *set;
    struct data_container *data;
    static char function_name[] = "_dispdata_array_create_dataobject2";

    if ( ( set = m_calloc (sizeof *set) ) == NULL )
    {
	m_error_notify (function_name, "dataset object");
	return (FALSE);
    }
    if ( ( set->data = m_calloc (sizeof *set->data) ) == NULL )
    {
	m_error_notify (function_name, "dataset container object");
	m_free (set);
	return (FALSE);
    }
    data = set->data;
    data->refcount = 1;
    data->counter = 1;
    if ( ( data->name = st_dup (name) ) == NULL )
    {
	m_error_notify (function_name, "name");
	_dispdata_array_destroy_dataobject (set);
	return (FALSE);
    }
    if (iarray_num_dim (some_arr) == 3)
    {
	data->num_xy_frames = iarray_dim_length (some_arr, 0);
	data->num_xz_frames = iarray_dim_length (some_arr, 1);
	data->num_zy_frames = iarray_dim_length (some_arr, 2);
    }
    data->object = iarray_get_named_string (some_arr, "OBJECT");
    if (pseudo)
    {
	flag compute_range = TRUE;
	flag scaled;
	double tmp;
	struct minmax minmax;

	if ( ( set->attributes.contour_neg_colour = st_dup ("grey70") ) ==
	     NULL ) m_abort (function_name, "grey70 name");
	if ( ( set->attributes.contour_pos_colour = st_dup ("green") ) ==
	     NULL ) m_abort (function_name, "green name");
	if ( ( set->attributes.renzogram_channels_string =
	       st_dup ("Imin:max+1") ) == NULL )
	    m_abort (function_name, "renzogram channels string");
	if ( ( set->attributes.profile_dark_colour = st_dup ("black") ) ==
	     NULL ) m_abort (function_name, "black name");
	if ( ( set->attributes.profile_light_colour = st_dup ("white") ) ==
	     NULL ) m_abort (function_name, "white name");
	scaled = iarray_get_data_scaling (pseudo,
					  &data->scale, &data->offset);
	if ( iarray_get_data_range (pseudo,
				    &minmax.min, &minmax.max) )
	{
	    /*  Range was provided  */
	    if (!quiet) fprintf (stderr, "Stated scaled min: %g  max: %g\n",
				 minmax.min, minmax.max);
	    if (minmax.min >= minmax.max)
	    {
		fprintf (stderr,"Min is not less than max: ignoring header\n");
	    }
	    else
	    {
		compute_range = FALSE;
		if (scaled)
		{
		    data->scaled = minmax;
		    data->raw.min = (minmax.min - data->offset) / data->scale;
		    data->raw.max = (minmax.max - data->offset) / data->scale;
		    if (data->scale < 0.0)
			swap (data->raw.min, data->raw.max, tmp);
		    if (!quiet)
			fprintf (stderr, "Stated unscaled min: %g  max: %g\n",
				 data->raw.min, data->raw.max);
		}
		else data->scaled = data->raw = minmax;
	    }
	}
	if (compute_range)
	{
	    if ( !iarray_min_max (pseudo, CONV1_REAL,
				  &data->raw.min, &data->raw.max) )
	    {
		fprintf (stderr, "Error computing min-max\n");
		_dispdata_array_destroy_dataobject (set);
		return (FALSE);
	    }
	    data->scaled.min = data->raw.min * data->scale + data->offset;
	    data->scaled.max = data->raw.max * data->scale + data->offset;
	    if (data->scale < 0.0)
		swap (data->scaled.min, data->scaled.max, tmp);
	    if (set_range_ok)
		iarray_set_data_range (pseudo,
				       data->scaled.min, data->scaled.max);
	    if (!quiet)
	    {
		if (scaled) fprintf (stderr,
				     "Computed scaled min: %g  max: %g\n",
				     data->scaled.min, data->scaled.max);
		fprintf (stderr, "Computed unscaled min: %g  max: %g\n",
			 data->raw.min, data->raw.max);
	    }
	}
	if (data->raw.min <= -TOOBIG)
	{
	    fprintf (stderr, "min: %e  is not greater than %e\n",
		     data->raw.min, -TOOBIG);
	    _dispdata_array_destroy_dataobject (set);
	    return (FALSE);
	}
	if ( DOUBLE_IS_BLANK (data->raw.max) )
	{
	    fprintf (stderr, "max: %e  is not less than %e\n",
		     data->raw.max, TOOBIG);
	    _dispdata_array_destroy_dataobject (set);
	    return (FALSE);
	}
	if (data->raw.min >= data->raw.max)
	{
	    fprintf (stderr, "Min is not less than max: bad data\n");
	    _dispdata_array_destroy_dataobject (set);
	    return (FALSE);
	}
	if (data->raw.min == data->raw.max)
	{
	    double step = 1.0;

	    fprintf (stderr,
		     "WARNING: min: %e is same as max: may cause problems\n",
		     data->raw.min);
	    while (data->raw.min == data->raw.max)
	    {
		data->raw.min -= step;
		step *= 10.0;
	    }
	    data->scaled.min = data->raw.min * data->scale + data->offset;
	    data->scaled.max = data->raw.max * data->scale + data->offset;
	}
	set->attributes.input_lower_clip = data->scaled.min;
	set->attributes.input_upper_clip = data->scaled.max;
	set->attributes.image_lower_clip = data->scaled.min;
	set->attributes.image_upper_clip = data->scaled.max;
	set->attributes.profile_lower_clip = data->scaled.min;
	set->attributes.profile_upper_clip = data->scaled.max;
	if ( !_dispdata_array_create_cimages (set, pseudo) )
	{
	    _dispdata_array_destroy_dataobject (set);
	    return (FALSE);
	}
    }
    data->ap = iarray_get_associated_wcs_astro (some_arr);
    /*  Compute frame co-ordinates before creating generic dataobject, which
	creates viewables  */
    data->pseudo = pseudo;
    data->red = red;
    data->green = green;
    data->blue = blue;
    _dispdata_array_compute_frame_coord (set, 0);
    _dispdata_array_compute_frame_coord (set, 1);
    _dispdata_array_compute_frame_coord (set, 2);
    if ( ( set->superclass = dispdata_create_dataobject (dataclass, domain,
							 name, set) )
	 == NULL )
    {
	m_error_notify (function_name, "dataobject");
	data->pseudo = NULL;
	data->red = NULL;
	data->green = NULL;
	data->blue = NULL;
	_dispdata_array_destroy_dataobject (set);
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function _dispdata_array_create_dataobject2  */

static flag _dispdata_array_split_hypercube (iarray pseudo, iarray red,
					     iarray green, iarray blue,
					     CONST char *domain,
					     CONST char *name)
/*  [SUMMARY] Split a hypercube into cube slices.
    <pseudo> The PseudoColour Intelligent Array.
    <red> The red component Intelligent Array for RGB data.
    <green> The green component Intelligent Array for RGB data.
    <blue> The blue component Intelligent Array for RGB data.
    <domain> The domain from where the event was generated. The string contents
    are copied.
    <name> The name of the event in its domain. The string contents are copied.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    iarray some_arr = pseudo ? pseudo : red;
    int precision;
    unsigned int num_cubes, count;
    /*static char function_name[] = "_dispdata_array_split_hypercube";*/

    num_cubes = iarray_dim_length (some_arr, 0);
    precision = ceil ( log10 (num_cubes) );
    for (count = 0; count < num_cubes; ++count)
    {
	char cubename[STRING_LENGTH];

	sprintf (cubename, "%s[cube%.*u]", name, precision, count);
	if (pseudo)
	{
	    iarray pseudo_cube;

	    if ( ( pseudo_cube =
		   iarray_get_3D_slice_from_4D (pseudo, 1, 2, 3, count) )
		 == NULL ) break;
	    if ( !_dispdata_array_create_dataobject2 (pseudo_cube,
						      NULL, NULL, NULL,
						      domain, cubename, TRUE,
						      FALSE) )
		iarray_dealloc (pseudo_cube);
	}
	else
	{
	    iarray red_cube, green_cube, blue_cube;

	    if ( ( red_cube =
		   iarray_get_3D_slice_from_4D (red, 1, 2, 3, count) )
		 == NULL ) break;
	    if ( ( green_cube =
		   iarray_get_3D_slice_from_4D (green, 1, 2, 3, count) )
		 == NULL )
	    {
		iarray_dealloc (red_cube);
		break;
	    }
	    if ( ( blue_cube =
		   iarray_get_3D_slice_from_4D (blue, 1, 2, 3, count) )
		 == NULL )
	    {
		iarray_dealloc (green_cube);
		iarray_dealloc (red_cube);
		break;
	    }
	    if ( !_dispdata_array_create_dataobject2 (NULL, red_cube,
						      green_cube, blue_cube,
						      domain, cubename, TRUE,
						      FALSE) )
	    {
		iarray_dealloc (blue_cube);
		iarray_dealloc (green_cube);
		iarray_dealloc (red_cube);
		break;
	    }
	}
    }
    if (pseudo) iarray_dealloc (pseudo);
    if (red) iarray_dealloc (red);
    if (green) iarray_dealloc (green);
    if (blue) iarray_dealloc (blue);
    return (count >= num_cubes) ? TRUE : FALSE;
}   /*  End Function _dispdata_array_split_hypercube  */

static flag _dispdata_array_create_cimages (struct dataset *set, iarray pseudo)
/*  [SUMMARY] Create ContourImage object(s) for a dataset.
    <set> The dataset.
    <pseudo> The PseudoColour Intelligent Array.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    if ( ( set->map_cimage_group = contour_create_group () ) == NULL )
	return (FALSE);
    if ( ( set->renzo_cimage_group = contour_create_group () ) == NULL )
	return (FALSE);
    if (iarray_num_dim (pseudo) == 2)
    {
	if ( !( set->map_cimage =
		contour_create_from_iarray (pseudo, FALSE,
					    set->map_cimage_group) ) )
	    return (FALSE);
    }
    else
    {
	if ( !( set->xy.map_cimages =
		contour_create_sequence_from_iarray
		(pseudo, 2, 1, 0, set->map_cimage_group) )
	     || !( set->xy.renzo_cimages =
		   contour_create_sequence_from_iarray
		   (pseudo, 2, 1, 0, set->renzo_cimage_group) )
	     || !( set->xz.map_cimages =
		   contour_create_sequence_from_iarray
		   (pseudo, 2, 0, 1, set->map_cimage_group) )
	     || !( set->xz.renzo_cimages =
		   contour_create_sequence_from_iarray
		   (pseudo, 2, 0, 1, set->renzo_cimage_group) )
	     || !( set->zy.map_cimages =
		   contour_create_sequence_from_iarray
		   (pseudo, 0, 1, 2, set->map_cimage_group) )
	     || !( set->zy.renzo_cimages =
		   contour_create_sequence_from_iarray
		   (pseudo, 0, 1, 2, set->renzo_cimage_group) ) )
	    return (FALSE);
    }
    return (TRUE);
}   /*  End Function _dispdata_array_create_cimages  */

static flag _dispdata_array_setup_2D_pseudo (struct viewable *viewable,
					     iarray array)
/*  [SUMMARY] Setup viewables for a 2D PsuedoColour array.
    <viewable> The cache for data on this window.
    <array> The Intelligent Array.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "_dispdata_array_setup_2D_pseudo";

    if ( ( viewable->vimage =
	   viewimg_create_from_iarray (viewable->window->pc.main, array,
				       FALSE) )  == NULL )
    {
	m_error_notify (function_name, "vimage");
	return (FALSE);
    }
    if ( ( viewable->mag_vimage =
	   viewimg_create_from_iarray (viewable->window->pc.magnifier,
				       array, FALSE) ) == NULL )
    {
	m_error_notify (function_name, "magnified vimage");
	return (FALSE);
    }
    if ( ( viewable->pan_vimage =
	   viewimg_create_from_iarray (viewable->window->pc.panner, array,
				       FALSE) ) == NULL )
    {
	m_error_notify (function_name, "magnified vimage");
	return (FALSE);
    }
    if ( ( viewable->map_pc_vcimage =
	   contour_create_viewable (viewable->dataset->map_cimage,
				    viewable->window->pc.main,
				    viewable->map_pc_vcimage_group) )
	 == NULL )
    {
	m_error_notify (function_name, "PC vcimage");
	return (FALSE);
    }
    if (viewable->window->rgb.main)
    {
	if ( ( viewable->map_rgb_vcimage =
	       contour_create_viewable (viewable->dataset->map_cimage,
					viewable->window->rgb.main,
					viewable->map_rgb_vcimage_group) )
	     == NULL )
	{
	    m_error_notify (function_name, "RGB vcimage");
	    return (FALSE);
	}
    }
    return (TRUE);
}   /*  End Function _dispdata_array_setup_2D_pseudo  */

static flag _dispdata_array_setup_3D_pseudo (struct viewable *viewable,
					     iarray array)
/*  [SUMMARY] Setup viewables for a 3D PsuedoColour array.
    <viewable> The cache for data on this window.
    <array> The Intelligent Array.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "_dispdata_array_setup_3D_pseudo";

    if ( ( viewable->xy.vimages = viewimg_create_sequence_from_iarray
	   (viewable->window->pc.main, array, 2, 1, 0) ) == NULL )
    {
	m_error_notify (function_name, "xy vimages");
	return (FALSE);
    }
    if ( ( viewable->xz.vimages = viewimg_create_sequence_from_iarray
	   (viewable->window->pc.main, array, 2, 0, 1) ) == NULL )
    {
	m_error_notify (function_name, "xz vimages");
	return (FALSE);
    }
    if ( ( viewable->zy.vimages = viewimg_create_sequence_from_iarray
	   (viewable->window->pc.main, array, 0, 1, 2) ) == NULL )
    {
	m_error_notify (function_name, "zy vimages");
	return (FALSE);
    }
    if ( ( viewable->xy.mag_vimages = viewimg_create_sequence_from_iarray
	   (viewable->window->pc.magnifier, array, 2, 1, 0) ) == NULL )
    {
	m_error_notify (function_name, "xy magnified vimages");
	return (FALSE);
    }
    if ( ( viewable->xz.mag_vimages = viewimg_create_sequence_from_iarray
	   (viewable->window->pc.magnifier, array, 2, 0, 1) ) == NULL )
    {
	m_error_notify (function_name, "xz magnified vimages");
	return (FALSE);
    }
    if ( ( viewable->zy.mag_vimages = viewimg_create_sequence_from_iarray
	   (viewable->window->pc.magnifier, array, 0, 1, 2) ) == NULL )
    {
	m_error_notify (function_name, "zy magnified vimages");
	return (FALSE);
    }
    if ( ( viewable->xy.pan_vimages = viewimg_create_sequence_from_iarray
	   (viewable->window->pc.panner, array, 2, 1, 0) ) == NULL )
    {
	m_error_notify (function_name, "xy panner vimages");
	return (FALSE);
    }
    if ( ( viewable->xz.pan_vimages = viewimg_create_sequence_from_iarray
	   (viewable->window->pc.panner, array, 2, 0, 1) ) == NULL )
    {
	m_error_notify (function_name, "xz panner vimages");
	return (FALSE);
    }
    if ( ( viewable->zy.pan_vimages = viewimg_create_sequence_from_iarray
	   (viewable->window->pc.panner, array, 0, 1, 2) ) == NULL )
    {
	m_error_notify (function_name, "zy panner vimages");
	return (FALSE);
    }
    if ( ( viewable->xy.map_pc_vcimages = contour_create_viewable_sequence
	   (viewable->dataset->xy.map_cimages,
	    viewable->dataset->data->num_xy_frames,
	    viewable->window->pc.main, viewable->map_pc_vcimage_group) )
	 == NULL )
    {
	m_error_notify (function_name, "xy vcimages");
	return (FALSE);
    }
    if ( ( viewable->xy.renzo_pc_vcimages = contour_create_viewable_sequence
	   (viewable->dataset->xy.renzo_cimages,
	    viewable->dataset->data->num_xy_frames,
	    viewable->window->pc.main, viewable->renzo_pc_vcimage_group) )
	 == NULL )
    {
	m_error_notify (function_name, "xy vcimages");
	return (FALSE);
    }
    if ( ( viewable->xz.map_pc_vcimages = contour_create_viewable_sequence
	   (viewable->dataset->xz.map_cimages,
	    viewable->dataset->data->num_xz_frames,
	    viewable->window->pc.main, viewable->map_pc_vcimage_group) )
	 == NULL )
    {
	m_error_notify (function_name, "xz vcimages");
	return (FALSE);
    }
    if ( ( viewable->xz.renzo_pc_vcimages = contour_create_viewable_sequence
	   (viewable->dataset->xz.renzo_cimages,
	    viewable->dataset->data->num_xz_frames,
	    viewable->window->pc.main, viewable->renzo_pc_vcimage_group) )
	 == NULL )
    {
	m_error_notify (function_name, "xz vcimages");
	return (FALSE);
    }
    if ( ( viewable->zy.map_pc_vcimages = contour_create_viewable_sequence
	   (viewable->dataset->zy.map_cimages,
	    viewable->dataset->data->num_zy_frames,
	    viewable->window->pc.main, viewable->map_pc_vcimage_group) )
	 == NULL )
    {
	m_error_notify (function_name, "zy vcimages");
	return (FALSE);
    }
    if ( ( viewable->zy.renzo_pc_vcimages = contour_create_viewable_sequence
	   (viewable->dataset->zy.renzo_cimages,
	    viewable->dataset->data->num_zy_frames,
	    viewable->window->pc.main, viewable->renzo_pc_vcimage_group) )
	 == NULL )
    {
	m_error_notify (function_name, "zy vcimages");
	return (FALSE);
    }
    if (viewable->window->rgb.main)
    {
	if ( ( viewable->xy.map_rgb_vcimages = contour_create_viewable_sequence
	       (viewable->dataset->xy.map_cimages,
		viewable->dataset->data->num_xy_frames,
		viewable->window->rgb.main, viewable->map_rgb_vcimage_group) )
	     == NULL )
	{
	    m_error_notify (function_name, "xy vcimages");
	    return (FALSE);
	}
	if ( ( viewable->xy.renzo_rgb_vcimages =
	       contour_create_viewable_sequence
	       (viewable->dataset->xy.renzo_cimages,
		viewable->dataset->data->num_xy_frames,
		viewable->window->rgb.main,viewable->renzo_rgb_vcimage_group) )
	     == NULL )
	{
	    m_error_notify (function_name, "xy vcimages");
	    return (FALSE);
	}
	if ( ( viewable->xz.map_rgb_vcimages = contour_create_viewable_sequence
	       (viewable->dataset->xz.map_cimages,
		viewable->dataset->data->num_xz_frames,
		viewable->window->rgb.main, viewable->map_rgb_vcimage_group) )
	     == NULL )
	{
	    m_error_notify (function_name, "xz vcimages");
	    return (FALSE);
	}
	if ( ( viewable->xz.renzo_rgb_vcimages =
	       contour_create_viewable_sequence
	       (viewable->dataset->xz.renzo_cimages,
		viewable->dataset->data->num_xz_frames,
		viewable->window->rgb.main,viewable->renzo_rgb_vcimage_group) )
	     == NULL )
	{
	    m_error_notify (function_name, "xz vcimages");
	    return (FALSE);
	}
	if ( ( viewable->zy.map_rgb_vcimages = contour_create_viewable_sequence
	       (viewable->dataset->zy.map_cimages,
		viewable->dataset->data->num_zy_frames,
		viewable->window->rgb.main, viewable->map_rgb_vcimage_group) )
	     == NULL )
	{
	    m_error_notify (function_name, "zy vcimages");
	    return (FALSE);
	}
	if ( ( viewable->zy.renzo_rgb_vcimages =
	       contour_create_viewable_sequence
	       (viewable->dataset->zy.renzo_cimages,
		viewable->dataset->data->num_zy_frames,
		viewable->window->rgb.main,viewable->renzo_rgb_vcimage_group) )
	     == NULL )
	{
	    m_error_notify (function_name, "zy vcimages");
	    return (FALSE);
	}
    }
    return (TRUE);
}   /*  End Function _dispdata_array_setup_3D_pseudo  */

static flag _dispdata_array_setup_2D_rgb (struct viewable *viewable,
					  iarray red, iarray green,iarray blue)
/*  [SUMMARY] Setup viewables for a 2D TrueColour array.
    <viewable> The cache for data on this window.
    <red> The red component array.
    <green> The green component array.
    <blue> The blue component array.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "_dispdata_array_setup_2D_rgb";

    if ( ( viewable->vimage =
	   viewimg_create_rgb2 (viewable->window->best_rgb->main,
				red->multi_desc, red->arr_desc,
				iarray_arrayp (red),
				red->data, 1, 0,
				red->elem_index, green->elem_index,
				blue->elem_index,
				0, (CONST char **) NULL,
				(CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "vimage");
	return (FALSE);
    }
    if ( ( viewable->mag_vimage =
	   viewimg_create_rgb2 (viewable->window->best_rgb->magnifier,
				red->multi_desc, red->arr_desc,
				iarray_arrayp (red),
				red->data, 1, 0,
				red->elem_index, green->elem_index,
				blue->elem_index,
				0, (CONST char **) NULL,
				(CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "vimage");
	return (FALSE);
    }
    if ( ( viewable->pan_vimage =
	   viewimg_create_rgb2 (viewable->window->best_rgb->panner,
				red->multi_desc, red->arr_desc,
				iarray_arrayp (red),
				red->data, 1, 0,
				red->elem_index, green->elem_index,
				blue->elem_index,
				0, (CONST char **) NULL,
				(CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "vimage");
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function _dispdata_array_setup_2D_rgb  */

static flag _dispdata_array_setup_3D_rgb (struct viewable *viewable,
					  iarray red, iarray green,iarray blue)
/*  [SUMMARY] Setup viewables for a 3D TrueColour array.
    <viewable> The cache for data on this window.
    <red> The red component array.
    <green> The green component array.
    <blue> The blue component array.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "_dispdata_array_setup_3D_rgb";

    if ( ( viewable->xy.vimages =
	   viewimg_create_rgb_sequence2 (viewable->window->best_rgb->main,
					 red->multi_desc, red->arr_desc,
					 iarray_arrayp (red),
					 red->data, 2, 1, 0,
					 red->elem_index, green->elem_index,
					 blue->elem_index,
					 0, (CONST char **) NULL,
					 (CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "xy vimages");
	return (FALSE);
    }
    if ( ( viewable->xz.vimages =
	   viewimg_create_rgb_sequence2 (viewable->window->best_rgb->main,
					 red->multi_desc, red->arr_desc,
					 iarray_arrayp (red),
					 red->data, 2, 0, 1,
					 red->elem_index, green->elem_index,
					 blue->elem_index,
					 0, (CONST char **) NULL,
					 (CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "xz vimages");
	return (FALSE);
    }
    if ( ( viewable->zy.vimages =
	   viewimg_create_rgb_sequence2 (viewable->window->best_rgb->main,
					 red->multi_desc, red->arr_desc,
					 iarray_arrayp (red),
					 red->data, 0, 1, 2,
					 red->elem_index, green->elem_index,
					 blue->elem_index,
					 0, (CONST char **) NULL,
					 (CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "zy vimages");
	return (FALSE);
    }
    if ( ( viewable->xy.mag_vimages =
	   viewimg_create_rgb_sequence2 (viewable->window->best_rgb->magnifier,
					 red->multi_desc, red->arr_desc,
					 iarray_arrayp (red),
					 red->data, 2, 1, 0,
					 red->elem_index, green->elem_index,
					 blue->elem_index,
					 0, (CONST char **) NULL,
					 (CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "xy magnified vimages");
	return (FALSE);
    }
    if ( ( viewable->xz.mag_vimages =
	   viewimg_create_rgb_sequence2 (viewable->window->best_rgb->magnifier,
					 red->multi_desc, red->arr_desc,
					 iarray_arrayp (red),
					 red->data, 2, 0, 1,
					 red->elem_index, green->elem_index,
					 blue->elem_index,
					 0, (CONST char **) NULL,
					 (CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "xz magnified vimages");
	return (FALSE);
    }
    if ( ( viewable->zy.mag_vimages =
	   viewimg_create_rgb_sequence2 (viewable->window->best_rgb->magnifier,
					 red->multi_desc, red->arr_desc,
					 iarray_arrayp (red),
					 red->data, 0, 1, 2,
					 red->elem_index, green->elem_index,
					 blue->elem_index,
					 0, (CONST char **) NULL,
					 (CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "zy magnified vimages");
	return (FALSE);
    }
    if ( ( viewable->xy.pan_vimages =
	   viewimg_create_rgb_sequence2 (viewable->window->best_rgb->panner,
					 red->multi_desc, red->arr_desc,
					 iarray_arrayp (red),
					 red->data, 2, 1, 0,
					 red->elem_index, green->elem_index,
					 blue->elem_index,
					 0, (CONST char **) NULL,
					 (CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "xy magnified vimages");
	return (FALSE);
    }
    if ( ( viewable->xz.pan_vimages =
	   viewimg_create_rgb_sequence2 (viewable->window->best_rgb->panner,
					 red->multi_desc, red->arr_desc,
					 iarray_arrayp (red),
					 red->data, 2, 0, 1,
					 red->elem_index, green->elem_index,
					 blue->elem_index,
					 0, (CONST char **) NULL,
					 (CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "xz magnified vimages");
	return (FALSE);
    }
    if ( ( viewable->zy.pan_vimages =
	   viewimg_create_rgb_sequence2 (viewable->window->best_rgb->panner,
					 red->multi_desc, red->arr_desc,
					 iarray_arrayp (red),
					 red->data, 0, 1, 2,
					 red->elem_index, green->elem_index,
					 blue->elem_index,
					 0, (CONST char **) NULL,
					 (CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "zy magnified vimages");
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function _dispdata_array_setup_3D_rgb  */

static void _dispdata_array_destroy_data_attributes
    (struct data_attributes *att)
{
    if (att->contour_neg_colour)       m_free (att->contour_neg_colour);
    if (att->contour_pos_colour)       m_free (att->contour_pos_colour);
    if (att->contour_levels_string)    m_free (att->contour_levels_string);
    if (att->renzogram_levels_string)  m_free (att->renzogram_levels_string);
    if (att->renzogram_channels_string)m_free (att->renzogram_channels_string);
    if (att->profile_dark_colour)      m_free (att->profile_dark_colour);
    if (att->profile_light_colour)     m_free (att->profile_light_colour);
}   /*  End Function _dispdata_array_destroy_data_attributes  */

static void _dispdata_array_dealloc_direction_viewable
    (struct viewable_direction *dir, unsigned int num_frames)
/*  [SUMMARY] Deallocate the cache for data on a window for a direction.
    [RETURNS] Nothing.
*/
{
    unsigned int count;

    for (count = 0; count < num_frames; ++count)
    {
	if (dir->vimages && dir->vimages[count])
	    viewimg_destroy (dir->vimages[count]);
	if (dir->mag_vimages && dir->mag_vimages[count])
	    viewimg_destroy (dir->mag_vimages[count]);
	if (dir->pan_vimages && dir->pan_vimages[count])
	    viewimg_destroy (dir->pan_vimages[count]);
    }
    m_free2 (dir->vimages);
    m_free2 (dir->mag_vimages);
    m_free2 (dir->pan_vimages);
    m_free2 (dir->map_pc_vcimages);
    m_free2 (dir->map_rgb_vcimages);
    m_free2 (dir->renzo_pc_vcimages);
    m_free2 (dir->renzo_rgb_vcimages);
}   /*  End Function _dispdata_array_dealloc_direction_viewable  */

static void _dispdata_array_set_contour_levels (struct dataset *dataset,
						flag set_map_levels,
						flag set_renzogram_levels)
{
    unsigned int num_levels;
    struct data_attributes *att = &dataset->attributes;
    flag dashes[MAX_CONTOUR_LEVELS];
    double clevels[MAX_CONTOUR_LEVELS], linewidths[MAX_CONTOUR_LEVELS];

    if (set_map_levels && att->contour_levels_string)
    {
	num_levels = contour_parse_levels (clevels, MAX_CONTOUR_LEVELS,
					   att->contour_levels_string,
					   dataset->data->scaled.min,
					   dataset->data->scaled.max);
	contour_unscale_levels (clevels, dashes, linewidths, num_levels,
				dataset->data->scale, dataset->data->offset,
				att->contour_neg_linewidth,
				att->contour_pos_linewidth);
	contour_set_group_levels (dataset->map_cimage_group, num_levels,
				  clevels);
    }
    if (set_renzogram_levels && att->renzogram_levels_string)
    {
	num_levels = contour_parse_levels (clevels, MAX_CONTOUR_LEVELS,
					   att->renzogram_levels_string,
					   dataset->data->scaled.min,
					   dataset->data->scaled.max);
	contour_unscale_levels (clevels, dashes, linewidths, num_levels,
				dataset->data->scale, dataset->data->offset,
				att->contour_neg_linewidth,
				att->contour_pos_linewidth);
	contour_set_group_levels (dataset->renzo_cimage_group, num_levels,
				  clevels);
    }
}   /*  End Function _dispdata_array_set_contour_levels  */

static void _dispdata_array_set_contour_styles (struct viewable *viewable)
{
    unsigned int count, num_levels;
    struct dataset *set = viewable->dataset;
    struct window *window = viewable->window;
    struct data_attributes *att = &set->attributes;
    flag dashes[MAX_CONTOUR_LEVELS];
    unsigned long pixel_values[MAX_CONTOUR_LEVELS];
    double clevels[MAX_CONTOUR_LEVELS], linewidths[MAX_CONTOUR_LEVELS];

    if (!att->contour_levels_string) return;
    num_levels = contour_parse_levels (clevels, MAX_CONTOUR_LEVELS,
				       att->contour_levels_string,
				       set->data->scaled.min,
				       set->data->scaled.max);
    if (num_levels < 1) return;
    contour_unscale_levels (clevels, dashes, linewidths, num_levels,
			    set->data->scale, set->data->offset,
			    att->contour_neg_linewidth,
			    att->contour_pos_linewidth);
    if (window->pc.main)
    {
	unsigned long neg_pixel, pos_pixel;

	kwin_get_colour_harder (canvas_get_pixcanvas (window->pc.main),
				att->contour_neg_colour, &neg_pixel);
	kwin_get_colour_harder (canvas_get_pixcanvas (window->pc.main),
				att->contour_pos_colour, &pos_pixel);
	for (count = 0; count < num_levels; ++count)
	    pixel_values[count] = dashes[count] ? neg_pixel : pos_pixel;
	contour_set_group_level_styles (viewable->map_pc_vcimage_group,
					pixel_values, NULL, linewidths);
    }
    if (window->rgb.main)
    {
	unsigned long neg_pixel, pos_pixel;

	kwin_get_colour_harder (canvas_get_pixcanvas (window->rgb.main),
				att->contour_neg_colour, &neg_pixel);
	kwin_get_colour_harder (canvas_get_pixcanvas (window->rgb.main),
				att->contour_pos_colour, &pos_pixel);
	for (count = 0; count < num_levels; ++count)
	    pixel_values[count] = dashes[count] ? neg_pixel : pos_pixel;
	contour_set_group_level_styles (viewable->map_rgb_vcimage_group,
					pixel_values, NULL, linewidths);
    }
}   /*  End Function _dispdata_array_set_contour_styles  */

static void _dispdata_array_set_renzogram_styles (struct viewable *viewable)
{
    unsigned int count, num_pixels, num_levels;
    struct dataset *set = viewable->dataset;
    struct window *window = viewable->window;
    struct data_attributes *att = &set->attributes;
    unsigned long *pixel_values;
    flag dashes[MAX_CONTOUR_LEVELS];
    unsigned long level_pixels[MAX_CONTOUR_LEVELS];
    double clevels[MAX_CONTOUR_LEVELS], linewidths[MAX_CONTOUR_LEVELS];

    if (!att->renzogram_levels_string) return;
    num_levels = contour_parse_levels (clevels, MAX_CONTOUR_LEVELS,
				       att->renzogram_levels_string,
				       set->data->scaled.min,
				       set->data->scaled.max);
    if (num_levels < 1) return;
    contour_unscale_levels (clevels, dashes, linewidths, num_levels,
			    set->data->scale, set->data->offset,
			    att->contour_neg_linewidth,
			    att->contour_pos_linewidth);
    if (window->pc.main)
    {
	num_pixels = kcmap_get_pixels (window->pc.renzogram_cmap,
				       &pixel_values);
	for (count = 0; count < set->data->num_xy_frames; ++count)
	{
	    unsigned int levelnum;
	    unsigned long pixel;

	    pixel = pixel_values[count * (num_pixels - 1) /
				(set->data->num_xy_frames - 1)];
	    for (levelnum = 0; levelnum < num_levels; ++levelnum)
		level_pixels[levelnum] = pixel;
	    contour_set_level_styles (viewable->xy.renzo_pc_vcimages[count],
				      level_pixels, NULL, NULL);
	}
	for (count = 0; count < set->data->num_xz_frames; ++count)
	{
	    unsigned int levelnum;
	    unsigned long pixel;

	    pixel = pixel_values[count * (num_pixels - 1) /
				(set->data->num_xz_frames - 1)];
	    for (levelnum = 0; levelnum < num_levels; ++levelnum)
		level_pixels[levelnum] = pixel;
	    contour_set_level_styles (viewable->xz.renzo_pc_vcimages[count],
				      level_pixels, NULL, NULL);
	}
	for (count = 0; count < set->data->num_zy_frames; ++count)
	{
	    unsigned int levelnum;
	    unsigned long pixel;

	    pixel = pixel_values[count * (num_pixels - 1) /
				(set->data->num_zy_frames - 1)];
	    for (levelnum = 0; levelnum < num_levels; ++levelnum)
		level_pixels[levelnum] = pixel;
	    contour_set_level_styles (viewable->zy.renzo_pc_vcimages[count],
				      level_pixels, NULL, NULL);
	}
    }
    if (window->rgb.main)
    {
	num_pixels = kcmap_get_pixels (window->rgb.renzogram_cmap,
				       &pixel_values);
	for (count = 0; count < set->data->num_xy_frames; ++count)
	{
	    unsigned int levelnum;
	    unsigned long pixel;

	    pixel = pixel_values[count * (num_pixels - 1) /
				(set->data->num_xy_frames - 1)];
	    for (levelnum = 0; levelnum < num_levels; ++levelnum)
		level_pixels[levelnum] = pixel;
	    contour_set_level_styles (viewable->xy.renzo_rgb_vcimages[count],
				      level_pixels, NULL, NULL);
	}
	for (count = 0; count < set->data->num_xz_frames; ++count)
	{
	    unsigned int levelnum;
	    unsigned long pixel;

	    pixel = pixel_values[count * (num_pixels - 1) /
				(set->data->num_xz_frames - 1)];
	    for (levelnum = 0; levelnum < num_levels; ++levelnum)
		level_pixels[levelnum] = pixel;
	    contour_set_level_styles (viewable->xz.renzo_rgb_vcimages[count],
				      level_pixels, NULL, NULL);
	}
	for (count = 0; count < set->data->num_zy_frames; ++count)
	{
	    unsigned int levelnum;
	    unsigned long pixel;

	    pixel = pixel_values[count * (num_pixels - 1) /
				(set->data->num_zy_frames - 1)];
	    for (levelnum = 0; levelnum < num_levels; ++levelnum)
		level_pixels[levelnum] = pixel;
	    contour_set_level_styles (viewable->zy.renzo_rgb_vcimages[count],
				      level_pixels, NULL, NULL);
	}
    }
}   /*  End Function _dispdata_array_set_renzogram_styles  */

static void _dispdata_array_cmap_modify_func (Kcolourmap cmap, void **info)
/*  [SUMMARY] This routine registers a change in the colourmap values.
    <cmap> The colourmap.
    <info> A pointer to the arbitrary colourmap information pointer.
    [RETURNS] Nothing.
*/
{
    ViewableImage vimage;
    flag refresh = FALSE;
    struct canvas *canvas = *info;
    struct window *window = canvas->window;
    struct viewable *viewable;

    if (!window->first) return;
    for (viewable = window->first; viewable;
	 viewable = viewable->window_chain.next)
    {
	if (!viewable->active_blinkentry) continue;
	if (viewable->active_blinkentry->contour_mode !=
	    KDISPLAYDATA_CONTOUR_RENZOGRAM) continue;
	_dispdata_array_set_renzogram_styles (viewable);
	refresh = TRUE;
    }
    viewable = window->active_blinkstate->image_viewable;
    if (!viewable) return;
    vimage = viewable->vimage ? viewable->vimage : viewable->xy.vimages[0];
    if (viewimg_get_worldcanvas (vimage) != canvas->canvas) return;
    if (refresh) dispdata_refresh_window (window->superclass, NULL);
}   /*  End Function _dispdata_array_cmap_modify_func  */

static void _dispdata_array_activate_frames (struct viewable *viewable)
{
    flag bool;
    unsigned int count;
    unsigned int frame_number = 0;
    unsigned int num_frames = 0;
    unsigned int axis = 0;
    unsigned int num_channels;
    struct dataset *dataset = viewable->dataset;
    struct blinkentry *blinkentry = viewable->active_blinkentry;
    struct viewable_direction *dir = NULL;
    char *flags;
    char *string = dataset->attributes.renzogram_channels_string;
    double *channels;
    static char function_name[] = "_dispdata_array_activate_frames";

    switch (viewable->window->slice_direction)
    {
      case KDISPLAYDATA_SLICE_DIRECTION_XY:
	dir = &viewable->xy;
	frame_number = dataset->attributes.xy_frame.index;
	num_frames = dataset->data->num_xy_frames;
	axis = 0;
	break;
      case KDISPLAYDATA_SLICE_DIRECTION_XZ:
	dir = &viewable->xz;
	frame_number = dataset->attributes.xz_frame.index;
	num_frames = dataset->data->num_xz_frames;
	axis = 1;
	break;
      case KDISPLAYDATA_SLICE_DIRECTION_ZY:
	dir = &viewable->zy;
	frame_number = dataset->attributes.zy_frame.index;
	num_frames = dataset->data->num_zy_frames;
	axis = 2;
	break;
    }
    if (blinkentry->blinkstate->image_viewable == viewable)
    {
	if (viewable->vimage)
	{
	    viewimg_set_active (viewable->vimage, FALSE);
	    viewimg_set_active (viewable->mag_vimage, FALSE);
	    viewimg_set_active (viewable->pan_vimage, FALSE);
	}
	else
	{
	    viewimg_set_active (dir->vimages[frame_number], FALSE);
	    viewimg_set_active (dir->mag_vimages[frame_number], FALSE);
	    viewimg_set_active (dir->pan_vimages[frame_number], FALSE);
	}
    }
    if (!viewable->dataset->data->pseudo) return;
    bool = (blinkentry->contour_mode ==
	    KDISPLAYDATA_CONTOUR_NORMAL) ? TRUE : FALSE;
    if (viewable->map_pc_vcimage)
	contour_set_active (viewable->map_pc_vcimage, bool, FALSE, FALSE);
    if (viewable->map_rgb_vcimage)
	contour_set_active (viewable->map_rgb_vcimage, bool, FALSE, FALSE);
    if (viewable->vimage) return;  /*  This is a 2D dataset  */
    /*  Disable contours in all directions  */
    if (viewable->map_pc_vcimage_group)
	contour_set_all_inactive (viewable->map_pc_vcimage_group, FALSE);
    if (viewable->renzo_pc_vcimage_group)
	contour_set_all_inactive (viewable->renzo_pc_vcimage_group, FALSE);
    if (viewable->map_rgb_vcimage_group)
	contour_set_all_inactive (viewable->map_rgb_vcimage_group, FALSE);
    if (viewable->renzo_rgb_vcimage_group)
	contour_set_all_inactive (viewable->renzo_rgb_vcimage_group, FALSE);
    /*  Now activate contours in the current direction  */
    if (blinkentry->contour_mode == KDISPLAYDATA_CONTOUR_NORMAL)
    {
	contour_set_active (dir->map_pc_vcimages[frame_number], TRUE,
			    FALSE, FALSE);
	if (dir->map_rgb_vcimages)
	    contour_set_active (dir->map_rgb_vcimages[frame_number], TRUE,
				FALSE, FALSE);
	return;
    }
    if (blinkentry->contour_mode != KDISPLAYDATA_CONTOUR_RENZOGRAM) return;
    flags = m_alloc_scratch (num_frames * (sizeof *flags + sizeof *channels),
			     function_name);
    channels = (void *) (flags + num_frames);
    memset (flags, 0, num_frames);
    if (toupper (string[0]) == 'I')
	num_channels = contour_parse_levels (channels, num_frames,
					     string + 1, 0, num_frames -1);
    else
    {
	iarray arr = dataset->data->pseudo;
	double tmp;
	dim_desc *dim = iarray_get_dim_desc (arr, axis);
	double *ptr;
	double x[2], y[2], z[2];

	iarray_get_world_coords (arr, 2, x, x + 1);
	iarray_get_world_coords (arr, 1, y, y + 1);
	iarray_get_world_coords (arr, 0, z, z + 1);
	if (viewable->dataset->data->ap)
	{
	    if ( !wcs_astro_transform3 (viewable->dataset->data->ap, FALSE,
					2,
					iarray_dim_name (arr, 0), x, FALSE,
					iarray_dim_name (arr, 1), y, FALSE,
					iarray_dim_name (arr, 2), z, FALSE,
					0, NULL, NULL) ) return;
	}
	if (axis == 0) ptr = x;
	else if (axis == 1) ptr = y;
	else ptr = z;
	if (ptr[1] < ptr[0]) swap (ptr[1], ptr[0], tmp);
	num_channels = contour_parse_levels (channels, num_frames,
					     string + 1, ptr[0], ptr[1]);
	if (viewable->dataset->data->ap)
	{
	    if ( !wcs_astro_transform3 (viewable->dataset->data->ap, FALSE,
					num_channels,
					iarray_dim_name (arr, axis),
					channels, TRUE,
					NULL, NULL, FALSE,
					NULL, NULL, FALSE,
					0, NULL, NULL) ) return;
	}
	ds_convert_coordinates (dim, num_channels, channels, TRUE, TRUE,
				TRUE);
    }
    for (count = 0; count < num_channels; ++count)
    {
	if ( DOUBLE_IS_BLANK (channels[count]) ) continue;
	flags[ (int) channels[count] ] = TRUE;
    }
    for (count = 0; count < num_frames; ++count)
    {
	contour_set_active (dir->renzo_pc_vcimages[count], flags[count], FALSE,
			    FALSE);
	if (dir->renzo_rgb_vcimages)
	    contour_set_active (dir->renzo_rgb_vcimages[count], flags[count],
				FALSE, FALSE);
    }
    m_free_scratch ();
}   /*  End Function _dispdata_array_activate_frames  */

static flag _dispdata_array_canvas_position_func (KWorldCanvas canvas,
						  double x, double y,
						  unsigned int event_code,
						  void *e_info,
						  void **f_info,
						  double x_lin, double y_lin,
						  int x_pix, int y_pix)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas.
    <canvas> The canvas on which the event occurred.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    <event_code> The arbitrary event code.
    <e_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    <x_lin> The horizontal linear world co-ordinate prior to the transform
    function being called.
    <y_lin> The vertical linear world co-ordinate prior to the transform
    function being called.
    <x_pix> The horizontal pixel co-ordinate.
    <y_pix> The vertical pixel co-ordinate.
    [NOTE] The world co-ordinate values will have been transformed by the
    registered transform function (see [<canvas_register_transforms_func>]).
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    flag copy_image_h_to_profile_h = FALSE;
    flag copy_image_v_to_profile_h = FALSE;
    struct window *window = *f_info;
    struct blinkstate *blinkstate = window->active_blinkstate;
    struct viewable *viewable;
    double *coord = NULL;

    if (!blinkstate) return (FALSE);
    if (!window->profile_canvas) return (FALSE);
    if (window->profile_mode != KDISPLAYDATA_PROFILE_MODE_LINE) return (FALSE);
    if (!blinkstate->profile_master) return (FALSE);
    switch (event_code)
    {
      case K_CANVAS_EVENT_PLAIN_KEY_PRESS:
      case K_CANVAS_EVENT_MODIFIED_KEY_PRESS:
      case K_CANVAS_EVENT_KEYSYM_PRESS:
      case K_CANVAS_EVENT_POINTER_ENTER:
      case K_CANVAS_EVENT_POINTER_LEAVE:
	return (FALSE);
	/*break;*/
      default:
	break;
    }
    for (viewable = window->first; viewable;
	 viewable = viewable->window_chain.next)
    {
	if (viewable == blinkstate->profile_master) continue;
	if (!viewable->active_blinkentry->profile_slave) continue;
	_dispdata_array_line_profile (viewable, x, y);
    }
    _dispdata_array_line_profile (blinkstate->profile_master, x, y);
    graph1d_orthogonal_configure_canvas (blinkstate->profile_master->profile);
    switch (window->slice_direction)
    {
      case KDISPLAYDATA_SLICE_DIRECTION_XY:
	switch (window->profile_direction)
	{
	  case KDISPLAYDATA_PROFILE_DIR_X:
	    coord = &x;
	    copy_image_h_to_profile_h = TRUE;
	    break;
	  case KDISPLAYDATA_PROFILE_DIR_Y:
	    coord = &y;
	    copy_image_v_to_profile_h = TRUE;
	    break;
	}
	break;
      case KDISPLAYDATA_SLICE_DIRECTION_XZ:
	switch (window->profile_direction)
	{
	  case KDISPLAYDATA_PROFILE_DIR_X:
	    coord = &x;
	    copy_image_h_to_profile_h = TRUE;
	    break;
	  case KDISPLAYDATA_PROFILE_DIR_Z:
	    coord = &y;
	    copy_image_v_to_profile_h = TRUE;
	    break;
	}
	break;
      case KDISPLAYDATA_SLICE_DIRECTION_ZY:
	switch (window->profile_direction)
	{
	  case KDISPLAYDATA_PROFILE_DIR_Z:
	    coord = &x;
	    copy_image_h_to_profile_h = TRUE;
	    break;
	  case KDISPLAYDATA_PROFILE_DIR_Y:
	    coord = &y;
	    copy_image_v_to_profile_h = TRUE;
	    break;
	}
	break;
    }
    if ( slave_profile_h_zoom
	 && (copy_image_h_to_profile_h || copy_image_v_to_profile_h) )
    {
	double left_x, right_x;

	if (copy_image_h_to_profile_h)
	    canvas_get_attributes (canvas,
				   CANVAS_ATT_LEFT_X, &left_x,
				   CANVAS_ATT_RIGHT_X, &right_x,
				   CANVAS_ATT_END);
	else canvas_get_attributes (canvas,
				    CANVAS_ATT_BOTTOM_Y, &left_x,
				    CANVAS_ATT_TOP_Y, &right_x,
				    CANVAS_ATT_END);
	canvas_set_attributes (window->profile_canvas,
			       CANVAS_ATT_LEFT_X, left_x,
			       CANVAS_ATT_RIGHT_X, right_x,
			       CANVAS_ATT_END);
    }
    if (coord)
	graph1d_set_vgraph_attributes (blinkstate->profile_master->vgraph,
				       GRAPH1D_VGRAPH_ATT_VLINE_X, *coord,
				       GRAPH1D_VGRAPH_ATT_END);
    if (window->profile_unzoom)
    {
	graph1d_unzoom (window->profile_canvas, TRUE, FALSE, FALSE);
	window->profile_unzoom = FALSE;
    }
    kwin_refresh_if_visible (window->profile_pixcanvas, TRUE);
    return (FALSE);
}   /*  End Function _dispdata_array_canvas_position_func */

static flag _dispdata_array_middle_geom_func (KPixCanvasEditorContext context,
					      KPixCanvas canvas,
					      unsigned int object_type,
					      unsigned int num_coords,
					      CONST int *x, CONST int *y,
					      unsigned int raw_event_code,
					      void *raw_e_info, void **f_info)
/*  [SUMMARY] Process a geometry editor event on a pixel canvas.
    <context> The KPixCanvasEditorContext.
    <canvas> The pixel canvas on which the event occurred.
    <object_type> The type of the geometric object.
    <num_coords> The number of co-ordinates in the geometric object.
    <x> The array of horizontal co-ordinates for the geometric object.
    <y> The array of vertical co-ordinates for the geometric object.
    <raw_event_code> The raw arbitrary event code.
    <raw_e_info> A pointer to the raw arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KWorldCanvas worldcanvas = NULL;
    struct window *window = *f_info;
    struct blinkstate *blinkstate = window->active_blinkstate;
    struct viewable *viewable;
    double lx[3], ly[3];
    void (*compute_profile) (struct viewable *viewable,
			     CONST double *x, CONST double *y, flag configure);
    /*static char function_name[] = "_dispdata_array_middle_geom_func";*/

    if (raw_event_code != K_CANVAS_EVENT_MIDDLE_MOUSE_RELEASE) return (FALSE);
    if (!blinkstate) return (FALSE);
    if (!window->profile_canvas) return (FALSE);
    if (!blinkstate->profile_master) return (FALSE);
    /*  Convert screen pixel co-ordinates to more appropriate co-ordinate
	systems  */
    if (window->pc.main)
    {
	if (canvas_get_pixcanvas (window->pc.main) == canvas)
	    worldcanvas = window->pc.main;
    }
    if (window->rgb.main)
    {
	if (canvas_get_pixcanvas (window->rgb.main) == canvas)
	    worldcanvas = window->rgb.main;
    }
    compute_profile = NULL;  /*  Keep compiler happy  */
    switch (window->profile_mode)
    {
      case KDISPLAYDATA_PROFILE_MODE_BOX_SUM:
      case KDISPLAYDATA_PROFILE_MODE_BOX_AVERAGE:
	if ( !_dispdata_array_test_profile_perpendicular (window) )
	{
	    fprintf (stderr, "Incompatible profile/slice directions\n");
	    return (TRUE);
	}
	/*  Fall through  */
      case KDISPLAYDATA_PROFILE_MODE_BOXED_HORIZONTAL:
      case KDISPLAYDATA_PROFILE_MODE_BOXED_VERTICAL:
	/*  Want the box corners in world co-ordinates  */
	lx[0] = x[0];
	ly[0] = y[1];
	lx[1] = x[1];
	ly[1] = y[0];
	canvas_convert_to_canvas_coords (worldcanvas, TRUE, 2,
					 lx, ly, NULL, NULL, lx, ly);
	compute_profile = _dispdata_array_box_event;
	break;
      case KDISPLAYDATA_PROFILE_MODE_RADIAL:
      case KDISPLAYDATA_PROFILE_MODE_RADIAL_GAUSS:
      case KDISPLAYDATA_PROFILE_MODE_AZIMUTHAL_AVERAGE:
	/*  Want centre in world co-ordinates  */
	lx[0] = x[0];
	ly[0] = y[0];
	lx[1] = x[0] + x[1];
	ly[1] = y[0];
	lx[2] = x[0];
	ly[2] = y[0] + y[1];
	canvas_convert_to_canvas_coords (worldcanvas, TRUE, 3,
					 lx, ly, NULL, NULL, lx, ly);
	compute_profile = _dispdata_array_circle_event;
	break;
    }
    if ( !graph1d_test_if_zoomed_horizontally (window->profile_canvas) )
	window->profile_unzoom = TRUE;
    for (viewable = window->first; viewable;
	 viewable = viewable->window_chain.next)
    {
	if (viewable == blinkstate->profile_master) continue;
	if (!viewable->active_blinkentry->profile_slave) continue;
	(*compute_profile) (viewable, lx, ly, FALSE);
    }
    (*compute_profile) (blinkstate->profile_master, lx, ly, TRUE);
    if (window->profile_unzoom)
    {
	graph1d_unzoom (window->profile_canvas, TRUE, FALSE, FALSE);
	window->profile_unzoom = FALSE;
    }
    kwin_refresh_if_visible (window->profile_pixcanvas, TRUE);
    return (TRUE);
}   /*  End Function middle_geom_func  */

static flag _dispdata_array_change_frame (struct window *sourcewin,
					  unsigned int framenum)
/*  [SUMMARY] Change the frame number for a window.
    <sourcewin> The window where the frame change event came from.
    <framenum> The frame number to change to.
    [RETURNS] TRUE if the source window should be refreshed, else FALSE.
*/
{
    unsigned int dim_num = 10000;  /*  Shut the compiler up  */
    struct window *window;
    struct viewable *sourceviewable;
    struct viewable *master = sourcewin->active_blinkstate->movie_master;
    struct frame_pos *sframe = NULL;

    /*  Compute world co-ordinate of frame  */
    switch (sourcewin->slice_direction)
    {
      case KDISPLAYDATA_SLICE_DIRECTION_XY:
	dim_num = 0;
	sframe = &master->dataset->attributes.xy_frame;
	break;
      case KDISPLAYDATA_SLICE_DIRECTION_XZ:
	dim_num = 1;
	sframe = &master->dataset->attributes.xz_frame;
	break;
      case KDISPLAYDATA_SLICE_DIRECTION_ZY:
	dim_num = 2;
	sframe = &master->dataset->attributes.zy_frame;
	break;
    }
    _dispdata_array_compute_frame_coord (master->dataset, dim_num);
    /*  Now clear the refresh flags for all windows. Do this below stuff above
	so that the <<window>> variable isn't accidentally used  */
    for (window = first_window; window; window = window->next)
	window->movie_refresh = FALSE;
    /*  Loop over sourcewin viewables, looking for master/slaves to update  */
    for (sourceviewable = sourcewin->first; sourceviewable;
	 sourceviewable = sourceviewable->window_chain.next)
    {
	struct dataset *dataset = sourceviewable->dataset;
	struct viewable *viewable;
	struct frame_pos *dframe = NULL;

	if (master == sourceviewable)
	{}  /*  Nothing needs to be done, as it was done earlier  */
	else if (sourceviewable->active_blinkentry->movie_slave)
	{
	    if ( !_dispdata_array_set_matching_frame (sourceviewable,
						      master, sframe) )
		continue;
	}
	else continue;
	/*  Loop over windows for the dataset corresponding to this viewable,
	    looking for frames to activate and profiles to update  */
	switch (sourcewin->slice_direction)
	{
	  case KDISPLAYDATA_SLICE_DIRECTION_XY:
	    dframe = &dataset->attributes.xy_frame;
	    break;
	  case KDISPLAYDATA_SLICE_DIRECTION_XZ:
	    dframe = &dataset->attributes.xz_frame;
	    break;
	  case KDISPLAYDATA_SLICE_DIRECTION_ZY:
	    dframe = &dataset->attributes.zy_frame;
	    break;
	}
	for (viewable = dataset->first; viewable;
	     viewable = viewable->dataset_chain.next)
	{
	    flag can_show_vlines = FALSE;
	    struct blinkentry *blinkentry = viewable->active_blinkentry;

	    window = viewable->window;
	    if (window->slice_direction != sourcewin->slice_direction)continue;
	    if ( (window->active_blinkstate->image_viewable == viewable) ||
		 (blinkentry->contour_mode == KDISPLAYDATA_CONTOUR_NORMAL) )
	    {
		_dispdata_array_activate_frames (viewable);
		window->movie_refresh = TRUE;
	    }
	    if ( _dispdata_array_test_profile_perpendicular (window) )
		can_show_vlines = TRUE;
	    switch (window->profile_mode)
	    {
	      case KDISPLAYDATA_PROFILE_MODE_LINE:
	      case KDISPLAYDATA_PROFILE_MODE_BOX_SUM:
	      case KDISPLAYDATA_PROFILE_MODE_BOX_AVERAGE:
		if (can_show_vlines && viewable->vgraph)
		    graph1d_set_vgraph_attributes (viewable->vgraph,
						   GRAPH1D_VGRAPH_ATT_VLINE_X,
					               dframe->coord,
						   GRAPH1D_VGRAPH_ATT_END);
		break;
	    }
	}
    }
    /*  Loop over windows, looking for refresh candidates  */
    for (window = first_window; window; window = window->next)
    {
	if (window == sourcewin) continue;  /*  Done later  */
	if (window->movie_refresh)
	    dispdata_refresh_window (window->superclass, NULL);
	if (window->profile_canvas)
	    graph1d_refresh_if_needed (window->profile_canvas);
    }
    return (sourcewin->movie_refresh);
}   /*  End Function _dispdata_array_change_frame  */

static flag _dispdata_array_set_matching_frame (struct viewable *dest,
						struct viewable *source,
						CONST struct frame_pos *frame)
{
    KwcsAstro dest_ap = dest->dataset->data->ap;
    KwcsAstro source_ap = source->dataset->data->ap;
    iarray dest_arr = dest->dataset->data->pseudo;
    iarray source_arr = source->dataset->data->pseudo;
    unsigned int dim_num = 10000;
    unsigned int num_frames = 0;
    unsigned int frame_index;
    unsigned int *frameptr = NULL;
    dim_desc *source_dim, *dest_dim;

    if (!dest_arr) dest_arr = dest->dataset->data->red;
    if (!source_arr) source_arr = source->dataset->data->red;
    /*  Copy over frame number, preferably by matching real units  */
    switch (source->window->slice_direction)
    {
      case KDISPLAYDATA_SLICE_DIRECTION_XY:
	dim_num = 0;
	frameptr = &dest->dataset->attributes.xy_frame.index;
	num_frames = dest->dataset->data->num_xy_frames;
	break;
      case KDISPLAYDATA_SLICE_DIRECTION_XZ:
	dim_num = 1;
	frameptr = &dest->dataset->attributes.xz_frame.index;
	num_frames = dest->dataset->data->num_xz_frames;
	break;
      case KDISPLAYDATA_SLICE_DIRECTION_ZY:
	dim_num = 2;
	frameptr = &dest->dataset->attributes.zy_frame.index;
	num_frames = dest->dataset->data->num_zy_frames;
	break;
    }
    source_dim = iarray_get_dim_desc (source_arr, dim_num);
    dest_dim = iarray_get_dim_desc (dest_arr, dim_num);
    if ( wcs_astro_compare_axes (source_ap, source_dim->name,
				 dest_ap, dest_dim->name) )
    {   /*  Match using real units  */
	unsigned int count;
	double coords[3];
	CONST char *names[3];

	coords[0] = dest->dataset->attributes.xy_frame.coord;
	coords[1] = dest->dataset->attributes.xz_frame.coord;
	coords[2] = dest->dataset->attributes.zy_frame.coord;
	coords[dim_num] = frame->coord;
	for (count = 0; count < 3; ++count)
	    names[count] = iarray_dim_name (dest_arr, count);
	if (dest_ap)
	    wcs_astro_transform3 (dest_ap, FALSE, 1,
				  names[0], coords, TRUE,
				  names[1], coords + 1, TRUE,
				  names[2], coords + 2, TRUE,
				  0, NULL, NULL);
	ds_convert_coordinates (dest_dim, 1, coords + dim_num,
				TRUE, FALSE, TRUE);
	if (coords[dim_num] < 0.0) return (FALSE);
	frame_index = coords[dim_num];
	if (frame_index >= num_frames) return (FALSE);
    }
    else frame_index = frame->index;
    if (frame_index >= num_frames) frame_index = num_frames - 1;
    *frameptr = frame_index;
    /*  Convert frame number to world co-ordinates  */
    _dispdata_array_compute_frame_coord (dest->dataset, dim_num);
    return (TRUE);
}   /*  End Function _dispdata_array_set_matching_frame  */

static void _dispdata_array_compute_frame_coord (struct dataset *dataset,
						 unsigned int dim_num)
/*  [SUMMARY] Compute all the frame co-ordinate from the frame index.
    <dataset> The dataset to compute for.
    <dim_num> The number of the dimension.
    [RETURNS] Nothing.
*/
{
    iarray arr = dataset->data->pseudo;
    unsigned int count;
    struct frame_pos *frame = NULL;
    CONST char *names[3];
    double coords[3];

    if (!arr) arr = dataset->data->red;
    if (iarray_num_dim (arr) != 3) return;
    /*  Convert frame number to world co-ordinates  */
    switch (dim_num)
    {
      case 0:
	frame = &dataset->attributes.xy_frame;
	break;
      case 1:
	frame = &dataset->attributes.xz_frame;
	break;
      case 2:
	frame = &dataset->attributes.zy_frame;
	break;
    }
    for (count = 0; count < 3; ++count)
    {
	dim_desc *dim = iarray_get_dim_desc (arr, count);

	names[count] = dim->name;
	coords[count] = (count == dim_num) ? frame->index : (dim->length / 2);
	ds_convert_coordinates (dim, 1, coords + count, FALSE, FALSE, FALSE);
    }
    if (dataset->data->ap) wcs_astro_transform3 (dataset->data->ap, FALSE, 1,
						 names[0], coords, FALSE,
						 names[1], coords + 1, FALSE,
						 names[2], coords + 2, FALSE,
						 0, NULL, NULL);
    frame->coord = coords[dim_num];
}   /*  End Function _dispdata_array_compute_frame_coords  */

static void _dispdata_array_set_canvas_iscale (struct viewable *viewable)
{
    double min, max;
    struct window *window = viewable->window;
    struct dataset *dataset = viewable->dataset;

    min = (window->min - dataset->data->offset) / dataset->data->scale;
    max = (window->max - dataset->data->offset) / dataset->data->scale;
    if (min > max)
    {
	double tmp;

	swap (min, max, tmp);
    }
    canvas_set_attributes (window->pc.main,
			   CANVAS_ATT_VALUE_MIN, min,
			   CANVAS_ATT_VALUE_MAX, max,
			   CANVAS_ATT_END);
    canvas_set_attributes (window->pc.magnifier,
			   CANVAS_ATT_VALUE_MIN, min,
			   CANVAS_ATT_VALUE_MAX, max,
			   CANVAS_ATT_END);
    canvas_set_attributes (window->pc.panner,
			   CANVAS_ATT_VALUE_MIN, min,
			   CANVAS_ATT_VALUE_MAX, max,
			   CANVAS_ATT_END);
}   /*  End Function _dispdata_array_set_canvas_iscale  */

static flag _dispdata_array_set_profile_vscale (struct viewable *viewable)
{
    double scale, offset;
    struct window *window = viewable->window;
    struct data_attributes *master;
    struct data_attributes *att;

    if (!window->active_blinkstate) return (FALSE);
    if (!window->active_blinkstate->profile_master) return (FALSE);
    master = &window->active_blinkstate->profile_master->dataset->attributes;
    att = &viewable->dataset->attributes;
    if (master != att)
    {   /*  Easy case: fiddling only my own scale  */
	if (!viewable->active_blinkentry->profile_slave) return (FALSE);
	scale = (master->profile_upper_clip - master->profile_lower_clip) /
	    (att->profile_upper_clip - att->profile_lower_clip);
	offset = master->profile_lower_clip - att->profile_lower_clip * scale;
	graph1d_set_vgraph_attributes (viewable->vgraph,
				       GRAPH1D_VGRAPH_ATT_SCALE, scale,
				       GRAPH1D_VGRAPH_ATT_OFFSET, offset,
				       GRAPH1D_VGRAPH_ATT_END);
	graph1d_set_vgraph_attributes (viewable->gausscurve,
				       GRAPH1D_VGRAPH_ATT_SCALE, scale,
				       GRAPH1D_VGRAPH_ATT_OFFSET, offset,
				       GRAPH1D_VGRAPH_ATT_END);
	return (TRUE);
    }
    /*  I am the master, so set scales for myself and all slaves  */
    graph1d_set_vgraph_attributes (viewable->vgraph,
				   GRAPH1D_VGRAPH_ATT_SCALE, 1.0,
				   GRAPH1D_VGRAPH_ATT_OFFSET, 0.0,
				   GRAPH1D_VGRAPH_ATT_END);
    graph1d_set_vgraph_attributes (viewable->gausscurve,
				   GRAPH1D_VGRAPH_ATT_SCALE, 1.0,
				   GRAPH1D_VGRAPH_ATT_OFFSET, 0.0,
				   GRAPH1D_VGRAPH_ATT_END);
    canvas_set_attributes (window->profile_canvas,
			   CANVAS_ATT_BOTTOM_Y, master->profile_lower_clip,
			   CANVAS_ATT_TOP_Y, master->profile_upper_clip,
			   CANVAS_ATT_END);
    graph1d_set_canvas_attributes
	(window->profile_canvas,
	 GRAPH1D_CANVAS_ATT_BOTTOM_Y, master->profile_lower_clip,
	 GRAPH1D_CANVAS_ATT_TOP_Y, master->profile_upper_clip,
	 GRAPH1D_CANVAS_ATT_END);
    for (viewable = window->first; viewable;
	 viewable = viewable->window_chain.next)
    {
	att = &viewable->dataset->attributes;
	if (att == master) continue;
	if (!viewable->active_blinkentry) continue;
	if (!viewable->active_blinkentry->profile_slave) continue;
	scale = (master->profile_upper_clip - master->profile_lower_clip) /
	    (att->profile_upper_clip - att->profile_lower_clip);
	offset = master->profile_lower_clip - att->profile_lower_clip * scale;
	graph1d_set_vgraph_attributes (viewable->vgraph,
				       GRAPH1D_VGRAPH_ATT_SCALE, scale,
				       GRAPH1D_VGRAPH_ATT_OFFSET, offset,
				       GRAPH1D_VGRAPH_ATT_END);
	graph1d_set_vgraph_attributes (viewable->gausscurve,
				       GRAPH1D_VGRAPH_ATT_SCALE, scale,
				       GRAPH1D_VGRAPH_ATT_OFFSET, offset,
				       GRAPH1D_VGRAPH_ATT_END);
    }
    return (TRUE);
}   /*  End Function _dispdata_array_set_profile_vscale  */

static ViewableImage _dispdata_array_get_vimage (struct viewable *viewable)
/*  [SUMMARY] Get the current ViewableImage for a viewable.
    <viewable> The viewable object.
    [RETURNS] The ViewableImage.
*/
{
    unsigned int frame_number = 0;
    struct viewable_direction *dir = NULL;

    if (viewable->vimage) return (viewable->vimage);
    switch (viewable->window->slice_direction)
    {
      case KDISPLAYDATA_SLICE_DIRECTION_XY:
	dir = &viewable->xy;
	frame_number = viewable->dataset->attributes.xy_frame.index;
	break;
      case KDISPLAYDATA_SLICE_DIRECTION_XZ:
	dir = &viewable->xz;
	frame_number = viewable->dataset->attributes.xz_frame.index;
	break;
      case KDISPLAYDATA_SLICE_DIRECTION_ZY:
	dir = &viewable->zy;
	frame_number = viewable->dataset->attributes.zy_frame.index;
	break;
    }
    return (dir->vimages[frame_number]);
}   /*  End Function _dispdata_array_get_vimage  */

static flag _dispdata_array_convert_3coord (struct dataset *dataset,
					    unsigned int slice_direction,
					    double *x, double *y, double *z,
					    double h, double v)
/*  [SUMMARY] Convert canvas co-ordinate to 3D dataset co-ordinate.
    <dataset> The dataset.
    <slice_direction> The slice direction.
    <x> The dataset X image pixel co-ordinate is written here.
    <y> The dataset Y image pixel co-ordinate is written here.
    <z> The dataset Z image pixel co-ordinate is written here.
    <h> The horizontal co-ordinate on the world canvas.
    <v> The vertical co-ordinate on the world canvas.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    iarray array;
    flag x_to_linear = TRUE;
    flag y_to_linear = TRUE;
    flag z_to_linear = TRUE;
    dim_desc *dim;
    CONST char *x_name = NULL;
    CONST char *y_name = NULL;
    CONST char *z_name = NULL;

    if (dataset->data->pseudo)
	array = dataset->data->pseudo;
    else array = dataset->data->red;
    if (iarray_num_dim (array) == 3)
    {
	/*  Get the full 3-dimensional position of the point  */
	x_name = iarray_dim_name (array, 2);
	y_name = iarray_dim_name (array, 1);
	z_name = iarray_dim_name (array, 0);
	switch (slice_direction)
	{
	  case KDISPLAYDATA_SLICE_DIRECTION_XY:
	    *x = h;
	    *y = v;
	    *z = dataset->attributes.xy_frame.index;
	    z_to_linear = FALSE;
	    break;
	  case KDISPLAYDATA_SLICE_DIRECTION_XZ:
	    *x = h;
	    *y = dataset->attributes.xz_frame.index;
	    y_to_linear = FALSE;
	    *z = v;
	    break;
	  case KDISPLAYDATA_SLICE_DIRECTION_ZY:
	    *x = dataset->attributes.zy_frame.index;
	    x_to_linear = FALSE;
	    *y = v;
	    *z = h;
	    break;
	}
    }
    else
    {
	x_name = iarray_dim_name (array, 1);
	y_name = iarray_dim_name (array, 0);
	*x = h;
	*y = v;
    }
    if (dataset->data->ap)
    {
	if ( !wcs_astro_transform3 (dataset->data->ap, FALSE, 1,
				    x_name, x, x_to_linear,
				    y_name, y, y_to_linear,
				    z_name, z, z_to_linear,
				    0, NULL, NULL) ) return (FALSE);
    }
    /*  Convert from linear co-ordinates to image pixel co-ordinates  */
    if (iarray_num_dim (array) == 3)
    {
	/*  Also undo damage done by transform  */
	switch (slice_direction)
	{
	  case KDISPLAYDATA_SLICE_DIRECTION_XY:
	    dim = iarray_get_dim_desc (array, 2);
	    *x = ds_get_coord_num (dim, *x, SEARCH_BIAS_CLOSEST);
	    dim = iarray_get_dim_desc (array, 1);
	    *y = ds_get_coord_num (dim, *y, SEARCH_BIAS_CLOSEST);
	    *z = dataset->attributes.xy_frame.index;
	    break;
	  case KDISPLAYDATA_SLICE_DIRECTION_XZ:
	    dim = iarray_get_dim_desc (array, 2);
	    *x = ds_get_coord_num (dim, *x, SEARCH_BIAS_CLOSEST);
	    *y = dataset->attributes.xz_frame.index;
	    dim = iarray_get_dim_desc (array, 0);
	    *z = ds_get_coord_num (dim, *z, SEARCH_BIAS_CLOSEST);
	    break;
	  case KDISPLAYDATA_SLICE_DIRECTION_ZY:
	    *x = dataset->attributes.zy_frame.index;
	    dim = iarray_get_dim_desc (array, 1);
	    *y = ds_get_coord_num (dim, *y, SEARCH_BIAS_CLOSEST);
	    dim = iarray_get_dim_desc (array, 0);
	    *z = ds_get_coord_num (dim, *z, SEARCH_BIAS_CLOSEST);
	    break;
	}
    }
    else
    {
	dim = iarray_get_dim_desc (array, 1);
	*x = ds_get_coord_num (dim, *x, SEARCH_BIAS_CLOSEST);
	dim = iarray_get_dim_desc (array, 0);
	*y = ds_get_coord_num (dim, *y, SEARCH_BIAS_CLOSEST);
	*z = 0.0;
    }
    return (TRUE);
}   /*  End Function _dispdata_array_convert_3coord  */

static flag _dispdata_array_convert_2coord (struct viewable *viewable,
					    double *h, double *v)
/*  [SUMMARY] Convert canvas co-ordinate to 2D dataset co-ordinate.
    <viewable> The dataset and canvas information.
    <h> The horizontal co-ordinate on the world canvas. This is modified with
    the horizontal image pixel co-ordinate.
    <v> The vertical co-ordinate on the world canvas. This is modified with
    the vertical image pixel co-ordinate.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    iarray array;
    double x, y, z;

    if ( !_dispdata_array_convert_3coord (viewable->dataset,
					  viewable->window->slice_direction,
					  &x, &y, &z, *h, *v) )
	return (FALSE);
    if (viewable->dataset->data->pseudo)
	array = viewable->dataset->data->pseudo;
    else array = viewable->dataset->data->red;
    if (iarray_num_dim (array) == 3)
    {
	switch (viewable->window->slice_direction)
	{
	  case KDISPLAYDATA_SLICE_DIRECTION_XY:
	    *h = x;
	    *v = y;
	    break;
	  case KDISPLAYDATA_SLICE_DIRECTION_XZ:
	    *h = x;
	    *v = z;
	    break;
	  case KDISPLAYDATA_SLICE_DIRECTION_ZY:
	    *h = z;
	    *v = y;
	    break;
	}
    }
    else
    {
	*h = x;
	*v = y;
    }
    return (TRUE);
}   /*  End Function _dispdata_array_convert_2coord  */

static void _dispdata_array_line_profile (struct viewable *viewable,
					  double x, double y)
/*  [SUMMARY] Draw a line profile for a dataset.
    <viewable> The viewable.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    [RETURNS] Nothing.
*/
{
    KOrthogonalProfile oprofile = viewable->profile;
    flag correct_aberration = FALSE;
    uaddr refpos = 0;  /*  Initialise to keep compiler happy  */
    double x_coord, y_coord, z_coord;
    uaddr coord_indices[2];
    unsigned int dim_indices[2];

    if ( !_dispdata_array_convert_3coord (viewable->dataset,
					  viewable->window->slice_direction,
					  &x_coord, &y_coord, &z_coord, x, y) )
	return;
    if (iarray_num_dim (viewable->dataset->data->pseudo) == 2)
    {
	/*  Can only get X or Y profile  */
	switch (viewable->window->profile_direction)
	{
	  case KDISPLAYDATA_PROFILE_DIR_Z:
	    fprintf (stderr, "Z profile not possible with 2D dataset\n");
	    return;
	    break;
	  case KDISPLAYDATA_PROFILE_DIR_Y:
	    dim_indices[0] = 1;
	    coord_indices[0] = x_coord;
	    refpos = y_coord;
	    break;
	  case KDISPLAYDATA_PROFILE_DIR_X:
	    dim_indices[0] = 0;
	    coord_indices[0] = y_coord;
	    refpos = x_coord;
	    break;
	}
    }
    else
    {
	/*  Can get X, Y or Z profile  */
	switch (viewable->window->profile_direction)
	{
	  case KDISPLAYDATA_PROFILE_DIR_Z:
	    dim_indices[0] = 2;
	    dim_indices[1] = 1;
	    coord_indices[0] = x_coord;
	    coord_indices[1] = y_coord;
	    refpos = z_coord;
	    if (viewable->window->slice_direction
		== KDISPLAYDATA_SLICE_DIRECTION_XY)
		correct_aberration = TRUE;
	    break;
	  case KDISPLAYDATA_PROFILE_DIR_Y:
	    dim_indices[0] = 2;
	    dim_indices[1] = 0;
	    coord_indices[0] = x_coord;
	    coord_indices[1] = z_coord;
	    refpos = y_coord;
	    if (viewable->window->slice_direction
		== KDISPLAYDATA_SLICE_DIRECTION_XZ)
		correct_aberration = TRUE;
	    break;
	  case KDISPLAYDATA_PROFILE_DIR_X:
	    dim_indices[0] = 0;
	    dim_indices[1] = 1;
	    coord_indices[0] = z_coord;
	    coord_indices[1] = y_coord;
	    refpos = x_coord;
	    if (viewable->window->slice_direction
		== KDISPLAYDATA_SLICE_DIRECTION_ZY)
		correct_aberration = TRUE;
	    break;
	}
    }
    graph1d_orthogonal_show_trace (oprofile, dim_indices, correct_aberration,
				   coord_indices, refpos);
}   /*  End Function _dispdata_array_line_profile  */

static void _dispdata_array_box_event (struct viewable *viewable,
				       CONST double *h, CONST double *v,
				       flag configure)
/*  [SUMMARY] Process a box event for a dataset.
    <viewable> The viewable.
    <x> The horizontal world co-ordinates of the box.
    <y> The vertical world co-ordinates of the box.
    <configure> If TRUE, the canvas should be configured.
    [RETURNS] Nothing.
*/
{
    switch (viewable->window->profile_mode)
    {
      case KDISPLAYDATA_PROFILE_MODE_BOX_SUM:
	_dispdata_array_box_profile (viewable, TRUE, h, v, configure);
	break;
      case KDISPLAYDATA_PROFILE_MODE_BOX_AVERAGE:
	_dispdata_array_box_profile (viewable, FALSE, h, v, configure);
	break;
      case KDISPLAYDATA_PROFILE_MODE_BOXED_HORIZONTAL:
	_dispdata_array_boxed_profile (viewable, TRUE, h, v, configure);
	break;
      case KDISPLAYDATA_PROFILE_MODE_BOXED_VERTICAL:
	_dispdata_array_boxed_profile (viewable, FALSE, h, v, configure);
	break;
      default:
	break;
    }
}   /*  End Function _dispdata_array_box_event  */

static void _dispdata_array_box_profile (struct viewable *viewable, flag sum,
					 CONST double *h, CONST double *v,
					 flag configure)
/*  [SUMMARY] Draw a box profile for a 3D dataset.
    <viewable> The viewable.
    <sum> If TRUE, a sum profile is computed instead of an average.
    <x> The horizontal world co-ordinates of the box.
    <y> The vertical world co-ordinates of the box.
    <configure> If TRUE, the canvas should be configured.
    [RETURNS] Nothing.
*/
{
    uaddr refpos = 0;  /*  Initialise to keep compiler happy  */
    unsigned int dim_indices[2];
    uaddr start_indices[2], stop_indices[2];
    double d_tmp;
    double x[2], y[2], z[2];

    if (iarray_num_dim (viewable->dataset->data->pseudo) != 3)
    {
	fprintf (stderr, "Box profile not possible with 2D dataset\n");
	return;
    }
    if ( !_dispdata_array_convert_3coord (viewable->dataset,
					  viewable->window->slice_direction,
					  x, y, z, h[0], v[0]) )
	return;
    if ( !_dispdata_array_convert_3coord (viewable->dataset,
					  viewable->window->slice_direction,
					  x + 1, y + 1, z + 1, h[1], v[1]) )
	return;
    if (x[0] > x[1]) swap (x[0], x[1], d_tmp);
    if (y[0] > y[1]) swap (y[0], y[1], d_tmp);
    if (z[0] > z[1]) swap (z[0], z[1], d_tmp);
    /*  Convert linear world co-ordinates to array indices and display  */
    switch (viewable->window->profile_direction)
    {
      case KDISPLAYDATA_PROFILE_DIR_Z:
	dim_indices[0] = 2;
	dim_indices[1] = 1;
	start_indices[0] = x[0];
	start_indices[1] = y[0];
	stop_indices[0] = x[1] + 1.0;
	stop_indices[1] = y[1] + 1.0;
	refpos = z[0];
	break;
      case KDISPLAYDATA_PROFILE_DIR_Y:
	dim_indices[0] = 2;
	dim_indices[1] = 0;
	start_indices[0] = x[0];
	start_indices[1] = z[0];
	stop_indices[0] = x[1] + 1.0;
	stop_indices[1] = z[1] + 1.0;
	refpos = y[0];
	break;
      case KDISPLAYDATA_PROFILE_DIR_X:
	dim_indices[0] = 0;
	dim_indices[1] = 1;
	start_indices[0] = z[0];
	start_indices[1] = y[0];
	stop_indices[0] = z[1] + 1.0;
	stop_indices[1] = y[1] + 1.0;
	refpos = x[0];
	break;
    }
    graph1d_orthogonal_show_boxed_trace (viewable->profile, dim_indices, sum,
					 start_indices, stop_indices, refpos);
    if (configure)
    {
	graph1d_orthogonal_configure_canvas (viewable->profile);
	if (sum)
	    graph1d_unzoom (viewable->window->profile_canvas, FALSE, FALSE,
			    TRUE);
    }
}   /*  End Function _dispdata_array_box_profile  */

static void _dispdata_array_boxed_profile (struct viewable *viewable,
					   flag horizontal,
					   CONST double *h, CONST double *v,
					   flag configure)
/*  [SUMMARY] Draw a (2D) boxed profile for a dataset.
    <viewable> The viewable.
    <horizontal> If TRUE, the profile is computed along the horizontal axis and
    vertical averaging is performed, else the profile is computed along the
    vertical axis and horizontal averaging is performed
    <h> The horizontal world co-ordinates of the box.
    <v> The vertical world co-ordinates of the box.
    <configure> If TRUE, the canvas should be configured.
    [RETURNS] Nothing.
*/
{
    ViewableImage vimage;
    unsigned int prof_len, type, hdim, vdim, elem_index, count, low, high;
    unsigned long hist_peak, hist_mode;
    double d_tmp, background, val, sum, wsum, peak_height, fwhm, mag_fit;
    double min_y = TOOBIG;
    double max_y = -TOOBIG;
    struct data_container *data = viewable->dataset->data;
    struct data_attributes *att = &viewable->dataset->attributes;
    double *vg_x, *vg_y, *curve_x, *curve_y;
    array_desc *arr_desc;
    dim_desc *hd, *vd;
    CONST char *array;
    unsigned long hist_array[MAX_HISTOGRAM_BINS];
    double hi[2], vi[2];
    static char function_name[] = "_dispdata_array_boxed_profile";

    hi[0] = h[0];
    hi[1] = h[1];
    vi[0] = v[0];
    vi[1] = v[1];
    if ( !_dispdata_array_convert_2coord (viewable, hi, vi) ) return;
    if ( !_dispdata_array_convert_2coord (viewable, hi + 1, vi + 1) ) return;
    if (hi[0] > hi[1]) swap (hi[0], hi[1], d_tmp);
    if (vi[0] > vi[1]) swap (vi[0], vi[1], d_tmp);
    if (horizontal) prof_len = (unsigned int) hi[1] - (unsigned int) hi[0] + 1;
    else prof_len = (unsigned int) vi[1] - (unsigned int) vi[0] + 1;
    if ( !graph1d_allocate_coords (viewable->vgraph, prof_len, &vg_x, &vg_y) )
	m_abort (function_name, "profile co-ordinates");
    if ( !graph1d_allocate_coords (viewable->gausscurve, GAUSSCURVE_LENGTH,
				   &curve_x, &curve_y) )
	m_abort (function_name, "gausscurve co-ordinates");
    vimage = _dispdata_array_get_vimage (viewable);
    viewimg_get_attributes (vimage,
			    VIEWIMG_VATT_ARRAY_DESC, &arr_desc,
			    VIEWIMG_VATT_SLICE, &array,
			    VIEWIMG_VATT_HDIM, &hdim,
			    VIEWIMG_VATT_VDIM, &vdim,
			    VIEWIMG_VATT_PSEUDO_INDEX, &elem_index,
			    VIEWIMG_VATT_END);
    hd = arr_desc->dimensions[hdim];
    vd = arr_desc->dimensions[vdim];
    array += ds_get_element_offset (arr_desc->packet, elem_index);
    hist_peak = 0;
    hist_mode = 0;
    m_clear (hist_array, sizeof hist_array);
    type = arr_desc->packet->element_types[elem_index];
    if (horizontal)
    {
	if ( !_dispdata_array_collapse_dim (vg_y, array, type,
					    hi[0], hi[1],
					    arr_desc->offsets[hdim],
					    vi[0], vi[1],
					    arr_desc->offsets[vdim],
					    data->scale, data->offset ) )
	    return;
	count = (unsigned int) vi[1] - (unsigned int) vi[0] + 1;
	ds_find_single_histogram (array +
				  arr_desc->offsets[hdim][ (int) hi[0] ],
				  type, CONV_CtoR_REAL, count,
				  arr_desc->offsets[vdim] + (int) vi[0], 0,
				  att->profile_lower_clip,
				  att->profile_upper_clip,
				  MAX_HISTOGRAM_BINS, hist_array, &hist_peak,
				  &hist_mode);
	ds_find_single_histogram (array +
				  arr_desc->offsets[hdim][ (int) hi[1] ],
				  type, CONV_CtoR_REAL, count,
				  arr_desc->offsets[vdim] + (int) vi[0], 0,
				  att->profile_lower_clip,
				  att->profile_upper_clip,
				  MAX_HISTOGRAM_BINS, hist_array, &hist_peak,
				  &hist_mode);
    }
    else
    {
	if ( !_dispdata_array_collapse_dim (vg_y, array, type,
					    vi[0], vi[1],
					    arr_desc->offsets[vdim],
					    hi[0], hi[1],
					    arr_desc->offsets[hdim],
					    data->scale, data->offset ) )
	    return;
	count = (unsigned int) hi[1] - (unsigned int) hi[0] + 1;
	ds_find_single_histogram (array +
				  arr_desc->offsets[vdim][ (int) vi[0] ],
				  type, CONV_CtoR_REAL, count,
				  arr_desc->offsets[hdim] + (int) hi[0], 0,
				  att->profile_lower_clip,
				  att->profile_upper_clip,
				  MAX_HISTOGRAM_BINS, hist_array, &hist_peak,
				  &hist_mode);
	ds_find_single_histogram (array +
				  arr_desc->offsets[vdim][ (int) vi[1] ],
				  type, CONV_CtoR_REAL, count,
				  arr_desc->offsets[hdim] + (int) hi[0], 0,
				  att->profile_lower_clip,
				  att->profile_upper_clip,
				  MAX_HISTOGRAM_BINS, hist_array, &hist_peak,
				  &hist_mode);
    }
    background = (double) hist_mode / MAX_HISTOGRAM_BINS;
    background = att->profile_lower_clip + (att->profile_upper_clip -
					    att->profile_lower_clip) *
	background;
    background = background * data->scale + data->offset;
    /*  Find the peak in the profile  */
    for (count = 0, low = 0, val = -TOOBIG; count < prof_len; ++count)
    {
	d_tmp = vg_y[count];
	if ( DOUBLE_IS_VALID (d_tmp) )
	{
	    if (d_tmp > val)
	    {
		val = d_tmp;
		low = count;
	    }
	    if (d_tmp < min_y) min_y = d_tmp;
	    if (d_tmp > max_y) max_y = d_tmp;
	}
    }
    val -= background;
    high = low;
    /*  Find the lower 25% point  */
    while ( (low > 0) && (vg_y[low] - background > val * 0.25) ) --low;
    /*  Find the upper 25% point  */
    while ( (high < prof_len - 1) &&
	    (vg_y[high] - background > val * 0.25) ) ++high;
    /*  Now find centre: use a weighted mean algorithm  */
    for (count = low, sum = 0.0, wsum = 0.0; count <= high; ++count)
    {
	val = vg_y[count] - background;
	sum += val;
	wsum += val * (double) count;
    }
    wsum = wsum / sum;
    for (count = 0; count < prof_len; ++count)
	vg_x[count] = (double) count - wsum;
    /*  TODO: remove assumption that profile is positive  */
    if ( !ds_fitgauss_list (vg_x, vg_y, prof_len, background, FALSE,
			    0.0, &peak_height, &fwhm, &mag_fit) )
    {
	fprintf (stderr, "Error fitting gaussian\n");
	return;
    }
    peak_height += background;
    mag_fit = 30.0 - 2.5 * log10 (mag_fit);
    fprintf (stderr,
	     "Raw peak  Peak      Background Opt. mag. FWHM  C%c\n",
	     horizontal ? 'h' : 'v');
    fprintf (stderr, "%-9g %-9g %-10g %-9g %-5.2f %-7.1f\n",
	     peak_height, peak_height - background, background, mag_fit,
	     fwhm,
	     floor (horizontal ? hi[0] : vi[0]) + wsum);
    d_tmp = (double) (prof_len - 1) / (double) (GAUSSCURVE_LENGTH - 1);
    for (count = 0; count < GAUSSCURVE_LENGTH; ++count)
	curve_x[count] = d_tmp * (double) count - wsum;
    ds_gausscurve (curve_y, NULL, GAUSSCURVE_LENGTH, wsum / d_tmp,
		   peak_height, fwhm / d_tmp, background, FALSE);
    graph1d_set_size (viewable->vgraph, prof_len);
    graph1d_set_active (viewable->vgraph, TRUE, FALSE, FALSE, FALSE, 0);
    graph1d_set_size (viewable->gausscurve, GAUSSCURVE_LENGTH);
    graph1d_set_active (viewable->gausscurve, TRUE, FALSE, FALSE, FALSE, 0);
    if (configure)
    {
	canvas_specify (viewable->window->profile_canvas,
			"Offset (pixels)", iarray_value_name (data->pseudo),
			0, NULL, NULL);
	if (peak_height > max_y) max_y = peak_height;
	graph1d_set_canvas_attributes
	    (viewable->window->profile_canvas,
	     GRAPH1D_CANVAS_ATT_LEFT_X, vg_x[0],
	     GRAPH1D_CANVAS_ATT_RIGHT_X, vg_x[prof_len - 1],
	     GRAPH1D_CANVAS_ATT_MINIMUM_Y, min_y,
	     GRAPH1D_CANVAS_ATT_MAXIMUM_Y, max_y,
	     GRAPH1D_CANVAS_ATT_END);
	canvas_set_attributes (viewable->window->profile_canvas,
			       CANVAS_ATT_BOTTOM_Y, min_y,
			       CANVAS_ATT_TOP_Y, max_y,
			       CANVAS_ATT_END);
    }
}   /*  End Function _dispdata_array_boxed_profile  */

static flag _dispdata_array_collapse_dim (double *out, CONST char *in,
					  unsigned int in_type, uaddr start_x,
					  uaddr end_x, CONST uaddr *xoffsets,
					  uaddr start_y, uaddr end_y,
					  CONST uaddr *yoffsets,
					  double scale, double offset)
/*  [SUMMARY] Collapse a dimension along Y.
    <out> The output array.
    <in> The input image.
    <in_type> The type of the input image.
    <start_x> The starting index along X.
    <end_x> The end index along X.
    <xoffsets> The array of horizontal byte offsets
    <start_y> The starting index along Y.
    <end_y> The end index along Y.
    <yoffsets> The array of vertical byte offsets
    <scale> The data scale.
    <offset> The data offset.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag some_data = FALSE;
    uaddr x, y, xo, count;
    double dval, sum;
    CONST char *ptr;
    double value[2];
    static char function_name[] = "_dispdata_array_collapse_dim";

    for (x = start_x, xo = 0; x <= end_x; ++x, ++xo)
    {
	sum = 0.0;
	count = 0;
	ptr = in + xoffsets[x];
	for (y = start_y; y <= end_y; ++y)
	{
	    ds_get_element (ptr + yoffsets[y], in_type, value, NULL);
	    dval = value[0];
	    if ( DOUBLE_IS_VALID (dval) )
	    {
		some_data = TRUE;
		sum += dval;
		++count;
	    }
	}
	if (count < 1) dval = TOOBIG;
	else dval = sum / (double) count * scale + offset;
	out[xo] = dval;
    }
    if (some_data) return (TRUE);
    fprintf (stderr, "%s: all blanks!\n", function_name);
    return (FALSE);
}   /*  End Function _dispdata_array_collapse_dim  */

static void _dispdata_array_circle_event (struct viewable *viewable,
					  CONST double *h, CONST double *v,
					  flag configure)
/*  [SUMMARY] Process a circle event for a dataset.
    <viewable> The viewable.
    <x> The horizontal world co-ordinates of the circle.
    <y> The vertical world co-ordinates of the circle.
    <configure> If TRUE, the canvas should be configured.
    [RETURNS] Nothing.
*/
{
    switch (viewable->window->profile_mode)
    {
      case KDISPLAYDATA_PROFILE_MODE_RADIAL:
	_dispdata_array_radial_profile (viewable, FALSE, h, v, configure);
	break;
      case KDISPLAYDATA_PROFILE_MODE_RADIAL_GAUSS:
	_dispdata_array_radial_profile (viewable, TRUE, h, v, configure);
	break;
      case KDISPLAYDATA_PROFILE_MODE_AZIMUTHAL_AVERAGE:
	_dispdata_array_azimuthal_avg_profile (viewable, h, v, configure);
	break;
      default:
	break;
    }
}   /*  End Function _dispdata_array_circle_event  */

static void _dispdata_array_radial_profile (struct viewable *viewable,
					    flag do_fit,
					    CONST double *h, CONST double *v,
					    flag configure)
/*  [SUMMARY] Draw a radial profile for a dataset.
    <viewable> The viewable.
    <do_fit> If TRUE, fit a Gaussian.
    <h> The horizontal world co-ordinates of the profile.
    <v> The vertical world co-ordinates of the profile.
    <configure> If TRUE, the canvas should be configured.
    [RETURNS] Nothing.
*/
{
    ViewableImage vimage;
    KWorldCanvas worldcanvas;
    KPixCanvas pixcanvas;
    flag inverted;
    unsigned int hdim, vdim, elem_index, num_points, count;
    unsigned long pixel_value = 0;
    double d_tmp, radius, background, px, py, hw, vw;
    double sum_raw, sum_subtracted, mag_raw, mag_subtracted;
    double peak_height = -TOOBIG;
    double flux_raw = TOOBIG;
    double flux_subtracted = TOOBIG;
    double weight = TOOBIG;
    struct data_container *data = viewable->dataset->data;
    double *vg_x, *vg_y;
    array_desc *arr_desc;
    dim_desc *hd, *vd;
    CONST char *array;
    double hi[3], vi[3];
    static char function_name[] = "_dispdata_array_radial_profile";

    hi[0] = h[0];
    hi[1] = h[1];
    hi[2] = h[2];
    vi[0] = v[0];
    vi[1] = v[1];
    vi[2] = v[2];
    if ( !_dispdata_array_convert_2coord (viewable, hi, vi) ) return;
    if ( !_dispdata_array_convert_2coord (viewable, hi + 1, vi + 1) ) return;
    if ( !_dispdata_array_convert_2coord (viewable, hi + 2, vi + 2) ) return;
    radius = (hi[1] - hi[0]) * (hi[1] - hi[0]) + (vi[2] - vi[0]) * (vi[2] -
								    vi[0]);
    radius = ceil ( sqrt (radius * 0.5) );
    if (radius < 2.0)
    {
	fprintf (stderr, "%s: radius: %g too small\n", function_name, radius);
	return;
    }
    vimage = _dispdata_array_get_vimage (viewable);
    viewimg_get_attributes (vimage,
			    VIEWIMG_VATT_ARRAY_DESC, &arr_desc,
			    VIEWIMG_VATT_SLICE, &array,
			    VIEWIMG_VATT_HDIM, &hdim,
			    VIEWIMG_VATT_VDIM, &vdim,
			    VIEWIMG_VATT_PSEUDO_INDEX, &elem_index,
			    VIEWIMG_VATT_END);
    array += ds_get_element_offset (arr_desc->packet, elem_index);
    hd = arr_desc->dimensions[hdim];
    vd = arr_desc->dimensions[vdim];
    if ( !ds_find_centroid (array, arr_desc->packet->element_types[elem_index],
			    &background, &inverted,
			    hd->length, arr_desc->offsets[hdim], hi, radius,
			    vd->length, arr_desc->offsets[vdim], vi, radius) )
	return;
    /*  Convert centroid to screen pixel co-ordinates and draw  */
    px = hi[0];
    py = vi[0];
    ds_convert_coordinates (hd, 1, &px, FALSE, FALSE, FALSE);
    ds_convert_coordinates (vd, 1, &py, FALSE, FALSE, FALSE);
    worldcanvas = viewimg_get_worldcanvas (vimage);
    pixcanvas = canvas_get_pixcanvas (worldcanvas);
    canvas_convert_from_canvas_coords (worldcanvas, FALSE, TRUE, 1,
				       &px, &py, &px, &py);
    kwin_get_colour_harder (pixcanvas, "red", &pixel_value);
    kwin_draw_line (pixcanvas, px - 1, py, px + 1, py, pixel_value);
    kwin_draw_line (pixcanvas, px, py - 1, px, py + 1, pixel_value);
    hi[1] = hi[0];
    vi[1] = vi[0];
    hi[2] = vi[2] = radius;
    d_tmp = 0.0;
    canvas_convert_ellipse_coords (worldcanvas, FALSE, TRUE,
				   hi + 1, vi + 1, hi + 2, vi + 2, &d_tmp);
    kwin_draw_ellipse (pixcanvas, hi[1], vi[1], hi[2], vi[2], pixel_value);
    background = background * data->scale + data->offset;
    if ( !_dispdata_array_collect_radial_points
	 (viewable->vgraph, array, arr_desc->packet->element_types[elem_index],
	  hd->length, arr_desc->offsets[hdim], hi[0], radius,
	  vd->length, arr_desc->offsets[vdim], vi[0], radius,
	  data->scale, data->offset) ) return;
    /*  Convert centroid to world co-ordinates  */
    hw = hi[0];
    vw = vi[0];
    canvas_coords_transform (worldcanvas, 1, &hw, FALSE, &vw, FALSE);
    num_points = graph1d_get_size (viewable->vgraph, &vg_x, &vg_y);
    /*  Compute sum and magnitude  */
    for (count = 0, sum_raw = 0.0, sum_subtracted = 0.0; count < num_points;
	 ++count)
    {
	sum_raw += vg_y[count];
	sum_subtracted += vg_y[count] - background;
    }
    mag_raw = 30.0 - 2.5 * log10 (inverted ? -sum_raw : sum_raw);
    mag_subtracted = 30.0 - 2.5 * log10 (inverted ? -sum_subtracted :
					 sum_subtracted);
    if (strcmp (iarray_value_name (data->pseudo), "JY/PIX") == 0)
	weight = 1.0;
    else if (strcmp (iarray_value_name (data->pseudo), "JY/BEAM") == 0)
	weight = iarray_get_beam_correction (data->pseudo, hdim, vdim);
    if ( DOUBLE_IS_VALID (weight) )
    {
	flux_raw = sum_raw * weight;
	flux_subtracted = sum_subtracted * weight;
    }
    if (do_fit)
    {
	double sum_fit, mag_fit, fwhm;
	double *curve_x, *curve_y;
	char flux_str[STRING_LENGTH];
	char fwhm_unit[STRING_LENGTH], fwhm_str[STRING_LENGTH];
	char h_unit[STRING_LENGTH], h_str[STRING_LENGTH];
	char v_unit[STRING_LENGTH], v_str[STRING_LENGTH];

	if ( !graph1d_allocate_coords (viewable->gausscurve, GAUSSCURVE_LENGTH,
				       &curve_x, &curve_y) )
	    m_abort (function_name, "gausscurve co-ordinates");
	if ( !ds_fitgauss_list (vg_x, vg_y, num_points,
				background, inverted, 0.0, &peak_height,
				&fwhm, &sum_fit) )
	{
	    fprintf (stderr, "Error fitting gaussian\n");
	    return;
	}
	mag_fit = 30.0 - 2.5 * log10 (sum_fit);
#if 0
	fprintf (stderr,
		 "Raw sum Adj sum Adj mag Backgnd Cx      Cy      Fit peak Adj peak Fit Mag FWHM\n");
	fprintf (stderr,
		 "%-7g %-7g %-7g %-7g %-7.1f %-7.1f %-8g %-8g %-7g %-5.2f\n",
		 sum_raw, sum_subtracted, mag_subtracted, background,
		 hi[0], vi[0], peak_height + background, peak_height, mag_fit,
		 fwhm);
	fprintf (stderr, "Npoints: %u  ", num_points);
	if ( DOUBLE_IS_VALID (flux_raw) )
	{
	    double scale;
	    char unit[STRING_LENGTH], format[STRING_LENGTH];

	    ds_format_unit (unit, format, &scale, "JY", flux_raw);
	    fprintf (stderr, "Raw flux: ");
	    fprintf (stderr, format, flux_raw * scale);
	    fprintf (stderr, " %s  Adj flux: ", unit);
	    fprintf (stderr, format, flux_subtracted * scale);
	    fprintf (stderr, " %s\n", unit);
	}
	_dispdata_array_show_fwhm (vimage, hi[0], vi[0], fwhm);
#else
	if ( DOUBLE_IS_VALID (flux_raw) )
	{
	    sprintf (flux_str, "%5.2f", flux_raw);
	}
	else strcpy (flux_str, "   -");
	_dispdata_array_show_fwhm2 (vimage, hi[0], vi[0], fwhm,
				    fwhm_unit, fwhm_str);
	if (data->ap)
	{
	    char *ptr1, *ptr2;

	    wcs_astro_format (data->ap, hd->name, h_unit, hw);
	    ptr1 = strchr (h_unit, ' ');
	    *ptr1++ = '\0';
	    for (ptr2 = h_str; *ptr1; ++ptr1)
	    {
		if ( isalpha (*ptr1) ) continue;
		if ( (ptr1[0] == ' ') && (ptr1[1] == ' ') ) continue;
		*ptr2++ = (*ptr1 == ' ') ? ':' : *ptr1;
	    }
	    *ptr2 = '\0';
	    wcs_astro_format (data->ap, vd->name, v_unit, vw);
	    ptr1 = strchr (v_unit, ' ');
	    *ptr1++ = '\0';
	    for (ptr2 = v_str; *ptr1; ++ptr1)
	    {
		if ( isalpha (*ptr1) ) continue;
		if ( (ptr1[0] == ' ') && (ptr1[1] == ' ') ) continue;
		*ptr2++ = (*ptr1 == ' ') ? ':' : *ptr1;
	    }
	    *ptr2 = '\0';
	}
	else
	{
	    strcpy (h_unit, hd->name);
	    sprintf (h_str, "%g", hw);
	    strcpy (v_unit, vd->name);
	    sprintf (v_str, "%g", vw);
	}
	fprintf (stderr,
		 "        Centre-x      Centre-y     FWHM |       Peak       Sum      Flux   Mag\n");
	fprintf (stderr,
		 "Pixel:   %-6.2f        %-6.2f     %5.2f |Raw:     -     %-9.4g %-7s %-5.2f\n",
		 hi[0], vi[0], fwhm, sum_raw, flux_str, mag_raw);
	if ( DOUBLE_IS_VALID (flux_raw) )
	{
	    sprintf (flux_str, "%5.2f", flux_subtracted);
	}
	fprintf (stderr,
		 "        %-12s %-12s %6s|Adj:     -     %-9.4g %-7s %-5.2f\n",
		 h_unit, v_unit, fwhm_unit, sum_subtracted, flux_str, mag_subtracted);
	if ( DOUBLE_IS_VALID (flux_raw) )
	{
	    sprintf (flux_str, "%5.2f", sum_fit * weight);
	}
	fprintf (stderr,
		 "World:  %-12s %-12s %6s|Fit: %-9.4g %-9.4g %-7s %-5.2f\n",
		 h_str, v_str, fwhm_str, peak_height, sum_fit, flux_str, mag_fit);
	fprintf (stderr,
		 "Npoints: %u  Baselevel for adjustment: %g  Units: %s\n",
		 num_points, background,
		 arr_desc->packet->element_desc[elem_index]);
#endif
	d_tmp = radius / (double) (GAUSSCURVE_LENGTH - 1);
	for (count = 0; count < GAUSSCURVE_LENGTH; ++count)
	    curve_x[count] = d_tmp * (double) count;
	ds_gausscurve (curve_y, NULL, GAUSSCURVE_LENGTH, 0.0,
		       peak_height + background, fwhm / d_tmp, background,
		       inverted);
	graph1d_set_size (viewable->gausscurve, GAUSSCURVE_LENGTH);
	graph1d_set_active (viewable->gausscurve, TRUE, FALSE, FALSE, FALSE,0);
    }
    else
    {
	fprintf (stderr, "Raw sum Adj sum Adj mag Backgnd Cx      Cy\n");
	fprintf (stderr, "%-7g %-7g %-7g %-7g %-7.1f %-7.1f\n",
		 sum_raw, sum_subtracted, mag_subtracted, background,
		 hi[0], vi[0]);
	fprintf (stderr, "Npoints: %u", num_points);
	if ( DOUBLE_IS_VALID (flux_raw) )
	{
	    double scale;
	    char unit[STRING_LENGTH], format[STRING_LENGTH];

	    ds_format_unit (unit, format, &scale, "JY", flux_raw);
	    fprintf (stderr, "  Raw flux: ");
	    fprintf (stderr, format, flux_raw * scale);
	    fprintf (stderr, " %s  Adj flux: ", unit);
	    fprintf (stderr, format, flux_subtracted * scale);
	    fprintf (stderr, " %s", unit);
	}
	putc ('\n', stderr);
    }
    if (configure)
    {
	double min_y = TOOBIG;
	double max_y = -TOOBIG;

	canvas_specify (viewable->window->profile_canvas,
			"Radius (pixels)", iarray_value_name (data->pseudo),
			0, NULL, NULL);
	for (count = 0; count < num_points; ++count)
	{
	    d_tmp = vg_y[count];
	    if (d_tmp < min_y) min_y = d_tmp;
	    if (d_tmp > max_y) max_y = d_tmp;
	}
	if (peak_height + background > max_y) max_y = peak_height + background;
	graph1d_set_canvas_attributes
	    (viewable->window->profile_canvas,
	     GRAPH1D_CANVAS_ATT_LEFT_X, 0.0,
	     GRAPH1D_CANVAS_ATT_RIGHT_X, radius,
	     GRAPH1D_CANVAS_ATT_MINIMUM_Y, min_y,
	     GRAPH1D_CANVAS_ATT_MAXIMUM_Y, max_y,
	     GRAPH1D_CANVAS_ATT_END);
	canvas_set_attributes (viewable->window->profile_canvas,
			       CANVAS_ATT_BOTTOM_Y, min_y,
			       CANVAS_ATT_TOP_Y, max_y,
			       CANVAS_ATT_END);
    }
}   /*  End Function _dispdata_array_radial_profile  */

static flag _dispdata_array_collect_radial_points (KViewableGraph vgraph,
						   CONST char *in,
						   unsigned int in_type,
						   uaddr xlen,
						   CONST uaddr *xoffsets,
						   double xcentre,
						   signed int xradius,
						   uaddr ylen,
						   CONST uaddr *yoffsets,
						   double ycentre,
						   signed int yradius,
						   float scale, float offset)
/*  [SUMMARY] Compute a list of radial profile points.
    <vgraph> The KViewableGraph to write the points to.
    <in> The input image.
    <in_type> The type of the input image.
    <xlen> The horizontal image width.
    <xoffsets> The array of horizontal byte offsets
    <xcentre> The horizontal ellipse centre. 0.0 is the centre of the origin.
    <xradius> The horizontal radius of the ellipse.
    <ylen> The vertical image width.
    <yoffsets> The array of vertical byte offsets
    <ycentre> The vertical ellipse centre. 0.0 is the centre of the origin.
    <yradius> The vertical radius of the ellipse.
    <scale> The data scale.
    <offset> The data offset.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    signed int x, y, xstart, xstop, xc, yc;
    unsigned int tpoints = (xradius * 2 + 1) * (yradius * 2 + 1);
    double max_radius = (xradius > yradius) ? xradius : yradius;
    double *xarr, *yarr;
    CONST char *ptr;
    static char function_name[] = "_dispdata_array_collect_radial_points";

    if ( !graph1d_allocate_coords (vgraph, tpoints, &xarr, &yarr) )
    {
	m_error_notify (function_name, "graph space");
	return (FALSE);
    }
    xc = rint (xcentre);
    yc = rint (ycentre);
    xstart = xc - xradius;
    xstop = xc + xradius + 1;
    tpoints = 0;
    for (y = yc - yradius; y < yc + yradius + 1; ++y)
	for (x = xstart; x < xstop; ++x)
	{
	    double radius;
	    double value[2];

	    if ( (x < 0) || (x >= xlen) || (y < 0) || (y >= ylen) ) continue;
	    ptr = in + xoffsets[x] + yoffsets[y];
	    ds_get_element (ptr, in_type, value, NULL);
	    if ( DOUBLE_IS_BLANK (value[0]) ) continue;
	    radius = sqrt ( (xcentre - x) * (xcentre - x) +
			    (ycentre - y) * (ycentre - y) );
	    if (radius > max_radius) continue;
	    xarr[tpoints] = radius;
	    yarr[tpoints] = value[0] * scale + offset;
	    ++tpoints;
	}
    graph1d_set_size (vgraph, tpoints);
    graph1d_set_active (vgraph, TRUE, FALSE, FALSE, FALSE, 0);
    return (TRUE);
}   /*  End Function _dispdata_array_collect_radial_points  */

static void _dispdata_array_show_fwhm (ViewableImage vimage,
				       double x, double y, double pixels)
/*  [SUMMARY] Show a FWHM value.
    <vimage> The ViewableImage.
    <x> The X position of the object in image pixels.
    <y> The Y position of the object in image pixels.
    <pixels> The FWHM of the object in image pixels.
    [RETURNS] Nothing.
*/
{
    KwcsAstro ap;
    unsigned int hd, vd;
    double fwhm, scale;
    array_desc *arr_desc;
    dim_desc *hdim, *vdim;
    char unit[STRING_LENGTH], format[STRING_LENGTH];
    double xarr[2], yarr[2];

    viewimg_get_attributes (vimage,
			    VIEWIMG_VATT_ARRAY_DESC, &arr_desc,
			    VIEWIMG_VATT_HDIM, &hd,
			    VIEWIMG_VATT_VDIM, &vd,
			    VIEWIMG_VATT_WCS_ASTRO, &ap,
			    VIEWIMG_VATT_END);
    hdim = arr_desc->dimensions[hd];
    vdim = arr_desc->dimensions[vd];
    xarr[0] = xarr[1] = x;
    yarr[0] = y;
    yarr[1] = y + pixels;
    if (ap) wcs_astro_transform3 (ap, FALSE, 2,
				  hdim->name, xarr, FALSE,
				  vdim->name, yarr, FALSE,
				  NULL, NULL, FALSE,
				  0, NULL, NULL);
    if ( wcs_astro_test_if_lon (ap, hdim->name) &&
	 wcs_astro_test_if_lat (ap, vdim->name) )
    {
	fwhm = (xarr[1] - xarr[0]) * cos (PION180 * yarr[0]);
	fwhm = sqrt ( fwhm * fwhm + (yarr[1] - yarr[0]) * (yarr[1] -yarr[0]) );
    }
    else if ( wcs_astro_test_if_lon (ap, vdim->name) &&
	      wcs_astro_test_if_lat (ap, hdim->name) )
    {
	fwhm = (yarr[1] - yarr[0]) * cos (PION180 * xarr[0]);
	fwhm = sqrt ( fwhm * fwhm + (xarr[1] - xarr[0]) * (xarr[1] -xarr[0]) );
    }
    else return;
    ds_format_unit (unit, format, &scale, "ASIZE", fwhm);
    fprintf (stderr, "FWHM: ");
    fprintf (stderr, format, fwhm * scale);
    fprintf (stderr, " %s\n", unit);
}   /*  End Function _dispdata_array_show_fwhm  */

static void _dispdata_array_show_fwhm2 (ViewableImage vimage,
					double x, double y, double pixels,
					char *unit_str, char *val_str)
/*  [SUMMARY] Show a FWHM value.
    <vimage> The ViewableImage.
    <x> The X position of the object in image pixels.
    <y> The Y position of the object in image pixels.
    <pixels> The FWHM of the object in image pixels.
    [RETURNS] Nothing.
*/
{
    KwcsAstro ap;
    unsigned int hd, vd;
    double fwhm, scale;
    array_desc *arr_desc;
    dim_desc *hdim, *vdim;
    char format[STRING_LENGTH];
    double xarr[2], yarr[2];

    unit_str[0] = '\0';
    val_str[0] = '\0';
    viewimg_get_attributes (vimage,
			    VIEWIMG_VATT_ARRAY_DESC, &arr_desc,
			    VIEWIMG_VATT_HDIM, &hd,
			    VIEWIMG_VATT_VDIM, &vd,
			    VIEWIMG_VATT_WCS_ASTRO, &ap,
			    VIEWIMG_VATT_END);
    hdim = arr_desc->dimensions[hd];
    vdim = arr_desc->dimensions[vd];
    xarr[0] = xarr[1] = x;
    yarr[0] = y;
    yarr[1] = y + pixels;
    if (ap) wcs_astro_transform3 (ap, FALSE, 2,
				  hdim->name, xarr, FALSE,
				  vdim->name, yarr, FALSE,
				  NULL, NULL, FALSE,
				  0, NULL, NULL);
    if ( wcs_astro_test_if_lon (ap, hdim->name) &&
	 wcs_astro_test_if_lat (ap, vdim->name) )
    {
	fwhm = (xarr[1] - xarr[0]) * cos (PION180 * yarr[0]);
	fwhm = sqrt ( fwhm * fwhm + (yarr[1] - yarr[0]) * (yarr[1] -yarr[0]) );
    }
    else if ( wcs_astro_test_if_lon (ap, vdim->name) &&
	      wcs_astro_test_if_lat (ap, hdim->name) )
    {
	fwhm = (yarr[1] - yarr[0]) * cos (PION180 * xarr[0]);
	fwhm = sqrt ( fwhm * fwhm + (xarr[1] - xarr[0]) * (xarr[1] -xarr[0]) );
    }
    else return;
    ds_format_unit (unit_str, format, &scale, "ASIZE", fwhm);
    sprintf (val_str, format, fwhm * scale);
}   /*  End Function _dispdata_array_show_fwhm2  */

static void _dispdata_array_azimuthal_avg_profile (struct viewable *viewable,
						   CONST double *h,
						   CONST double *v,
						   flag configure)
/*  [SUMMARY] Draw an azimuthal average profile for a dataset.
    <viewable> The viewable.
    <do_fit> If TRUE, fit a Gaussian.
    <h> The horizontal world co-ordinates of the profile.
    <v> The vertical world co-ordinates of the profile.
    <configure> If TRUE, the canvas should be configured.
    [RETURNS] Nothing.
*/
{
    ViewableImage vimage;
    unsigned int hdim, vdim, elem_index, num_points, count;
    double d_tmp, radius;
    struct data_container *data = viewable->dataset->data;
    double *vg_x, *vg_y;
    array_desc *arr_desc;
    dim_desc *hd, *vd;
    CONST char *array;
    double hi[3], vi[3];
    static char function_name[] = "_dispdata_array_azimuthal_avg_profile";

    hi[0] = h[0];
    hi[1] = h[1];
    hi[2] = h[2];
    vi[0] = v[0];
    vi[1] = v[1];
    vi[2] = v[2];
    if ( !_dispdata_array_convert_2coord (viewable, hi, vi) ) return;
    if ( !_dispdata_array_convert_2coord (viewable, hi + 1, vi + 1) ) return;
    if ( !_dispdata_array_convert_2coord (viewable, hi + 2, vi + 2) ) return;
    radius = (hi[1] - hi[0]) * (hi[1] - hi[0]) + (vi[2] - vi[0]) * (vi[2] -
								    vi[0]);
    radius = ceil ( sqrt (radius * 0.5) );
    if (radius < 2.0)
    {
	fprintf (stderr, "%s: radius: %g too small\n", function_name, radius);
	return;
    }
    vimage = _dispdata_array_get_vimage (viewable);
    viewimg_get_attributes (vimage,
			    VIEWIMG_VATT_ARRAY_DESC, &arr_desc,
			    VIEWIMG_VATT_SLICE, &array,
			    VIEWIMG_VATT_HDIM, &hdim,
			    VIEWIMG_VATT_VDIM, &vdim,
			    VIEWIMG_VATT_PSEUDO_INDEX, &elem_index,
			    VIEWIMG_VATT_END);
    array += ds_get_element_offset (arr_desc->packet, elem_index);
    hd = arr_desc->dimensions[hdim];
    vd = arr_desc->dimensions[vdim];
    if ( !_dispdata_array_azimuthally_average
	 (viewable->vgraph, array, arr_desc->packet->element_types[elem_index],
	  hd->length, arr_desc->offsets[hdim], hi[0], radius,
	  vd->length, arr_desc->offsets[vdim], vi[0], radius,
	  0.0, data->scale, data->offset) ) return;
    num_points = graph1d_get_size (viewable->vgraph, &vg_x, &vg_y);
    if (configure)
    {
	double min_y = TOOBIG;
	double max_y = -TOOBIG;

	canvas_specify (viewable->window->profile_canvas,
			"Radius (pixels)", iarray_value_name (data->pseudo),
			0, NULL, NULL);
	for (count = 0; count < num_points; ++count)
	{
	    d_tmp = vg_y[count];
	    if (d_tmp < min_y) min_y = d_tmp;
	    if (d_tmp > max_y) max_y = d_tmp;
	}
	graph1d_set_canvas_attributes
	    (viewable->window->profile_canvas,
	     GRAPH1D_CANVAS_ATT_LEFT_X, 0.0,
	     GRAPH1D_CANVAS_ATT_RIGHT_X, radius,
	     GRAPH1D_CANVAS_ATT_MINIMUM_Y, min_y,
	     GRAPH1D_CANVAS_ATT_MAXIMUM_Y, max_y,
	     GRAPH1D_CANVAS_ATT_END);
	canvas_set_attributes (viewable->window->profile_canvas,
			       CANVAS_ATT_BOTTOM_Y, min_y,
			       CANVAS_ATT_TOP_Y, max_y,
			       CANVAS_ATT_END);
    }
}   /*  End Function _dispdata_array_azimuthal_avg_profile  */

static flag _dispdata_array_azimuthally_average (KViewableGraph vgraph,
						 CONST char *in,
						 unsigned int in_type,
						 uaddr xlen,
						 CONST uaddr *xoffsets,
						 double xcentre,double xradius,
						 uaddr ylen,
						 CONST uaddr *yoffsets,
						 double ycentre,double yradius,
						 double rotation,
						 float scale, float offset)
/*  [SUMMARY] Azimuthally average an image.
    <vgraph> The KViewableGraph to write the profile to.
    <in> The input image.
    <in_type> The type of the input image.
    <xlen> The horizontal image width.
    <xoffsets> The array of horizontal byte offsets
    <xcentre> The horizontal ellipse centre. 0.0 is the centre of the origin.
    <xradius> The horizontal radius of the ellipse.
    <ylen> The vertical image width.
    <yoffsets> The array of vertical byte offsets
    <ycentre> The vertical ellipse centre. 0.0 is the centre of the origin.
    <yradius> The vertical radius of the ellipse.
    <rotation> The rotation of the ellipse.
    <scale> The data scale.
    <offset> The data offset.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int out_len;
    uaddr rcount, acount, num_angle_steps;
    double val = 0.0;
    double radius, radius_step, angle, angle_step;
    double xmax, ymax, cos_rot, sin_rot, cos_ang, sin_ang;
    CONST char *ptr;
    double *xarr, *yarr;
    static unsigned int count_arr_len = 0;
    static unsigned int *count_arr = NULL;
    static char function_name[] = "_dispdata_array_azimuthally_average";

    radius = xradius;
    if (yradius > radius) radius = yradius;
    out_len = radius + 1.1;
    if ( !graph1d_allocate_coords (vgraph, out_len, &xarr, &yarr) )
    {
	m_error_notify (function_name, "graph space");
	return (FALSE);
    }
    if (count_arr_len < out_len)
    {
	if (count_arr) m_free (count_arr);
	if ( ( count_arr = m_alloc (sizeof *count_arr * out_len) ) == NULL )
	{
	    m_abort (function_name, "count array");
	}
	count_arr_len = out_len;
    }
    for (rcount = 0; rcount < out_len; ++rcount)
    {
	xarr[rcount] = rcount;
	yarr[rcount] = 0.0;
	count_arr[rcount] = 0;
    }
    radius_step = 1.0 / (double) (out_len - 1);
    num_angle_steps = TWOPI * sqrt (xradius * xradius + yradius * yradius);
    angle_step = TWOPI / (double) num_angle_steps;
    xmax = xlen - 1;
    ymax = ylen - 1;
    cos_rot = cos (rotation * PION180);
    sin_rot = sin (rotation * PION180);
    for (acount = 0, angle = 0.0; acount < num_angle_steps;
	 ++acount, angle += angle_step)
    {
	cos_ang = cos (angle);
	sin_ang = sin (angle);
	for (rcount = 0, radius = 0.0; rcount < out_len;
	     ++rcount, radius += radius_step)
	{
	    int toobig_count;
	    float xg, yg, x0, x1, y0, y1, dx, dy, dx_r, dy_r;
	    double val, val_00, val_10, val_01, val_11;
	    float one = 1.0;
	    double xp, yp;
	    double zero = 0.0;
	    double value[2];

	    xp = xradius * radius * cos_ang;
	    yp = yradius * radius * sin_ang;
	    xg = xcentre + xp * cos_rot - yp * sin_rot;
	    yg = ycentre + yp * cos_rot + xp * sin_rot;
	    if ( (xg < zero) || (xg > xmax) || (yg < zero) || (yg > ymax) )
		continue;
	    /*  Take contribution from 4 neighbouring points  */
	    toobig_count = 0;
	    x0 = floor (xg);
	    x1 = ceil (xg);
	    dx = xg - x0;
	    dx_r = one - dx;
	    y0 = floor (yg);
	    y1 = ceil (yg);
	    dy = yg - y0;
	    dy_r = one - dy;
	    ptr = in + xoffsets[(int) x0] + yoffsets[(int) y0];
	    ds_get_element (ptr, in_type, value, NULL);
	    if ( DOUBLE_IS_BLANK (value[0]) )
	    {
		val_00 = zero;
		++toobig_count;
	    }
	    else val_00 = value[0];
	    ptr = in + xoffsets[(int) x1] + yoffsets[(int) y0];
	    ds_get_element (ptr, in_type, value, NULL);
	    if ( DOUBLE_IS_BLANK (value[0]) )
	    {
		val_10 = zero;
		++toobig_count;
	    }
	    else val_10 = value[0];
	    ptr = in + xoffsets[(int) x0] + yoffsets[(int) y1];
	    ds_get_element (ptr, in_type, value, NULL);
	    if ( DOUBLE_IS_BLANK (value[0]) )
	    {
		val_01 = zero;
		++toobig_count;
	    }
	    else val_01 = value[0];
	    ptr = in + xoffsets[(int) x1] + yoffsets[(int) y1];
	    ds_get_element (ptr, in_type, value, NULL);
	    if ( DOUBLE_IS_BLANK (value[0]) )
	    {
		val_11 = zero;
		++toobig_count;
	    }
	    else val_11 = value[0];
	    if (toobig_count > 3) continue;
	    val = val_00 * dx_r * dy_r + val_10 * dx * dy_r +
		val_01 * dx_r * dy + val_11 * dx * dy;
	    if ( DOUBLE_IS_BLANK (val) ) continue;
	    yarr[rcount] += val;
	    ++count_arr[rcount];
	}
    }
    /*  Average computation  */
    for (rcount = 0; rcount < out_len; ++rcount)
    {
	if (count_arr[rcount] < 1) val = TOOBIG;
	else
	{
	    val = yarr[rcount] / (double) count_arr[rcount];
	    val = val * scale + offset;
	}
	yarr[rcount] = val;
    }
    if ( DOUBLE_IS_BLANK (val) ) fprintf (stderr, "Trailing blank(s)\n");
    graph1d_set_size (vgraph, out_len);
    graph1d_set_active (vgraph, TRUE, FALSE, FALSE, FALSE, 0);
    return (TRUE);
}   /*  End Function _dispdata_array_azimuthally_average  */

static void _dispdata_array_set_title (struct window *window)
{
    flag first_contour = TRUE;
    int pos;
    char *name;
    struct blinkstate *blinkstate = window->active_blinkstate;
    struct viewable *viewable;
    struct data_container *data;
    char title[STRING_LENGTH];
    static char overlaid[] = " overlaid with ";

    if (!blinkstate || !blinkstate->image_viewable) return;
    data = blinkstate->image_viewable->dataset->data;
    name = data->object ? data->object : data->name;
    strncpy (title, name, STRING_LENGTH);
    title[STRING_LENGTH - 1] = '\0';  /*  Need this for strlen()  */
    pos = strlen (title);
    if (pos >= STRING_LENGTH) return;
    for (viewable = window->first; viewable;
	 viewable = viewable->window_chain.next)
    {
	struct blinkentry *blinkentry = viewable->active_blinkentry;

	if (blinkentry->contour_mode == KDISPLAYDATA_CONTOUR_OFF) continue;
	if (first_contour)
	{
	    strncpy (title + pos, overlaid, STRING_LENGTH - pos);
	    pos += sizeof overlaid - 1;
	    first_contour = FALSE;
	}
	else
	{
	    strncpy (title + pos, " and ", STRING_LENGTH - pos);
	    pos += 5;
	}
	data = viewable->dataset->data;
	name = data->object ? data->object : data->name;
	if (pos >= STRING_LENGTH) return;
	strncpy (title + pos, name, STRING_LENGTH - pos);
	title[STRING_LENGTH - 1] = '\0';  /*  Need this for strlen()  */
	pos = strlen (title);
	if (pos >= STRING_LENGTH) return;
    }
    if (window->pc.main) canvas_set_dressing (window->pc.main,
					      KCD_ATT_TITLE, title,
					      KCD_ATT_END);
    if (window->rgb.main) canvas_set_dressing (window->rgb.main,
					       KCD_ATT_TITLE, title,
					       KCD_ATT_END);
}   /*  End Function _dispdata_array_set_title  */

static flag _dispdata_array_irange_change_func (void *object,
						void *client1_data,
						void *call_data,
						void *client2_data)
/*  [SUMMARY] Generic callback.
    [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    void (*func) (KDisplayDataObject data, void *info);

    func = ( void (*) (KDisplayDataObject data, void *info) ) client1_data;
    (*func) (object, client2_data);
    return (FALSE);
}   /*  End Function _dispdata_array_irange_change_func  */

static flag _dispdata_array_test_profile_perpendicular
    (const struct window *window)
/*  [SUMMARY] Test if the profile direction is perpendicular to the slice.
    <window> The window.
    [RETURNS] TRUE if perpendicular, else FALSE.
*/
{
    switch (window->slice_direction)
    {
      case KDISPLAYDATA_SLICE_DIRECTION_XY:
	if (window->profile_direction == KDISPLAYDATA_PROFILE_DIR_Z)
	    return (TRUE);
	break;
      case KDISPLAYDATA_SLICE_DIRECTION_XZ:
	if (window->profile_direction == KDISPLAYDATA_PROFILE_DIR_Y)
	    return (TRUE);
	break;
      case KDISPLAYDATA_SLICE_DIRECTION_ZY:
	if (window->profile_direction == KDISPLAYDATA_PROFILE_DIR_X)
	    return (TRUE);
	break;
    }
    return (FALSE);
}   /*  End Function _dispdata_array_test_profile_perpendicular  */
